/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation

/**
 * ActivationManager handles client activation
 */
class ActivationManager {
    
    static let MESSAGE_DISPATCHER_URN: String =
        "urn:oracle:iot:dcd:capability:message_dispatcher"
    static let DIAGNOSTICS_URN: String =
        "urn:oracle:iot:dcd:capability:diagnostics"
    static let DIRECT_ACTIVATION_URN: String =
        "urn:oracle:iot:dcd:capability:direct_activation"
    static let INDIRECT_ACTIVATION_URN: String =
        "urn:oracle:iot:dcd:capability:indirect_activation"
    
    
    static func getInstance() -> ActivationManager {
        return INSTANCE
    }
    
    private static var INSTANCE: ActivationManager = ActivationManager()
    
    private init() {
    }
    
    private static let DEFAULT_MESSAGE_DIGEST_ALGORITHM: String = "HmacSHA256"
    
    /**
     * REST resource for activation policy
     */
    private static let REST_ACTIVATION_POLICY: String = RestApi.V2.getReqRoot()+"/activation/policy";
    /**
     * REST resource for direct activation of a device
     */
    private static let REST_DIRECT_ACTIVATION: String = RestApi.V2.getReqRoot()+"/activation/direct" + (Config.ALLOW_DRAFT_DEVICE_MODELS ? "" : "?createDraft=false")
    /**
     * REST resource for indirect activation of a device
     */
    private static let REST_INDIRECT_ACTIVATION: String = RestApi.V2.getReqRoot()+"/activation/indirect/device" + (Config.ALLOW_DRAFT_DEVICE_MODELS ? "" : "?createDraft=false")
    /**
     * Current OS name
     */
    private static let OS_NAME = "iOS"
    /**
     * Current OS version
     */
    private static let OS_VERSION: String = String(ProcessInfo().operatingSystemVersion.majorVersion)
            + "." + String(ProcessInfo().operatingSystemVersion.minorVersion)
    
    func getActivationPolicy(secureConnection: HttpSecureConnection,
                             deviceId: String,
                             callback: @escaping ((ActivationPolicyResponse?, ClientError?) -> Void)) throws -> Void {
        var headers: [String : String] = [ : ]
        headers["X-ActivationId"] = deviceId
        headers["Content-Type"] = "application/json"
        headers["Accept"] = "application/json"
    
        let policyRequest: ActivationPolicyRequest = createActivationPolicyRequest()!
    
        let restRsc: String = ActivationManager.REST_ACTIVATION_POLICY + policyRequest.toQuery()
    
        var status: Int = 0
        var data: Foundation.Data?
    
        try secureConnection.get(restApi: restRsc, headers: headers, response: { (httpResponse, error) -> Void in
            if (error != nil) {
                callback(nil, error)
                return
            } else {
                if let httpObject = httpResponse {
                    status = httpObject.getStatus()

                    if status == StatusCode.OK.getCode() {
                        data = httpObject.getData()
                    } else if status == StatusCode.UNAUTHORIZED.getCode() {
                        // Assume the endpoint is already activated.
                        callback(nil, ClientError.state(NSLocalizedString("Endpoint already activated, HTTP status 401", comment: "")))
                        
                        return
                    } else {
                        callback(nil, ClientError.state(NSLocalizedString("HTTP error in getActivationPolicy, status", comment: "")))
                        return
                    }
                    
                    var activationPolicyResponse: ActivationPolicyResponse?
                    do {
                        activationPolicyResponse = try ActivationPolicyResponse.fromJson(data: data)!
                        callback(activationPolicyResponse, nil)
                        return
                    } catch {
                        callback(nil, ClientError.state(NSLocalizedString("Issue while retrieving ActivationPolicyResponse from JSON", comment: "")))
                        return
                    }
                } else {
                    callback(nil, ClientError.argument(NSLocalizedString("ActivationManager getActivationPolicy() httpResponse should not be nil", comment: "")))
                    return
                }
            }
        })
    }
    
    private func createActivationPolicyRequest() -> ActivationPolicyRequest? {
        return ActivationPolicyRequest(osName: ActivationManager.OS_NAME, osVersion: ActivationManager.OS_VERSION)
    }
    
    /**
     *
     * -throws: `ClientError.security` if the request cannot be signed.
     */
    func createDirectActivationRequest(trustedAssetsManager: TrustedAssetsManager, hashAlgorithm: String , deviceModels: [String]) throws -> DirectActivationRequest {
        let subjectPublicKeyInfo: SubjectPublicKeyInfo = SubjectPublicKeyInfo()
    
        let certificationRequestInfo: CertificationRequestInfo = CertificationRequestInfo()
        certificationRequestInfo.setSubjectPublicKeyInfo(subjectPublicKeyInfo: subjectPublicKeyInfo)
        certificationRequestInfo.setSubject(subject: trustedAssetsManager.getClientId())
    
        let request: DirectActivationRequest = DirectActivationRequest()
        request.setCertificationRequestInfo(certificationRequestInfo: certificationRequestInfo)
        for dm in deviceModels {
            request.addDeviceModel(deviceModel: dm)
        }
    
        do {
            try signRequest(directActivationRequest: request,
                            trustedAssetsManager: trustedAssetsManager,
                            hashAlgorithm: hashAlgorithm)
        } catch {
            throw ClientError.security(NSLocalizedString(
                "Security issue during while signing request", comment: ""))
        }
        return request;
    }
    
    func  postDirectActivationRequest(secureConnection: HttpSecureConnection,
                deviceId: String, directActivationRequest: DirectActivationRequest,
                callback: @escaping ((DirectActivationResponse?, ClientError?) -> Void) ) throws -> Void {
        
        let payloadString: String = try directActivationRequest.toJsonString()!
        let payload: Foundation.Data = payloadString.data(using: .utf8)!
        // Send public key here...
        var headers: [String : String] = [ : ]
        headers["X-ActivationId"] = deviceId
        headers["Content-Type"] = "application/json"
        headers["Accept"] = "application/json"
        
        var status: Int = 0
        var data: Foundation.Data?

        try secureConnection.post(restApi: ActivationManager.REST_DIRECT_ACTIVATION, payload: payload, headers: headers, response: { httpResponse, error in
            if (error != nil) {
                callback(nil, error)
            } else {
                if let httpObject = httpResponse {
                    status = httpObject.getStatus()
                    
                    if status == StatusCode.OK.getCode() {
                        data = httpObject.getData()
                    } else if status == StatusCode.UNAUTHORIZED.getCode() {
                        callback(nil, ClientError.argument(NSLocalizedString("Endpoint already activated, HTTP status 401", comment: "")))
                        return
                    } else {
                        callback(nil, ClientError.argument(NSLocalizedString("HTTP error in postDirectActivationRequest , status " + status.description, comment: "")))
                        return
                    }
                    
                    var directActivationResponse: DirectActivationResponse?
                    do {
                        directActivationResponse = try DirectActivationResponse.fromJson(data: data)!
                        callback(directActivationResponse!, nil)
                        return
                    } catch {
                        callback(nil, ClientError.network(NSLocalizedString("Issue while retrieving DirectActivationResponse from JSON", comment: "")))
                        return
                    }
                } else {
                    callback(nil, ClientError.argument(NSLocalizedString("ActivationManager postDirectActivationRequest() httpResponse should not be nil", comment: "")))
                    return
                }
            }
        })
    }
    
    func  createIndirectActivationRequest(hardwareId: String,
            metadata: [String : String], deviceModels: [String],
            signature:String? = nil) -> IndirectActivationRequest? {
        return IndirectActivationRequest(hardwareId: hardwareId,
                                         metadata: metadata,
                                         deviceModels: deviceModels,
                                         signature: signature)
    }
    
    func postIndirectActivationRequest(secureConnection: HttpSecureConnection,
                                       indirectActivationRequest:IndirectActivationRequest,
                                       endpointId: String,
                                       callback: @escaping ((IndirectActivationResponse?, ClientError?) -> Void))
        throws -> Void {
            let payloadString: String = indirectActivationRequest.toJsonString()!
            let payload: Foundation.Data = payloadString.data(using: .utf8)!
            var headers: [String : String] = [ : ]
            headers["Content-Type"] = "application/json"
            headers["Accept"] = "application/json"
            
            var status: Int = 0
            var data: Foundation.Data?
            
            try secureConnection.post(restApi: ActivationManager.REST_INDIRECT_ACTIVATION, payload: payload, headers: headers, response: { httpResponse, error in
                if (error != nil) {
                    callback(nil, error)
                } else {
                    if let httpObject = httpResponse {
                        status = httpObject.getStatus()
                        if status == StatusCode.OK.getCode() {
                            data = httpObject.getData()
                        } else if status == StatusCode.UNAUTHORIZED.getCode() {
                            callback(nil, ClientError.argument(NSLocalizedString("Endpoint already activated, HTTP status 401", comment: "")))
                            return
                        } else {
                            callback(nil, ClientError.argument(NSLocalizedString("HTTP error in postDirectActivationRequest , status " + status.description, comment: "")))
                            return
                        }
                        
                        var indirectActivationResponse: IndirectActivationResponse?
                        indirectActivationResponse = IndirectActivationResponse.fromJson(data: data)!
                        callback(indirectActivationResponse, nil)
                        return
                    } else {
                        callback(nil, ClientError.network(NSLocalizedString("Issue while retrieving IndirectActivationResponse http response", comment: "")))
                        return
                    }
                }
            })
    }
    
    private func signRequest(directActivationRequest: DirectActivationRequest,
                             trustedAssetsManager: TrustedAssetsManager,
                             hashAlgorithm: String) throws  {
        let certificationRequestInfo: CertificationRequestInfo  = directActivationRequest.getCertificationRequestInfo()
        let subjectPublicKeyInfo: SubjectPublicKeyInfo  = certificationRequestInfo.getSubjectPublicKeyInfo()
        subjectPublicKeyInfo.setSecretHashAlgorithm(secretHashAlgorithm: ActivationManager.DEFAULT_MESSAGE_DIGEST_ALGORITHM)
        // RSA is the only public key algorithm currently supported by iOS and X.509 the format
        subjectPublicKeyInfo.setAlgorithm(algorithm: "RSA")
        subjectPublicKeyInfo.setFormat(format: "X.509")
        subjectPublicKeyInfo.setPublicKey(subjectPublicKey: trustedAssetsManager.getPublicKey())
        certificationRequestInfo.setSubjectPublicKeyInfo(subjectPublicKeyInfo: subjectPublicKeyInfo)
        
        let clientSecretString:String =
            try trustedAssetsManager.signWithSharedSecretString(
                data: trustedAssetsManager.getClientId(),
                algorithm: subjectPublicKeyInfo.getSecretHashAlgorithm(),
                activationId: nil)
        let clientSecretStringData = dataFromHexadecimalString(str: clientSecretString)
        let privateKey: Foundation.Data = getSignaturePayload(requestInfo: certificationRequestInfo, clientSecretData: clientSecretStringData!)
        let signature: Foundation.Data? = try trustedAssetsManager.signWithPrivateKey(data: privateKey, algorithm: hashAlgorithm)
        if signature == nil {
            throw ClientError.security(NSLocalizedString("ERROR: signature is nil", comment: ""))
        }
    
        directActivationRequest.setSignatureAlgorithm(signatureAlgorithm: hashAlgorithm)
        directActivationRequest.setSignature(signature: signature!)
    }
    
    private func dataFromHexadecimalString(str: String) -> Foundation.Data? {
        let trimmedString = str.trimmingCharacters(in: CharacterSet(charactersIn: "<> ")).replacingOccurrences(of: " ", with: "")
        
        // make sure the cleaned up string consists solely of hex digits, and 
        // that we have even number of them
        
        let regex = try! NSRegularExpression(pattern: "^[0-9a-f]*$", options: .caseInsensitive)
        
        let found = regex.firstMatch(in: trimmedString, options: [], range: NSMakeRange(0, trimmedString.count))
        if found == nil || found?.range.location == NSNotFound || trimmedString.count % 2 != 0 {
            return nil
        }
        
        // everything ok, so now let's build NSData
        var data = Foundation.Data(capacity: trimmedString.count / 2)
        
        var index = trimmedString.startIndex
        while index < trimmedString.index(before: trimmedString.endIndex) {
            let offset = trimmedString.index(index, offsetBy: 2)
            let byteString = trimmedString[index..<offset]
            let num = UInt8(byteString.withCString { strtoul($0, nil, 16) })
            data.append([num] as [UInt8], count: 1)
            index = offset
        }
        
        return data as Foundation.Data?
    }
    
    private func getSignaturePayload(requestInfo: CertificationRequestInfo, clientSecretData: Foundation.Data) -> Foundation.Data {
        let subjectPublicKeyInfo: SubjectPublicKeyInfo = requestInfo.getSubjectPublicKeyInfo()
        
        var payload: String = requestInfo.getSubject() + "\n"
        payload = payload + subjectPublicKeyInfo.getAlgorithm() + "\n"
        payload = payload + subjectPublicKeyInfo.getFormat() + "\n"
        payload = payload + subjectPublicKeyInfo.getSecretHashAlgorithm() + "\n"
        
        let map: [String : AnyObject?]? = requestInfo.getAttributes()
        if map != nil {
            for (key, val) in map! {
                payload = payload + key + "="
                if val != nil {
                    payload = payload + "\'" + (val?.description)! + "\'"
                } else {
                    payload = payload + "nil"
                }
                payload = payload + "\n"
            }
        }
        
        var payloadData: Foundation.Data = payload.data(using: .utf8,
                                                    allowLossyConversion: true)!
        
        payloadData.append(clientSecretData)
        payloadData.append(subjectPublicKeyInfo.getPublicKey())
        
        return payloadData
    }
}
