/*
 * Copyright (c) 2018, Oracle and/or its affiliates.  All rights reserved.
 *
 * This software is dual-licensed to you under the MIT License (MIT) and
 * the Universal Permissive License (UPL).  See the LICENSE file in the root
 * directory for license terms.  You may choose either license, or both.
 */

// com.oracle.iot.client.impl.device.persistence

import Foundation
import SQLite3
import SQLite3Swift
import SwiftyJSON
import os

fileprivate let TABLE_NAME:String = "BATCH_BY"

// TODO: Should be in external file
// <tablename>, <maxuuidlength>, <maxendpointidlength>
fileprivate let CREATE_TABLE_SQL:String = 
    "CREATE TABLE IF NOT EXISTS %@ (" +
        "TIMESTAMP BIGINT NOT NULL," +
        "ENDPOINT_ID VARCHAR(%d) NOT NULL," +
        "MESSAGE_ID VARCHAR(%d) NOT NULL," +
        "MESSAGE BLOB, PRIMARY KEY (MESSAGE_ID)" +
    ")"

fileprivate let ENDPOINT_ID_INDEX_SQL:String =
    "CREATE INDEX IF NOT EXISTS endpoint_id ON %@ (ENDPOINT_ID)"

// Explicitly list the desired select columns to guarantee the order
// <tablename>
fileprivate let GET_TABLE_SQL:String =
    "SELECT TIMESTAMP,ENDPOINT_ID,MESSAGE_ID,MESSAGE FROM %@ " +
    "WHERE ENDPOINT_ID = ? ORDER BY TIMESTAMP"
// The first bound parameter index
fileprivate let GET_TABLE_ENDPOINT_ID_IDX:Int32 = 1
// Result set indexes, order of columns in SELECT
fileprivate let GET_TABLE_TIMESTAMP_RESULT_IDX:Int32 = 0
fileprivate let GET_TABLE_ENDPOINT_ID_RESULT_IDX:Int32 = 1
fileprivate let GET_TABLE_MESSAGE_ID_RESULT_IDX:Int32 = 2
fileprivate let GET_TABLE_MESSAGE_RESULT_IDX:Int32 = 3

// <tablename>
fileprivate let LOAD_MESSAGE_SQL:String =
    "SELECT MESSAGE FROM %@ " +
"WHERE ENDPOINT_ID = ? ORDER BY TIMESTAMP"
// The first bound parameter index
fileprivate let LOAD_MESSAGE_ENDPOINT_ID_IDX:Int32 = 1
fileprivate let LOAD_MESSAGE_RESULT_IDX:Int32 = 0

// <tablename>
fileprivate let SAVE_SQL:String =
    "INSERT INTO %@ (TIMESTAMP,ENDPOINT_ID,MESSAGE_ID,MESSAGE) VALUES (?, ?, ?, ?)"
fileprivate let SAVE_TIMESTAMP_IDX:Int32 = 1
fileprivate let SAVE_ENDPOINT_ID_IDX:Int32 = 2
fileprivate let SAVE_MESSAGE_ID_IDX:Int32 = 3
fileprivate let SAVE_MESSAGE_IDX:Int32 = 4

// <tablename>
fileprivate let DELETE_SQL:String = "DELETE FROM %@ WHERE MESSAGE_ID IN (%@)"
fileprivate let DELETE_MESSAGE_ID_IDX:Int32 = 1

/**
 * Support for persisting messages for guaranteed delivery.
 */
class BatchByPersistence {
    
    // Default name of the database
    let DB_NAME:String = Config.BATCH_BY_PERSISTENCE_DBNAME
    let MESSAGE_ID_LENGTH:Int32 = Config.PERSISTENCE_MESSAGE_ID_LENGTH
    let ENDPOINT_ID_LENGTH:Int32 = Config.PERSISTENCE_ENDPOINT_ID_LENGTH
    private static var PERSISTENCE_ENABLED:Bool {
        get {
            return Config.BATCH_BY_PERSISTENCE_ENABLED
        }
    }

    private static var instance:BatchByPersistence? = nil
    let writeQueue:DispatchQueue =
        DispatchQueue(label:"BatchByPersistence writeQueue",
                      qos:.utility)
    
    let readQueue:DispatchQueue =
        DispatchQueue(label:"BatchByPersistence readQueue",
                      qos:.utility)
    
    private static let getInstanceQueue:DispatchQueue =
        DispatchQueue(label:"BatchByPersistence getInstanceQueue",
                      qos:.utility)
    
    public static func getInstance() -> BatchByPersistence? {
        if !BatchByPersistence.PERSISTENCE_ENABLED {
            return nil
        }
        return getInstanceQueue.sync(execute: {
            if BatchByPersistence.instance == nil {
                // TODO: Should this be fatal if persistence is enabled
                // but database cannot be created ?
                // instance is set or not in constructor
                BatchByPersistence.instance = BatchByPersistence()
            }
            return BatchByPersistence.instance
        })
    }

    private init?() {
        let schema:String = String(format: CREATE_TABLE_SQL + ";" +
                                           ENDPOINT_ID_INDEX_SQL + ";",
                TABLE_NAME, ENDPOINT_ID_LENGTH, MESSAGE_ID_LENGTH, TABLE_NAME)
        
        let rc = SQLite3_createdb(dbName:DB_NAME, schema:schema)
        if rc != SQLITE_OK {
            log(type:.fault,
                message:"%@: failed to create %@.",
                args: "BatchByPersistence():SQLite3_createdb()", DB_NAME)
            return nil
        }
    }
    
    // used in test before deleting db
    static func reset() {
        BatchByPersistence.getInstanceQueue.sync() {
            BatchByPersistence.instance = nil
        }
    }
    
    // used in testing
    static var defaultSaveCallback:(()->())? =  nil
    static func setDefaultSaveCallback(callback: @escaping ()->()) {
        BatchByPersistence.defaultSaveCallback = callback
    }

    /**
     * Insert a record into the BATCH_BY_PERSISTENCE_STORE table and associate
     * it with `endpointId`.
     *
     * - parameter messages: the message to persist
     *
     */
    public func save(messages:[Message], endpointId:String,
                     callback: (()->())? = nil) {
        if !BatchByPersistence.PERSISTENCE_ENABLED || messages.isEmpty {
            return
        }
        writeQueue.async() {
            self.readQueue.sync() {
                self._save(messages:messages, endpointId:endpointId)
                if let callback = callback {
                    callback()
                }
                if let callback = BatchByPersistence.defaultSaveCallback {
                    callback()
                }
            }
        }
    }
        
    // TODO: JCL single message
    func _save(messages:[Message], endpointId:String) {
        /*
        print("### Entering save() ###")
        defer {
            print("### Leaving save() ###")
        }
        */
        
        var db:SQLite3_connection! = nil
        do {
            db = try SQLite3_open(dbName:DB_NAME)
        } catch {
            // WARNING
            log(type:.fault,
                message: "%@: failed to open %@. %@",
                args: "load():SQLite3_open()", DB_NAME, "\(error)")
            return
        }

        // Create a block so defer isn't called on method exit
        // if open failed
        do {
            let ppStmt:UnsafeMutablePointer<OpaquePointer?> =
                UnsafeMutablePointer<OpaquePointer?>.allocate(capacity: 1)
            ppStmt.pointee = nil
            var rc:SQLite3_returnCode = SQLITE_ERROR
            
            defer {
                // Ensure that db is closed and statement is finalized
                if ppStmt.pointee != nil {
                    rc = sqlite3_finalize(ppStmt.pointee)
                    if rc != SQLITE_OK {
                        log(type:.fault,
                            message:"%@: endpointId %@, failed. (%d) %@",
                            args:"save():sqlite3_finalize()", endpointId, rc,
                                SQLite3_getErrorMsg(db:db))
                    }
                }
                rc = SQLite3_close(db:db)
                if rc != SQLITE_OK {
                    log(type:.fault,
                        message:"%@: endpointId %@, failed. (%d) %@",
                        args:"save():SQLite3_close()", endpointId, rc)
                }
            }
            
            let zSql:String = String(format:SAVE_SQL, TABLE_NAME)
            
            rc = sqlite3_prepare_v2(db, zSql, -1, ppStmt, nil)
            if rc != SQLITE_OK {
                log(type:.fault,
                    message: "%@: endpointId %@, failed. (%d) %@",
                    args: "save():sqlite3_prepare()", endpointId, rc,
                        SQLite3_getErrorMsg(db:db))
                return
            }

            for message in messages {
                
                let messageId:String = message.getClientId()
                let timestamp:Int64? = message.getEventTime() ??
                        Int64(Date().timeIntervalSince1970 * 1000.0)
                rc = sqlite3_bind_int64(ppStmt.pointee, SAVE_TIMESTAMP_IDX, timestamp!)
                if rc != SQLITE_OK {
                    log(type:.fault,
                        message: "%@: endpointId %@, messageId %@, failed to bind TIMESTAMP. (%d) %@",
                        args:"save():sqlite3_bind_int64()", endpointId, messageId, rc,
                            SQLite3_getErrorMsg(db:db))
                    continue
                }

                guard let ENDPOINT_ID:SQLite3_stringBinding =
                        SQLite3_getStringBinding(value:endpointId) else {
                    log(type:.fault,message:"%@: endpointId %@, messageId %@, failed to get binding for endpointId %@.",
                        args:"save():SQLite3_getStringBinding()", endpointId, messageId,
                            endpointId)
                    continue
                }
                rc = sqlite3_bind_text(ppStmt.pointee, SAVE_ENDPOINT_ID_IDX, ENDPOINT_ID, Int32(ENDPOINT_ID.count), nil)
                if rc != SQLITE_OK {
                    log(type:.fault,
                        message: "%@: endpointId %@, messageId %@, failed to bind ENDPOINT_ID. (%d) %@",
                        args:"save():sqlite3_bind_text()", endpointId, messageId, rc,
                            SQLite3_getErrorMsg(db:db))
                    continue
                }
                
                guard let MESSAGE_ID:SQLite3_stringBinding =
                        SQLite3_getStringBinding(value:messageId) else {
                    log(type:.fault,
                        message:"%@: endpointId %@, messageId %@, failed to get binding for messageId %@.",
                        args:"save():SQLite3_getStringBinding()", endpointId, messageId,
                            messageId)
                    continue
                }
                rc = sqlite3_bind_text(ppStmt.pointee, SAVE_MESSAGE_ID_IDX, MESSAGE_ID,
                                       Int32(MESSAGE_ID.count), nil)
                if rc != SQLITE_OK {
                    log(type:.fault,
                        message: "%@: endpointId %@, messageId %@, failed to bind MESSAGE_ID. (%d) %@",
                        args:"save():sqlite3_bind_text()", endpointId, messageId, rc,
                        SQLite3_getErrorMsg(db:db))
                    continue
                }
                
                
                var messageJson:String!
                do {
                    messageJson = try message.toJsonString()
                } catch {
                    log(type:.fault,
                        message: "%@: endpointId %@, messageId %@, threw exception. %@",
                        args:"save():toJsonString()", endpointId, messageId, "\(error)")
                    continue
                }
                guard let message:Foundation.Data = messageJson.data(using:.utf8),
                        let MESSAGE:SQLite3_blobBinding =
                            SQLite3_getBlobBinding(value:message) else {
                    log(type:.fault,
                        message:"%@: endpointId %@, messageId %@, failed to get binding for message.",
                        args:"save():SQLite3_getBlobBinding()", endpointId, messageId)
                    continue
                }
                rc = sqlite3_bind_blob(ppStmt.pointee, SAVE_MESSAGE_IDX,
                                       MESSAGE, Int32(MESSAGE.count), nil)
                if rc != SQLITE_OK {
                    log(type:.fault,
                        message: "%@: endpointId %@, messageId %@, failed to bind MESSAGE. (%d) %@",
                        args:"save():sqlite3_bind_blob()", endpointId, messageId, rc,
                            SQLite3_getErrorMsg(db:db))
                        continue
                }
                
                //executing the query to insert values
                rc = sqlite3_step(ppStmt.pointee)
                if rc != SQLITE_DONE {
                    log(type:.fault,
                        message:"%@: endpointId %@, messageId %@, failed. (%d) %@",
                        args: "save():sqlite3_step()", endpointId, messageId, rc,
                            SQLite3_getErrorMsg(db:db))
                    // TODO: Should we just retry and try again ?
                    //continue
                }
                
                rc = sqlite3_reset(ppStmt.pointee)
                if rc != SQLITE_OK {
                    log(type:.fault,
                        message:"%@: endpointId %@, failed after messageId %@. (%d) %@",
                        args:"save():sqlite3_reset()", endpointId, messageId, rc,
                        SQLite3_getErrorMsg(db:db))
                    // Try and finalize the prepared statement and call
                    // prepare again
                    rc = sqlite3_finalize(ppStmt.pointee)
                    if rc != SQLITE_OK {
                        log(type:.fault,
                            message:"%@: endpointId %@, failed after messageId %@ exiting save(), " +
                            "all messages may not have been saved. (%d) %@",
                            args:"save():sqlite3_finalize()", endpointId, messageId, rc,
                            SQLite3_getErrorMsg(db:db))
                        break
                    }

                    rc = sqlite3_prepare_v2(db, zSql, -1, ppStmt, nil)
                    if rc != SQLITE_OK {
                        log(type:.fault,
                            message:"%@: endpointId %@, failed after a failed reset, " +
                            "exiting save() all messages may not have been saved. (%d) %@",
                            args: "save():sqlite3_prepare()", endpointId, rc,
                            SQLite3_getErrorMsg(db:db))
                        ppStmt.pointee = nil
                        break
                    }
                }
                rc = sqlite3_clear_bindings(ppStmt.pointee)
                if rc != SQLITE_OK {
                    log(type:.fault,
                        message:"%@: endpointId %@, messageId %@, failed. (%d) %@",
                        args: "save():sqlite3_clear_bindings()", endpointId,
                            messageId, rc, SQLite3_getErrorMsg(db:db))
                    break
                }
            }
        }
    }
    
    /**
     * Delete the record associated with `message`. The `clientId` is used
     * as the query to select the record to delete. The `clientId` must
     * match the `MESSAGE_ID` column of the BATCH_BY_PERSISTENCE_STORE table.
     *
     * - parameter messages: the messages to delete
     */
    public func delete(messages:[Message], callback: (()->())? = nil) {
        if !BatchByPersistence.PERSISTENCE_ENABLED || messages.isEmpty {
            return
        }
        // Serialize deletes
        writeQueue.async() {
            self._delete(messages:messages)
            if let callback = callback {
                callback()
            }
        }
    }
        
    func _delete(messages:[Message]) {
        /*
        print("### Entering delete() ###")
        defer {
            print("### Leaving delete() ###")
        }
        */
        if messages.count == 0 {
            return
        }
        var db:SQLite3_connection! = nil
        do {
            db = try SQLite3_open(dbName:DB_NAME)
        } catch {
            // WARNING
            log(type:.fault,
                message: "%@: failed to open %@. %@",
                args: "load():SQLite3_open()", DB_NAME, "\(error)")
            return
        }

        var rc:SQLite3_returnCode = SQLITE_ERROR
        
        // Create a block so defer isn't called on method exit
        // if open failed
        do {
            defer {
                rc = SQLite3_close(db:db)
                if rc != SQLITE_OK {
                    log(type:.fault,
                        message:"%@: failed. (%d) %@",
                        args:"save():SQLite3_close()", rc)
                }
            }
            // Create the delete list
            // We know we have at least one message
            var inArgs:String = "'\(messages[0].getClientId())'"
            
            for i in 1..<messages.count {
                let message:Message = messages[i]
                inArgs.append(",'\(message.getClientId())'")
            }
            let zSql:String = String(format:DELETE_SQL, TABLE_NAME, inArgs)

            rc = sqlite3_exec(db, zSql, nil, nil, nil)
            if rc != SQLITE_OK {
                log(type:.fault,
                    message:"%@: exec delete failed. (%d) %@",
                    args: "delete():sqlite3_exec()", rc,
                    SQLite3_getErrorMsg(db:db))
                return
            }
        }
    }
    
    /**
     * Perform a query based on the `endpointId`. The `endpointId` must
     * match the `ENDPOINT_ID` column of the MESSAGE_PERSISTENCE_STORE table.
     *
     * - parameter endpointId: The query parameter associated with the message
     *                         record
     *
     * - returns: an array of `Message` or nil if an error occurs performing
     *            the query
     */
    public func load(endpointId:String) -> [Message]? {
        if !BatchByPersistence.PERSISTENCE_ENABLED {
            return nil
        }
        return readQueue.sync(execute: {
            return _load(endpointId:endpointId)
        })
    }
    
    func _load(endpointId:String) -> [Message]? {

        /*
        print("### Entering load() ###")
        defer {
            print("### Leaving laod() ###")
        }
        */
        var db:SQLite3_connection! = nil
        do {
            db = try SQLite3_open(dbName:DB_NAME)
        } catch {
            // WARNING
            log(type:.fault,
                message: "%@: failed to open %@. %@",
                args: "load():SQLite3_open()", DB_NAME, "\(error)")
            return nil
        }
        
        let ppStmt:UnsafeMutablePointer<OpaquePointer?> =
            UnsafeMutablePointer<OpaquePointer?>.allocate(capacity: 1)
        ppStmt.pointee = nil
        var rc:SQLite3_returnCode = SQLITE_ERROR
        
        // Create a block so defer isn't called on method exit
        // if open failed
        do {
            
            defer {
                // Ensure that db is closed and statement is finalized
                if ppStmt.pointee != nil {
                    rc = sqlite3_finalize(ppStmt.pointee)
                    if rc != SQLITE_OK {
                        log(type:.fault,
                            message:"%@: failed. (%d) %@",
                            args:"load():sqlite3_finalize()", rc,
                                 SQLite3_getErrorMsg(db:db))
                    }
                }
                rc = SQLite3_close(db:db)
                if rc != SQLITE_OK {
                    log(type:.fault,
                        message:"%@: failed. (%d) %@",
                        args:"load():SQLite3_close()", rc)
                }
            }
            
            let zSql:String = String(format:LOAD_MESSAGE_SQL, TABLE_NAME)
            
            rc = sqlite3_prepare_v2(db, zSql, -1, ppStmt, nil)
            if rc != SQLITE_OK {
                log(type:.fault,
                    message:"%@: failed:(%d) %@",
                    args: "load()", rc, SQLite3_getErrorMsg(db:db))
                return nil
            }
            guard let ENDPOINT_ID:SQLite3_stringBinding =
                    SQLite3_getStringBinding(value:endpointId) else {
                log(type:.fault,
                    message:"%@: failed for %@.",
                    args: "load():SQLite3_getStringBinding()",
                          endpointId)
                return nil
            }
            rc = sqlite3_bind_text(ppStmt.pointee, LOAD_MESSAGE_ENDPOINT_ID_IDX,
                                   ENDPOINT_ID,
                                   Int32(ENDPOINT_ID.count), nil)
            if rc != SQLITE_OK {
                log(type:.fault,
                    message:"%@: failed for %@. (%d) %@",
                    args:"load():sqlite3_bind_text()", endpointId, rc,
                        SQLite3_getErrorMsg(db:db))
                return nil
            }
            
            var messages:[Message] = [Message]()
            var rows:Int = 0
            var haveRows:Bool = true
            while haveRows {
                var messageJson:Foundation.Data? = nil
                rc = sqlite3_step(ppStmt.pointee)
                if rc == SQLITE_ROW  {
                    // Only MESSAGE is needed
                    
                    let MESSAGE:UnsafeRawPointer? =
                        sqlite3_column_blob(ppStmt.pointee, LOAD_MESSAGE_RESULT_IDX)
                    if MESSAGE == nil {
                        log(type:.fault,
                            message:"%@: endpointId %@, row %d, MESSAGE is zero length blob",
                            args:"load():sqlite3_column_blob()", endpointId, rows)
                    } else {
                        let length = Int(sqlite3_column_bytes(ppStmt.pointee,
                                                              LOAD_MESSAGE_RESULT_IDX))
                        let bufptr = UnsafeBufferPointer(start:
                            MESSAGE!.assumingMemoryBound(to: UInt8.self), count: length)
                        messageJson = Foundation.Data(buffer:bufptr)
                    }
                } else if rc != SQLITE_DONE {
                    log(type:.fault,
                        message:"%@: failed on row %d. (%d) %@",
                        args:["load():sqlite3_step()", rows, rc,
                            SQLite3_getErrorMsg(db:db)])
                }
                haveRows = rc == SQLITE_ROW
                rows += 1
                if let messageJson = messageJson {
                    do {
                        // TODO: Is there ever more than one message in the blob ?
                        // Why return array ?
                        // TODO: THe calls in Message and MessageBuilder are messed up.
                        // duplicate methods that that make no sense. Also they are
                        // not consistent with respect to JCL, in that Message has
                        // some methods and MessageBuilder has others.
                        let message:Message = try getMessage(jsonData:messageJson)
                        messages.append(message)
                    } catch {
                        log(type:.fault,
                            message:"%@: endpointId %@, row %d, threw exception. %@",
                            args:"load():Message.fromJSON()", endpointId, rows,
                            "\(error)")
                    }
                }
            }
            return messages
        }
    }
}
