/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import os

/**
 * Data is an object that represents a set of custom data fields (key/value
 * pairs) defined in the formats section of the device model. The custom data
 * is sent by calling the `submit` method.
 *
 * The `set` method returns the `Data` instance to allow
 * the data fields to be set in fluent style.
 *
 */
public protocol Data {
    
    /**
     * Set the value of a field in the `Data`. The fields are
     * determined by the format given when the Data is created.
     * The value is validated according to the constraints in the format.
     * If the value is not valid, a `ClientError.argument` exception is thrown.
     *
     * All fields defined in the format that are not "optional"
     * must be set before calling `submit()`
     *
     * - parameter field: name of a field from the custom data format
     *
     * - parameter value: the value to set
     *
     * - returns: `Data` object that can be used for function chaining
     *
     * - throws: `ClientError.argument` if `field` is `nil` or `value` is not
     *           valid
     *
     */
    func set(field: String, value: AnyObject) throws -> Data
    
    /**
     * This method is used to actually send the custom data fields to the 
     * server. All mandatory fields (according to the device model definition) 
     * must be set before sending, otherwise an error will be thrown. Any error 
     * that can arise while sending will be handled by the 
     * `VirtualDevice.onError` handler, if set.
     *
     * After a successful send all the values are reset so to send again the
     * values must be set again.
     *
     * - throws: 
     *      - `ClientError.argument` if a required field is not set
     *      - `ClientError.network` if a `MessageDispatcher` cannot be
     *           obtained.
     */
    func submit() throws -> Void
    
    /**
     * Set a callback that is invoked if an error occurs when submitting the
     * `Data`. The callback may be `nil`, which will un-set
     * the callback.
     * - parameter callback: a callback to invoke, or `nil`
     */
    func setOnError(callback: ((ErrorEvent)->())?)
}
    
class DataImpl:NSObject,Data {
    private var  virtualDevice: VirtualDeviceImpl?
    private var  deviceModelFormat: DeviceModelFormat?
    private var fieldValues: [String : AnyObject]?
    private var errorCallback:((ErrorEvent)->())?
    
    /**
     * Creates instance of a `Data` implementation.
     *
     * - parameter virtualDevice: the virtual device for which the `Data` is 
     *                            created
     *
     * - parameter formatUrn: the URN for the data
     *
     * - throws: `ClientError.argument` if a device model format cannot be 
     *           obtained from `formatUrn` or if the format is not
     *           `DeviceModelFormatType.DATA`
     *
     */
    init(virtualDevice: VirtualDeviceImpl, formatUrn: String) throws {
        self.deviceModelFormat =
            DataImpl.getDeviceModelFormat(virtualDevice: virtualDevice,
                                          formatUrn: formatUrn)
        
        if (self.deviceModelFormat == nil) {
            throw ClientError.argument(String(format: NSLocalizedString(
                "%@ not found", comment: ""), formatUrn))
        }
        
        if (self.deviceModelFormat!.getType() != .DATA) {
            throw ClientError.argument(String(format: NSLocalizedString(
                "%@ is not a data format", comment: ""), formatUrn))
        }
        
        self.virtualDevice = virtualDevice
        self.fieldValues = [ : ]
    }
    
    /**
     * - throws: `ClientError.argument` if `field` is not a valid field name
     *           or `value` is not the field type.
     */
    func set(field: String, value: AnyObject) throws -> Data {
        let dmField: DeviceModelFormatField? = getField(fieldName: field)
        if (dmField == nil) {
            throw ClientError.argument(String(format: NSLocalizedString(
                "%@ is not in model", comment: ""), field))
        }
        
        switch (dmField!.getType()) {
        case .NUMBER:
            if (!(value is NSNumber)) {
                throw ClientError.argument(String(format: NSLocalizedString(
                    "%@ is not a number", comment: ""), field))
            }
            break
        case .INTEGER:
            if (!(value is Int)) {
                throw ClientError.argument(String(format: NSLocalizedString(
                    "%@ is not an Int", comment: ""), field))
            }
            break
        case .DATETIME:
            if (!(value is Int64)) {
                throw ClientError.argument(String(format: NSLocalizedString(
                    "%@ is not a DateTime", comment: ""), field))
            }
            break
        case .BOOLEAN:
            if (!(value is Bool)) {
                throw ClientError.argument(String(format: NSLocalizedString(
                    "%@ is not a Bool", comment: ""), field))
            }
            break
        case .STRING:
            if (!(value is String)) {
                throw ClientError.argument(String(format: NSLocalizedString(
                    "%@ is not a String", comment: ""), field))
            }
            break
        case .URI:
            if (!(value is ExternalObject)) {
                throw ClientError.argument(String(format: NSLocalizedString(
                    "value for '%@' is not an ExternalObject", comment: ""),
                                                  field))
            }
            break
        }
        
        fieldValues![field] = value
        
        return self
    }
    
    /**
     * - throws: 
     *      - `ClientError.argument` if a required field is not set
     *      - `ClientError.network` if a `MessageDispatcher` cannot be
     *        obtained.
     */
    func submit() throws -> Void {
        let builder: DataMessageBuilder = DataMessageBuilder()
        let fieldList: [DeviceModelFormatField] = deviceModelFormat!.getFields()
        var storageObjects = [StorageObjectImpl]()
        for field in fieldList {
            let key: String = field.getName()
            let value: AnyObject? = fieldValues![key]
            fieldValues![key] = nil
            if (value != nil) {
                switch (field.getType()) {
                case .NUMBER:
                    try builder.addDataItem(key: key, value: value as! Double)
                    break
                case .INTEGER:
                    try builder.addDataItem(key: key, value: value as! Int)
                    break
                case .DATETIME:
                    if value is Date {
                        try builder.addDataItem(key: key, value: (value as! Date).timeIntervalSince1970)
                    } else {
                        try builder.addDataItem(key: key, value: Double(truncating:value as! NSNumber))
                    }
                    break
                case .BOOLEAN:
                    try builder.addDataItem(key: key, value: value as! Bool)
                    break
                case .STRING:
                    try builder.addDataItem(key: key, value: value as! String)
                    break
                case .URI:
                    if value is StorageObjectImpl {
                        let so = value as! StorageObjectImpl
                        if ((so.getSyncStatus() == SyncStatus.notInSync) ||
                            (so.getSyncStatus() == SyncStatus.syncPending)) {
                            storageObjects.append(so);
                        }
                        so.setSyncEventInfo(virtualDevice: virtualDevice!,
                                            nameForSyncEvent: key);
                        try so.sync();
                    }
                    let uri = (value as! ExternalObject).getURI()
                    try builder.addDataItem(key: key, value: uri)
                    break
                }
                
            } else if (!field.isOptional()) {
                fieldValues = [ : ]
                throw ClientError.argument(String(format: NSLocalizedString(
                    "Non-optional field %@ is not set", comment: ""), key))
            }
        }
        
        fieldValues = [ : ]
        
        builder.setFormat(format: deviceModelFormat!.getURN())
        builder.setSource(source: virtualDevice!.getEndpointId())
        
        let message: DataMessage = try builder.build()
        let messageDispatcher: MessageDispatcherImpl? =
            MessageDispatcher.getMessageDispatcher(msgDirectlyConnectedDevice: virtualDevice!.msgDirectlyConnectedDevice!) as? MessageDispatcherImpl
        
        do {
            for so:StorageObjectImpl in storageObjects {
                messageDispatcher?.addStorageObjectDependency(storageObject:so,
                                            clientMsgId: message.getClientId())
            }
            // MessageDispatcherImpl ends up calling VirtualDeviceImpl.ErrorCallbackBridge
            // if the message fails to be sent. Here, we tell VirtualDeviceImpl about
            // the callback for this message so that the ErrorCallbackBridge
            // can invoke it.
            if let cb = self.errorCallback {
                VirtualDeviceImpl.addOnErrorCallback(message:self as AnyHashable,
                                                     messageId:message.getClientId(),
                                                     callback: cb)
            }
            try messageDispatcher?.queue(message: message)
        } catch {
            // SEVERE
            log(type:OSLogType.error,
                message:"@: error: %@",
                args: "DataImpl.submit()", "\(error)")
        }
    }
    
    func setOnError(callback: ((ErrorEvent)->())?) {
        self.errorCallback = callback
        VirtualDeviceImpl.replaceOnErrorCallback(message:self as AnyHashable,
                                                 callback:self.errorCallback)
    }
    
    private func getField(fieldName: String) -> DeviceModelFormatField? {
        if (deviceModelFormat == nil) {
            return nil
        }
        
        let fieldList: [DeviceModelFormatField] = deviceModelFormat!.getFields()
        for field in fieldList {
            if field.getName() == fieldName {
                return field
            }
        }
        return nil
    }
    
    private static func getDeviceModelFormat(virtualDevice: VirtualDeviceImpl,
                                             formatUrn: String) -> DeviceModelFormat? {
        
        let deviceModel: DeviceModel = virtualDevice.getDeviceModel()
        return (deviceModel as! DeviceModelImpl).getDeviceModelFormats()[formatUrn]
    }

}
