/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation

public protocol DataItemProtocol {
    func getValue() -> Self
}

extension Int: DataItemProtocol {
    public func getValue() -> Int {
        return self
    }
}

/**
 * Implementation of protocol Equatable
 *
 * - parameter lhs: one item to be compared
 *
 * - parameter rhs: the other item to be compared
 *
 * - returns: *true* if equal, *false* otherwise
 */
public func ==(lhs: DataItemType, rhs: DataItemType) -> Bool {
    return lhs.rawValue == rhs.rawValue
}

/**
 * The data type of the data item.
 */
public enum DataItemType: String {
    case STRING
    case DOUBLE
    case BOOLEAN
    
}

/**
 * Implementation of protocol Equatable
 *
 * - parameter lhs: one item to be compared
 *
 * - parameter rhs: the other item to be compared
 *
 * - returns: *true* if equal, *false* otherwise
 */
public func ==(lhs: DataItem, rhs: DataItem) -> Bool {
    if lhs === rhs {
        return true
    }
    if lhs.key != rhs.key {
       return false
    }
    if lhs.type != rhs.type {
        return false
    }
    switch lhs.type {
    case .STRING:
        return (lhs.value as! String) == (rhs.value as! String)
    case .BOOLEAN:
        return (lhs.value as! Bool) == (rhs.value as! Bool)
    case .DOUBLE:
        return Double(truncating: (lhs.value as! NSNumber)) == Double(truncating: (rhs.value as! NSNumber))
    }
}

/**
 * `DataItem` is a key/value/type triplet for data in `DataMessage` along
 * with the type of the data. `Key` cannot be `nil` or empty, it is mandatory
 * field. String values cannot be `nil`, other values are primitive data types.
 * `Value` is mandatory field. `Type` can be `STRING`, `BOOLEAN` or `DOUBLE`, 
 * see `DataItemType`; it is assigned automatically by constructor.
 */
public class DataItem: Hashable {
    
    /**
     * Implementation of protocol Hashable
     *
     * - returns: the hash value
     */
    public var hashValue : Int {
        var result: Int = key.hashValue
        switch type {
        case .STRING:
            result = 31 &* result &+ (value as! String).hashValue
            break
        case .BOOLEAN:
            result = 31 &* result &+ (value as! Bool).hashValue
            break
        case .DOUBLE:
            result = 31 &* result &+ Double(truncating: value as! NSNumber).hashValue
            break
        }
        return result
    }
    
    /** Data item key */
    fileprivate let key: String
    
    /** Data item value */
    fileprivate let value: AnyObject
    
    /** Type of the value */
    fileprivate let type: DataItemType
    
    /**
     * Constructor that takes `String key` and `Double value`.
     * This is to reinforce the value type.
     *
     * - parameter key: data item key
     *
     * - parameter value: data item double value
     *
     * - throws: `ClientError.argument` if `key` is `nil` or is `Double.NaN` or
     *           `Double.infinite`
     */
    public convenience init(key: String, value: Double) throws {
        if (value.isFinite) {
            try self.init(key: key, value: value as AnyObject, type: .DOUBLE)
        } else {
            throw ClientError.argument(NSLocalizedString(
                "Invalid argument value: Double.NaN or Double.infinity",
                comment: ""))
        }
    }
    
    /**
     * Constructor that takes `String key` and `Int value`.
     * This is to reinforce the value type.
     *
     * - parameter key: data item key
     *
     * - parameter value: data item Int value
     *
     * - throws: `ClientError.argument` if `key` is `nil` or is the empty string
     */
    public convenience init(key: String, value: Int) throws {
        try self.init(key: key, value: value as AnyObject, type: .DOUBLE)
    }
    
    /**
     * Constructor that takes `String key` and `Bool value`.
     * This is to reinforce the value type.
     *
     * - parameter key: data item key
     *
     * - parameter value: data item boolean value
     *
     * - throws: `ClientError.argument` if `key` is `nil` or is the empty string
     */
    public convenience init(key: String, value: Bool) throws {
        try self.init(key: key, value: value as AnyObject, type: .BOOLEAN)
    }
    
    /**
     * Constructor that takes `String key` and `String value`.
     * This is to reinforce the value type
     *
     * - parameter key:  data item key
     *
     * - parameter value:  data item String value
     *
     * - throws: `ClientError.argument` if `value` or `key` is `nil` or is the
     *            empty string
     */
    public convenience init(key: String, value: String?) throws {
        // TODO:  why is value here optional if this will throw an exception 
        // anyway
        guard let value = value, !value.isEmpty else {
            throw ClientError.argument(NSLocalizedString(
                "Data item of type String has illegal value", comment: ""))
        }
        try self.init(key: key, value: value as AnyObject,
                      type: DataItemType.STRING)
    }
    
    // TODO: Why have optional key if exception is thrown if it is nil ?
    
    /**
     * Private constructor for data item.
     *
     * - parameter key: data item key
     *
     * - parameter value: data item value
     *
     * - parameter type: type of the data
     *
     * - throws: `ClientError.argument` if `key` is `nil` or is the empty string
     */
    private init(key: String?, value: AnyObject, type: DataItemType) throws {
        // TODO: description says private constructor but this is public
        guard let key = key, !key.isEmpty, key.count <= Config.MAX_KEY_LENGTH else {
            throw ClientError.argument(NSLocalizedString(
                "Key of Data item illegal value", comment: ""))
        }
        self.key = key
        self.type = type
        self.value = value
    }
    
    /**
     * Get data item key.
     *
     * - returns: key of this *DataItem*, never `nil`.
     */
    public func getKey() -> String {
        return self.key
    }
    
    /**
     * Get data item value.
     *
     * - returns: value of this *DataItem*, never `nil`.
     */
    public func getValue() -> AnyObject {
        return self.value
    }
    
    /**
     * Get data item type.
     *
     * - returns: type of this *DataItem*, never `nil`.
     */
    public func getType() -> DataItemType {
        return self.type
    }
    

    public func toString() -> String {
        let val:String = type == DataItemType.STRING ?
            "\"\(value)\"" : "\(value)"
        return "{\"" + key + "\"," + val + "}";
    }
    
}
