/*
 * Copyright (c) 2018, Oracle and/or its affiliates.  All rights reserved.
 *
 * This software is dual-licensed to you under the MIT License (MIT) and
 * the Universal Permissive License (UPL).  See the LICENSE file in the root
 * directory for license terms.  You may choose either license, or both.
 */

// com.oracle.iot.client.impl.device

import Foundation
import os

/**
 * Device
 */
class DeviceAnalogImpl :  DeviceAnalog {

    private let directlyConnectedDevice:AdvancedDirectlyConnectedDevice!
    private let deviceModel:DeviceModelImpl!
    private var attributeValueMap:Dictionary<String,AnyObject>!
    private var endpointId:String!

    init(directlyConnectedDevice:AdvancedDirectlyConnectedDevice,
                deviceModel:DeviceModelImpl,
                endpointId:String) {
        self.directlyConnectedDevice = directlyConnectedDevice
        self.deviceModel = deviceModel
        self.attributeValueMap = Dictionary<String,AnyObject>()
        self.endpointId = endpointId
    }

    public func getEndpointId() -> String {
        return endpointId
    }

    public func getDeviceModel() -> DeviceModel {
        return deviceModel
    }

    public func setAttributeValue(attribute:String, value:AnyObject) throws {

        guard let deviceModelAttribute:DeviceModelAttribute =
            deviceModel.getDeviceModelAttributes()[attribute] else {
            throw ClientError.argument(String(format:NSLocalizedString(
                "%@ does not contain attribute %@", comment:""),
                    deviceModel.getURN(), attribute))
        }

        let type:DeviceModelAttributeType = deviceModelAttribute.getType()
        var badValue:Bool = false
        switch (type) {
            case .NUMBER:
                badValue = !(value is NSNumber)
                break
            case .STRING, .URI:
                badValue = !(value is String)
                break
            case .BOOLEAN:
                badValue = !(value is Bool)
                break
            case .INTEGER:
                badValue = !(value is Int64)
                break
            case .DATETIME:
                badValue = !(value is Double)
                break
        }

        if badValue {
            throw ClientError.argument(String(format:NSLocalizedString(
                "cannot set '%@:attribute/%@' to %@", comment:""),
                deviceModel.getURN(), attribute, "\(value)"))
        }

        attributeValueMap[attribute] = value
    }

    public func getAttributeValue(attribute:String) throws -> AnyObject? {

        guard let deviceModelAttribute:DeviceModelAttribute =
                deviceModel.getDeviceModelAttributes()[attribute] else {
            throw ClientError.argument(String(format:NSLocalizedString(
                "%@ does not contain attribute %@", comment:""),
                deviceModel.getURN(), attribute))
        }

        if let value:AnyObject = attributeValueMap[attribute] {
            return value
        } else {
            return deviceModelAttribute.getDefaultValue()
        }
    }
    
    // TODO: call may need a callback it calls "send" which requires a
    // callback. But there are other methods that call "call" that must
    // return a value. For now just call send, which returns immeditately
    // and assume it does not affect the rest of the call implementation
    public func call(actionName:String, args:[Any?]?
        /*,callback: @escaping (ClientError?)->()*/) {
        
        let deviceModelActionMap:Dictionary<String,DeviceModelAction> =
            deviceModel.getDeviceModelActions()
        guard let deviceModelAction:DeviceModelAction =
                deviceModelActionMap[actionName] else {
            // WARNING
            log(type:OSLogType.fault,
                message:"%@ does not contain action '%@'",
                args: deviceModel.getURN(), actionName)
            /*
            callback(ClientError.argument(String(format:
                     NSLocalizedString("%@ does not contain action '%@'",
                                       comment:""),
                            deviceModel.getURN(), actionName)))
            */
            return
        }
        
        let argType:DeviceModelAttributeType? = deviceModelAction.getArgType()
        
        // TODO: currently, call only supports one arg
        let arg:Any? = args != nil && !args!.isEmpty ? args![0] : nil
        
        // What this is doing is pushing a request message into the message
        // queue for the requested action. To the LL, it is handled just like
        // any other RequestMessage. But we don't want this RM going to the
        // server, so the source and destination are set to be the same
        // endpoint id. In SendReceiveImpl, if there is a RequestMessage
        // with a source == destination, it is treated it specially.
        let requestMessageBuilder:RequestMessageBuilder =
            RequestMessageBuilder()
        requestMessageBuilder.setSource(source:getEndpointId())
        requestMessageBuilder.setDestination(destination:getEndpointId())
        requestMessageBuilder.setUrl(url:"deviceModels/" +
            getDeviceModel().getURN() + "/actions/" + actionName)
        requestMessageBuilder.setMethod(method:"POST")
        
        // check arg for correct type
        if argType != nil {
            if arg == nil {
                // WARNING
                log(type:OSLogType.fault,
                    message:"%@ action '%@' requires an argument",
                    args: getDeviceModel().getURN(), actionName)
                /*
                callback(ClientError.argument(String(format:NSLocalizedString(
                    "%@ action '%@' requires an argument", comment:""),
                    getDeviceModel().getURN(), actionName)))
                */
                return
            }
            
            var goodArg:Bool = false
            switch (argType!) {
            case .INTEGER,.NUMBER:
                goodArg = arg is NSNumber
                if goodArg {
                    let number:NSNumber = arg as! NSNumber
                    var value:String!
                    if argType == DeviceModelAttributeType.INTEGER {
                        let roundedValue:Int64 = Int64(round(Double(truncating: number)))
                        value = "\(roundedValue)"
                    } else {
                        value = "\(number)"
                    }
                    requestMessageBuilder.setBody(body:
                        "{\"value\":\(value)}".data(using:.utf8)!)
                    
                    // Assumption here is that lowerBound <= upperBound
                    let val:NSNumber = arg as! NSNumber
                    if let upper:NSNumber = deviceModelAction.getUpperBound() {
                        if Double(truncating: val) > Double(truncating: upper) {
                            // WARNING
                            log(type:OSLogType.fault,
                                message:"%@ action '%@' arg out of range: %f > %f",
                                args: getDeviceModel().getURN(),
                                    actionName, val, upper)
                            // even though the arg is out of range, pass it to the action
                            // TODO is this the right thing to do?
                        }
                    }
                    if let lower = deviceModelAction.getLowerBound() {
                        if Double(truncating: val) < Double(truncating: lower) {
                            // WARNING
                            log(type:OSLogType.fault,
                                message:"%@ action '%@' arg out of range: %f < %f",
                                args: getDeviceModel().getURN(),
                                actionName, val, lower)
                            // even though the arg is out of range, pass it to the action
                            // TODO is this the right thing to do?
                        }
                    }
                }
            break
            case .DATETIME:
                goodArg = arg is Date || arg is NSNumber
                if goodArg {
                    var value:String!
                    if arg is Date {
                        let date:Date = arg as! Date
                        value = date.description
                    } else {
                        let lval:Double = Double(truncating: arg as! NSNumber)
                        value = "\(lval)"
                    }
                    requestMessageBuilder.setBody(body:
                        "{\"value\":\(value)}".data(using:.utf8)!)
                }
            break
            case.BOOLEAN:
                goodArg = arg is Bool
                if goodArg {
                    let value:Bool = arg as! Bool
                    requestMessageBuilder.setBody(body:
                        "{\"value\":\(value)}".data(using:.utf8)!)
                }
            break
            case .STRING,.URI:
                goodArg = arg is String
                if goodArg {
                    let value:String = arg as! String
                    requestMessageBuilder.setBody(body:
                        "{\"value\":\(value)}".data(using:.utf8)!)
                }
            break
            }
            
            if !goodArg {
                // WARNING
                log(type:OSLogType.fault,
                    message:"%@ action '%@': wrong argument type. " +
                        "Expected %@ found %@",
                    args: getDeviceModel().getURN(), actionName,
                        "\(argType!)", "\(arg!)")
                /*
                callback(ClientError.argument(String(format:NSLocalizedString(
                    "%@ action '%@': wrong argument type. Expected %@ found %@",
                    comment: ""), getDeviceModel().getURN(), actionName,
                                  "\(argType!)", "\(arg!)")))
                */
                return
            }
        }
        do {
            let requestMessage:RequestMessage = try requestMessageBuilder.build()
            let useLongPolling:Bool = !Config.DISABLE_LONG_POLLING
        
            // Assumption here is that, if you are using long polling, you are using message dispatcher.
            // This could be a bad assumption. But if long polling is disabled, putting the message on
            // the request buffer will work regardless of whether message dispatcher is used.
        
            if useLongPolling {
                let requestDispatcher:RequestDispatcher = RequestDispatcher.getInstance()
                let responseMessage:ResponseMessage = try requestDispatcher.dispatch(requestMessage: requestMessage)
                if (responseMessage.getStatusCode() == StatusCode.NOT_FOUND) {
                    // INFO
                    log(type:.debug,
                        message: "DeviceAnalogImpl:call(): Endpoint %@ has no handler for %@",
                        args: getEndpointId(), requestMessage.getURL())
                }
            } else {
                // TODO: When call becomes async, uss callback here
                // Not long polling, push request message back on request buffer
                try directlyConnectedDevice.send(messages: requestMessage,
                                                 callback: { error in
                    if error != nil {
                        log(type:.fault,
                            message:"DeviceAnalogImpl:call(): send failed with %@",
                            args: "\(error!)")
                    }
                })
            }
        } catch {
            // WARNING
            log(type:.fault,
                message: "DeviceAnalogImpl:call(): %@",
                args: "\(error)")
        }
    }

    // DeviceAnalog API
    // TODO: Need callback
    public func queueMessage(message:Message) {
        do {
            try directlyConnectedDevice.offer(message:message,
                                              callback: { error in })
        } catch {
            // info
            log(type:OSLogType.info, message: "\(error)")
        }
    }
}
