/*
 * Copyright (c) 2017, Oracle and/or its affiliates.  All rights reserved.
 *
 * This software is dual-licensed to you under the MIT License (MIT) and 
 * the Universal Permissive License (UPL).  See the LICENSE file in the root
 * directory for license terms.  You may choose either license, or both.
 */
// com.oracle.iot.client.impl.device

import Foundation

/**
 * Detailed information on a device policy.
 */
public class DevicePolicy : Hashable {
    
    /**
     * A policy with an attribute equal to this string implies
     * the policy is applied to all attributes
     */
    public class func ALL_ATTRIBUTES() -> String {
        return "*"
    }

    public class Function : Hashable {

        public var hashValue:Int {
            get {
                return self.id.hashValue
            }
        }
        
        public static func ==(lhs:DevicePolicy.Function, 
                              rhs:DevicePolicy.Function) -> Bool {
            // TODO: Why aren't parameters compared ?
            return lhs === rhs || lhs.id == rhs.id
        }

        fileprivate let id:String!
        private let parameters:Dictionary<String,Any>!

        public init(id:String, parameters:Dictionary<String,Any>?) {
            self.id = id
            if let parameters = parameters, !parameters.isEmpty {
                // passed by value
                self.parameters = parameters
            } else {
                self.parameters = Dictionary<String,Any>()
            }
        }

        public func getId() -> String {
            return id
        }
        
        public func getParameters() -> Dictionary<String,Any> {
            return parameters
        }
        
        public func toString() -> String {
            var details:String = "id: \(id), parameters: ["
            var comma = ""
            for (k,v) in parameters {
                details += comma + String(format:"%@ : %@",
                    k, String(describing: v))
                comma = ","
            }
            details += "]"
            return details
        }
    }

    fileprivate let id:String!
    private let deviceModelURN:String!
    private let deviceIds:Set<String>!
    private var pipelines:[String:[Function]]!
    private let description:String?
    private let lastModified:Double!
    private let enabled:Bool!

    public var hashValue:Int {
        get {
            return self.id.hashValue
        }
    }
    
    public static func ==(lhs:DevicePolicy, rhs:DevicePolicy) -> Bool {
        return lhs === rhs || lhs.id == rhs.id
    }

    public init(id:String,
                deviceModelURN:String,
                pipelines:[String:[Function]]?,
                description:String,
                lastModified:Double,
                enabled:Bool) {

        self.id = id
        self.deviceModelURN = deviceModelURN
        // TODO: IOT-46187 - removed ?
        self.deviceIds = Set<String>()
        self.description = description
        self.lastModified = lastModified
        self.enabled = enabled
        
        self.pipelines = [String:[Function]]()
        if let pipelines = pipelines, !pipelines.isEmpty {
            // Probably not necessar Structs passed by value
            pipelines.forEach({ k,v in
                self.pipelines[k] = v
            })
        }
    }
    
    /**
     * Get the policy id.
     *
     * - returns: the policy id
     */
    public func getId() -> String {
        return id
    }
    
    /**
     * Get the target device model URN. 
     *
     * - returns: the URN of the target device model
     */
    public func getDeviceModelURN() -> String {
        return deviceModelURN
    }

    /**
     * Get the function pipeline of this policy for an attribute.
     *
     * - returns: a read-only list of `DevicePolicy.Function`.
     */
    public func getPipeline(attributeName:String?) -> [Function]? {
        // TODO: IOT-46187 are there still model pipelies ?
        // What does nil stand for now ?
        // policy applies to all attributes ?
        return attributeName != nil ? pipelines[attributeName!]
                                    : pipelines[DevicePolicy.ALL_ATTRIBUTES()]
    }
    
    /**
     * Get all the pipelines of this policy. The map key is
     * an attribute name, the value is the pipeline for that attribute.
     * 
     * - returns: the pipelines of this policy
     */
    public func getPipelines() -> [String:[Function]] {
        return pipelines
    }

    /**
     * Get the free form description of the device policy
     *
     * - returns: the description of the model
     */
    public func getDescription() -> String? {
        return description
    }

    /**
     * Get the `enabled` state of the device policy
     *
     * - returns: `true` if the policy is enabled, `false` otherwise
     */
    public func isEnabled() -> Bool {
        return enabled
    }

    /**
     * Get the date of last modification.
     *
     * - returns: the date of last modification
     */
    public func getLastModified() -> Double {
        return lastModified
    }
    
    public func toString() -> String {
        var pipelineString:String = ""
        var comma:String = ","
        let pipelines:[String:[DevicePolicy.Function]] = getPipelines()
        for (attributeName,pipeline)in pipelines {
            let dumpedPipeline:String = dumpPipeline(attributeName:attributeName,
                         pipeline:pipeline)
            pipelineString.append(comma)
            pipelineString.append(dumpedPipeline)
            comma = ","
        }
        return pipelineString
    }
    
    func dumpPipeline(attributeName:String,
                      pipeline:[DevicePolicy.Function]) -> String {
        let format:String =
            "\n  {" +
            "\n    \"attributeName\":\"%@\"," +
            "\n    \"pipeline\":" +
            "\n    [%@" +
            "\n    ]" +
            "\n  }"
        
        var pipelineString:String = ""
        for function in pipeline {
            let id:String = function.getId()
            let parameters:[String:Any] = function.getParameters()
            if let deviceFunction:DeviceFunction =
                    DeviceFunction.getDeviceFunction(functionId: id) {
                pipelineString.append(String(format:"\n      %@",
                        deviceFunction.getDetails(configuration: parameters)))
            } else {
                pipelineString.append(String(format:"\n      '%@' not found!",
                                             id))
            }
        }
        return String(format:format, attributeName, pipelineString)
    }
}
