/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import IPTools
import UIKit

/**
 * Message handler for device diagnostics.
 */
class Diagnostics: RequestHandler {
    
    private let startTime: Date
    private let version: String
    private let fixedIpAddress: String?
    private let fixedMacAddress: String
    
    override init() {
        self.startTime = Date();
        self.version = Config.VERSION ?? "Unknown"
        self.fixedIpAddress = Config.IP_ADDRESS
        self.fixedMacAddress = Config.MAC_ADDRESS ?? "Unknown"
    }
    
    override func handleRequest(requestMessage: RequestMessage?) throws -> ResponseMessage {
        if (requestMessage == nil) {
            throw ClientError.argument(NSLocalizedString(
                "Request must not be nil", comment: ""))
        }
        
        if (requestMessage!.getMethod() != nil) {
            if "GET" == requestMessage!.getMethod()!.uppercased() {
                do {
                    var job: [String : AnyObject] = [ : ]
                    var jsonValue: String = "{"
                    let fds: Int64 = self.getFreeDiskSpace()
                    job["freeDiskSpace"] =  fds as AnyObject?
                    jsonValue = jsonValue + "\"freeDiskSpace\":" + fds.description
                    
                    // diagnostics rest API definition currently only
                    // supports a single address, should consider extending
                    // this to include an array of addresses
                    // (for multiple interfaces)
                    if fixedIpAddress == nil {
                        // init address retrieve for being able to get IP and MAC addresses
                        InitAddresses();
                        GetIPAddresses()
                        GetHWAddresses()
                        
                        let ip: String = self.getIpAddress()!
                        let mac: String = self.getMacAddress()!
                        job["ipAddress"] = ip as AnyObject?
                        jsonValue = jsonValue + ",\"ipAddress\":\"" + ip + "\""
                        job["macAddress"] = mac as AnyObject?
                        jsonValue = jsonValue + ",\"macAddress\":\"" + mac + "\""
                    } else {
                        job["ipAddress"] = fixedIpAddress as AnyObject?
                        jsonValue = jsonValue + ",\"ipAddress\":\"" + fixedIpAddress! + "\""
                        job["macAddress"] = fixedMacAddress as AnyObject?
                        jsonValue = jsonValue + ",\"macAddress\":\"" + fixedMacAddress + "\""
                    }
                    let st: Int64 = getStartTimeMs()
                    let tds: Int64 = getTotalDiskSpace()
                    job["startTime"] =  st as AnyObject?
                    jsonValue = jsonValue + ",\"startTime\":" + st.description
                    job["totalDiskSpace"] =  tds as AnyObject?
                    jsonValue = jsonValue + ",\"totalDiskSpace\":" + tds.description
                    job["version"] = self.version as AnyObject?
                    jsonValue = jsonValue + ",\"version\":\"" + self.version + "\"}"
                    return try getResponseMessage(requestMessage: requestMessage!, body: jsonValue, statusCode: .OK)
                } catch {
                    // Will pass through to return bad request.
                }
                
                return try getBadRequestResponse(requestMessage: requestMessage!)
            }
            return try getMethodNotAllowedResponse(requestMessage: requestMessage!)
        } else {
            return try getNilRequestResponse(requestMessage: requestMessage!)
        }
    }
    
    func getFreeDiskSpace() -> Int64 {
        let documentDirectory = NSSearchPathForDirectoriesInDomains(.documentDirectory, .userDomainMask, true).last!
        guard
            let systemAttributes = try? FileManager.default.attributesOfFileSystem(forPath: documentDirectory),
            let freeSize = systemAttributes[FileAttributeKey.systemFreeSize] as? NSNumber
            else {
                // something failed
                return 0
        }
        return freeSize.int64Value
    }

    private func convertArr<T>(count: Int, data: UnsafePointer<T>) -> [T] {
        let buffer = UnsafeBufferPointer(start: data, count: count)
        return Array(buffer)
    }
    
    var hints = addrinfo(
        ai_flags: AI_PASSIVE,       // Assign the address of my local host to the socket structures
        ai_family: AF_UNSPEC,       // Either IPv4 or IPv6
        ai_socktype: SOCK_STREAM,   // TCP
        ai_protocol: 0,
        ai_addrlen: 0,
        ai_canonname: nil,
        ai_addr: nil,
        ai_next: nil)

    /**
     * - returns: the first IPv4 address of this device.
     */
    func getIpAddress() -> String? {
        var status: Int32 = 0
        var servinfo: UnsafeMutablePointer<addrinfo>? = nil
        status = getaddrinfo(
            "localhost",                // Any interface
            nil,                        // The port on which will be listened
            &hints,                     // Protocol configuration as per above
            &servinfo)                  // The created information
        
        if status != 0 {
            return nil
        }  else {
            var info = servinfo
            while info != nil {
                let (clientIp, _ ) = sockaddrDescription(addr: (info?.pointee.ai_addr)!)
                if info?.pointee.ai_family == AF_INET {
                   return clientIp
                }
                info = info?.pointee.ai_next
            }
        }
        return nil
    }
    
    private func sockaddrDescription(addr: UnsafePointer<sockaddr>) -> (String?, String?) {
        
        var host : String?
        var service : String?
        
        var hostBuffer = [CChar](repeating: 0, count: Int(NI_MAXHOST))
        var serviceBuffer = [CChar](repeating: 0, count: Int(NI_MAXSERV))
        
        if getnameinfo(
            addr,
            socklen_t(addr.pointee.sa_len),
            &hostBuffer,
            socklen_t(hostBuffer.count),
            &serviceBuffer,
            socklen_t(serviceBuffer.count),
            NI_NUMERICHOST | NI_NUMERICSERV)
            
            == 0 {
            
            host = String(cString: hostBuffer)
            service = String(cString: serviceBuffer)
        }
        return (host, service)
    }

    
    /**
     * Return the UUID number of device, the MAC address is not available on iOS.
     *
     * - returns: the UUID as 16 byte long String.
     */
    func getMacAddress() -> String? {
        return UIDevice.current.identifierForVendor!.uuidString
    }
 
    
    /**
     * - returns: the start time in ms.
     */
    func getStartTimeMs() -> Int64 {
        return Int64(round(startTime.timeIntervalSince1970 * 1000))
    }
    
    func getTotalDiskSpace() -> Int64 {
        do {
            let systemAttributes = try FileManager.default.attributesOfFileSystem(forPath: NSHomeDirectory() as String)
            let space = (systemAttributes[FileAttributeKey.systemSize] as? NSNumber)?.int64Value
            return space!
        } catch {
            return 0
        }
    }
    
    private func getBadRequestResponse(requestMessage: RequestMessage) throws -> ResponseMessage {
        return try getResponseMessage(requestMessage: requestMessage, body: "Unable to retrieve resource: state.", statusCode: .BAD_REQUEST)
    }
    
    private func getMethodNotAllowedResponse(requestMessage: RequestMessage) throws -> ResponseMessage {
        return try getResponseMessage(requestMessage: requestMessage, body: "Unsupported request: " + requestMessage.toString(), statusCode: .METHOD_NOT_ALLOWED)
    }
    
    private func getNilRequestResponse(requestMessage: RequestMessage) throws -> ResponseMessage {
        return try getResponseMessage(requestMessage: requestMessage, body: "Unsupported request.", statusCode: .METHOD_NOT_ALLOWED)
    }
    
    private func getResponseMessage(requestMessage: RequestMessage, body: String, statusCode: StatusCode) throws -> ResponseMessage {
        let rmb: ResponseMessageBuilder = ResponseMessageBuilder(requestMessage: requestMessage)
        rmb.setBody(body: body)
        rmb.setStatusCode(statusCode: statusCode)
        return try rmb.build()
    }
}
