/*
 *  Copyright © 2016, 2018, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import os

// FIXME: comparison operators with optionals were removed from the 
// Swift Standard Libary.
// Consider refactoring the code to use the non-optional operators.
fileprivate func < <T : Comparable>(lhs: T?, rhs: T?) -> Bool {
  switch (lhs, rhs) {
  case let (l?, r?):
    return l < r
  case (nil, _?):
    return true
  default:
    return false
  }
}


/*************************************************************************
 *
 * Handling of the blocking queue, servicing the queue,
 * and dispatching dequeued entries.
 *
 *************************************************************************/
class Transmitter : NSObject {
    
    private var impl: MessageDispatcherImpl
    
    init(impl: MessageDispatcherImpl) {
        self.impl = impl
        super.init()
    }
    
    @objc func run() {
        //self.impl.group.enter()
        self.impl.enterGroup(by: "Transmitter run")
        self.impl.transmitQueue.async(execute: {
            while (true) {
                self.impl.messageQueued.lock()
                while (!self.impl.requestClose && self.impl.outgoingMessageQueue.isEmpty) {
                    // transmit dispatch queue blocks waiting for an outgoing message
                    // TODO: Should we use a timed wait ?
                    self.impl.messageQueued.wait(until: Date().addingTimeInterval(30.0))
                }
                // TODO: Shouldn't this lock be for the length of the message
                // processing loop ? Is this taken care of by the "messageQueue"
                // dispatch queue ?
                self.impl.messageQueued.unlock()
                // Check again for close or empty queue
                // TODO: Shouldn't this be in the lock ?
                if (self.impl.requestClose && self.impl.outgoingMessageQueue.isEmpty) {
                    //self.impl.group.leave()
                    self.impl.leaveGroup(by:"Transmitter run - requestClosed && messageQueue.isEmpty")
                    break
                }
                
                var messageList: [Message]!
                var errorList:[Message] = [Message]()
                self.impl.messageQueue.sync(execute: {
                    let size: Int = min(
                        self.impl.outgoingMessageQueue.count,
                        self.impl.maximumMessagesPerConnection)
                    
                    if (size <= 0) {
                        return
                    }
                    messageList = []
                    
                    // TODO: Guaranteed delivery
                    // MessageStore.Transaction transaction = messageStore.openTransaction()
                    
                    var inProgressSources:Set<String> = Set<String>()
                    var waitList:[Message] = [Message]()
                    //var sendList:[Message] = [Message]()

                    for _ in 0..<size {
                        let m:Message = self.impl.outgoingMessageQueue.pop()!
                        let clientId:String = m.getClientId();
                        let source:String = m.getSource();
                        
                        if self.impl.failedContentIds.contains(clientId) {
                            errorList.append(m);
                            continue;
                        }
                        
                        if m.getType() == Message.MsgType.RESPONSE ||
                                !(inProgressSources.contains(source) ||
                                    self.impl.isContentDependent(clientId: clientId)) {
                            //sendList.append(m);
                            m.setRemainingRetries(
                                remainingRetries:MessageDispatcherImpl.getBasicNumberOfRetries()) // TODO - why?
                            messageList.append(m)
                            //if sendList.count == size {
                            if messageList.count == size {
                                break;
                            }
                        } else {
                            inProgressSources.update(with: source)
                            waitList.append(m)
                        }
                    }
                    //if sendList.count > 0 {
                    //    messageList = sendList
                    //}
                    for m:Message in waitList {
                        let _ = self.impl.outgoingMessageQueue.push(new: m)
                    }
                })
                
                self.send(messages: messageList)
                
                if !errorList.isEmpty && self.impl.errorCallback != nil {
                    do {
                        try self.impl.errorCallback!.failed(messages: errorList,
                            error: ClientError.network(NSLocalizedString(
                                "Content sync failed", comment: "")))
                    } catch {
                        print("errorcallback for " + NSLocalizedString(
                            "Content sync failed", comment: "") +
                            " threw exception")
                    }
                }
            }
        })
    }
    
    private func handleDispatchError(msgErr: String, messages: [Message]?) {
        if (impl.errorCallback != nil) {
            do {
                try impl.errorCallback!.failed(messages: messages!,
                                               error:ClientError.network(msgErr))
            } catch {
                // TODO: Need to find out why "failed" is throwing exception
                // For now do not log this, it is flooding the output on
                // unit tests, likely is cause is phantom threads and
                // unit tests that do not activate virtual devices
                //print(msgErr + ": errorCallback threw exception " +
                //                                    error.localizedDescription)
            }
        }
        impl.counterQueue.sync {
            impl.totalProtocolErrors = impl.totalProtocolErrors + 1
        }
    }

    // TODO: Does the paraemter have to accept nil ?
    private func send(messages: [Message]?) {
        guard let messages = messages, !messages.isEmpty else {
            return
        }
        do {
            self.impl.group.enter()
            try impl.deviceClient.send(messages: messages, callback: {error in
                if error != nil {
                    self.handleDispatchError(msgErr: String(format:NSLocalizedString(
                        "%@: received error in callback: %@", comment: ""),
                        "Transmitter.send()", "\(error!)"),
                        messages: messages)
                    switch (error!) {
                    case .security:
                        break
                    default:
                        self.handleRemainingRetries(messages:messages)
                        break
                    }
                } else {

                    if self.impl.deliveryCallback != nil {
                        self.impl.deliveryCallback!.delivered(
                            messages: messages)
                    }
                    
                    var persistedMessages:[Message] = [Message]()
                    
                    self.impl.counterQueue.sync(execute: {
                        self.impl.totalMessagesSent =
                            self.impl.totalMessagesSent + messages.count
                        var jsonStr: String = ""
                        for message in messages {
                            // Guaranteed delivery messages need to be deleted from
                            // message persistence. If the message is not guaranteed
                            // delivery, remove it from the list of messages. At
                            // the end of the loop, only GUARANTEED_DELIVERY messages
                            // will be in the original list and we can call delete on that.
                            //
                            if let _ = self.impl.messagePersistence,
                                    message.getReliability() ==
                                        Message.Reliability.guaranteed_delivery{
                                persistedMessages.append(message)
                            }
                            // Catch and log any json errors to ensure that
                            // persisted messages are processed
                            // TODO: There must be a better way to handle totalBytesSent
                            do {
                                try jsonStr = jsonStr + message.toJsonString()
                            } catch {
                                print(String(format:NSLocalizedString(
                                    "%@: caught exception counting message bytes sent: %@",
                                    comment:""), "Transmitter.send(): callback", "\(error)"))
                            }
                        }
                        // TODO: There must be a better way to handle totalBytesSent
                        self.impl.totalBytesSent +=
                            Int64((jsonStr.data(using: .utf8)?.count)!)
                    })
                    
                    // TODO: call delete from separate thread?
                    // If persisted messages is not empty then messagePersistence
                    // is not nil
                    if let mp = self.impl.messagePersistence,
                            !persistedMessages.isEmpty {
                        mp.delete(messages:persistedMessages)
                    }
                }
                // TODO: JCL signals receive thread why isn't that done here ?
                self.impl.group.leave()
            })
        } catch {
            handleDispatchError(msgErr: String(format:NSLocalizedString(
                "%@: threw exception: %@", comment:""),
                "Transmitter.send(): deviceClient.send()", "\(error)"),
                messages: messages)
            defer {
                self.impl.group.leave()
            }
            if error is ClientError {
                switch (error as! ClientError) {
                case .security:
                    return
                default:
                    break
                }
            }
            // For all other errors
            handleRemainingRetries(messages:messages)
        }
    }
    
    func handleRemainingRetries(messages:[Message]) {
        self.impl.totalProtocolErrors += 1
        for message in messages {
            if message.getRemainingRetries() > 0 {
                if self.impl.outgoingMessageQueue.count < self.impl.maximumQueueSize {
                    self.impl.messageQueue.sync() {
                        message.setRemainingRetries(remainingRetries:message.getRemainingRetries() - 1)
                        _ = self.impl.outgoingMessageQueue.push(new: message)
                        self.impl.totalMessagesRetried += 1
                    }
                } else if message.getReliability() != Message.Reliability.guaranteed_delivery {
                    // INFO
                    print(String(format:NSLocalizedString(
                        "%@: Cannot queue message for retry. Message discarded: %@",
                        comment:""), "Transmitter.handlePersistedMessages()",
                        message.getClientId()))
                    continue
                }
            } else {
                // INFO
                print(String(format:NSLocalizedString(
                    "%@: Maximum number message retries reached. Message discarded: %@",
                    comment:""),"Transmitter.handlerPersistedMessages()",
                    message.getClientId()))
            }
        }
    }
}

class Dispatcher {
    var requestMessage: RequestMessage?
    var mdi: MessageDispatcherImpl
    
    init(requestMessage: RequestMessage, mdi: MessageDispatcherImpl) {
        self.requestMessage = requestMessage
        self.mdi = mdi
    }

    func run() {
        do {
            try mdi.getRequestDispatcher().dispatch(requestMessage: requestMessage,
                        callback: { (responseMessage, error) in
                do {
                    if error == nil {
                        try self.mdi.queue(message: responseMessage!)
                    } else {
                        print(error!)
                    }
                } catch {
                    print("Issue while dispatching: \(error)")
                }
            })
        } catch {
            print("Issue while dispatching: \(error)")
        }
    }
}

class Receiver : NSObject {
    private var impl: MessageDispatcherImpl
    
    init(impl: MessageDispatcherImpl) {
        self.impl = impl
        super.init()
    }
    
    @objc func run() {
        //self.impl.group.enter()
        self.impl.enterGroup(by: "Receiver.run()")
        self.impl.receiveQueue.async(execute: {
            
            while (!self.impl.requestClose) {
                if (Config.DISABLE_LONG_POLLING) {
                    sleep(UInt32(Config.MESSAGE_POLLING_INTERVAL/1000))
                }
                do {
                    try self.impl.deviceClient.receive(timeout: -1, callback: { requestMessage, error in
                        defer {
                            self.impl.receiveSem.signal()
                        }
                        if (error != nil) {
                            // TODO: Can cause infinite loop if network receive
                            // error does not get corrected.
                            print("Message receive returned error: \(error!)")
                        } else if (requestMessage != nil) {
                            self.impl.counterQueue.sync(execute: {
                                self.impl.totalMessagesReceived = self.impl.totalMessagesReceived + 1
                                self.impl.totalBytesReceived = self.impl.totalBytesReceived +
                                    Int64(((requestMessage!.toJsonString()).data(using: String.Encoding(rawValue: String.Encoding.utf8.rawValue))?.count)!)
                            })
                            if (!self.impl.requestClose) {
                                //new reference for later use when processed from the queue as requestMessage changes in the loop
                                let msg = requestMessage
                                Dispatcher(requestMessage: msg!, mdi: self.impl).run()
                            }
                        }
                    })
                    // Only wait if a call to receivce occurred
                    // Only want one receive attempt at a time.
                    _ = self.impl.receiveSem.wait(timeout:.now() +
                        Double(Config.MESSAGE_POLLING_INTERVAL/1000))
                } catch {
                    //print("Message receive dispatcher threw exception : \(error)")
                    // Nothing waiting at this point.
                    //self.impl.receiveSem.signal()
                }
                //self.impl.receiveSem.wait()
            }
            //self.impl.group.leave()
            self.impl.leaveGroup(by: "Reciever.run(): requestClose == \(self.impl.requestClose)")
        })
    }
}

class CounterHandler: RequestHandler {
    
    private var impl: MessageDispatcherImpl
    
    init(impl: MessageDispatcherImpl) {
        self.impl = impl
    }
    
    override func handleRequest(requestMessage: RequestMessage?) throws -> ResponseMessage {
        if (requestMessage == nil) {
            throw ClientError.argument(NSLocalizedString(
                "Request must not be nil", comment: ""))
        }
        
        var status: StatusCode
        if (requestMessage!.getMethod() != nil) {
            if ("GET" == requestMessage!.getMethod()?.uppercased()) {
                do {
                    var jobDict: [String : AnyObject] = [ : ]
                    var jsonBody: String = "{"
                    var size: Float = Float(0)
                    impl.messageQueue.sync(execute: {
                        size = Float(impl.outgoingMessageQueue.count)
                    })
                    let max: Float = Float(impl.maximumQueueSize)
                    // we only want 1 decimal point precision
                    let load: Int = Int(size / max * 1000)
                    self.impl.counterQueue.sync(execute: {
                        jobDict["load"] = Float(load / 10) as AnyObject?
                        jsonBody = jsonBody + "\"load\":" + (Float(load / 10)).description
                        jobDict["totalBytesSent"] = impl.totalBytesSent as AnyObject
                        jsonBody = jsonBody + ",\"totalBytesSent\":" + impl.totalBytesSent.description
                        jobDict["totalBytesReceived"] = impl.totalBytesReceived as AnyObject
                        jsonBody = jsonBody + ",\"totalBytesReceived\":" + impl.totalBytesReceived.description
                        jobDict["totalMessagesReceived"] = impl.totalMessagesReceived as AnyObject
                        jsonBody = jsonBody + ",\"totalMessagesReceived\":" + impl.totalMessagesReceived.description
                        jobDict["totalMessagesRetried"] = impl.totalMessagesRetried as AnyObject
                        jsonBody = jsonBody + ",\"totalMessagesRetried\":" + impl.totalMessagesRetried.description
                        jobDict["totalMessagesSent"] = impl.totalMessagesSent as AnyObject
                        jsonBody = jsonBody + ",\"totalMessagesSent\":" + impl.totalMessagesSent.description
                        jobDict["totalProtocolErrors"] = impl.totalProtocolErrors as AnyObject
                        jsonBody = jsonBody + ",\"totalProtocolErrors\":" + impl.totalProtocolErrors.description
                        jsonBody = jsonBody + "}"
                    })
                    let rmb: ResponseMessageBuilder =
                        ResponseMessageBuilder(requestMessage: requestMessage!)
                    rmb.setStatusCode(statusCode: .OK)
                    rmb.setBody(body: jsonBody)
                    return try rmb.build()
                } catch {
                    status = .BAD_REQUEST;
                }
            } // GET
            else  {
                //Unsupported request method
                status = .METHOD_NOT_ALLOWED;
            }
        }
        else {
            //Unsupported request method
            status = .METHOD_NOT_ALLOWED;
        }
        let rmb: ResponseMessageBuilder =
            ResponseMessageBuilder(requestMessage: requestMessage!)
        rmb.setStatusCode(statusCode: status)
        return try rmb.build()
    }
}


class ResetHandler: RequestHandler {
    
    private var impl: MessageDispatcherImpl
    
    init(impl: MessageDispatcherImpl) {
        self.impl = impl
    }
    
    override func handleRequest(requestMessage: RequestMessage?) throws -> ResponseMessage {
        if (requestMessage == nil) {
            throw ClientError.argument(NSLocalizedString(
                "Request must not be nil", comment: ""))
        }
        
        var status: StatusCode
        if (requestMessage!.getMethod() != nil) {
            if ("PUT" == requestMessage!.getMethod()?.uppercased()) {
                impl.totalBytesSent = 0
                impl.totalBytesReceived = 0
                impl.totalProtocolErrors = 0
                impl.totalMessagesSent = 0
                impl.totalMessagesReceived = 0
                impl.totalMessagesRetried = 0
                status = .OK
            }
            else  {
                //Unsupported request method
                status = .METHOD_NOT_ALLOWED
            }
        }
        else {
            status = .METHOD_NOT_ALLOWED
        }
        let rmb: ResponseMessageBuilder =
            ResponseMessageBuilder(requestMessage: requestMessage!)
        rmb.setStatusCode(statusCode: status)
        return try rmb.build()
    }
}

class PollingIntervalHandler: RequestHandler {
    
    private var impl: MessageDispatcherImpl
    
    init(impl: MessageDispatcherImpl) {
        self.impl = impl
    }
    
    override func handleRequest(requestMessage: RequestMessage?) throws -> ResponseMessage {
        
        if (requestMessage == nil) {
            throw ClientError.argument(NSLocalizedString(
                "Request must not be nil", comment: ""))
        }
        
        var status: StatusCode
        if (requestMessage!.getMethod() != nil) {
            if ("GET" == requestMessage!.getMethod()?.uppercased()) {
                do {
                    let jobStr: String = "{\"pollingInterval\":" +
                        (impl.pollingInterval / 1000).description + "}"
                    
                    let rmb: ResponseMessageBuilder =
                        ResponseMessageBuilder(requestMessage: requestMessage!)
                    rmb.setStatusCode(statusCode: .OK)
                    rmb.setBody(body: jobStr)
                    let messRet = try rmb.build()
                    return messRet
                } catch {
                    status = .BAD_REQUEST;
                }
            }
            else if ("PUT" == requestMessage!.getMethod()?.uppercased()) {
                do {
                    let errors: String = ""
                    // Use Long objects here in case there was a problem 
                    // getting the parameters.
                    impl.pollingInterval = impl.getParam(data: requestMessage!.getBody() as Foundation.Data,
                                    paramName: "value", errors: errors) * 1000
                    
                    // TODO: What's this for errors is "" ?
                    // Is errors supposed to be "inout" ?
                    // Also not that getParam always returns 0
                    if (errors.count != 0) {
                        return try impl.getBadRequestResponse(requestMessage: requestMessage!,
                                        message: impl.endpointId, src: errors)
                    }
                    status = .OK;
                } catch {
                    status = .BAD_REQUEST
                }
            }
            else  {
                //Unsupported request method
                status = .METHOD_NOT_ALLOWED
            }
        }
        else {
            status = .METHOD_NOT_ALLOWED
        }
        let rmb: ResponseMessageBuilder = ResponseMessageBuilder(requestMessage: requestMessage!)
        rmb.setStatusCode(statusCode: status)
        return try rmb.build()
    }
}

/**
 * MessageDispatcherImpl
 */
class MessageDispatcherImpl: MessageDispatcher {
    
    private static var DEFAULT_MAXIMUM_MESSAGES_TO_QUEUE: Int = 1000
    private static var DEFAULT_MAXIMUM_MESSAGES_PER_CONNECTION: Int = 100
    private static var DEFAULT_POLLING_INTERVAL: Int64 = 1000 // milliseconds
    
    private static let MAXIMUM_MESSAGES_TO_QUEUE_PROPERTY: String =
        "oracle.iot.client.device.dispatcher_max_queue_size"
    private static let MAXIMUM_MESSAGES_PER_CONNECTION_PROPERTY: String  =
        "oracle.iot.client.device.dispatcher_max_messages_per_connection"
    private static let POLLING_INTERVAL_PROPERTY: String  =
        "oracle.iot.client.device.dispatcher_polling_interval"
    
    static let backslash = "\\"
    
    static let COUNTERS_URL: String = "deviceModels" + "/" + ActivationManager.MESSAGE_DISPATCHER_URN + "/counters"
    static let RESET_URL: String = "deviceModels" + "/" + ActivationManager.MESSAGE_DISPATCHER_URN + "/reset"
    static let POLLING_INTERVAL_URL: String = "deviceModels" + "/" + ActivationManager.MESSAGE_DISPATCHER_URN + "/pollingInterval"
    static let DIAGNOSTICS_URL: String = "deviceModels" + "/" + ActivationManager.DIAGNOSTICS_URN + "/info"
    static let TEST_CONNECTIVITY_URL: String = "deviceModels" + "/" + ActivationManager.DIAGNOSTICS_URN + "/testConnectivity"
    
    static let DIAGNOSTICS_URL_STATUS: String = "deviceModels" + "/" + ActivationManager.DIAGNOSTICS_URN + "/testConnectivity/status"
    static let DIAGNOSTICS_URL_START: String = "deviceModels" + "/" + ActivationManager.DIAGNOSTICS_URN + "/testConnectivity/start"
    static let DIAGNOSTICS_URL_STOP: String = "deviceModels" + "/" + ActivationManager.DIAGNOSTICS_URN + "/testConnectivity/stop"
    
    /**
     * Serial dispatch queues for message dispatching activities
     */
    let messageQueue  = DispatchQueue(label: "com.oracle.client.impl.device.messageQueue",
        qos: .utility)
    let counterQueue  = DispatchQueue(label: "com.oracle.client.impl.device.counterQueue",
        qos: .utility)
    let transmitQueue = DispatchQueue(label: "com.oracle.client.impl.device.transmitQueue",
        qos: .utility)
    let receiveQueue  = DispatchQueue(label: "com.oracle.client.impl.device.receiveQueue",
        qos: .utility)
    fileprivate let group = DispatchGroup()
    
    fileprivate func enterGroup(by:String) {
        //print("### Entering group: " + by)
        group.enter()
    }
    fileprivate func leaveGroup(by:String) {
        //print("### Leaving group: " + by)
        group.leave()
    }
    
    // Need implicit optional because not initialized before super is called.
    private var counterHandler: RequestHandler!
    private var resetHandler: RequestHandler!
    private var pollingIntervalHandler: RequestHandler!
    private var diagnosticsHandler: RequestHandler!
    private var testConnectivityHandler: RequestHandler!
    var transmitter: Transmitter!
    var receiver: Receiver!
    private var resourceMessage:ResourceMessage!
    
    fileprivate var contentMap:Dictionary<StorageObject, Set<String>>
    fileprivate var failedContentIds:Set<String>

    let receiveSem = DispatchSemaphore(value:0)
    
    /*
     * Queue of outgoing messages to be dispatched to server
     * (with thread for servicing the queue).
     * Note: it has package access for unit test access.
     */
    var outgoingMessageQueue: PriorityQueue<Message>
    
    private var contentLock:NSRecursiveLock = NSRecursiveLock()
    //private var closeLock: AnyObject?
    fileprivate var messageQueued:NSCondition = NSCondition()
    
    let messagingPolicyImplQueue:DispatchQueue =
        DispatchQueue(label:"messagingPolicyImpl queue", qos:.utility)
    var messagingPolicyImpl:MessagingPolicyImpl? = nil
    
    /*
     * maximum number of messages allowed in the outgoing message queue
     */
    fileprivate var maximumQueueSize: Int = 0
    
    /*
     * maximum number of messages to send in one connection
     */
    fileprivate var maximumMessagesPerConnection: Int = 0
    
    // Counter indicating total number of messages that have been delivered
    fileprivate var totalMessagesSent: Int = 0
    
    // Counter indicating total number of messages that have been delivered
    fileprivate var totalMessagesReceived: Int = 0
    
    // Counter indicating total number of messages that were retried for delivery
    fileprivate var totalMessagesRetried: Int = 0
    
    // Counter indicating total number of bytes that have been delivered
    fileprivate var totalBytesSent: Int64 = 0
    
    // Counter indicating total number of bytes that have been delivered
    fileprivate var totalBytesReceived: Int64 = 0
    
    // Counter indicating total number of protocol errors.
    fileprivate var totalProtocolErrors: Int64 = 0
    
    fileprivate var endpointId: String
    
    /*
     * Maximum time in milliseconds to wait for a message to be queued
     * before the send thread will poll the server
     */
    fileprivate var pollingInterval: Int64 = 0
    
    fileprivate var deviceClient: AdvancedDirectlyConnectedDevice!
    
    fileprivate var messagePersistence:MessagePersistence?
    
    /*
     * Set to true if close method has been called.
     */
    private var _closed:Bool = false
    private let closedQueue:DispatchQueue =
        DispatchQueue(label:"MessageDispatcher close queue",
                      qos: .userInitiated)
    fileprivate var closed: Bool {
        get {
            return closedQueue.sync() {
                return self._closed
            }
        }
        set {
            closedQueue.sync(execute: {
                self._closed = newValue
            })
        }
    }
    private var _requestClose: Bool = false
    private let requestCloseQueue:DispatchQueue =
        DispatchQueue(label:"MessageDispatcher requestClose queue",
                      qos: .userInitiated)
    fileprivate var requestClose: Bool {
        get {
            return requestCloseQueue.sync() {
                return self._requestClose
            }
        }
        set {
            requestCloseQueue.sync(execute: {
                self._requestClose = newValue
            })
        }
    }
    
    fileprivate var deliveryCallback: MessageDispatcher.DeliveryCallback?
    fileprivate var errorCallback: MessageDispatcher.ErrorCallback?
    
    private let mesDispTime:Double = 1.0
    
    override func setOnDelivery(callback: MessageDispatcher.DeliveryCallback) {
        self.deliveryCallback = callback
    }
    
    override func setOnError(callback: MessageDispatcher.ErrorCallback) {
        self.errorCallback = callback
    }
    
    override func getRequestDispatcher() -> RequestDispatcher{
        return RequestDispatcher.getInstance()
    }
// ------------------------------------------
    /**
     * Implmentation of MessageDispatcher.
     *
     * - parameter deviceClient: The `DirectlyConnectedDevice` that uses this
     *                           message dispatcher.
     * - parameter messagePersistence: A `MessagePersistence` instanc
     */
    init(deviceClient: AdvancedDirectlyConnectedDevice) throws {
    
        self.deviceClient = deviceClient
        self.pollingInterval = MessageDispatcherImpl.getPollingInterval()
        self.maximumQueueSize = MessageDispatcherImpl.getQueueSize()
        self.maximumMessagesPerConnection =
            MessageDispatcherImpl.getMaximumMessagesPerConnection();
        
        self.endpointId = deviceClient.getEndpointId()
        
        self.contentMap = Dictionary<StorageObject, Set<String>>()
        self.failedContentIds = Set<String>()
        
        self.outgoingMessageQueue = PriorityQueue<Message>(
            compare: {
                // Note this implementation is not consistent with equals. It 
                // is possible that a.compareTo(b) == 0 is not that same 
                // boolean value as a.equals(b)
                
                // The natural order of enum is the enum's ordinal, i.e.,
                // x.getPriority().compareTo(y.getPriority() will give {x,y}
                // if x is a lower priority. What we want is to sort by the
                // higher priority.
                return ($0.getPriority().rawValue < $1.getPriority().rawValue)
                    // If they are the same priority, take the one that was 
                    // created first
                    || ($0.getEventTime() < $1.getEventTime())
                    
                    // If they are still the same, take the one with higher 
                    //reliability.
                    || ($0.getReliability().rawValue < $1.getReliability().rawValue)
        })
        
        self.totalMessagesSent = 0
        self.totalMessagesReceived = 0
        self.totalMessagesRetried = 0
        self.totalBytesSent = 0
        self.totalBytesReceived = 0
        self.totalProtocolErrors = 0
        
        super.init()
        
        let resetHandler: RequestHandler = ResetHandler(impl: self)
        let counterHandler: RequestHandler = CounterHandler(impl: self)
        
        /*
         * resource definition for pollingInterval: {
         *   "get": {
         *     "schema": {
         *       "properties": {
         *         "value": {
         *           "type": "number",
         *           "description": "The incoming message polling interval
         *              in seconds on the directly connected device."
         *         }
         *       }
         *     },
         *   "put" : {
         *     "parameters": [
         *       {"name": "value",
         *        "type": "integer",
         *        "in": "body",
         *        "description": "The incoming message polling interval in
         *          seconds.",
         *        "required": "true"},
         *       ]
         *   }
         */
        self.pollingIntervalHandler = PollingIntervalHandler(impl: self)
        
        do {
            diagnosticsHandler = Diagnostics()
            let testConnectivity: TestConnectivity =
                TestConnectivity(endpointId: endpointId, messageDispatcher: self)
            testConnectivityHandler = testConnectivity.getTestConnectivityHandler()
            let rmb: ResourceMessageBuilder = ResourceMessageBuilder()
            rmb.setEndpointName(endpointName: endpointId)
            rmb.setSource(source: endpointId)
            rmb.setSender(sender: endpointId)
            // build a list String that can be hashed
            var stringList:[String] = Array()
            stringList.append(MessageDispatcherImpl.COUNTERS_URL)
            stringList.append(MessageDispatcherImpl.RESET_URL)
            stringList.append(MessageDispatcherImpl.POLLING_INTERVAL_URL)
            stringList.append(MessageDispatcherImpl.DIAGNOSTICS_URL)
            stringList.append(MessageDispatcherImpl.DIAGNOSTICS_URL_STATUS)
            stringList.append(MessageDispatcherImpl.DIAGNOSTICS_URL_START)
            stringList.append(MessageDispatcherImpl.DIAGNOSTICS_URL_STOP)
            let reconMark = ResourceMessage.getMD5ofList(resStrings: stringList)
            // calculate the reconciliation mark as MD5 hash over all path information (in the same order)
            rmb.setReconciliationMark(rM: reconMark!)
            //
            try rmb.register(ri: getResource(endpointId: endpointId,
                                             name: "counters",
                                             path: MessageDispatcherImpl.COUNTERS_URL,
                                             requestHandler: counterHandler,
                                             methods: Array(arrayLiteral: Method.GET)))
            try rmb.register(ri: getResource(endpointId: endpointId,
                                             name: "reset",
                                             path: MessageDispatcherImpl.RESET_URL,
                                             requestHandler: resetHandler,
                                             methods: Array(arrayLiteral: Method.PUT)))
            try rmb.register(ri: getResource(endpointId: endpointId,
                                             name: "polling_interval",
                                             path: MessageDispatcherImpl.POLLING_INTERVAL_URL,
                                             requestHandler: pollingIntervalHandler,
                                             methods: Array(arrayLiteral: Method.GET, Method.PUT)))
            try rmb.register(ri: getResource(endpointId: endpointId,
                                             name: "dignostic_info",
                                             path:MessageDispatcherImpl.DIAGNOSTICS_URL,
                                             requestHandler: diagnosticsHandler,
                                             methods: Array(arrayLiteral: Method.GET)))
            // additional resources like in the Posix API
            try rmb.register(ri: getResource(endpointId: endpointId,
                                             name: "diagnostic_connectivity_status",
                                             path: MessageDispatcherImpl.DIAGNOSTICS_URL_STATUS,
                                             requestHandler: testConnectivityHandler,
                                             methods: Array(arrayLiteral: Method.GET, Method.PUT)))
            try rmb.register(ri: getResource(endpointId: endpointId,
                                             name: "diagnostic_connectivity_start",
                                             path: MessageDispatcherImpl.DIAGNOSTICS_URL_START,
                                             requestHandler: testConnectivityHandler,
                                             methods: Array(arrayLiteral: Method.GET, Method.PUT)))
            try rmb.register(ri: getResource(endpointId: endpointId,
                                             name: "diagnostic_connectivity_stop",
                                             path: MessageDispatcherImpl.DIAGNOSTICS_URL_STOP,
                                             requestHandler: testConnectivityHandler,
                                             methods: Array(arrayLiteral: Method.GET, Method.PUT)))
            self.resourceMessage = try rmb.build()
            // TODO: handle error
        } catch {
            throw ClientError.network(String(format:NSLocalizedString(
                "Illegal Resource Message: %@", comment: ""), "\(error)"))
        }
        
        transmitter = Transmitter(impl: self)
        receiver = Receiver(impl: self)
        
        self.transmitter.run()
        self.receiver.run()
        // TODO: IOT-37051
        // Fails if device not active, Does this message need to be sent ?
        // It is "diagnostic_connectivity_stop"
        try self.queue(message: self.resourceMessage)
        
        // Do this last after everything else is established.
        // Populate outgoing message queue from persisted messages, but leave the
        // messages in persistence. The messages are removed from persistence when
        // they are delivered successfully.
        // TODO: IOT-49043
        self.messagePersistence = MessagePersistence.getInstance()
        guard let mp = self.messagePersistence else {
            return
        }
        if let messages:[Message] = mp.load(endpointId: deviceClient.getEndpointId()),
                !messages.isEmpty {
            messageQueued.lock()
            self.outgoingMessageQueue.addAll(elements:messages)
            messageQueued.signal()
            messageQueued.unlock()
        }
    }
    
    private static func getQueueSize() -> Int {
        let size: Int = MessageDispatcherImpl.DEFAULT_MAXIMUM_MESSAGES_TO_QUEUE
        // size must be at least 1 or PriorityQueue constructor will throw 
        // exception
        return (size > 0 ? size : MessageDispatcherImpl.DEFAULT_MAXIMUM_MESSAGES_TO_QUEUE)
    }
    
    private static func getMaximumMessagesPerConnection() -> Int {
        let max: Int = MessageDispatcherImpl.DEFAULT_MAXIMUM_MESSAGES_PER_CONNECTION
        // size must be at least 1
        return (max > 0 ? max : MessageDispatcherImpl.DEFAULT_MAXIMUM_MESSAGES_PER_CONNECTION)
    }
    
    private static func getPollingInterval() -> Int64 {
        let interval: Int64 = MessageDispatcherImpl.DEFAULT_POLLING_INTERVAL
        // polling interval may be zero, which means wait forever
        return (interval >= 0 ? interval : MessageDispatcherImpl.DEFAULT_POLLING_INTERVAL)
    }
    
    private static let DEFAULT_BASIC_NUMBER_OF_RETRIES:Int = 3
    static func getBasicNumberOfRetries() -> Int {
        let max:Int = Config.BASIC_NUMBER_OF_RETRIES
        // number of retries must be at least 3
        return (max > 0 ? max : DEFAULT_BASIC_NUMBER_OF_RETRIES)
    }
    
    /**
     * Returns the Int64 parameter specified by `paramName` in the request if 
     * it's available.
     *
     * - parameter data: a `Foundation.Data` object containing the request data.
     *
     * - parameter paramName: the name of the parameter to get.
     *
     * - parameter errors: a `String` of errors.  Any errors produced from
     *                     retrieving the parameter will be appendeto this.
     *
     * - returns: the parameter if it can be retrieved, otherwise `nil`
     */
    fileprivate func getParam(data: Foundation.Data, paramName: String,
                              errors: String) -> Int64 {
       return 0
    }
    
    /**
     * Appends spaces (error separators) to the `errors` list if it's not empty.
     *
     * - parameter errors: a list of errors in a String
     */
    private func appendSpacesToErrors(errors: String) -> String {
        if !errors.isEmpty {
            return errors + "  "
        } else {
            return errors
        }
    }
    
    /**
     * Returns an appropriate response if the request is bad.
     *
     * - parameter requestMessage: the request for this capability.
     *
     * - returns an appropriate response if the request is bad.
     */
    fileprivate func getBadRequestResponse(requestMessage: RequestMessage,
                    message: String, src: String) throws -> ResponseMessage {
        return try getResponseMessage(requestMessage: requestMessage,
                                      body: message, statusCode: .BAD_REQUEST)
    }
    
    /**
     * Returns a *RequestMessage* with the supplied parameters.
     *
     * - parameter requestMessage: the request for this capability.
     *
     * - parameter body:           the body for the response.
     *
     * - parameter statusCode:     the status code of the response.
     *
     * - returns: an appropriate response if the request is *nil*
     */
    private func getResponseMessage(requestMessage: RequestMessage, body: String, statusCode: StatusCode) throws -> ResponseMessage {
        let rmb: ResponseMessageBuilder = ResponseMessageBuilder(requestMessage: requestMessage)
        rmb.setBody(body: body)
        rmb.setStatusCode(statusCode: statusCode)
        return try rmb.build()
    }
    
    // Get current software version
    private func getSoftwareVersion() -> String {
        // TODO: Need better mechanism for obtaining the version
        return Config.VERSION!
    }
    // TODO: Use Dispatch queues instead of locks.
    public func addStorageObjectDependency(storageObject:StorageObject, clientMsgId:String) {
        contentLock.lock();
        defer {
            contentLock.unlock()
        }
        var msgIds:Set<String>? = contentMap[storageObject]
        if msgIds == nil {
            msgIds = Set<String>()
            contentMap[storageObject] = msgIds
        }
        msgIds!.update(with: clientMsgId)
    }
    
    public func removeStorageObjectDependency(storageObject:StorageObject) {
        let completed:Bool = storageObject.getSyncStatus() == SyncStatus.inSync
        contentLock.lock()
        defer {
            contentLock.unlock()
        }
        if let ids:Set<String> = contentMap.removeValue(forKey: storageObject) {
            if !completed {
                failedContentIds.formUnion(ids)
            }
        }
    }
    
    func isContentDependent(clientId:String) -> Bool {
        contentLock.lock();
        defer {
            contentLock.unlock()
        }
        for set:Set<String> in contentMap.values {
            if set.contains(clientId) {
                return true
            }
        }
        return false
    }
    override func queue(message: Message) throws {
        try queue(messages:[message])
    }

    override func queue(messages: [Message]) throws {
        if messages.isEmpty {
            throw ClientError.argument(NSLocalizedString("messages array is empty",
                                                         comment: ""))
        }
        if let mp = self.messagePersistence {
            var mpMessages:[Message] = [Message]()
            for message in messages {
                if message.getReliability() == Message.Reliability.guaranteed_delivery {
                    mpMessages.append(message)
                }
            }
            if !mpMessages.isEmpty {
                // TODO: call save from separate thread?
                mp.save(messages: mpMessages,
                        endpointId:deviceClient.getEndpointId())
            }
        }

        self.messageQueue.async(execute: {
            // serial dispatch queue to read/write "outgoingMessageQueue"
            for message in messages {
                defer {
                    self.messageQueued.unlock()
                }
                var queued: Bool = false
                self.messageQueued.lock()
                // TODO: No indication that the queue has exceeeded max queue size
                queued = (self.outgoingMessageQueue.count < self.maximumQueueSize) &&
                    (self.outgoingMessageQueue.push(new: message))
                
                if !queued && self.errorCallback != nil {
                    do {
                        try self.errorCallback!.failed(messages: Array<Message>(arrayLiteral: message))
                    } catch {
                        print("Issue when calling error callback")
                    }
                } else {
                    self.messageQueued.signal()
                }
            }
        })
    }
    
    override
    public func offer(messages:[Message]) throws {
        if messages.isEmpty {
            throw ClientError.argument(NSLocalizedString(
                "messages cannot be empty", comment:""))
        }
        
        // TODO: get from persistenece store
        var messagingPolicy:MessagingPolicyImpl? = nil
        messagingPolicyImplQueue.sync() {
            if messagingPolicyImpl == nil {
                messagingPolicyImpl = MessagingPolicyImpl(directlyConnectedDevice:deviceClient!)
            }
            messagingPolicy = messagingPolicyImpl
        }
        // TODO: get from persistenece store
        if let devicePolicyManager:DevicePolicyManager =
                DevicePolicyManager.getDevicePolicyManager(dcd:deviceClient!) {
            devicePolicyManager.addChangeListener(changeListener:messagingPolicy!)
        }
        
        var messagesToQueue:[Message] = [Message]()
        do {
            for message in messages {
                let messagesFromPolicy:[Message] =
                    try messagingPolicy!.applyPolicies(message:message)
                messagesToQueue.append(contentsOf:messagesFromPolicy)
            }
        } catch {
            // TODO: retry?
            // WARNING
            log(type:OSLogType.fault,
                message: "\(error)")
            
            if errorCallback != nil {
                var messageList:[Message] = [Message]()
                messageList.append(contentsOf:messages)
                try errorCallback!.failed(messages:messageList,
                                     error: error as? ClientError)
            }
        }
        
        if !messagesToQueue.isEmpty {
            try queue(messages:messagesToQueue)
        }
    }
    
    /** @see oracle.iot.client.Client#close() */
    override func close(callback: @escaping () -> Void) {
        if closed {
            callback()
        }
        // All these operations have to be atomic
        self.messageQueued.lock()
        requestCloseQueue.sync() {
            // the callback can't assume the callback will be on the main thread.
            group.notify(queue: DispatchQueue(label:"MessageDispatcherImpl request close queue",
                                              qos:.utility)) {
                self.closed = true
                callback()
            }
            _requestClose = true
            self.receiveSem.signal()
            self.messageQueued.signal()
        }
        self.messageQueued.unlock()
        
    }
    
    func isClosed() -> Bool {
        return closed
    }
    
    /* package */

    /**
     * - throws: `ClientError.argument` if `requestHandler` is `nil`
     */
    func getResource(endpointId: String, name: String, path: String,
                     requestHandler: RequestHandler,
                     methods: [Method]) throws -> Resource {
        
        let messageDispatcher: MessageDispatcher = self
        let requestDispatcher: RequestDispatcher = messageDispatcher.getRequestDispatcher()
        try requestDispatcher.registerRequestHandler(endpointId: endpointId,
                                                     path: path,
                                                     handler: requestHandler)
        
        let builder: ResourceBuilder = ResourceBuilder()
        builder.setEndpointName(endpointName: endpointId)
        for m in methods {
            builder.addMethod(method: m)
        }
        builder.setPath(path: path)
        builder.setName(name: name)
        return try builder.build()
    }
}
