/*
 * Copyright (c) 2018, Oracle and/or its affiliates.  All rights reserved.
 *
 * This software is dual-licensed to you under the MIT License (MIT) and
 * the Universal Permissive License (UPL).  See the LICENSE file in the root
 * directory for license terms.  You may choose either license, or both.
 */

// com.oracle.iot.client.impl.device

import Foundation
import SQLite3
import SQLite3Swift
import SwiftyJSON

// Hack because Message is shared with EnterpriseLib, and it does not
// know about DataMessageBuilder, or the other builders.
func getMessage(jsonData:Foundation.Data) throws -> Message {
    let jsonObject:JSON = try JSON(data:jsonData)
    let jsonDict:[String:JSON] = jsonObject.dictionaryValue
    guard let jsonDataType:String = jsonDict["type"]?.string  else {
        throw ClientError.state(String(format:NSLocalizedString(
            "%@: The MsgType is required.", comment: ""),
                                       "MessageBuilder.build()"))
    }
    var message:Message? = nil
    do {
        switch (jsonDataType.uppercased()) {
        case "DATA":
            let bldr:DataMessageBuilder = DataMessageBuilder()
            _ = try bldr.fromJSON(data:jsonData)
            message = try bldr.build()
            break
        case "REQUEST":
            let bldr:RequestMessageBuilder = RequestMessageBuilder()
            _ = try bldr.fromJSON(data:jsonData)
            message = try bldr.build()
            break
        case "RESPONSE":
            let bldr:ResponseMessageBuilder = ResponseMessageBuilder()
            _ = try bldr.fromJSON(data:jsonData)
            message = try bldr.build()
            break
        case "RESOURCE":
            let bldr:ResponseMessageBuilder = ResponseMessageBuilder()
            _ = try bldr.fromJSON(data:jsonData)
            message = try bldr.build()
            break
        case "ALERT":
            let bldr:AlertMessageBuilder = AlertMessageBuilder()
            _ = try bldr.fromJSON(data:jsonData)
            message = try bldr.build()
            break
        default:
            break
        }
    } catch {
        throw error
    }
    if message == nil {
        throw ClientError.state(String(format:NSLocalizedString(
            "%@: The MsgType `%@` is not valid.", comment: ""),
                                       "MessageBuilder.build()", jsonDataType))
    }
    return message!
}


fileprivate let TABLE_NAME:String = "MESSAGE_PERSISTENT_STORE"

// TODO: Should be in external file
// <tablename>, <maxuuidlength>, <maxendpointidlength>
fileprivate let CREATE_TABLE_SQL:String =
    "CREATE TABLE IF NOT EXISTS %@ " +
    "(TIMESTAMP BIGINT NOT NULL," +
    "UUID VARCHAR(%d) NOT NULL," +
    "ENDPOINT_ID VARCHAR(%d) NOT NULL," +
    "MESSAGE BLOB, PRIMARY KEY (UUID))"

fileprivate let ENDPOINT_ID_INDEX_SQL:String =
    "CREATE INDEX IF NOT EXISTS endpoint_id ON %@ (ENDPOINT_ID)"

// Explicitly list the desired select columns to guarantee the order
// <tablename>
fileprivate let LOAD_TABLE_SQL:String =
    "SELECT TIMESTAMP,UUID,ENDPOINT_ID,MESSAGE FROM %@ " +
    "WHERE ENDPOINT_ID = ? ORDER BY TIMESTAMP"
// The first bound parameter index
fileprivate let LOAD_TABLE_ENDPOINT_ID_IDX:Int32 = 1
// Result set indexes, order of columns in SELECT
fileprivate let LOAD_TABLE_TIMESTAMP_RESULT_IDX:Int32 = 0
fileprivate let LOAD_TABLE_UUID_RESULT_IDX:Int32 = 1
fileprivate let LOAD_TABLE_ENDPOINT_ID_RESULT_IDX:Int32 = 2
fileprivate let LOAD_TABLE_MESSAGE_RESULT_IDX:Int32 = 3
// <tablename>
fileprivate let LOAD_MESSAGE_SQL:String =
    "SELECT MESSAGE FROM %@ " +
    "WHERE ENDPOINT_ID = ? ORDER BY TIMESTAMP"
// The first bound parameter index
fileprivate let LOAD_MESSAGE_ENDPOINT_ID_IDX:Int32 = 1
fileprivate let LOAD_MESSAGE_RESULT_IDX:Int32 = 0

// <tablename>
fileprivate let SAVE_SQL:String =
    "INSERT INTO %@ (TIMESTAMP,UUID,ENDPOINT_ID,MESSAGE) VALUES (?, ?, ?, ?)"
fileprivate let SAVE_TIMESTAMP_IDX:Int32 = 1
fileprivate let SAVE_UUID_IDX:Int32 = 2
fileprivate let SAVE_ENDPOINT_ID_IDX:Int32 = 3
fileprivate let SAVE_MESSAGE_IDX:Int32 = 4

// TODO: Make this delete like BatchByPersistence with where clause using
// "IN" a set of UUID's to delete.
// <tablename>
fileprivate let DELETE_SQL:String = "DELETE FROM %@ WHERE UUID = ?"
fileprivate let DELETE_UUID_IDX:Int32 = 1

/**
 * Support for persisting messages for guaranteed delivery.
 */
class MessagePersistence {
    
    // Default name of the database
    let DB_NAME:String = Config.MESSAGE_PERSISTENCE_DBNAME
    let UUID_LENGTH:Int32 = Config.PERSISTENCE_UUID_LENGTH
    let ENDPOINT_ID_LENGTH:Int32 = Config.PERSISTENCE_ENDPOINT_ID_LENGTH
    private static var PERSISTENCE_ENABLED:Bool {
        get {
            return Config.MESSAGE_PERSISTENCE_ENABLED
        }
    }

    private static var instance:MessagePersistence? = nil
    let writeQueue:DispatchQueue =
        DispatchQueue(label:"MessagePersistence writeQueue",
                      qos:.utility)
    
    let readQueue:DispatchQueue =
        DispatchQueue(label:"BatchByPersistence readQueue",
                      qos:.utility)
    
    private static let getInstanceQueue:DispatchQueue =
        DispatchQueue(label:"MessagePersistence getInstanceQueue",
                      qos:.utility)
    
    public static func getInstance() -> MessagePersistence? {
        if !MessagePersistence.PERSISTENCE_ENABLED {
            return nil
        }
        return MessagePersistence.getInstanceQueue.sync(execute: {
            if MessagePersistence.instance == nil {
                // TODO: Should this be fatal if persistence is enabled
                // but database cannot be created ?
                // instance is set or not in constructor
                MessagePersistence.instance = MessagePersistence()
            }
            return MessagePersistence.instance
        })
    }

    private init?() {
        let schema:String = String(format: CREATE_TABLE_SQL + ";" +
                                           ENDPOINT_ID_INDEX_SQL + ";",
                TABLE_NAME, UUID_LENGTH, ENDPOINT_ID_LENGTH, TABLE_NAME)
        
        let rc = SQLite3_createdb(dbName:DB_NAME, schema:schema)
        if rc != SQLITE_OK {
            log(format:"%@: failed to create %@.",
                args: "MessagePersistence():SQLite3_createdb()", DB_NAME)
            return nil
        }
    }
    
    // used in before deleting db
    static func reset() {
        MessagePersistence.getInstanceQueue.sync() {
            MessagePersistence.instance = nil
        }
    }

    /**
     * Insert a record into the MESSAGE_PERSISTENCE_STORE table and associate
     * it with `endpointId`.
     *
     * - parameter messages: the message to persist
     *
     */
    public func save(messages:[Message], endpointId:String,
                     callback: (()->())? = nil) {
        if !MessagePersistence.PERSISTENCE_ENABLED || messages.isEmpty {
            return
        }
        writeQueue.async() {
            self.readQueue.sync() {
                self._save(messages:messages, endpointId:endpointId)
                if let callback = callback {
                    callback()
                }
            }
        }
    }
        
    func _save(messages:[Message], endpointId:String) {
        /*
        print("### Entering save() ###")
        defer {
            print("### Leaving save() ###")
        }
        */
        
        var db:SQLite3_connection! = nil
        do {
            db = try SQLite3_open(dbName:DB_NAME)
        } catch {
            // WARNING
            log(format: "%@: failed to open %@. %@",
                args: "load():SQLite3_open()", DB_NAME, "\(error)")
            return
        }

        // Create a block so defer isn't called on method exit
        // if open failed
        do {
            let ppStmt:UnsafeMutablePointer<OpaquePointer?> =
                UnsafeMutablePointer<OpaquePointer?>.allocate(capacity: 1)
            ppStmt.pointee = nil
            var rc:SQLite3_returnCode = SQLITE_ERROR
            
            defer {
                // Ensure that db is closed and statement is finalized
                if ppStmt.pointee != nil {
                    rc = sqlite3_finalize(ppStmt.pointee)
                    if rc != SQLITE_OK {
                        log(format:"%@: endpointId %@, failed. (%d) %@",
                            args:"save():sqlite3_finalize()", endpointId, rc,
                                SQLite3_getErrorMsg(db:db))
                    }
                }
                rc = SQLite3_close(db:db)
                if rc != SQLITE_OK {
                    log(format:"%@: endpointId %@, failed. (%d) %@",
                        args:"save():SQLite3_close()", endpointId, rc)
                }
            }
            
            let zSql:String = String(format:SAVE_SQL, TABLE_NAME)
            rc = sqlite3_prepare_v2(db, zSql, -1, ppStmt, nil)
            if rc != SQLITE_OK {
                log(format:"%@: endpointId %@, failed. (%d) %@",
                    args: "save():sqlite3_prepare()", endpointId, rc,
                        SQLite3_getErrorMsg(db:db))
                return
            }

            for message in messages {
                
                // Message with reliability set to NO_GUARANTEE should be
                // never persisted
                if message.getReliability() ==
                        Message.Reliability.no_guarantee {
                    continue
                }
                
                let clientId:String = message.getClientId()
                let timestamp:Int64? = message.getEventTime() ??
                        Int64(Date().timeIntervalSince1970 * 1000.0)
                rc = sqlite3_bind_int64(ppStmt.pointee, SAVE_TIMESTAMP_IDX, timestamp!)
                if rc != SQLITE_OK {
                    log(format: "%@: endpointId %@, clientId %@, failed to bind TIMESTAMP. (%d) %@",
                        args:"save():sqlite3_bind_int64()", endpointId, clientId, rc,
                            SQLite3_getErrorMsg(db:db))
                    continue
                }
                
                guard let _UUID:SQLite3_stringBinding =
                        SQLite3_getStringBinding(value:clientId) else {
                    log(format:"%@: endpointId %@, clientId %@, failed to get binding for clientId %@.",
                        args:"save():SQLite3_getStringBinding()", endpointId, clientId,
                            clientId)
                    continue
                }
                rc = sqlite3_bind_text(ppStmt.pointee, SAVE_UUID_IDX, _UUID,
                                       Int32(_UUID.count), nil)
                if rc != SQLITE_OK {
                    log(format: "%@: endpointId %@, clientId %@, failed to bind UUID. (%d) %@",
                        args:"save():sqlite3_bind_text()", endpointId, clientId, rc,
                        SQLite3_getErrorMsg(db:db))
                    continue
                }
                
                guard let ENDPOINT_ID:SQLite3_stringBinding =
                        SQLite3_getStringBinding(value:endpointId) else {
                    log(format:"%@: endpointId %@, clientId %@, failed to get binding for endpointId %@.",
                        args:"save():SQLite3_getStringBinding()", endpointId, clientId,
                            endpointId)
                    continue
                }
                rc = sqlite3_bind_text(ppStmt.pointee, SAVE_ENDPOINT_ID_IDX, ENDPOINT_ID, Int32(ENDPOINT_ID.count), nil)
                if rc != SQLITE_OK {
                    log(format: "%@: endpointId %@, clientId %@, failed to bind ENDPOINT_ID. (%d) %@",
                        args:"save():sqlite3_bind_text()", endpointId, clientId, rc,
                            SQLite3_getErrorMsg(db:db))
                    continue
                }
                
                var messageJson:String!
                do {
                    messageJson = try message.toJsonString()
                } catch {
                    log(format: "%@: endpointId %@, clientId %@, threw exception. %@",
                        args:"save():toJsonString()", endpointId, clientId, "\(error)")
                    continue
                }
                guard let message:Foundation.Data = messageJson.data(using:.utf8),
                        let MESSAGE:SQLite3_blobBinding =
                            SQLite3_getBlobBinding(value:message) else {
                    log(format:"%@: endpointId %@, clientId %@, failed to get binding for message.",
                        args:"save():SQLite3_getBlobBinding()", endpointId, clientId)
                    continue
                }
                rc = sqlite3_bind_blob(ppStmt.pointee, SAVE_MESSAGE_IDX,
                                       MESSAGE, Int32(MESSAGE.count), nil)
                if rc != SQLITE_OK {
                    log(format: "%@: endpointId %@, clientId %@, failed to bind MESSAGE. (%d) %@",
                        args:"save():sqlite3_bind_blob()", endpointId, clientId, rc,
                            SQLite3_getErrorMsg(db:db))
                        continue
                }
                
                //executing the query to insert values
                rc = sqlite3_step(ppStmt.pointee)
                if rc != SQLITE_DONE {
                    log(format:"%@: endpointId %@, clientId %@, failed. (%d) %@",
                        args: "save():sqlite3_step()", endpointId, clientId, rc,
                            SQLite3_getErrorMsg(db:db))
                    // TODO: Should we just retry and try again ?
                    //continue
                }
                
                rc = sqlite3_reset(ppStmt.pointee)
                if rc != SQLITE_OK {
                    log(format:"%@: endpointId %@, failed after clientId %@. (%d) %@",
                        args:"save():sqlite3_reset()", endpointId, clientId, rc,
                        SQLite3_getErrorMsg(db:db))
                    // Try and finalize the prepared statement and call
                    // prepare again
                    rc = sqlite3_finalize(ppStmt.pointee)
                    if rc != SQLITE_OK {
                        log(format:"%@: endpointId %@, failed after clientId %@ exiting save(), " +
                            "all messages may not have been saved. (%d) %@",
                            args:"save():sqlite3_finalize()", endpointId, clientId, rc,
                            SQLite3_getErrorMsg(db:db))
                        break
                    }
                    rc = sqlite3_prepare_v2(db, zSql, -1, ppStmt, nil)
                    if rc != SQLITE_OK {
                        log(format:"%@: endpointId %@, failed after a failed reset, " +
                            "exiting save() all messages may not have been saved. (%d) %@",
                            args: "save():sqlite3_prepare()", endpointId, rc,
                            SQLite3_getErrorMsg(db:db))
                        ppStmt.pointee = nil
                        break
                    }
                }
                rc = sqlite3_clear_bindings(ppStmt.pointee)
                if rc != SQLITE_OK {
                    log(format:"%@: endpointId %@, clientId %@, failed. (%d) %@",
                        args: "save():sqlite3_clear_bindings()", endpointId,
                            clientId, rc, SQLite3_getErrorMsg(db:db))
                    break
                }
            }
        }
    }
    /*
    let retryQueue:DispatchQueue = DispatchQueue(label: "MessagePersistence retryQueue",
                                                 qos:.utility)
    func retry(callback: @escaping () -> SQLite3_returnCode) -> SQLite3_returnCode {
        var retryCount:Int = 3 // Config.MESSAGE_PERSISTENCE_RETRY_COUNT
        // in ms
        let retryInterval:Int = 10 // Config.MESSAGE_PERSISTENCE_RETRY_INTERVAL
        let retrySemaphore:DispatchSemaphore = DispatchSemaphore(value:0)
        var result:SQLite3_returnCode = SQLITE_ERROR
        while retryCount != 0 {
            retryQueue.asyncAfter(deadline:DispatchTime.now() +
                    .milliseconds(retryInterval)) {
                result = callback()
                retrySemaphore.signal()
            }
            // Add a slight offset to the interval
            let retryResult:DispatchTimeoutResult =
                retrySemaphore.wait(timeout:DispatchTime.now() +
                    .milliseconds(retryInterval + 5))
            switch (retryResult) {
            case .success:
                break
            case .timedOut:
                log(format:"%@: retry timedout, stopping retries",
                    args: "MessagePersistence.retry()")
                break
            }
            if result == SQLITE_BUSY || result == SQLITE_LOCKED {
                retryCount -= 1
                continue
            } else {
                break
            }
        }
        if retryCount == 0 {
            log(format:"%@: retries exceeded",
                args: "MessagePersistence.retry()")
        }
        return result
    }
    */

    
    /**
     * Delete the record associated with `message`. The `clientId` is used
     * as the query to select the record to delete. The `clientId` must
     * match the `UUID` column of the MESSAGE_PERSISTENCE_STORE table.
     *
     * - parameter message: the message to delete
     */
    public func delete(messages:[Message], callback: (()->())? = nil) {
        if !MessagePersistence.PERSISTENCE_ENABLED || messages.isEmpty {
            return
        }
        // Serialize deletes
        writeQueue.async() {
            self._delete(messages:messages)
            if let callback = callback {
                callback()
            }
        }
    }
        
    func _delete(messages:[Message]) {
        /*
        print("### Entering delete() ###")
        defer {
            print("### Leaving delete() ###")
        }
        */
        var db:SQLite3_connection! = nil
        do {
            db = try SQLite3_open(dbName:DB_NAME)
        } catch {
            // WARNING
            log(format: "%@: failed to open %@. %@",
                args: "load():SQLite3_open()", DB_NAME, "\(error)")
            return
        }
        
        let ppStmt:UnsafeMutablePointer<OpaquePointer?> =
            UnsafeMutablePointer<OpaquePointer?>.allocate(capacity: 1)
        ppStmt.pointee = nil
        var rc:SQLite3_returnCode = SQLITE_ERROR
        
        // Create a block so defer isn't called on method exit
        // if open failed
        do {
            
            defer {
                // Ensure that db is closed and statement is finalized
                if ppStmt.pointee != nil {
                    rc = sqlite3_finalize(ppStmt.pointee)
                    if rc != SQLITE_OK {
                        log(format:"%@: failed. (%d) %@",
                            args:"delete():sqlite3_finalize()", rc,
                            SQLite3_getErrorMsg(db:db))
                    }
                }
                rc = SQLite3_close(db:db)
                if rc != SQLITE_OK {
                    log(format:"%@: failed. (%d) %@",
                        args:"save():SQLite3_close()", rc)
                }
            }

            let zSql:String = String(format:DELETE_SQL, TABLE_NAME)
            rc = sqlite3_prepare_v2(db, zSql, -1, ppStmt, nil)
            if rc != SQLITE_OK {
                log(format:"%@: initial prepare failed. (%d) %@",
                    args: "delete():sqlite3_prepare()", rc,
                        SQLite3_getErrorMsg(db:db))
                return
            }
            
            for message in messages {
                // Message with reliability set to NO_GUARANTEE should be 
                // never persisted
                if message.getReliability() == 
                        Message.Reliability.no_guarantee {
                    continue
                }
                
                let clientId:String = message.getClientId()
                guard let _UUID:SQLite3_stringBinding =
                        SQLite3_getStringBinding(value:clientId) else {
                    log(format:"%@: failed to get binding for clientId %@.",
                        args:"delete():SQLite3_getStringBinding()",
                            clientId)
                    continue
                }
                rc = sqlite3_bind_text(ppStmt.pointee, DELETE_UUID_IDX, _UUID,
                                       Int32(_UUID.count), nil)
                if rc != SQLITE_OK {
                    log(format: "%@: clientId %@, failed to bind UUID. (%d) %@",
                        args:"delete():sqlite3_bind_text()", clientId, rc,
                            SQLite3_getErrorMsg(db:db))
                    continue
                }
                
                rc = sqlite3_step(ppStmt.pointee)
                if rc != SQLITE_DONE {
                    log(format:"%@: failed for clientId %@. (%d) %@",
                        args:"delete():sqlite3_step()", clientId, rc,
                            SQLite3_getErrorMsg(db:db))
                    // TODO: Should we just reset and try again ?
                    // Doesn't look like you can, reset just returns the
                    // same error
                    break
                }
                rc = sqlite3_reset(ppStmt.pointee)
                if rc != SQLITE_OK {
                    log(format:"%@: failed after clientId %@. (%d) %@",
                        args:"delete():sqlite3_reset()", clientId, rc,
                            SQLite3_getErrorMsg(db:db))
                    // Try and finalize the prepared statement and call
                    // prepare again
                    rc = sqlite3_finalize(ppStmt.pointee)
                    if rc != SQLITE_OK {
                        log(format:"%@: failed after id %@ exiting delete, " +
                                "all messages may not have been deleted. (%d) %@",
                            args:"delete():sqlite3_finalize()", clientId, rc,
                                SQLite3_getErrorMsg(db:db))
                        break
                    }
                    
                    rc = sqlite3_prepare_v2(db, zSql, -1, ppStmt, nil)
                    if rc != SQLITE_OK {
                        log(format:"%@: failed after a failed reset, " +
                            "exiting delete all messages may not have been deleted. (%d) %@",
                            args: "delete():sqlite3_prepare()", rc,
                            SQLite3_getErrorMsg(db:db))
                        ppStmt.pointee = nil
                        break
                    }
                }
                rc = sqlite3_clear_bindings(ppStmt.pointee)
                if rc != SQLITE_OK {
                    log(format:"%@: clientId %@, failed. (%d) %@",
                        args: "delete():sqlite3_clear_bindings()", clientId, rc,
                            SQLite3_getErrorMsg(db:db))
                    break
                }
            }
        }
    }
    
    /**
     * Perform a query based on the `endpointId`. The `endpointId` must
     * match the `ENDPOINT_ID` column of the MESSAGE_PERSISTENCE_STORE table.
     *
     * - parameter endpointId: The query parameter associated with the message
     *                         record
     *
     * - returns: an array of `Message` or nil if an error occurs performing
     *            the query
     */
    public func load(endpointId:String) -> [Message]? {
        if !MessagePersistence.PERSISTENCE_ENABLED {
            return nil
        }
        return readQueue.sync(execute: {
            return _load(endpointId:endpointId)
        })
    }
    
    func _load(endpointId:String) -> [Message]? {
        /*
        print("### Entering load() ###")
        defer {
            print("### Leaving laod() ###")
        }
        */
        var db:SQLite3_connection! = nil
        do {
            db = try SQLite3_open(dbName:DB_NAME)
        } catch {
            // WARNING
            log(format: "%@: failed to open %@. %@",
                args: "load():SQLite3_open()", DB_NAME, "\(error)")
            return nil
        }
        
        let ppStmt:UnsafeMutablePointer<OpaquePointer?> =
            UnsafeMutablePointer<OpaquePointer?>.allocate(capacity: 1)
        ppStmt.pointee = nil
        var rc:SQLite3_returnCode = SQLITE_ERROR
        
        // Create a block so defer isn't called on method exit
        // if open failed
        do {
            
            defer {
                // Ensure that db is closed and statement is finalized
                if ppStmt.pointee != nil {
                    rc = sqlite3_finalize(ppStmt.pointee)
                    if rc != SQLITE_OK {
                        log(format:"%@: failed. (%d) %@",
                            args:"load():sqlite3_finalize()", rc,
                                 SQLite3_getErrorMsg(db:db))
                    }
                }
                rc = SQLite3_close(db:db)
                if rc != SQLITE_OK {
                    log(format:"%@: failed. (%d) %@",
                        args:"load():SQLite3_close()", rc)
                }
            }
            
            let zSql:String = String(format:LOAD_MESSAGE_SQL, TABLE_NAME)
            rc = sqlite3_prepare_v2(db, zSql, -1, ppStmt, nil)
            if rc != SQLITE_OK {
                log(format:"%@: failed:(%d) %@",
                    args: "load()", rc, SQLite3_getErrorMsg(db:db))
                return nil
            }
            guard let ENDPOINT_ID:SQLite3_stringBinding =
                    SQLite3_getStringBinding(value:endpointId) else {
                log(format:"%@: failed for %@.",
                    args: "load():SQLite3_getStringBinding()",
                          endpointId)
                return nil
            }
            rc = sqlite3_bind_text(ppStmt.pointee, LOAD_MESSAGE_ENDPOINT_ID_IDX,
                                   ENDPOINT_ID,
                                   Int32(ENDPOINT_ID.count), nil)
            if rc != SQLITE_OK {
                log(format:"%@: failed for %@. (%d) %@",
                    args:"load():sqlite3_bind_text()", endpointId, rc,
                        SQLite3_getErrorMsg(db:db))
                return nil
            }
            
            var messages:[Message] = [Message]()
            var rows:Int = 0
            var haveRows:Bool = true
            while haveRows {
                var messageJson:Foundation.Data? = nil
                rc = sqlite3_step(ppStmt.pointee)
                if rc == SQLITE_ROW  {
                    // Only MESSAGE is needed
                    /*
                    let TIMESTAMP:Int64 =
                        sqlite3_column_int64(ppStmt.pointee, TIMESTAMP_RESULT_IDX)
                    let _UUID:String =
                        String(cString:sqlite3_column_text(ppStmt.pointee,
                                                           UUID_RESULT_IDX))
                    let ENDPOINT_ID:String =
                        String(cString:sqlite3_column_text(ppStmt.pointee,
                                                           ENDPOINT_ID_RESULT_IDX))
                    */
                    
                    let MESSAGE:UnsafeRawPointer? =
                        sqlite3_column_blob(ppStmt.pointee, LOAD_MESSAGE_RESULT_IDX)
                    if MESSAGE == nil {
                        log(format:"%@: endpointId %@, row %d, MESSAGE is zero length blob",
                            args:"load():sqlite3_column_blob()", endpointId, rows)
                    } else {
                        let length = Int(sqlite3_column_bytes(ppStmt.pointee,
                                                              LOAD_MESSAGE_RESULT_IDX))
                        let bufptr = UnsafeBufferPointer(start:
                            MESSAGE!.assumingMemoryBound(to: UInt8.self), count: length)
                        messageJson = Foundation.Data(buffer:bufptr)
                    }
                } else if rc != SQLITE_DONE {
                    log(format:"%@: failed on row %d. (%d) %@",
                        args:["load():sqlite3_step()", rows, rc,
                            SQLite3_getErrorMsg(db:db)])
                }
                haveRows = rc == SQLITE_ROW
                rows += 1
                if let messageJson = messageJson {
                    do {
                        // TODO: Is there ever more than one message in the blob ?
                        // Why return array ?
                        // TODO: THe calls in Message and MessageBuilder are messed up.
                        // duplicate methods that that make no sense. Also they are
                        // not consistent with respect to JCL, in that Message has
                        // some methods and MessageBuilder has others.
                        let message:Message = try getMessage(jsonData:messageJson)
                        messages.append(message)
                    } catch {
                        log(format:"%@: endpointId %@, row %d, threw exception. %@",
                            args:"load():Message.fromJSON()", endpointId, rows,
                            "\(error)")
                    }
                }
            }
            return messages
        }
        

    }

    private func log(format:String, args:CVarArg...) {
        print(logMsg(format:format, args:args))
    }
    private func logMsg(format:String, args:[CVarArg]) -> String {
        return String(format:NSLocalizedString(format, comment:""), arguments: args)
    }
}

