/*
 * Copyright (c) 2017, 2018, Oracle and/or its affiliates.  All rights reserved.
 *
 * This software is dual-licensed to you under the MIT License (MIT) and 
 * the Universal Permissive License (UPL).  See the LICENSE file in the root
 * directory for license terms.  You may choose either license, or both.
 */

// com.oracle.iot.client.impl.device
import Foundation
import os

/**
 * MessagingPolicyImpl is the code that supports the `DirectlyConnectedDevice`
 * and the `MessageDispatcher` `offer(Message...)`
 * methods. This class uses the devicePolicy REST API
 * 
 * The `DirectlyConnectedDevice.offer(Message...)`
 * API invokes the `applyPolicies(Message)` method and then calls
 * `DirectlyConnectedDevice.send(Message...)` * with the resulting messages.
 * 
 * The `MessageDispatcher.offer(Message...)` API invokes the
 * `applyPolicies(Message)` method and then calls
 * `MessageDispatcher.queue(Message...)` with the resulting messages.
 */
public class MessagingPolicyImpl : DevicePolicyManagerChangeListener {

    private let directlyConnectedDevice:AdvancedDirectlyConnectedDevice

    // Data pertaining to this virtual device and its attributes for computing 
    // policies. The key is attribute name (or nil for the device model 
    // policies), value is a list. Each element in the list corresponds to a 
    // function in the pipeline for the attribute, and the map is used by the 
    // function to hold data between calls.
    // Note that there is a 1:1 correspondence between the pipeline 
    // configuration data for the attribute, and the pipeline data for the
    // attribute.
    private let pipelineDataCacheQueue:DispatchQueue =
        DispatchQueue(label:"pipelineDataCache queue [messagingPolicyImpl]",
                      qos:.utility)
    private var 
        pipelineDataCache:Dictionary<String,[Dictionary<String,AnyObject>]> =
            Dictionary<String,[Dictionary<String,AnyObject>]>()
    private func getPipelineDataCache(attribute:String) -> [Dictionary<String,AnyObject>]? {
        return pipelineDataCacheQueue.sync(execute: {
            return pipelineDataCache[attribute]
        })
    }
    private func putPipelineDataCache(attribute:String,
                                      pipelineData:[Dictionary<String,AnyObject>]) {
        pipelineDataCacheQueue.sync() {
            pipelineDataCache[attribute] = pipelineData
        }
    }

    // deviceModelUrn -> DeviceAnalog
    // We need more than one DeviceAnalog because the DCD can have more than 
    // one device model.
    private let deviceAnalogMapQueue:DispatchQueue =
        DispatchQueue(label:"deviceAnalogMap queue", qos:.utility)
    private var deviceAnalogMap:Dictionary<String,DeviceAnalog> =
            Dictionary<String,DeviceAnalog>()

    // deviceModelUrn:attribute:deviceFunctionId -> start time of last window
    // For a window policy, this maps the policy target plus the function to
    // when the window started. When the attribute for a timed function is in
    // the message, we can compare this start time to the elapsed time to
    // determine if the window has expired. If the window has expired, the value
    // computed by the function is passed to the remaining functions in the 
    // pipeline.
    //
    private let windowMapQueue:DispatchQueue =
        DispatchQueue(label:"windowMap queue", qos:.utility)
    private var windowMap:Dictionary<String,Int64> =
            Dictionary<String,Int64>()

    // Key is device model urn, value is attribute -> trigger attributes
    // (a trigger attribute is a referenced attribute in a computedMetric 
    // formula).
    private let computedMetricTriggersQueue:DispatchQueue =
        DispatchQueue(label:"computedMetricTriggers queue", qos:.utility)
    private var computedMetricTriggers:
        Dictionary<String,Dictionary<String,Set<String>>> =
            Dictionary<String,Dictionary<String,Set<String>>>()

    private let messagesFromExpiredPoliciesQueue:DispatchQueue =
        DispatchQueue(label:"messagesFromExpiredPolicies queue", qos:.utility)
    private var messagesFromExpiredPolicies:[Message] = [Message]()

    /**
     * Constructed in com.oracle.iot.client.device.DirectlyConnectedDevice
     *
     * - parameter directlyConnectedDevice: the DirectlyConnectedDevice
     */
    public init(directlyConnectedDevice:AdvancedDirectlyConnectedDevice) {
        self.directlyConnectedDevice = directlyConnectedDevice
    }

    /**
     * This is the method that applies whatever policies there may be to 
     * the message. The method returns zero or more messages, depending on 
     * the policies that have been
     * applied to the message. The caller is responsible for sending or 
     * queuing the returned messages. The data items in the returned are 
     * messages are possibly modified  by some policy; for example, a message 
     * with a temperature value goes in, a copy of the same message is 
     * returned with the temperature value replaced by the avearage 
     * temperature. A returned message may also be one that is created by a
     * policy function (such as a computedMetric). Or the returned messages 
     * may be messages that have been batched. If no policy applies to the 
     * message, the message itself is returned.
     *
     * - parmeter message: A message of any kind
     * 
     * - returns: messages to be delivered
     *
     * - throws: `ClientError.network` possibly thrown during network 
     *                               communications
     *      `ClientError.security`: possibly thrown accessing or using the
     *                              secure connection
     */
    public func applyPolicies(message:Message?) throws -> [Message] {

        guard let message = message else {
            return [Message]()
        }

        // Pass the current millisecond time down so that functions that have the same
        // or intersecting timeout get expired at the same time. This ensures the attributes
        // are grouped.
        let currentTimeMillis:Int64 = getCurrentTimeMillis()
        
        var resultingMessages:[Message] = [Message]()
        if message.getType() == Message.MsgType.DATA {
            if let dataMessage:DataMessage =
                    try applyAttributePolicies(dataMessage:message as! DataMessage,
                                               currentTimeMillis: currentTimeMillis) {
                resultingMessages.append(dataMessage)
            }
        } else {
            resultingMessages.append(message)
        }

        var messageList:[Message] = [Message]()
        if messagesFromExpiredPolicies.count > 0 {
            messagesFromExpiredPoliciesQueue.sync() {
                messageList.append(contentsOf:messagesFromExpiredPolicies)
                messagesFromExpiredPolicies.removeAll(keepingCapacity:false)
            }
        }
        for m in resultingMessages {
            if let messagesFromDevicePolicy:[Message] =
                    try applyDevicePolicies(message:m,
                                            currentTimeMillis: currentTimeMillis) {
                messageList.append(contentsOf:messagesFromDevicePolicy)
            }
        }
        return messageList
    }

    // ChangeListener methods
    override
    public func policyAssigned(devicePolicy:DevicePolicy,
                               assignedDevices:Set<String>) throws {
        // do nothing
    }
    
    override
    public func policyUnassigned(devicePolicy: DevicePolicy,
                                 unassignedDevices: Set<String>) throws {
        
        let currentTimeMillis:Int64 = getCurrentTimeMillis()
        let messages:[Message] =
                try expirePolicy(devicePolicy:devicePolicy,
                                   currentTimeMillis: currentTimeMillis)
        if !messages.isEmpty {
            messagesFromExpiredPoliciesQueue.sync() {
                messagesFromExpiredPolicies.append(contentsOf:messages)
            }
        }

        // TODO:  Need to figure out how to handle accumulated values.
        //        For now, just clear out the various maps, which
        //        effectively means "start from scratch"
        deviceAnalogMapQueue.sync() {
            deviceAnalogMap.removeAll(keepingCapacity:false)
        }
        pipelineDataCacheQueue.sync() {
            pipelineDataCache.removeAll(keepingCapacity:false)
        }
        computedMetricTriggersQueue.sync() {
            computedMetricTriggers.removeAll(keepingCapacity:false)
        }
        windowMapQueue.sync() {
            windowMap.removeAll(keepingCapacity:false)
        }
    }

    // Apply policies that are targeted to an attribute
    private func applyAttributePolicies(dataMessage:DataMessage,
                                        currentTimeMillis:Int64)
            throws -> DataMessage? {

        // A data message format cannot be nil or empty
        // (enforced in DataMessage(Builder) constructor
        let format:String = dataMessage.getFormat()
        let match:Range? = format.range(of:":attributes")
        let deviceModelUrn:String =
            String(format[format.startIndex..<match!.lowerBound])

        // TODO: implement callback
        var deviceModel:DeviceModelImpl? = nil
        var dm:DeviceModel? = nil
        // TODO: make asynchronous
        let getDeviceModelSemaphore:DispatchSemaphore =
            DispatchSemaphore(value:0)
        try directlyConnectedDevice.getDeviceModel(deviceModelUrn:deviceModelUrn,
                callback: { deviceModel, error in
                    if error != nil {
                        log(type:OSLogType.fault,
                            message:"MessagingPolicyImpl could not get device model for %@",
                            args:deviceModelUrn)
                    }
                    dm = deviceModel
                    getDeviceModelSemaphore.signal()
        })
        // TODO: wait with timeout
        getDeviceModelSemaphore.wait()
        if dm != nil && dm is DeviceModelImpl {
            deviceModel = dm as? DeviceModelImpl
        } else {
            return dataMessage
        }

        let endpointId:String = dataMessage.getSource()
        var deviceAnalog:DeviceAnalog? = nil
        deviceAnalogMapQueue.sync() {
            deviceAnalog = deviceAnalogMap[endpointId]
            if deviceAnalog == nil {
                deviceAnalog = DeviceAnalogImpl(
                    directlyConnectedDevice:directlyConnectedDevice,
                    deviceModel:deviceModel!, endpointId:endpointId)
                deviceAnalogMap[endpointId] = deviceAnalog
            }
        }
        var triggerMap:Dictionary<String,Set<String>>? = nil
        computedMetricTriggersQueue.sync() {
            triggerMap = computedMetricTriggers[deviceModelUrn]
        }

        if triggerMap == nil {
            triggerMap = Dictionary<String,Set<String>>()
            // TODO: Cannot add triggerMap until we are done adding
            // to it
            //computedMetricTriggers[deviceModelUrn] = triggerMap

            let deviceModelAttributeMap:Dictionary<String,DeviceModelAttribute>
                = deviceModel!.getDeviceModelAttributes()

            for (_,v) in deviceModelAttributeMap {
                let deviceModelAttribute:DeviceModelAttribute = v
                let attribute:String = deviceModelAttribute.getName()!

                let devicePolicyManager:DevicePolicyManager? =
                    DevicePolicyManager.getDevicePolicyManager(
                        dcd:directlyConnectedDevice)
                // TODO: Requires callback, doesn not return value
                let devicePolicy:DevicePolicy? =
                    try devicePolicyManager!.getPolicy(
                        deviceModelURN:deviceModelUrn,
                        deviceId:endpointId
                        /*,callback: { (policy,error) in }*/)
                if devicePolicy == nil {
                    continue
                }

                guard let pipeline:[DevicePolicy.Function] =
                        devicePolicy!.getPipeline(attributeName:attribute),
                        !pipeline.isEmpty else {
                    continue
                }

                // If a computedMetric is the first function in the pipeline,
                // then see if the formula refers to other attributes. If so,
                // then try this pipeline after all others.
                let function:DevicePolicy.Function = pipeline[0]
                let deviceFunctionId:String = function.getId()
                let parameters:Dictionary<String,Any> =
                    function.getParameters()
                // TODO: Use manifest constant not literal function name.
                if "computedMetric" == deviceFunctionId {
                    // TODO: Duplicate code in VirtualDeviceImpl
                    // TODO: Can this be nil ?
                    if var formula:String = parameters["formula"] as? String {
                        // If there is no formula or embedded variables
                        // then there aren't any triggerAttributes
                        var triggerAttributes:Set<String>? = nil
                        // Extract attributes where the syntax is $(attr)
                        // Range .upperBound points to after the "("
                        var attrvar:String? = nil
                        // If ther are no attributes, do nothing
                        var attrBegin = formula.range(of: "$(")
                        // Don't create triggerAttributes unles there is at least
                        // one embedded attribute
                        if attrBegin != nil {
                            triggerAttributes = Set<String>()
                        }
                        while attrBegin != nil {
                            // point formula to after "$("
                            attrvar = String(formula[attrBegin!.upperBound..<formula.endIndex])
                            // Find the end of attribute, nil if no ")"
                            if attrvar != nil {
                                if let attrEnd = attrvar!.range(of:")") {
                                    // the rest of formula after the attribute
                                    formula = String(attrvar![attrEnd.upperBound..<attrvar!.endIndex])
                                    attrvar = String(attrvar![attrvar!.startIndex..<attrEnd.lowerBound])
                                    if attrvar != nil && attrvar != attribute {
                                        triggerAttributes!.update(with:attrvar!)
                                    }
                                }
                            }
                            attrBegin = formula.range(of: "$(")
                        }
                        // If triggerAttributes is not nil, there it will not be
                        // empty unless the embedded attribute equals attributeName
                        if triggerAttributes != nil && !triggerAttributes!.isEmpty {
                            triggerMap![attribute] = triggerAttributes
                        }
                    }
                }
            }
            // Always add triggerMap even if empty
            computedMetricTriggersQueue.sync() {
                computedMetricTriggers[deviceModelUrn] = triggerMap
            }
        }

        // getDataItems returns an unmodifiable list
        let dataMessageDataItems:[DataItem] = dataMessage.getDataItems()

        // DataItems resulting from policies
        var policyDataItems:[DataItem] = [DataItem]()

        // DataItems that have no policies
        var skippedDataItems:[DataItem] = [DataItem]()

        // If no policies are found, we will return the original message
        var noPoliciesFound:Bool = true

        for dataItem in dataMessageDataItems {

            let attribute:String = dataItem.getKey()
            let deviceModelAttribute:DeviceModelAttribute? =
                    deviceModel?.getDeviceModelAttributes()[attribute]

            if deviceModelAttribute == nil {
                log(type: OSLogType.info, message: "%@ does not contain attribute %@",
                    args: deviceModel!.getURN(), attribute)
                skippedDataItems.append(dataItem)
                continue
            }

            let devicePolicyManager:DevicePolicyManager? =
                DevicePolicyManager.getDevicePolicyManager(
                        dcd:directlyConnectedDevice)
            // TODO: func does not return value use callback
            let devicePolicy:DevicePolicy? =
                try devicePolicyManager!.getPolicy(deviceModelURN:deviceModelUrn,
                        deviceId:endpointId
                        /*,callback: { (policy,error) in }*/)
            if devicePolicy == nil {
                try deviceAnalog!.setAttributeValue(attribute:attribute,
                    value:cast(attributeType:deviceModelAttribute!.getType(),
                            newValue:dataItem.getValue()))
                skippedDataItems.append(dataItem)
                continue
            }

            // No policies for this attribute, retain the data item.
            guard let pipeline:[DevicePolicy.Function] =
                    devicePolicy!.getPipeline(attributeName:attribute),
                    !pipeline.isEmpty else {
                try deviceAnalog!.setAttributeValue(attribute:attribute,
                    value:cast(attributeType:deviceModelAttribute!.getType(),
                            newValue:dataItem.getValue()))
                skippedDataItems.append(dataItem)
                continue
            }

            noPoliciesFound = false

            // if this is a computed metric, skip it for now
            if triggerMap![attribute] != nil {
                continue
            }

            if let policyDataItem:DataItem =
                    try applyAttributePolicy(deviceAnalog:deviceAnalog!,
                                         dataItem:dataItem, pipeline:pipeline,
                                         currentTimeMillis:currentTimeMillis) {
                policyDataItems.append(policyDataItem)
            }
        }

        // if no policies were found, return the original message
        if noPoliciesFound {
            return dataMessage
        }

        // if policies were found, but there are no policyDataItems
        // and no skipped data items, then return nil.
        if policyDataItems.isEmpty && skippedDataItems.isEmpty {
            return nil
        }

        // This looks like a good place to check for computed metrics, too.
        if !policyDataItems.isEmpty {
            try checkComputedMetrics(dataItems:&policyDataItems,
                deviceAnalog:deviceAnalog!, triggerMap:triggerMap!,
                currentTimeMillis:currentTimeMillis)
        }

        // Since we can't modify the original data message, we have to
        // create a copy.
        let dataMessageBuilder:DataMessageBuilder = DataMessageBuilder()
        dataMessageBuilder.setFormat(format:format)
        dataMessageBuilder.setClientId(clientId:dataMessage.getClientId())
        dataMessageBuilder.setSource(source:dataMessage.getSource())
        dataMessageBuilder.setDestination(destination:dataMessage.getDestination())
        dataMessageBuilder.setPriority(priority:dataMessage.getPriority())
        dataMessageBuilder.setReliability(reliability:dataMessage.getReliability())
        dataMessageBuilder.setEventTime(eventTime:dataMessage.getEventTime()!)
        dataMessageBuilder.setProperties(properties:dataMessage.getProperties())
        dataMessageBuilder.setSender(sender:dataMessage.getSender())

        let diagnostics:Dictionary<String,AnyObject> = dataMessage.getDiagnostics()
        for (k,v) in diagnostics {
            dataMessageBuilder.setDiagnostic(name:k, value:v)
        }

        dataMessageBuilder.addDataItems(dataItems:policyDataItems)
        dataMessageBuilder.addDataItems(dataItems:skippedDataItems)

        return try dataMessageBuilder.build()
    }

    // Return DataItem if it should be included in the Message, nil if it 
    // should not be.
    private func applyAttributePolicy(deviceAnalog:DeviceAnalog,
             dataItem:DataItem,
             pipeline:[DevicePolicy.Function],
             currentTimeMillis:Int64) throws -> DataItem? {

        let attribute:String = dataItem.getKey()
        let deviceModel:DeviceModelImpl = 
            deviceAnalog.getDeviceModel() as! DeviceModelImpl

        var policyValue:AnyObject? = dataItem.getValue()

        // Create the pipeline data for this attribute
        // TODO: Can attribute be nil here ?
        var pipelineData:[Dictionary<String,AnyObject>] =
            getPipelineDataCache(attribute:attribute) ?? [Dictionary<String,AnyObject>]()
    
        // SE does this in the for loop
        // create missing function maps in the pipeline data
        if pipelineData.count < pipeline.count {
            // create missing function data maps
            for _ in pipelineData.count..<pipeline.count {
                pipelineData.append(Dictionary<String,AnyObject>())
            }
        }
        
        // Add back pipelineData after modification
        putPipelineDataCache(attribute:attribute, pipelineData:pipelineData)

        DeviceFunction.putInProcessValue(
                endpointId:deviceAnalog.getEndpointId(), 
                deviceModelURN:deviceModel.getURN(),
                attribute:attribute, value:policyValue)
        
        for index in pipeline.startIndex..<pipeline.endIndex {

            let function:DevicePolicy.Function = pipeline[index]
            var functionData:Dictionary<String,AnyObject> = pipelineData[index]
            let key:String = function.getId()
            let parameters:Dictionary<String,Any> = function.getParameters()
            let deviceFunction:DeviceFunction? =
                    DeviceFunction.getDeviceFunction(functionId:key)
            if deviceFunction == nil {
                continue
            }

            var windowExpired:Bool = false
            // window in milliseconds
            let window:Int64 = 
                DeviceFunction.getWindow(configuration:parameters)
            let slide:Int64 = DeviceFunction.getSlide(configuration:parameters,
                                                      window:window)
            if window > 0 {

                // This could be more succinct, but it makes the key easy to
                // read in the debugger.
                let k:String = deviceModel.getURN() + ":" +
                               attribute + ":" + deviceFunction!.getId() + ".expiry"
                var expiry:Int64? = windowMapQueue.sync(execute:{
                    return windowMap[k]
                })
                if expiry == nil {
                    expiry = currentTimeMillis + window
                    windowMapQueue.sync() {
                        windowMap[k] = expiry!
                    }
                }

                windowExpired = expiry! <= currentTimeMillis
                if windowExpired {
                    windowMapQueue.sync() {
                        windowMap[k] = expiry! + slide
                    }
                }
            }
            
            if deviceFunction!.apply(deviceAnalog:deviceAnalog,
                    attribute:attribute,
                    configuration:parameters,
                    data:&functionData, value:policyValue) || windowExpired {
                
                // Add back pipelineData
                pipelineData[index] = functionData
                putPipelineDataCache(attribute:attribute,
                                     pipelineData:pipelineData)
                
                if let valueFromPolicy:AnyObject = deviceFunction?.get(
                        deviceAnalog:deviceAnalog,
                        attribute:attribute,
                        configuration:parameters,
                        data:&functionData) {

                    policyValue = valueFromPolicy
                    DeviceFunction.putInProcessValue(
                        endpointId:deviceAnalog.getEndpointId(), 
                        deviceModelURN:deviceModel.getURN(),
                        attribute:attribute, 
                        value:policyValue!)
                } else {
                    // WARNING
                    log(type:OSLogType.fault, 
                        message:"%@ got nil value from policy %@",
                        args: attribute,
                              deviceFunction!.getDetails(
                                  configuration:parameters))
                    break
                }
                // Add back pipelineData
                pipelineData[index] = functionData
                putPipelineDataCache(attribute:attribute,
                                     pipelineData:pipelineData)

            } else {
                // Add back pipelineData
                pipelineData[index] = functionData
                putPipelineDataCache(attribute:attribute,
                                     pipelineData:pipelineData)
                // apply returned false
                policyValue = nil
                break
            }
        }

        // After the policy loop, if the policyValue is nil, then the policy
        // either filtered out the attribute, or the policy parameters have not
        // been met (e.g., sampleQuality rate is not met).
        // If it is not nil, then create a DataItem to replace the old
        // in the data message.
        var policyDataItem:DataItem? = nil
        if policyValue != nil {
            // TODO: Can this nil ?
            let deviceModelAttribute:DeviceModelAttribute =
                deviceModel.getDeviceModelAttributes()[attribute]!
            try deviceAnalog.setAttributeValue(attribute:attribute,
                value:cast(attributeType:deviceModelAttribute.getType(),
                           newValue:policyValue!))
            policyDataItem = try createDataItem(type:dataItem.getType(),
                                            key:dataItem.getKey(),
                                            newValue:policyValue!)
        } else {
            policyDataItem = nil
        }

        _ = DeviceFunction.removeInProcessValue(
            endpointId:deviceAnalog.getEndpointId(), 
            deviceModelURN:deviceModel.getURN(),
            attribute:attribute)

        return policyDataItem
    }

    private func checkComputedMetrics(dataItems:inout [DataItem],
            deviceAnalog:DeviceAnalog,
            triggerMap:Dictionary<String,Set<String>>,
            currentTimeMillis:Int64) throws {

        if triggerMap.isEmpty || dataItems.isEmpty {
            return
        }

        var updatedAttributes:Set<String> = Set<String>()
        for dataItem in dataItems {
            updatedAttributes.update(with:dataItem.getKey())
        }

        let endpointId:String = deviceAnalog.getEndpointId()
        let deviceModel:DeviceModelImpl = 
            deviceAnalog.getDeviceModel() as! DeviceModelImpl
        let deviceModelAttributes:Dictionary<String,DeviceModelAttribute> =
            deviceModel.getDeviceModelAttributes()
        let deviceModelUrn:String = deviceModel.getURN()

        for (k,v) in triggerMap {

            // This is the set of attributes that the formula refers to.
            let key:Set<String> = v

            // If the set of attributes that the formula refers to
            // is a subset of the updated attributes, then compute
            // the value of the computedMetric.
            if !updatedAttributes.isSubset(of:key) {
                continue
            }

            let attribute:String = k
            // TODO: Can this be nil ?
            let deviceModelAttribute:DeviceModelAttribute =
                deviceModelAttributes[attribute]!
            var attributeValue:AnyObject? =
                try deviceAnalog.getAttributeValue(attribute:attribute)
            if attributeValue == nil {
                attributeValue = deviceModelAttribute.getDefaultValue()
            }

            var dataItem:DataItem? = nil
            switch (deviceModelAttribute.getType()) {
                case .NUMBER, .INTEGER:
                    dataItem = try DataItem(key:attribute,
                            value: attributeValue as? Double ?? 0.0)
                    break
                case .STRING, .URI:
                    dataItem = try DataItem(key:attribute,
                            value:attributeValue as? String ?? "")
                    break
                case .BOOLEAN:
                    dataItem = try DataItem(key:attribute,
                            value:attributeValue as? Bool ?? false)
                    break
            case .DATETIME:
                    var value:Double? = nil
                    if attributeValue is Date {
                        value = Date().timeIntervalSince1970
                    } else {
                        value = attributeValue as? Double ?? 0.0
                    }
                    dataItem = try DataItem(key:attribute, value:value!)
                    break
            }

            let devicePolicyManager:DevicePolicyManager? =
                    DevicePolicyManager.getDevicePolicyManager(
                            dcd:directlyConnectedDevice)
            // TODO: requires callback does not return value
            let devicePolicy:DevicePolicy? =
                try devicePolicyManager?.getPolicy(deviceModelURN:deviceModelUrn,
                                deviceId:endpointId
                    /*,callback: { (policy,error) in }*/)
            if devicePolicy == nil {
                continue
            }

            guard let pipeline:[DevicePolicy.Function] =
                    devicePolicy!.getPipeline(attributeName:attribute),
                    !pipeline.isEmpty else {
                continue
            }

            if let policyDataItem:DataItem =
                    try applyAttributePolicy(deviceAnalog:deviceAnalog,
                        dataItem:dataItem!, pipeline:pipeline,
                        currentTimeMillis: currentTimeMillis) {
                dataItems.append(policyDataItem)
            }
        }
    }

    // Apply policies that are targeted to a device model
    private func applyDevicePolicies(message:Message,
                                currentTimeMillis:Int64) throws -> [Message]? {

        // A data message or alert format cannot be nil or empty
        // (enforced in Data/AlertMessage(Builder) constructor)
        var format:String!
        var deviceModelUrn:String? = nil
        let endpointId:String = message.getSource()
        if message is DataMessage {
            format = (message as! DataMessage).getFormat()
            let match:Range? = format.range(of:":attributes")
            deviceModelUrn = String(format[format.startIndex..<match!.lowerBound])
        } else if message is AlertMessage {
            format = (message as! AlertMessage).getFormat()
            let match:Range? = format.range(of:":",
                                   options:String.CompareOptions.backwards, 
                                   range:nil, locale:nil)
            deviceModelUrn = String(format[format.startIndex..<match!.lowerBound])
        } else {
            return [message]
        }

        var deviceAnalog:DeviceAnalog? =
            deviceAnalogMapQueue.sync(execute:{
                return deviceAnalogMap[endpointId]
            })
        if deviceAnalog == nil {
            var deviceModel:DeviceModel? = nil
            let deviceModelSemaphore:DispatchSemaphore =
                DispatchSemaphore(value:0)
            try directlyConnectedDevice.getDeviceModel(deviceModelUrn:deviceModelUrn!,
                callback: {(dm, error) in
                    if error != nil {
                        log(type:OSLogType.fault,
                            message:"applyDevicePolicies can not get device model for %@",
                            args: deviceModelUrn!)
                    }
                    deviceModel = dm
                    deviceModelSemaphore.signal()
            })
            // What else could it be ?
            if let deviceModel = deviceModel, deviceModel is DeviceModelImpl {
                deviceAnalog = DeviceAnalogImpl(
                        directlyConnectedDevice:directlyConnectedDevice, 
                        deviceModel:(deviceModel as! DeviceModelImpl),
                        endpointId: endpointId)
                deviceAnalogMapQueue.sync() {
                    deviceAnalogMap[endpointId] = deviceAnalog
                }
            }
            if deviceAnalog == nil {
                return [message]
            }
        }

        let devicePolicyManager:DevicePolicyManager? =
            DevicePolicyManager.getDevicePolicyManager(dcd:directlyConnectedDevice)
        // TODO: Does not return must use callback
        guard let devicePolicy:DevicePolicy =
            try devicePolicyManager?.getPolicy(deviceModelURN:deviceModelUrn!,
                        deviceId:endpointId
                        /*,callback:{ (policy,error) in }*/) else {
            return [message]
        }

        guard let pipeline:[DevicePolicy.Function] =
                devicePolicy.getPipeline(attributeName:DevicePolicy.ALL_ATTRIBUTES()),
                !pipeline.isEmpty else {
            return [message]
        }
        
        if pipeline.count > 1 {
            // TODO: handle more than one function in a pipeline for all-attributes
            // Print out a warning message about too many function for all-attributes pipeline
            for index in 0..<pipeline.count {
                
                let function:DevicePolicy.Function = pipeline[index]
                let id:String = function.getId()
                let parameters:[String:Any] = function.getParameters()
                let deviceFunction:DeviceFunction? =
                    DeviceFunction.getDeviceFunction(functionId: id)
                
                if index == 0 {
                    // WARNING
                    log(type:.fault,
                        message: "Only one function allowed for all-attribute pipeline. " +
                        "Applying: %@",
                        args: deviceFunction?.getDetails(configuration: parameters) ?? id)
                } else {
                    // WARNING
                    log(type:.fault,
                        message:"Ignoring: %@",
                        args: deviceFunction?.getDetails(configuration: parameters) ?? id)
                }
            }
        }

        // Create the pipeline data for this device model that applies to all
        // attributes
        var pipelineData:[Dictionary<String,AnyObject>]? = 
            getPipelineDataCache(attribute:DevicePolicy.ALL_ATTRIBUTES())
        if pipelineData == nil {
            pipelineData = [Dictionary<String,AnyObject>]()
            // Add back pipelineData
            putPipelineDataCache(attribute:DevicePolicy.ALL_ATTRIBUTES(),
                                 pipelineData:pipelineData!)
        }

        // SE does this in the for loop.
        // create missing function maps in the pipeline data
        if pipelineData!.count < pipeline.count {
            // create missing function data maps
            for _ in pipelineData!.count..<pipeline.count {
                pipelineData!.append(Dictionary<String,AnyObject>())
            }
            // Add back pipelineData
            putPipelineDataCache(attribute:DevicePolicy.ALL_ATTRIBUTES(),
                                 pipelineData:pipelineData!)
        }

        // Handle pipleline for device policy
        // TODO: handle more than one function in a pipeline for all-attributes
        // Note: A warning is printed out, several lines above this, if there
        // is more than one function for an all-attributes policy.
        // Leaving original line here, commented-out, for future reference.
        for index in 0..<1 /*pipeline.startIndex..<pipeline.endIndex*/ {

            let function:DevicePolicy.Function = pipeline[index]
            var functionData:Dictionary<String,AnyObject> = pipelineData![index]
            let key:String = function.getId()
            let parameters:Dictionary<String,Any> = function.getParameters()
            let deviceFunction:DeviceFunction? =
                    DeviceFunction.getDeviceFunction(functionId:key)
            if deviceFunction == nil {
                continue
            }
            // TODO: generalize code reduce duplication
            var windowExpired:Bool = false
            // in milliseconds
            let window:Int64 = DeviceFunction.getWindow(configuration:parameters)
            let slide:Int64 = DeviceFunction.getSlide(configuration:parameters,
                                                      window:window)
            if window > 0 {
                let k:String = deviceModelUrn! + "::" +
                    deviceFunction!.getId() + ".expiry"
                var expiry:Int64? = windowMapQueue.sync(execute:{
                    return windowMap[k]
                })
                if expiry == nil {
                    expiry = currentTimeMillis + window
                    windowMapQueue.sync() {
                        windowMap[k] = expiry!
                    }
                }
                windowExpired = expiry! <= currentTimeMillis
                if windowExpired {
                    windowMapQueue.sync() {
                        windowMap[k] = expiry! + slide
                    }
                }
            }

            var alertOverridesPolicy:Bool = false
            if message is AlertMessage {
                let alertMessage:AlertMessage = (message as! AlertMessage)
                let alertMessageSeverity:AlertMessageSeverity =
                    alertMessage.getSeverity()

                var configuredSeverity:AlertMessageSeverity =
                    AlertMessageSeverity.critical
                if let criterion:String =
                        (parameters["alertSeverity"] as? String) {
                    // default is .critial
                    configuredSeverity = 
                        AlertMessageSeverity.valueOf(severity:criterion)
                }
                alertOverridesPolicy = 
                    configuredSeverity.rawValue <= alertMessageSeverity.rawValue
            }

            // VirtualDevice stores Pair<Message,StorageObject>, do the same here.
            // TODO: be careful of nil value in Pair
            let pair:Pair<Message,StorageObject?> =
                Pair<Message,StorageObject?>(key:message, value:nil)
            
            // TODO: Make sure inout parameters are modified and reinserted
            // into structs they came from
            if deviceFunction!.apply(deviceAnalog:deviceAnalog!,
                                    attribute:nil,
                                    configuration:parameters,
                                    data:&functionData,
                                    value:pair)
                    || windowExpired
                    || alertOverridesPolicy {
                
                // Add back pipelineData
                pipelineData![index] = functionData
                putPipelineDataCache(attribute: DevicePolicy.ALL_ATTRIBUTES(),
                                     pipelineData: pipelineData!)

                if let valueFromPolicy:Any = deviceFunction!.get(
                        deviceAnalog:deviceAnalog!,
                        attribute:DevicePolicy.ALL_ATTRIBUTES(), //nil,
                        configuration:parameters,
                        data:&functionData) {
                    
                    // Add back pipelineData
                    pipelineData![index] = functionData
                    putPipelineDataCache(attribute: DevicePolicy.ALL_ATTRIBUTES(),
                                         pipelineData: pipelineData!)
                    
                    if let messageList:[Pair<Message,StorageObject?>] =
                            valueFromPolicy as? [Pair<Message,StorageObject?>] {
                        var messages:[Message] = [Message]()
                        for p in messageList {
                            // For LL, we don't care about StorageObject since it
                            // isn't queued with message.
                            // For HL, we do care because MessageDispatcherImpl
                            // logically ties the two together.
                            // But this is LL code here, so we only want the message.
                            // Besides, when this Pair is created (just before calling
                            // apply), the value side of the Pair is passed as null.
                            messages.append(p.getKey())
                        }
                        return messages
                    }
                }
            }
            // Add back pipelineData
            pipelineData![index] = functionData
            putPipelineDataCache(attribute: DevicePolicy.ALL_ATTRIBUTES(),
                                 pipelineData: pipelineData!)
            return [Message]()
        }
        return [Message]()
    }

    private func expirePolicy(devicePolicy:DevicePolicy,
                                currentTimeMillis:Int64) throws -> [Message] {


        let messageList:[Message] = try expirePolicy(devicePolicy:devicePolicy)
        var consolidatedMessageList:[Message] = [Message]()
        if messageList.isEmpty {
            return consolidatedMessageList
        }

        // consolidate messages
        var dataItemMap:Dictionary<String,[DataItem]> =
            Dictionary<String,[DataItem]>()
        for message in messageList {
            if message is DataMessage {
                let endpointId:String = message.getSource()
                var dataItems:[DataItem]? = dataItemMap[endpointId]
                if dataItems == nil {
                    dataItems = [DataItem]()
                    dataItemMap[endpointId] = dataItems
                }
                dataItems!.append(contentsOf:(message as! DataMessage).getDataItems())
            } else {
                consolidatedMessageList.append(message)
            }
        }

        for (k,v) in dataItemMap {
            let deviceAnalog:DeviceAnalog? = deviceAnalogMapQueue.sync(execute:{
                return deviceAnalogMap[k]
            })
            if deviceAnalog == nil {
                continue
            }
            var dataItems:[DataItem] = v
            let format:String = deviceAnalog!.getDeviceModel().getURN()

            if !computedMetricTriggers.isEmpty {

                if let triggerMap:Dictionary<String,Set<String>> =
                        computedMetricTriggersQueue.sync(execute:{
                            return computedMetricTriggers[format]}),
                                                        !triggerMap.isEmpty {
                    do {
                        try checkComputedMetrics(dataItems:&dataItems,
                                             deviceAnalog:deviceAnalog!,
                                             triggerMap:triggerMap,
                                             currentTimeMillis:currentTimeMillis)
                    } catch {
                        // WARNING
                        log(type:OSLogType.fault, message:"\(error)")
                    }
                }
            }

            let dataMessageBuilder:DataMessageBuilder = DataMessageBuilder()
            dataMessageBuilder.setSource(source:deviceAnalog!.getEndpointId())
            dataMessageBuilder.setFormat(format:deviceAnalog!.getDeviceModel().getURN())
            dataMessageBuilder.addDataItems(dataItems:dataItems)
            consolidatedMessageList.append(try dataMessageBuilder.build())
        }
        return consolidatedMessageList
    }

    private func expirePolicy(devicePolicy:DevicePolicy) throws -> [Message] {
        var messageList:[Message] = [Message]()
 
        for (_, deviceAnalog) in deviceAnalogMap {
            let messages:[Message] =
                    try expirePolicy(devicePolicy:devicePolicy,
                                 deviceAnalog:deviceAnalog)
            if !messages.isEmpty {
                messageList.append(contentsOf:messages)
            }
        }
    
        return messageList
    }
    
    private func expirePolicy(devicePolicy:DevicePolicy,
                              deviceAnalog:DeviceAnalog) throws -> [Message] {
    
        var messageList:[Message] = [Message]()
        for (attributeName,functions) in devicePolicy.getPipelines() {
            if let messages:[Message] = try expirePolicy(attributeName:attributeName,
                                                  pipeline:functions,
                                                  deviceAnalog:deviceAnalog) {
                messageList.append(contentsOf:messages)
            }
        }
        return messageList
    }

    private func expirePolicy(attributeName:String,
                              pipeline:[DevicePolicy.Function],
                              deviceAnalog:DeviceAnalog) throws  -> [Message]? {
        
        if pipeline.isEmpty {
            return nil
        }

        // Note that we are _removing_ the pipeline data cache for this
        // attribute
        guard var pipelineData:[Dictionary<String,AnyObject>] =
                pipelineDataCacheQueue.sync(execute:{
                    return pipelineDataCache.removeValue(forKey:attributeName)}) else {
            return nil
        }

        // Need to iterate using index to overwrite with modified data
        
        for index in pipeline.startIndex..<pipeline.endIndex {
            let function:DevicePolicy.Function = pipeline[index]
            guard let deviceFunction:DeviceFunction =
                    DeviceFunction.getDeviceFunction(functionId:function.getId()) else {
                continue
            }

            // Looking for the first policy function in the pipeline that has
            // a "window".
            // If there isn't one, we'll drop out of the loop and return nil.
            // If there is one, we process the remaining pipeline from there.
            let window:Int64 = 
                DeviceFunction.getWindow(configuration:function.getParameters())
            if window == -1 {
                continue
            }

            guard var functionData:Dictionary<String,AnyObject> =
                    index < pipelineData.count ? pipelineData[index] : nil else {
                // there is no data for this function, so return
                return nil
            }

            var valueFromPolicy:AnyObject? = deviceFunction.get(
                    deviceAnalog:deviceAnalog,
                    attribute:attributeName,
                    configuration:function.getParameters(),
                    data:&functionData)
            
            // Add back functionData
            pipelineData[index] = functionData
            
            if valueFromPolicy == nil {
                return nil
            }

            for next in pipeline.index(after: index)..<pipeline.endIndex {
                let function = pipeline[next]
                guard let deviceFunction = 
                        DeviceFunction.getDeviceFunction(functionId:function.getId()) else {
                    return nil
                }
                if next < pipelineData.count {
                    functionData = pipelineData[next]
                } else {
                    return nil
                }

                if deviceFunction.apply(
                        deviceAnalog:deviceAnalog,
                        attribute:attributeName,
                        configuration:function.getParameters(),
                        data:&functionData,
                        value:valueFromPolicy!) {
                    
                    // Add back functionData
                    pipelineData[next] = functionData
                    
                    valueFromPolicy = deviceFunction.get(
                        deviceAnalog:deviceAnalog,
                        attribute:attributeName,
                        configuration:function.getParameters(),
                        data:&functionData)
                    
                    // Add back functionData
                    pipelineData[next] = functionData
                    
                    if valueFromPolicy == nil {
                        return nil
                    }
                } else {
                    // Add back functionData
                    pipelineData[next] = functionData
                    return nil
                }

            }

            // If we get here, valueFromPolicy is not nil.
            if valueFromPolicy is Array<Message> {
                return valueFromPolicy! as? [Message]
            }

            let deviceModel:DeviceModelImpl = 
                deviceAnalog.getDeviceModel() as! DeviceModelImpl

            let dataMessageBuilder:DataMessageBuilder = DataMessageBuilder()
            dataMessageBuilder.setSource(source:deviceAnalog.getEndpointId())
            dataMessageBuilder.setFormat(format:deviceModel.getURN())


            if let deviceModelAttribute:DeviceModelAttribute =
                deviceModel.getDeviceModelAttributes()[attributeName] {

                switch (deviceModelAttribute.getType()) {
                    case .DATETIME, .NUMBER, .INTEGER:
                        try dataMessageBuilder.addDataItem(key:attributeName,
                                                       value:
                            cast(attributeType:DeviceModelAttributeType.NUMBER,
                                 newValue:valueFromPolicy!) as! Double)
                        break
                    case .STRING, .URI:
                        try dataMessageBuilder.addDataItem(key:attributeName,
                                                       value:
                            cast(attributeType:DeviceModelAttributeType.STRING,
                                 newValue:valueFromPolicy!) as! String)
                        break
                    case .BOOLEAN:
                        try dataMessageBuilder.addDataItem(key:attributeName,
                                                       value:
                            cast(attributeType:DeviceModelAttributeType.BOOLEAN,
                                 newValue:valueFromPolicy!) as! Bool)
                        break
                }
                var messages:[Message] = [Message]()
                messages.append(try dataMessageBuilder.build())
                return messages
            }
    }
        return nil
    }

    private func createDataItem(type:DataItemType, 
                                key:String, 
                                newValue:Any) throws -> DataItem {
        var dataItem:DataItem!
        switch (type) {
            case .DOUBLE:
                dataItem = try DataItem(key:key, value:Double(truncating: newValue as! NSNumber))
                break
            case .STRING:
                dataItem = try DataItem(key:key, value:String(describing:newValue))
                break
            case .BOOLEAN:
                var newBool:Bool = false
                if newValue is String {
                    newBool = Bool(newValue as! String)!
                } else if newValue is NSNumber {
                    newBool = Bool(truncating: newValue as! NSNumber)
                } else {
                    newBool = newValue as! Bool
                }
                dataItem = try DataItem(key:key, value:newBool) //(newValue as! Bool))
                break
        }
        return dataItem
    }

    // TODO: Maybe this should be a method in DeviceModelAttribute? 
    // Also duplicated in VirtualDeviceImpl
    private func cast(attributeType:DeviceModelAttributeType,
                      newValue:AnyObject) -> AnyObject {
        // TODO: Shouldn't this throw if type is invalid or value
        // can't be converted ?
        var castValue:AnyObject = newValue
        switch (attributeType) {
        case .INTEGER:
            castValue = 
                Int64(truncating: newValue as! NSNumber)/*roundedValue*/ as AnyObject
            break
        case .NUMBER:
            castValue = Double(truncating: newValue as! NSNumber) as AnyObject
            break
        case .STRING:
            castValue = (newValue as! String) as AnyObject
            break
        case .BOOLEAN:
            castValue = (newValue as! Bool) as AnyObject
            break
        case .DATETIME:
            if newValue is Date {
                castValue = 
                    (newValue as! Date).timeIntervalSince1970 as AnyObject
            } else {
                // TODO: In some cases this is Double
                castValue = Int64(truncating: newValue as! NSNumber) as AnyObject
            }
            break
        case .URI:
            castValue = (newValue as! String) as AnyObject
            break
        }
        return castValue
    }
}
