/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation

/**
 * This is the implementation of an unbounded priority queue based on a priority heap. The elements 
 * are ordered by a Comparator provided at queue construction time. 
 * The priority queue does not permit `nil` elements.
 * The head of this queue is the least element with respect to the specified ordering. If multiple 
 * elements are tied for least value, the head is one of those elements - ties are broken arbitrarily. 
 * The queue retrieval operations poll, remove, peek, and element access the element at the head of the queue.
 * <p>
 * A priority queue is unbounded, but has an internal capacity governing the size of an array used to store 
 * the elements on the queue. It is always at least as large as the queue size. As elements are added to a 
 * priority queue, its capacity grows automatically. The details of the growth policy are not specified.
 *
 */
class PriorityQueue<T> {
    fileprivate final var theHeap: [T]
    private let compare: (T, T) -> Bool
    
    fileprivate let lock: NSRecursiveLock = NSRecursiveLock()
    
    /**
     * Constructs a *PriorityQueue*
     *
     * - parameter compare: the comparator to compare queue items
     *
     */
    init(compare: @escaping (T, T) -> Bool) {
        theHeap = []
        self.compare = compare
    }
    
    /**
     * Add new element to the queue.
     *
     * - parameter new: the new element
     *
     */
    func push(new: T) -> Bool {
        lock.lock()
        defer {
            lock.unlock()
        }
        theHeap.append(new)
        let _ = shiftUp(index: theHeap.endIndex - 1)
        
        return true
    }
    
    /**
     * Remove and return the upper element of the queue.
     *
     * - returns: the removed element
     */
    func pop() -> T? {
        lock.lock()
        defer {
            lock.unlock()
        }
        if theHeap.count == 0 {
            return nil
        }
        if theHeap.count == 1 {
            return theHeap.removeLast()
        }
        
        theHeap.swapAt(0, theHeap.endIndex - 1)
        let upperElement: T = theHeap.removeLast()
        let _ = shiftDown(index: 0)
        
        return upperElement
    }
    
    fileprivate func shiftDown(index: Int) -> Bool {
        let left: Int = index * 2 + 1
        let right: Int = index * 2 + 2
        var smallest: Int = index
        
        if left < theHeap.count && compare(theHeap[left], theHeap[smallest]) {
            smallest = left
        }
        if right < theHeap.count && compare(theHeap[right], theHeap[smallest]) {
            smallest = right
        }
        
        if smallest != index {
            theHeap.swapAt(index, smallest)
            let _ = shiftDown(index: smallest)
            return true
        }
        
        return false
    }
    
    fileprivate func shiftUp(index: Int) -> Bool {
        if index == 0 {
            return false
        }
        
        let former: Int = (index - 1) >> 1
        
        if compare(theHeap[index], theHeap[former]) {
            theHeap.swapAt(index, former)
            let _ = shiftUp(index: former)
            return true
        }
        
        return false
    }
}

extension PriorityQueue {
    
    var count: Int {
        lock.lock()
        defer {
            lock.unlock()
        }
        return theHeap.count
    }
 
    
    var isEmpty: Bool {
        lock.lock()
        defer {
            lock.unlock()
        }
        return theHeap.isEmpty
    }
    
    func update<T2>(element: T2) -> T? where T2: Equatable {
        lock.lock()
        defer {
            lock.unlock()
        }
        for (index, item) in theHeap.enumerated() {
            if (item as! T2) == element {
                theHeap[index] = element as! T
                if shiftDown(index: index) || shiftUp(index: index) {
                    return item
                }
            }
        }
        return nil
    }
    
    func remove<T2>(element: T2) -> T? where T2: Equatable {
        lock.lock()
        defer {
            lock.unlock()
        }
        for (index, item) in theHeap.enumerated() {
            if (item as! T2) == element {
                if(index != theHeap.endIndex - 1) {
                    theHeap.swapAt(index, theHeap.endIndex - 1)
                }
                theHeap.removeLast()
                let _ = shiftDown(index: index)
                return item
            }
        }
        return nil
    }
    
    var heap: [T] {
        return theHeap
    }
    
    func addAll<T2>(elements:[T2]) where T2: Equatable {
        for element in elements {
            _ = update(element:element)
        }
    }
    
    func removeAll() {
        theHeap.removeAll()
    }
}

extension PriorityQueue: IteratorProtocol {
    
    func next() -> T? {
        return pop()
    }
}



extension PriorityQueue: Sequence {
    
    func makeIterator() -> PriorityQueue {
        return self
    }
}
