/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
// FIXME: comparison operators with optionals were removed from the 
// Swift Standard Libary.
// Consider refactoring the code to use the non-optional operators.
fileprivate func < <T : Comparable>(lhs: T?, rhs: T?) -> Bool {
  switch (lhs, rhs) {
  case let (l?, r?):
    return l < r
  case (nil, _?):
    return true
  default:
    return false
  }
}


/**
 * Implementation of a default RequestHandler. It implements
 * the default behavior for a resource that hasn't registered a handler.
 */
class DefaultRequestHandler : RequestHandler {
    
    override init() {
    }
    
    /**
     * Handles incoming Request message, creating a Response message.
     *
     * - parameter request: incoming Request message
     *
     * - returns: outgoing Response message
     * 
     * - Throws: WrongSource
     */
    override func handleRequest(requestMessage: RequestMessage?) throws -> ResponseMessage {
        let builder:ResponseMessageBuilder =
            ResponseMessageBuilder(requestMessage: requestMessage!)
        let url = requestMessage?.getURL()
        builder.setStatusCode(statusCode: .NOT_FOUND)
        builder.setUrl(url: url!)
        return try builder.build()
    }
}

/**
 * RequestDispatcher is a utility that allows the caller to register handlers
 * for requests coming from the IoT server. The RequestDispatcher handles all
 * of the logic of matching the resource to a handler and invoking the handler.
 */
public class RequestDispatcher {
    /**
     * This is the default behavior for a resource that hasn't registered a 
     * handler.
     */
    var DEFAULT_HANDLER:RequestHandler = DefaultRequestHandler()
    
    internal static var INSTANCE:RequestDispatcher = RequestDispatcher()
    
    /**
     * The `RequestDispatcher` is a singleton.
     *
     * - returns: The single instance of RequestDispatcher
     */
    public static func getInstance() -> RequestDispatcher {
        return INSTANCE
    }
    
    /**
     * Registered RequestHandlers
     */
    internal var requestHandlerRegistrations: [RequestHandlerRegistration]?
    
    /**
     * The current default handler
     */
    internal var defaultHandler: RequestHandler?
    
    private let registrationsQueue:DispatchQueue =
        DispatchQueue(label:"registrationQueue", qos: .utility)
    
    /**
     * Only singleton allowed
     */
    private init() {
        self.requestHandlerRegistrations = []
        self.defaultHandler = self.DEFAULT_HANDLER
    }
    
    /**
     * Register a handler for the given endpoint and path. If endpoint is `nil`,
     * empty, or a wildcard ("\*"), then the handler is registered as being used
     * for all endpoints. If `path` is `nil`, empty, or a wildcard ("\*")
     * then `handler` is registered as the handler for any requests that do
     * not have a more specific handler.
     * The `handler` parameter may not be `nil`.
     *
     * Registering a handler with
     * `registerRequestHandler("*", "*", handler)` will register a default,
     * catch-all handler. Note that this utility class already provides a
     * reasonable default handler that can be replaced with this technique.
     * The implementation ensures there is always a default handler.
     * </p>
     *
     * - parameter endpointId: the endpoint id that has the given the resource 
     *                         path, or a wildcard ("*") for any endpoint.
     *
     * - parameter path: the resource path, or a wildcard ("*") for any resource
     *
     * - parameter handler: the handler to invoke for requests to this endpoint 
     *                      and path
     *
     * - throws: `ClientError.argument` if `handler` is `nil`
     */
    public func registerRequestHandler(endpointId: String?, path: String?,
                                       handler: RequestHandler?) throws {
                
        if handler == nil {
            throw ClientError.argument(NSLocalizedString(
                "Request Handler must not be nil", comment: ""))
        }
        
        registrationsQueue.sync(execute: {
            let normalizedEndpointId: String = endpointId == nil ||
                    endpointId!.isEmpty ? "*" : endpointId!

            let normalizedPath: String =
                path == nil || path!.isEmpty ? "*" : path!
            
            // special case for default handler
            if("*" == normalizedEndpointId) && ("*" == normalizedPath) {
                self.defaultHandler = handler
                return
            }
            
            // RequestHandlerRegistration is immutable, so make a new one
            let registration: RequestHandlerRegistration =
                RequestHandlerRegistration(endpointId: normalizedEndpointId,
                                           path: normalizedPath, handler: handler!)
            
            var found: Bool = false
            for n in 0..<self.requestHandlerRegistrations!.count {
                let other = self.requestHandlerRegistrations![n]
                // assignment in the conditional is optional
                if registration == other {
                    // replace the existing registration
                    self.requestHandlerRegistrations![n] = registration
                    found = true
                    break
                }
            }
            if !found {
                self.requestHandlerRegistrations!.append(registration)
                self.requestHandlerRegistrations!.sort(by: {$0 > $1})
            }
        })
    }
    
    /**
     * Un-register the handler for the given endpoint and path. Either or both
     * of the parameters may be a wildcard ("*"). nil or empty parameters are
     * interpreted as a wildcard.
     * Note that the wildcard is not used to match any endpoint or path, but
     * is used to match a handler that was registered with the wildcard endpoint
     * and/or path. Thus, a handler registered by calling
     * *registerRequestHandler("0-AB", "resource", handler)*
     * will not be un-registered by calling
     * *unregisterRequestHandler("\*", "resource")*.
     * Nor will calling *unregisterRequestHandler("0-AB", "\*")* result
     * in un-registering all handlers for endpoint "0-AB". In short, the
     * arguments to *registerRequestHandler* should match the
     * arguments to *unregisterRequestHandler*.
     * <p>
     * A default handler can be registered by calling
     * *registerRequestHandler("\*", "\*", handler)*. Calling
     * *unregisterRequestHandler("\*", "\*")* will restore the
     * default handler. The implementation ensures there is always a
     * default handler.
     * </p>
     *
     * - parameter endpointId: the endpoint id that has the given the resource 
     *                         path, or a wildcard ("*").
     *
     * - parameter path:  the resource path, or a wildcard ("*")
     */
    public func unregisterRequestHandler(endpointId: String?, path: String?) {
        registrationsQueue.sync(execute: {
            let normalizedEndpointId: String = endpointId == nil ||
                    endpointId!.isEmpty ? "*" : endpointId!
            let normalizedPath: String =
                path == nil || path!.isEmpty ? "*" : path!
                
            // special case for default handler
            if("*" == normalizedEndpointId) && ("*" == normalizedPath) {
                self.defaultHandler = self.DEFAULT_HANDLER
                return
            }
                
            for n in 0..<self.requestHandlerRegistrations!.count {
                let other = self.requestHandlerRegistrations![n]
                    
                let otherEndpointId: String = other.endpointId!
                let otherPath: String = other.path!
                if (normalizedEndpointId == otherEndpointId) &&
                        (normalizedPath == otherPath) {
                    self.requestHandlerRegistrations!.remove(at: n)
                    return
                }
            }
        })
    }
    
    /**
     * Un-register the given handler for all endpoints and paths for which the
     * handler may be registered. The `handler` parameter may not be nil.
     *
     * - parameter handler: the handler to un-register
     */
    public func unregisterRequestHandler(handler: RequestHandler) {
        registrationsQueue.sync(execute: {
            // special case for default handler
            if (handler === self.defaultHandler!) &&
                    !(self.defaultHandler === self.DEFAULT_HANDLER) {
                self.defaultHandler = self.DEFAULT_HANDLER
                return
            }
                
            var n = self.requestHandlerRegistrations!.count - 1
            while 0 <= n {
                let other:RequestHandlerRegistration =
                    self.requestHandlerRegistrations![n]
                if other.handler === handler {
                    self.requestHandlerRegistrations!.remove(at: n)
                }
                n -= 1
            }
        })
    }
    
    /**
     * Match the request to a registered handler and invoke the handler.
     *
     * - parameter requestMessage: The request message to be dispatched
     *
     * - returns: The return value of the invoked handler. If the invoked
     * handler throws an exception or returns nil, a reasonable response
     * will be returned by this utility. This method will not return nil.
     *
     * - Throws: ClientError.argument if requestMessage is nil
     */
    public func dispatch(requestMessage: RequestMessage?) throws -> ResponseMessage {
                
        if (requestMessage == nil) {
            throw ClientError.argument(NSLocalizedString(
                "Request Message must not be nil", comment: ""))
        }
        
        let endpointId: String = requestMessage!.getDestination()
        let path: String = requestMessage!.getURL()
        let requestHandler: RequestHandler = getRequestHandler(endpointId: endpointId, path: path)
        
        do {
            let responseMessage: ResponseMessage =
                try requestHandler.handleRequest(requestMessage: requestMessage!)!
            return responseMessage
        } catch {
            let builder = ResponseMessageBuilder(requestMessage: requestMessage!)
            builder.setSource(source: requestMessage!.getDestination())
            builder.setRequestId(requestId: requestMessage!.getId())
            builder.setBody(body: "bla")
            builder.setStatusCode(statusCode: .INTERNAL_SERVER_ERROR)
            return try builder.build()
        }
    }
    
    private func failedDispatch(requestMessage:RequestMessage, msg:String,
                    callback: @escaping (ResponseMessage?, ClientError?)->()) {
        do {
            let builder = ResponseMessageBuilder(requestMessage: requestMessage)
            builder.setSource(source: requestMessage.getDestination())
            builder.setRequestId(requestId: requestMessage.getId())
            builder.setBody(body: msg)
            builder.setStatusCode(statusCode: .INTERNAL_SERVER_ERROR)
            callback(try builder.build(), nil)
        } catch {
            let errmsg =  NSLocalizedString(
                "Failed to create ResponseMessage for failed dispatch", comment: "")
            callback(nil, ClientError.network(msg + " : " + errmsg + ": \(error)"))
        }
    }
    
    // Internall used by MessageDispatcherImpl 
    // AsynRequestHandler used internally by VirtualDeviceImpl
    func dispatch(requestMessage: RequestMessage?,
            callback: @escaping (ResponseMessage?,ClientError?)->()) throws {
        
        guard let requestMessage = requestMessage else {
            throw ClientError.argument(NSLocalizedString(
                "Request Message must not be nil", comment: ""))
        }
        
        let endpointId: String = requestMessage.getDestination()
        let path: String = requestMessage.getURL()
        let requestHandler: RequestHandler = getRequestHandler(endpointId: endpointId,
                                                               path: path)
        
        do {
            if requestHandler is AsyncRequestHandler {
                try (requestHandler as! AsyncRequestHandler).handleRequest(
                    requestMessage: requestMessage, callback: callback)
            } else {
                if let responseMessage =
                    try requestHandler.handleRequest(requestMessage: requestMessage) {
                    callback(responseMessage, nil)
                } else {
                    failedDispatch(requestMessage:requestMessage,
                        msg: NSLocalizedString("Failed to dispatch, nil response",
                            comment: ""), callback: callback)
                }
            }
        } catch {
            failedDispatch(requestMessage:requestMessage, msg: "\(error)",
                 callback: callback)
        }
    }

    // private helper
    func copyResponseMessage(message: Message) throws -> ResponseMessageBuilder {
        var builder: ResponseMessageBuilder = ResponseMessageBuilder()
        if message is ResponseMessage {
            let responseMsg: ResponseMessage = message as! ResponseMessage
            builder = copyMessage(builder: &builder, message: responseMsg)
            let oldHeader: [String : [String?]?] = responseMsg.getHeaders()
            for (name, values) in oldHeader {
                    try builder.addHeader(name: name, values: values)
            }
            builder.setBody(body: responseMsg.getBody())
            builder.setStatusCode(statusCode: responseMsg.getStatusCode())
            builder.setUrl(url: responseMsg.getURL())
            builder.setRequestId(requestId: responseMsg.getRequestId())
        } else {
            throw ClientError.argument(NSLocalizedString(
                "Cannot copy a different type of message", comment: ""))
        }
        return builder
    }
    
    // private helper
    func copyMessage(builder: inout ResponseMessageBuilder, message: Message) -> ResponseMessageBuilder {
        builder.setId(id: message.getId())
        builder.setClientId(clientId: message.getClientId())
        builder.setSource(source: message.getSource())
        builder.setDestination(destination: message.getDestination())
        builder.setPriority(priority: message.getPriority())
        builder.setProperties(properties: message.getProperties())
        builder.setEventTime(eventTime: message.getEventTime()!)
        builder.setReliability(reliability: message.getReliability())
        builder.setSender(sender: message.getSender())
        builder = copyDict(builder: &builder, dict: message.getDiagnostics())
        builder.setDirection(direction: message.getDirection())
        builder.setReceivedTime(receivedTime: message.getReceivedTime()!)
        builder.setSentTime(sentTime: message.getSentTime()!)
        return builder
    }
    
    // private helper
    func copyDict(builder: inout ResponseMessageBuilder,
                  dict: [String : AnyObject]?) -> ResponseMessageBuilder {
        if dict == nil {
            return builder
        }
        for (name, value) in dict! {
            builder.setDiagnostic(name: name, value: value)
        }
        return builder
    }
    
    /**
     * Lookup a RequestHandler for the given endpointId and path. The lookup
     * will return the most specific handler for the endpointId and path. If
     * there is not a specific match for endpointId, the lookup will match
     * try to match *("\*", path)*. Failing that, the lookup will return
     * a handler for *("\*", "\*")*, or the default handler if no handler
     * for *("\*","\*")* has been registered.
     * Will not return nil.
     *
     * - parameter endpointId: the endpoint id that has the given the resource 
     *                         path, or a wildcard ("*").
     *
     * - parameter path: the resource path, or a wildcard ("*")
     *
     * - returns: The RequestHandler for the given endpointId and path.
     */
    public func getRequestHandler(endpointId: String?, path: String?) -> RequestHandler {
        return registrationsQueue.sync(execute: {
            let normalizedEndpointId: String = endpointId == nil ||
                    endpointId!.isEmpty ? "*" : endpointId!
            let normalizedPath: String =
                path == nil || path!.isEmpty ? "*" : path!
                
            // special case for default handler
            if("*" == normalizedEndpointId) && ("*" == normalizedPath) {
                return defaultHandler!
            }
            
            // TODO: Should loop over entries no need fro index
            
            // Note that this loop depends on requestHandlerRegistrations being
            // sorted such that longer paths come first in the list.
            for n in 0..<self.requestHandlerRegistrations!.count {
                let registration: RequestHandlerRegistration = self.requestHandlerRegistrations![n]
                let otherEndpointId: String = registration.endpointId!
                    
                if (normalizedEndpointId == otherEndpointId) || (otherEndpointId == "*") {
                    let otherPath: String = registration.path!
                    if otherPath == "*" {
                        return registration.handler!
                    }
                        
                    let diff:Int = otherPath.count - normalizedPath.count
                    if diff == 0 {
                        if normalizedPath == otherPath {
                            return registration.handler!
                        }
                    } else if diff < 0 {
                        if normalizedPath.hasPrefix(otherPath) {
                            return registration.handler!
                        }
                    }
                }
            }
            return defaultHandler!
        })
    }
}

/**
 * RequestHandlerRegistration class that has to be comparable and 
 * hashable
 */
class RequestHandlerRegistration : Comparable, Hashable {
    
    var endpointId: String?
    var path: String?
    var handler: RequestHandler?
    
    init(endpointId: String?, path: String?, handler: RequestHandler) {
        self.endpointId = endpointId == nil ||
            endpointId!.isEmpty ? "*" : endpointId
        self.path = path == nil || path!.isEmpty ? "*" : path
        self.handler = handler
    }
    
    // TODO: why is this hashCode() and duplicates hasgValue ?
    func hashCode() -> Int {
        return self.endpointId!.hashValue + self.path!.hashValue
    }
    
    var hashValue: Int {
        return self.endpointId!.hashValue + self.path!.hashValue
    }
    
    func toString() -> String {
        return "RequestHandlerRegistration{" + self.endpointId! + ", " +
                                           self.path! + "}"
    }
}

/**
 * Implementation of protocol `Equatable`
 *
 * - parameter lhs: one item to be compared
 *
 * - parameter rhs: the other item to be compared
 *
 * - returns: `true` if `lhs` is "less than" `rhs`, `false` otherwise
 */
func < (lhs: RequestHandlerRegistration, rhs: RequestHandlerRegistration) -> Bool {
    if lhs.endpointId != rhs.endpointId {
        if lhs.endpointId < rhs.endpointId {
            return true
        } else {
            return false
        }
    }
    
    // same endpoint
    let l1:Int = (lhs.path?.count)!
    let l2:Int = (rhs.path?.count)!
    if l1 != l2 {
        if l1 < l2 {
            return true
        } else {
            return false
        }
    }
    
    // paths have same length
    if lhs.path != rhs.path {
        if lhs.path < rhs.path {
            return true
        } else {
            return false
        }
    } else {
        return false
    }
}

/**
 * Implementation of protocol Equatable
 *
 * - parameter lhs: one item to be compared
 *
 * - parameter rhs: the other item to be compared
 *
 * - returns: `true` if equal, `false` otherwise
 */
func == (lhs: RequestHandlerRegistration, rhs: RequestHandlerRegistration) -> Bool {
    if lhs === rhs {
        return true
    }
    if lhs.endpointId != rhs.endpointId {
        return false
    }
    
    // same endpoint
    let l1:Int = (lhs.path?.count)!
    let l2:Int = (rhs.path?.count)!
    if l1 != l2 {
        return false
    }
    
    // paths have same length
    if lhs.path != rhs.path {
        return false
    }
    return true
}
