/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation

/**
 * Method supported by a Resource
 */
public enum Method: Int {
    case GET = 0x1
    case POST = 0x2
    case PUT = 0x4
    case DELETE = 0x8
    case PATCH = 0x10
    
    public static let allValues = [GET, POST, PUT, DELETE, PATCH]
    
    /**
     * Returns the Method value for the String
     *
     * - parameter name: String shall be "GET" or "POST"
     *
     * - returns: Method
     */
    public static func createValue(name: String) -> Method? {
        if name == "GET" {
            return .GET
        } else if name == "POST" {
            return .POST
        } else if name == "PUT" {
            return .PUT
        } else if name == "DELETE" {
            return .DELETE
        } else if name == "PATCH" {
            return .PATCH
        } else {
            return nil
        }
    }
    
    /**
     * Returns a list of the Method values for the bit string
     *
     * - parameter bitMask: bit string
     *
     * - returns: list of the Method values
     */
    public static func createValue(bitMask: Int) -> [Method] {
        var methods: [Method] = []
    
        for m in Method.allValues {
            if ((bitMask & m.rawValue) != 0) {
                methods.append(m)
            }
        }
    
        return methods
    }
    
    /**
     * Returns a bit mask for a list of the Method values
     *
     * - parameter methods: list of the Method values
     *
     * - returns: int bit mask
     */
    public static func getBitMask(methods: [Method]) -> Int {
        var result: Int = 0
    
        for m in methods {
            result = result + m.rawValue
        }
    
        return result
    }
    
    /**
     * Returns String for a list of the Method values
     *
     * - parameter methods: list of the Method values.
     *
     * - returns: String a comma-delimited string of method names or an empty 
     *            string if methods is empty or `nil`
     */
    public static func listToString(methods: [Method]?) -> String? {
        var result: String = ""
    
        if (methods != nil) && (methods!.count != 0) {
            var i: Int = 0
            for m in methods! {
                result = result + String(describing: m)
                if i < (methods!.count - 1) {
                    result = result + ","
                }
                i += 1
            }
        }
    
        return result
    }
    
    /**
     * Converts string of the method names into a list of the Method values.  
     * Returns an empty list if *methods* is `nil` or empty.
     *
     * - parameter methods: string of methods
     *
     * - returns: list of the Method values
     */
    public static func stringToList(methods: String?) -> [Method] {
        var result: [Method] = []
        guard let methods = methods, !methods.isEmpty else {
            return result
        }
        let methodNamesArr = methods.split{$0 == ","}.map(String.init)
        for mna in methodNamesArr {
            for m in Method.allValues {
                if mna == String(describing: m) {
                    result.append(m)
                    break
                }
            }
        }
        return result
    }
}

/**
 * Enumeration for choosing status of the resource directory item.  
 * Allowed statuses are ADDED, REMOVED.
 */
public enum Status: String {
    case ADDED
    case REMOVED
}

/**
 * *Resource* class is immutable. A builder is required when
 * creating new *Resource*
 */
public class ResourceBuilder {
    /**
     * Resource's endpoint name or ID.
     * RD TODO: Is endpointName needed?
     */
    fileprivate var endpointName: String?
    
    /**
     * URI at which the resource can be found.  This must be unique for the 
     * endpoint.
     */
    fileprivate var path: String?
    
    /**
     * Resource name or description in free form.
     * It could be used for resource lookup.
     */
    fileprivate var name: String?
    
    /**
     * Resource status. Defines what action shall be done with it.
     * *ADDED* or *REMOVED*
     */
    fileprivate var status: Status?
    
    fileprivate var methods: [Method]?
    
    /**
     * Constructor - does nothing.
     */
    public init() {
    }
    
    /**
     * Sets the name or ID of the resource's endpoint.
     *
     * - parameter endpointName: name or ID of the resource's endpoint.
     *
     */
    public func setEndpointName(endpointName: String) {
        self.endpointName = endpointName
    }
    
    /**
     * Sets the path of the resource.  path must be unique for all resources 
     * within an endpoint.
     *
     * - parameter path: Uri of the resource.  Must be unique and must not 
     *                   be `nil`
     */
    public func setPath(path: String) {
        self.path = path
    }
    
    /**
     * Sets the name of the resource.
     *
     * - parameter name: name of the resource.
     */
    public func setName(name: String) {
        self.name = name
    }
    
    /**
     * Sets the status value for the resource.  The default is *Status.ADDED*
     * if no status or a `nil` status is specified.
     *
     * - parameter status: the status of the resource.
     */
    public func setStatus(status: Status) {
        // TODO: does not comply with the description every status is here 
        // excepted
        self.status = status
    }
    
    /**
     * Sets the method value for the resource.  The default is an empty list if
     * there is no method or `nil` method is specified.
     *
     * - parameter method: status of the resource
     */
    public func addMethod(method: Method) {
        if self.methods == nil {
            self.methods = []
        }
        
        self.methods!.append(method)
    }
    
    /**
     * Sets the method values for the resource.
     * The default is an empty list if no method or `nil`
     * methods is specified.
     *
     * - parameter methods: an array of methods supported by the resource.
     */
    public func addMethods(methods: [Method]?) {
        if self.methods == nil {
            self.methods = []
        }
        
        if methods != nil {
            for m in methods! {
                self.methods!.append(m)
            }
        }
    }
    
    /**
     * Method to initialize the `Resource` from a JSON object.
     *
     * - parameter resObject: the jsonObject to read from
     */
    public func fromJSON(data: Foundation.Data?) throws {
        if data == nil {
            return
        }
        let jsonDict: Dictionary<String,Any>? =
            try JSONSerialization.jsonObject(with: data!,
                options: JSONSerialization.ReadingOptions()) as?
                    Dictionary<String,Any>
        fromData(jsonDict: jsonDict!)
    }
    
    private func parseJsonStr(str: String) throws -> [String : AnyObject]? {
        if let data = str.data(using: .utf8) {
            return try JSONSerialization.jsonObject(with: data,
                                        options: []) as? [String : AnyObject]
        }
        return nil
    }
    
    /**
     * Method to initialize the `resource` from a dictionary.
     *
     * - parameter item: a `[String : AnyObject]` dictionary with the resource
     *                   data
     */
    public func fromJson(jsonStr: String) throws {
        let jsonDict: [String: AnyObject]? = try parseJsonStr(str: jsonStr)
        fromData(jsonDict: jsonDict! as Dictionary)
    }
    
    /**
     * Method to initialize the `resource` from a dictionary.
     *
     * - parameter item: an `Dictionary`  with the resource data
     */
    public func fromData(jsonDict: Dictionary<String,Any>) {
        for (key, value) in jsonDict {
            // TODO: use switch
            if key == "status" {
                let st: String? = (value as! String)
                if st != nil {
                    self.status = (st == "ADDED") ? Status.ADDED : Status.REMOVED
                }
            } else
            if key == "methods" {
                self.methods = Method.stringToList(methods: value as? String)
            } else
            if key == "endpointName" {
                self.endpointName = value as? String
            } else
            if key == "name" {
                self.name = value as? String
            } else
            if key == "path" {
                self.path = value as? String
            }
        }
    }
    
    /**
     * Builds (creates) new instance of `Resource`.
     *
     * - returns: new instance of `Resource`
     */
    public func build() throws -> Resource {
        return try Resource(builder: self)
    }
    
}

/**
 * Implementation of protocol `Equatable`
 *
 * - parameter lhs: one item to be compared
 *
 * - parameter rhs: the other item to be compared
 *
 * - returns: `true` if equal, `false` otherwise
 */
public func ==(lhs: Resource, rhs: Resource) -> Bool {
    if lhs === rhs {    
        return true
    }
    if lhs.getName() != rhs.getName() {
        return false
    }
    if lhs.getPath() != rhs.getPath() {
        return false
    }
    if lhs.getStatus() != rhs.getStatus() {
        return false
    }
    let lhsMethods: [Method]? = lhs.getMethods()
    let rhsMethods: [Method]? = rhs.getMethods()
    if lhsMethods!.count != rhsMethods!.count {
        return false
    }

    for i in 0..<lhsMethods!.count {
        if lhsMethods![i] != rhsMethods![i] {
            return false
        }
    }
    return true
}

/**
 * This class represents a resource for an endpoint.
 */
public class Resource: Hashable {
    /**
     * The endpoint name or ID this resource is for.
     */
    private var endpointName: String?
    
    /**
     * Resource name. This name should describe the resource in free form.
     */
    private var name: String?
    
    /**
     * Resource status.
     */
    private var status: Status?
    
    /**
     * Path of the resource
     */
    private var path: String?
    
    /**
     * Methods supported by the resource
     */
    private var methods: [Method]?
    
    /**
     *
     * - throws: `ClientError.state` if builder name, path, or methods are `nil`
     *           or empty
     *
     */
    fileprivate init(builder: ResourceBuilder) throws {
        guard let name = builder.name, !name.isEmpty,
              let path = builder.path, !path.isEmpty else {
            throw ClientError.state(NSLocalizedString(
                "Resource name must not be nil or empty", comment: ""))
        }
    
        self.endpointName = builder.endpointName
        self.name = name
        self.path = path
        self.status = builder.status ?? .ADDED
    
        if self.status == Status.REMOVED {
            return
        }
        guard let methods = builder.methods, !methods.isEmpty else {
            throw ClientError.state(NSLocalizedString(
                "Resource methods must not be nil or empty", comment: ""))
        }
        self.methods = methods
    }
    

    private func toJsonString(newKeys: [String], newValues: [AnyObject]) -> String {
        var resultStr: String = "{"
        var z: Int = 0
        for i in 0..<newKeys.count {
            // TODO what if value[i] is not a String??
            if (newValues[i] is String) &&
                    ((newValues[i] as! String).hasPrefix("[")) {
                resultStr = resultStr + "\"" + newKeys[i] + "\":" +
                            newValues[i].description
            } else {
                resultStr = resultStr + "\"" + newKeys[i] + "\":\"" +
                            newValues[i].description + "\""
            }
            z += 1
            if z < newKeys.count {
                resultStr = resultStr + ","
            }
        }
        resultStr = resultStr + "}"
        return resultStr
    }
    
    /**
     * Builds JSON string from this *Resource*.
     *
     * - returns: JSON string
     */
    public func toJsonString() -> String {
        // TODO: is it intentional that the toString and to Dict methods 
        // don't write the endpoint Name
        // TODO: Why not write endpointName ? Mock server expects it
        // Writing it because of problems in IOT-46187
        let jsonStatus: String = (status == nil) ?
                                 Status.ADDED.rawValue : (status?.rawValue)!
        var result: String
        
        if methods == nil {
            result = toJsonString(newKeys: ["name", "path", "status", "endpointName"],
                                  newValues: [name! as AnyObject,
                                  path! as AnyObject,
                                  jsonStatus as AnyObject,
                                  endpointName as AnyObject])
        } else {
            result = toJsonString(newKeys: ["name", "path", "status", "endpointName", "methods"],
                                  newValues: [name! as AnyObject,
                                  path! as AnyObject,
                                  jsonStatus as AnyObject,
                                  endpointName as AnyObject,
                          Method.listToString(methods: methods)! as AnyObject])
        }
        return result
    }
    
    /**
     * Writes structure of this `Resource` into a dictionary.
     *
     * - returns: dictionary
     */
    public func toDict() -> [String : AnyObject] {
        var result: [String : AnyObject] = [ : ]
        result["name"] = name as AnyObject?
        result["path"] = path as AnyObject?
        result["status"] = ((status == nil) ?
                     Status.ADDED.rawValue : (status?.rawValue)!) as AnyObject
        if methods != nil {
            result["methods"] = (Method.listToString(methods: methods)!) as AnyObject
        }
        return result
    }
    
    /**
     * Writes this `Resource` into a JSON string.
     *
     * - returns: the JSON string
     */
    public func toString() -> String {
        return toJsonString()
    }
    
    /**
     * Sets resource's status.
     *
     * - parameter st: the status
     */
    public func setStatus(st: Status) {
        self.status = st
    }
    
    /**
     * Gets the resource's endpoint name.
     *
     * - returns: the endpoint name
     */
    public func getEndpointName() -> String {
        return self.endpointName!
    }
    
    /**
     * Gets the resource's name.
     *
     * - returns: the name
     */
    public func getName() -> String {
        return self.name!
    }
    
    /**
     * Gets the resource's path.
     *
     * - returns: the path
     */
    public func getPath() -> String {
        return self.path!
    }
    
    /**
     * Gets the resource's status.
     *
     * - returns: the status
     */
    public func getStatus() -> Status {
        return self.status!
    }
    
    /**
     * Gets the resource's methods as an array.
     *
     * - returns: the methods in an array
     */
    public func getMethods() -> [Method] {
        return self.methods!
    }
    
    
    /**
     * Implementation of protocol Hashable
     *
     * - returns: the hash value
     */
    public var hashValue : Int {
        var result: Int = (name?.hashValue)!
        result = 31 &* result &+ (path?.hashValue)!
        if status != nil {
            result = 31 &* result &+ (status?.hashValue)!
        }
        if methods != nil {
            var methHash: Int = 0
            for m in methods! {
                methHash = methHash + m.hashValue
            }
            result = 31 &* result &+ methHash
        }
        return result;
    }
}
