/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation

/**
 * An abstract class for adapting to the device or enterprise implementation
 * of an `AbstractVirtualDevice`
 */
protocol Adapter {
    
    /**
     * Sets the the value for attribute.
     * For enterprise client make a resource request to set the
     * the attribute. For device client calls the attribute's set method.
     * Invokes the on error handlers on error.
     *
     * - parameter deviceAttribute: the device attribute name
     *
     * - parameter value: the device attribute value
     *
     * - throws: `ClientError.argument` if `deviceAttribute` is `nil`
     */
    func setValue(deviceAttribute: VirtualDeviceAttribute?,
                  value: AnyObject) throws

    /**
     * Return a VirtualDeviceAttribute appropriate for either the enterprise
     * or device client.
     *
     * - parameter name: the name of the attribute
     *
     * - returns: The VirtualDeviceAttribute for the given name
     *
     * - throws: `ClientError.argument` if `name` is not a device attribute.
     */
    func getAttribute(name: String) throws -> VirtualDeviceAttribute
    
    /**
     * Set all the attributes in an update batch. Errors are handled in the
     * set call, including calling the on error handler.
     *
     * - parameter updatedAttributes: The attributes modified during an update
     */
    //func updateFields(updatedAttributes: [VirtualDeviceAttribute : AnyObject]) throws
    func updateFields(updatedAttributes: inout [Pair<VirtualDeviceAttribute,AnyObject>]) throws
}

/**
 * Implementation of VirtualDevice.
 */
class VirtualDeviceBase {
    
    // TODO: make sure qos is right
    private var updateLock:DispatchQueue = DispatchQueue.global(qos:.userInitiated)
    private var updateMode: Bool? = false
    
    /** The attributes modified during an update */
    //private var updatedAttributes: [VirtualDeviceAttribute : AnyObject]?
    private var updatedAttributes:[Pair<VirtualDeviceAttribute,AnyObject>]?
    
    private var adapter: VDIAdapter!
    
    private var endpointId: String?
    private var deviceModel: DeviceModel?
    
    // TODO: make sure qos is right
    private let changeCBLock:DispatchQueue = DispatchQueue.global(qos:.userInitiated)
    private var _changeCallback:((ChangeEvent) -> Void)?
    private var changeCallback:((ChangeEvent) -> Void)? {
        get {
            return changeCBLock.sync() {
                return _changeCallback
            }
        }
        set {
            changeCBLock.sync() {
                _changeCallback = newValue
            }
        }
    }
    
    // TODO: make sure qos is right
    private let errorCBLock:DispatchQueue = DispatchQueue.global(qos:.userInitiated)
    private var _errorCallback:((ErrorEvent) -> Void)?
    private var errorCallback: ((ErrorEvent) -> Void)? {
        get {
            return errorCBLock.sync() {
                return _errorCallback
            }
        }
        set {
            errorCBLock.sync() {
                _errorCallback = newValue
            }
        }
    }

    /**
     * - parameter adapter: the device or enterprise specific implementation of
     *                      the Adapter interface
     *
     * - parameter endpointId: the server's endpoint id for the device
     *
     * - parameter deviceModel: the device model the device implements
     *
     */
    init(adapter: VDIAdapter, endpointId: String, deviceModel: DeviceModel) {
        self.adapter = adapter
        self.endpointId = endpointId
        self.deviceModel = deviceModel
    }
    
    func getDeviceModel() -> DeviceModel {
        return deviceModel!
    }
    
    func getEndpointId() -> String {
        return endpointId!
    }
    
    /**
     * - throws: `ClientError.argument` if attempting to set a read-only
     *           attribute, or setting an invalid value
     */
    func set(attributeName: String, value: AnyObject) throws {
        // Consider moving validation from adapter.setValue and pass 
        // VirtualDeviceAttribute instance.
        // Validate here for updateSetField.
    
        // getAttribute throws ClientError.argument if the attribute
        // is not in the model.
        let df: VirtualDeviceAttribute =
                                try adapter.getAttribute(name: attributeName)
    
        if !df.isSettable() {
            throw ClientError.argument(NSLocalizedString(
                "Virtual Device: attempt to modify read-only attribute",
                comment: ""))
        }
    
        // Will throw ClientError.argument
        do {
            try df.validate(attribute: df.getDeviceModelAttribute()!,
                            value: value)
        } catch {
            throw ClientError.argument(NSLocalizedString(
                "Virtual Device: illegal value", comment: ""))
        }
    
        if !isUpdateMode() {
            try adapter.setValue(deviceAttribute: df, value: value);
        } else {
            updateSetField(attribute: df, value: value);
        }
    }
    
    func update() {
        updateLock.sync() {
            if updatedAttributes == nil {
                //updatedAttributes = [ : ]
                updatedAttributes = [Pair<VirtualDeviceAttribute,AnyObject>]()
            }
            updateMode = true;
        }
    }
    
    /**
     * - throws: `ClientError.network` if a *MessageDispatcher* cannot be 
     *           obtained.<br/>
     *           `ClientError.argument` if an attribute is `nil` or is the empty
     *           string or if an attribute type is string, if `newValue` is 
     *           `nil` or the empty string.
     */
    func finish() throws {
        try updateLock.sync() {
            if true == updateMode {
                //try adapter.updateFields(updatedAttributes: updatedAttributes!)
                guard  var updatedAttributesCopy:[Pair<VirtualDeviceAttribute,AnyObject>] =
                    updatedAttributes else {
                        return
                }
                try adapter.updateFields(updatedAttributes: &updatedAttributesCopy)
                updatedAttributes?.removeAll(keepingCapacity:false) // = [ : ]
                updateMode = false
            }
        }
    }
    
    func setOnChange(callback: @escaping (ChangeEvent) -> Void) {
        changeCallback = callback
    }

    func getOnChangeCallback() -> ((ChangeEvent) -> Void)? {
        return changeCallback
    }
    
    func setOnError(callback: @escaping (ErrorEvent) -> Void) {
        errorCallback = callback
    }
    
    func getOnErrorCallback() -> ((ErrorEvent) -> Void)? {
        return errorCallback
    }
    
    /**
     * Return `true` if the device is in update mode.
     *
     * - returns: `true` if in update mode else `false`
     */
    func isUpdateMode() -> Bool {
        return updateLock.sync() {
            let result: Bool = updateMode!
            return result
        }
    }
    
    /**
     * Add attribute `attribute` to the set of attributes for batch update.
     *
     * - param attribute: the attribute being set
     *
     * - param value the attribute value
     */
    private func updateSetField(attribute: VirtualDeviceAttribute,
                                value: AnyObject) {
        updateLock.sync() {
            //updatedAttributes![attribute] = value
            // TODO: Can this result in duplicates ?
            updatedAttributes?.append(Pair(key:attribute,value:value))
        }
    }
    
    func toString() -> String {
        return "{" + endpointId! + " , " + deviceModel!.getURN() + "}"
    }
    
}
