/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import SwiftyJSON

class Controller {
    private static let MONITOR_POLLING_INTERVAL: UInt32 = UInt32(Config.MESSAGE_POLLING_INTERVAL)/1000
    private static let controllerRetryCount = 20

    /**
     * Handle the asynchronous requests required by single/batch attribute 
     * control and actions.
     *
     * - parameter virtualDevice: connection to the server
     *
     * - parameter method: type of request (POST, PUT, PATCH) used at 
     *                     single/batch attribute control and actions
     *
     * - parameter resource: resource's uri
     *
     * - parameter errorCallback: callback called in case of errors
     *
     * - returns: void
     */
    static func update(virtualDevice: VirtualDeviceImpl, method: String,
                       resource: String, payload: String,
                       callback: @escaping (ClientError?) -> Void) -> Void {
        DispatchQueue.global(qos: .utility).async {
            
            let secureConnection = virtualDevice.getSecureConnection()
            
            var headers = ["Content-Type" : "application/json",
                           "Accept" : "application/json"]
            var myMethod = method
            if (method == "PATCH") {
                myMethod = "POST"
                headers = ["Content-Type" : "application/json",
                           "Accept" : "application/json",
                           "X-HTTP-Method-Override":"PATCH"]
            }
            var isExecuted : Bool = false
            var response : HttpResponse? = nil
            var responseError : ClientError? = nil
            do {
                try secureConnection.invoke(method: myMethod, restApi: resource,
                                            payload: payload.data(using: .utf8)!,
                                            headers: headers,
                                            response: { invokePost, err -> Void in
                    response = invokePost
                    responseError = err
                    isExecuted = true
                })
            } catch {
                print(error)
                response = nil
                responseError = ClientError.network(NSLocalizedString("\(error)", comment: ""))
                isExecuted = true
            }

            // wait for previous alamofire.invoke call to finish otherwise it 
            // hang up at next call inside handleResponse (looks like invoke() 
            // from alamofire is not reentrant)
            // TODO FIXME : async vs sync
            while (!isExecuted) {
            }
            
            if (responseError != nil) {
                callback(responseError)
            } else {
                if let jsonObj = response {
                    self.handleResponse(secureConnection: secureConnection,
                                        response: jsonObj, callback: { err in
                        callback(err)
                    })
                } else {
                    callback(ClientError.argument(NSLocalizedString("Controller update() invokePost should not be nil", comment: "")))
                }
            }
        }
    }
    
    /**
     * Handle the asynchronous request responses. Poll the server for COMPLETE 
     * response that indicates that the
     * request was made to the target client and the client has responded.
     *
     * - parameter secureConnection: connection to the server
     *
     * - parameter response: the initial server response
     *
     * - parameter msgCallback: callback called in case of message errors.
     */
    private static func handleResponse(secureConnection: HttpSecureConnection,
                                       response: HttpResponse,
                                       callback: (ClientError?) -> Void) -> Void {
        var status = StatusCode.OTHER.getCode()
        var myResponse = response
        var httpstatus = response.getStatus()
        
        var msg: String = ""

        // Async requests return ACCEPTED when the request reaches the server.
        // The response "state" is RECEIVED. When the request reaches the device
        //the response status code is OK and the "state" is COMPLETED.
        //Intermediate response contain the "callback" request url in "links":
        //[...,{"rel":"requests": <url>},...]
        for _ in 1...controllerRetryCount {
            if ((httpstatus != StatusCode.ACCEPTED.getCode()) &&
                (httpstatus != StatusCode.OK.getCode())) {
                status = httpstatus
                break
            }
            var jsonData:JSON!
            do {
                jsonData = try JSON(data: myResponse.getData() as Foundation.Data)
            } catch {
                print(String(format:NSLocalizedString(
                    "%@: JSON threw exception: %@", comment:""),
                    "Controller.handleResponse()", "\(error)"))
                status = StatusCode.INTERNAL_SERVER_ERROR.getCode()
                break
            }
            if let reqStatus = jsonData.dictionaryValue["status"]?.stringValue {
                if ((reqStatus == "COMPLETED") || (reqStatus == "FAILED")) {
                    status = getResponseStatusCode(jsonResponse: jsonData)
                    break
                }
            } else {
                msg = "LOG ERR: handleResponse: Unexpected json format (cannot get status)"
                print(msg)
                status = StatusCode.INTERNAL_SERVER_ERROR.getCode()
                break
            }
            
            let respReqHref = getLinksResponse(jsonResponse: jsonData)
            if (nil == respReqHref) {
                msg = "LOG ERR: handleResponse: Unexpected json format (cannot get request href)"
                print(msg)
                status = StatusCode.INTERNAL_SERVER_ERROR.getCode()
                break
            }
            
            sleep(MONITOR_POLLING_INTERVAL)
            
            let headers = ["Content-Type" : "application/json",
                           "Accept" : "application/json"]
            var isExecuted : Bool = false
            // forward exceptions (e.g user auth) in order to be handled by 
            // the caller
            do {
                try secureConnection.get(restApi: respReqHref!, headers: headers,
                                         response: { (httpResponse, err) -> Void in
                    
                    if (err != nil) {
                        msg = "LOG ERR: update error: " + err!.localizedDescription
                        httpstatus = StatusCode.INTERNAL_SERVER_ERROR.getCode()
                    } else {
                        if let jsonObj = httpResponse {
                            myResponse = jsonObj
                            httpstatus = myResponse.getStatus()
                        } else {
                            msg = "LOG ERR: Controller update() httpResponse should not be nil"
                            httpstatus = StatusCode.INTERNAL_SERVER_ERROR.getCode()
                        }
                    }
                    isExecuted = true
                })
            } catch {
                print(msg)
                msg = "LOG ERR: update: Connection manager failed"
                httpstatus = StatusCode.INTERNAL_SERVER_ERROR.getCode()
                break
            }
            
            while (!isExecuted) {
            }
        }
        
        if (status == StatusCode.OTHER.getCode()) {
            msg = "LOG ERR: handleResponse: Timeout reached!"
            print(msg)
            status = StatusCode.REQUEST_TIMEOUT.getCode()
        }

        if ( (status != StatusCode.ACCEPTED.getCode()) &&
            (status != StatusCode.OK.getCode()) ) {
            callback(ClientError.network(msg))
        } else {
            callback(nil)
        }
    }

    /**
     * Given the asynchronous response state, return the appropriate "links" 
     * href value
     *
     * - parameter jsonResponse: the cloud service response
     *
     * - returns: the uri to obtain the response for the given asynchronous 
     *            response state
     */
    private static func getLinksResponse(jsonResponse: JSON) -> String? {
        let ASYNC_RESPONSE_FORMAT : String = RestApi.V2.getReqRoot() +
                                             "/requests/%1$s"
        var resource : String? = nil
        
        if let id = jsonResponse.dictionaryValue["id"] {
            resource = ASYNC_RESPONSE_FORMAT.replacingOccurrences(of: "%1$s",
                                                        with: id.stringValue)
        }
        
        return resource
    }

    /**
     * Return the async request response status from the embedded json object
     *
     * - parameter jsonResponse: the requests api response
     *
     * - returns: the async request's status or StatusCode.INTERNAL_SERVER_ERROR
     * if the actual response cannot be obtained
     */
    private static func getResponseStatusCode(jsonResponse: JSON) -> Int {
        var statusCode : Int = StatusCode.INTERNAL_SERVER_ERROR.getCode()

        let asyncResponse = jsonResponse.dictionaryValue["response"]
        if let tmpStatus = asyncResponse?.dictionaryValue["statusCode"]?.intValue {
            statusCode = tmpStatus
        } else {
            print("LOG ERR: getResponseStatusCode: Unexpected json format (cannot get response and statusCode)")
        }

        return statusCode
    }
}
