/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation

class HttpSecureConnectionUserAuth : HttpSecureConnection {
    
    init(trustedAssetsManager: TrustedAssetsManager) {
        super.init(trustedAssetsManager: trustedAssetsManager, useOnlySharedSecret: false)
    }
    
    static func createUserAuthSecureConnection(tam: TrustedAssetsManager) -> HttpSecureConnectionUserAuth {
        return HttpSecureConnectionUserAuth(trustedAssetsManager: tam)
    }
    
    override func invoke(method: String, restApi: String, payload: Foundation.Data?, headers: [String : String], response: @escaping ((HttpResponse?, ClientError?) -> Void)) throws -> Void {
        let serverHost: String
        let serverPort: String
        let serverScheme: String
        var URL: String
        
        serverScheme = self.getTrustedAssetsManager().getServerScheme()
        serverHost = self.getTrustedAssetsManager().getServerHost()
        serverPort = self.getTrustedAssetsManager().getServerPort()
        var _headers = headers
        
        URL = "\(serverScheme)://\(serverHost):\(serverPort)\(restApi)"
        URL = URL.addingPercentEncoding(withAllowedCharacters: .urlQueryAllowed)!
        
        let httpClient = HttpClientUserAuth(url: URL, httpManager: self.defaultHttpManager!)
        
        httpClient.getCsrfToken(serverScheme: serverScheme, serverHost: serverHost, serverPort: serverPort, httpResponse: { (responseCsrf, error) in
            if (error != nil) {
                response(nil, error)
            } else {
                if let responseObject = responseCsrf {
                    let xCSRFToken = responseObject.allHeaderFields["X-CSRF-TOKEN"] as! String
                    _headers[httpClient.CSRF_TOKEN_NAME] = xCSRFToken
                    
                    self.invoke(httpClient: httpClient, method: method, payload: payload!, headers: _headers, response: { (invokeResult, error) -> Void in
                        if (error != nil) {
                            response(nil, error)
                        } else {
                            if let responseInvoke = invokeResult {
                                if responseInvoke.getStatus() == 302 {
                                    response(nil, ClientError.failedAuthentication(NSLocalizedString("User Authentication failed (REST call failed)!", comment: "")))
                                } else {
                                    response(responseInvoke, nil)
                                }
                            } else {
                                response(nil, ClientError.argument(NSLocalizedString("HttpSecureConnectionUserAuth invoke() responseInvoke argument should not be nil", comment: "")))
                            }
                        }
                    })
                } else {
                    response(nil, ClientError.argument(NSLocalizedString("HttpSecureConnectionUserAuth invoke() responseCsrf argument should not be nil", comment: "")))
                }
            }
        })
    }
    
    
    private func invoke(httpClient: HttpClientUserAuth, method: String, payload: Foundation.Data, headers: [String : String], response: @escaping ((HttpResponse?, ClientError?) -> Void)) -> Void {
        
        do {
            if method == "GET" {
                try httpClient.get(headers: headers, response: { (invokeResult, error) -> Void in
                    response(invokeResult, error)
                })
            } else if method == "POST" {
                try httpClient.post(data: payload, headers: headers, response: { (invokeResult, error) -> Void in
                    response(invokeResult, error)
                })
            } else if method == "DELETE" {
                try httpClient.delete(headers: headers, response: { (invokeResult, error) -> Void in
                    response(invokeResult, error)
                })
            } else { // Must be "PUT"
                try httpClient.put(data: payload, headers: headers, response: { (invokeResult, error) -> Void in
                    response(invokeResult, error)
                })
            }
        } catch {
            response(nil, ClientError.state(NSLocalizedString("Error while executing HTTP method: " + method, comment: "")))
        }
    }
}
