/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation

/**
 * Provides means to lookup for messages.
 */
public class MessageEnumerator { 
    

    /// The secure connection to use

    private let secureConnection: HttpSecureConnection
    

    /// The application identifier: namespace to use to look for messages

    private let appID: String
    
    /**
     * Creates a message enumerator to retrieve messages.
     *
     * - Parameter client: The client to use to retrieve messages
     *
     * - Throws: NullPointerException if the client is *nil*
     */
    public init(client: EnterpriseClient) {
        self.appID = client.getApplication().getId()
        self.secureConnection = SecureHttpConnectionMap.getSecureHttpConnection(client: client)
    }
    
    /**
     * Returns a list of messages stored by cloud service matching the given 
     * query criteria.
     *
     * A query criteria can be based on either a specific device or a specific 
     * type of messages. The messages are always returned in descending order of
     * message received-time (for messages sent by the device) or message
     * sent-time (for messages sent to the device) as recorded by the cloud 
     * service.
     *
     * - Parameter deviceID: The identifier of the device to get messages. 
     *                       If *nil*, messages coming from any device are listed.
     *
     * - Parameter type: The type of messages to enumerate. If *nil*, all 
     *                    message types are enumerated.
     *
     * - Parameter expand: Indicates that messages should include the message 
     *                     payload along with core message properties.
     *
     * - Parameter since: Epoch time in milliseconds used to retrieve the 
     *                    messages with message recorded time (either 
     *                    received-time or sent-time) by the cloud service that 
     *                    is greater than or equal to this time. If the value 
     *                    is 0 the parameter is ignored.
     *
     * - Parameter until: Epoch time in milliseconds used to retrieve the 
     *                    messages with message recorded time (either
     *                    received-time or sent-time) by the cloud service that 
     *                    is less than this time. If the value is 0 the 
     *                    parameter is ignored.
     *
     * - Throws: IOException if request for messages failed
     *
     * - Throws: If User Authentication is used, then 
     *           *UserAuthenticationException* will be thrown if the session 
     *           cookie has expired.
     *
     * - SeeAlso: UserAuthenticationException
     */
    public func getMessages(deviceID: String, type: Message.MsgType, expand: Bool, since: CLongLong, until: CLongLong, msgIterator: @escaping (MessageIterator?, ClientError?) -> Void) throws -> Void {
        let iotMsgIterator = MessageIterator(appID: self.appID, offset: 0, refillCount: 10, deviceID: deviceID, type: type, expand: expand, since: since, until: until, secureConnection: self.secureConnection, newestFirst: false)
        
        try iotMsgIterator.getMsgs(pageMsgResponse: { pageResponse, error in
            if (error != nil) {
                msgIterator(nil, error)
            } else {
                if pageResponse != nil {
                    msgIterator(iotMsgIterator, nil)
                } else {
                    msgIterator(nil, ClientError.argument(NSLocalizedString("MessageEnumerator getMessages() pageResponse should not be nil", comment: "")))
                }
            }
        })
    }
    
}
