/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import SwiftyJSON

/**
 * A *MessageIterator* can be used to enumerate messages matching a set of 
 * criteria. When iterating, it may generates REST calls to get the next chunk 
 * of messages.
 */

public class MessageIterator : PageableEnumerator<Message> {

    /// The application identifier: namespace to use to look for messages

    private let appID: String

    /// The *HttpSecureConnection* to handle requests.

    private let secureConnection: HttpSecureConnection
    

    /// The identifier of the device which is the source of messages

    private let deviceID: String
    

    /// The value of the parameter to put in the query.

    private let type: Message.MsgType
    
    /// The value of the expand parameter to put in the query, if true, the 
    /// default value, the value is not sent.

    private let expand: Bool
    

    /// Epoch time in milliseconds used to retrieve the messages with message 
    /// recorded time (either received-time or sent-time) by the cloud service 
    /// that is greater than or equal to this time. If the value is 0 the 
    /// parameter is not sent.

    private let since: CLongLong
    

    /// Epoch time in milliseconds used to retrieve the messages with message 
    /// recorded time (either received-time or sent-time) by the cloud service 
    /// that is less than this time. If the value is 0 the value is not sent.

    private let until: CLongLong
    

    /// Ordering of messages by received time: newest first (descending) or 
    /// oldest first (ascending)

    private let newestFirst: Bool
    
    private var firstPage: PagedResponse<Message>?
    
    /**
     * Instantiate a new iterator.
     *
     *
     * - Parameter appID: The application identifier
     *
     * - Parameter offset: The offset of the first item to retrieve
     *
     * - Parameter refillCount: The maximum number of items requested per 
     *                          REST call
     *
     * - Parameter deviceID: The identifier of the device to get messages. 
     *                       If *nil*, messages coming from any device are listed.
     *
     * - Parameter type: The type of messages to enumerate. If *nil*, all 
     *                   message types are enumerated.
     *
     * - Parameter expand: Indicates that messages should include the message 
     *                     payload along with core message properties
     *
     * - Parameter since: Epoch time in milliseconds used to retrieve the 
     *                    messages with message recorded time (either 
     *                    received-time or sent-time) by the cloud service that
     *                    is greater than or equal to this time. If the value 
     *                    is 0 the parameter is ignored.
     *
     * - Parameter until: Epoch time in milliseconds used to retrieve the 
     *                    messages with message recorded time (either 
     *                    received-time or sent-time) by the cloud service that 
     *                    is less than this time. If the value is 0 the 
     *                    parameter is ignored.
     *
     * - Parameter secureConnection: A connection to the server
     *
     * - Parameter newestFirst: Indicates message received time ordering: newest
     *                          first or oldest first.
     */

    init(appID: String, offset: Int, refillCount: Int, deviceID: String, type: Message.MsgType, expand: Bool,
                since: CLongLong, until: CLongLong, secureConnection: HttpSecureConnection, newestFirst: Bool) {
    
        self.appID = appID
        self.secureConnection = secureConnection
        self.deviceID = deviceID
        self.type = type
        self.expand = expand
        self.since = since
        self.until = until
        self.newestFirst = newestFirst
        
        super.init(offs: offset, lim: refillCount)
    
    }
    
    public func getMsgs(pageMsgResponse: @escaping (PagedResponse<Message>?, ClientError?) -> Void) throws -> Void {
        try load(offset: self.voffset, limit: self.vlimit, pageResponse: { msgResponse, error in
            if (error != nil) {
                pageMsgResponse(nil, error)
            } else {
                if let msgObject = msgResponse {
                    self.firstPage = msgObject
                    self.velements = msgObject.elements()
                    self.vhasMore = !(self.firstPage!.elements().isEmpty)
                    pageMsgResponse(msgObject, nil)
                } else {
                    pageMsgResponse(nil, ClientError.argument(NSLocalizedString("MessageIterator getMsgs() msgResponse should not be nil", comment: "")))
                }
            }
        })
    }
    
    func getSecureConnection() -> HttpSecureConnection {
        return self.secureConnection
    }
    
    func load(offset: Int, limit: Int, pageResponse: @escaping (MessageResponse?, ClientError?) -> Void) throws -> Void {
        let req = try createMessageRequest(aid: self.appID, limit: limit, offset: offset, deviceID: self.deviceID, type: self.type,
                                       expand: self.expand, since: self.since, until: self.until, newestFirst: self.newestFirst)
        let reqHeaders = ["Content-Type" : "application/json", "Accept" : "application/json"]
        
        try self.getSecureConnection().get(restApi: req, headers: reqHeaders, response: { getResponse, error in
            if (error != nil) {
                pageResponse(nil, error)
            } else {
                if let getObject = getResponse {
                    let responseData = getObject.getData()
                    do {
                        let jsonObject = try JSON(data: responseData)
                        let msgResponse: MessageResponse = try MessageResponse.fromJson(response: jsonObject, defaultOffset: offset)
                        pageResponse(msgResponse, nil)
                    } catch {
                        pageResponse(nil, ClientError.network(NSLocalizedString("Error while loading messages for offset " + offset.description, comment: "")))
                    }
                } else {
                     pageResponse(nil, ClientError.state(NSLocalizedString("MessageIterator load() getResponse should not be nil", comment: "")))
                }
            }
        })
    }
    
    
    private func createMessageRequest(aid: String, limit: Int, offset: Int, deviceID: String, type: Message.MsgType,
                                             expand: Bool, since: CLongLong, until: CLongLong, newestFirst: Bool) throws -> String {
        var query:String = "?expand=" + String(expand)
        
        if deviceID.isEmpty == false {
            query += "&" + "device=" + deviceID
        }
        
        if String(describing: type).isEmpty == false {
            query += "&" + "type=" + String(describing: type)
        }
        
        if limit > 0 {
            query += "&" + "limit=" + String(limit)
        }
        
        if offset > 0 {
            query += "&" + "offset=" + String(offset)
        }
        
        if since > 0 {
            query += "&" + "since=" + String(since)
        }
        
        if until > 0 {
            query += "&" + "until=" + String(until)
        }
        
        query += "&orderBy=deliveredTime:" + ( (newestFirst == true) ? "desc" : "asc")
        
        var request:String
        if (aid.isEmpty == true) {
            request = RestApi.V2.getReqRoot()+"/messages";
        } else {
            request = RestApi.V2.getReqRoot()+"/apps/" + aid + "/messages";
        }
        
        return (request + query)
    }
    
    func offset(offset: Int, messagesPageResponse: @escaping (MessageResponse?, ClientError?) -> Void) -> Void {
        do {
            try self.load(offset: offset, limit: self.vlimit, pageResponse: { returnPage, error in
                if (error != nil) {
                    messagesPageResponse(nil, error)
                } else {
                    if let pageObject = returnPage {
                        self.velements = pageObject.elements()
                        self.vhasMore = pageObject.hasMore()
                        self.voffset = pageObject.offset()
                        self.vsize = pageObject.total()
                        messagesPageResponse(pageObject, nil)
                    } else {
                        messagesPageResponse(nil, ClientError.argument(NSLocalizedString("MessageIterator offset() returnPage should not be nil", comment: "")))
                    }
                }
            })
        } catch {
            messagesPageResponse(nil, ClientError.network(NSLocalizedString("Error while retrieving messages page for offset " + offset.description, comment: "")))
        }
    }
    
    class MessageResponse : PagedResponse<Message> {
    
        private init(elements: [Message], offset: Int, total: Int, hasMore: Bool) {
            super.init(elements: elements, offset: offset, total: total, hasMore: hasMore, links: [String : String]())
        }

        static func fromJson(response: JSON, defaultOffset: Int) throws -> MessageResponse {
            let items = try response.dictionaryValue["items"]?.rawData()
            var hasMore = response.dictionaryValue["hasMore"]?.boolValue
            if hasMore == nil {
                hasMore = false
            }
            
            var offset = response.dictionaryValue["offset"]?.intValue
            if offset == nil {
                offset = defaultOffset
            }
            
            var total = response.dictionaryValue["total"]?.intValue
            if total == nil {
                total = -1
            }
            
            var list: [Message] = [Message]()
            list = Message.fromJSON(jsonString: String(data: items!, encoding: .utf8)!)
                        
            return MessageResponse(elements: list, offset: offset!, total: total!, hasMore: hasMore!)
        }

    }
}
