/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation


enum PageEnumeratorError: Error {
    case noSuchElement
}

public class PageableEnumerator<T> : Pageable<T> {
    var vsize: Int
    var voffset: Int
    var vlimit: Int
    var vhasMore: Bool
    var velements: [T]
    
    init(offs: Int, lim: Int) {
        self.voffset = offs
        self.vlimit = lim
        self.vhasMore = false
        self.velements = []
        self.vsize = 0
    }
    
    func load(offset: Int, limit: Int, callback: @escaping (PagedResponse<T>?, ClientError?) -> Void) throws {
        fatalError("NOT IMPLEMENTED")
    }
    
    public func setSize(size: Int) {
        self.vsize = size
        self.vhasMore = ( (size < 0) || (self.voffset + self.velements.count < self.vsize))
    }
    
    public override func size() -> Int {
        return self.vsize
    }
    
    public override func hasMore() -> Bool {
        return self.vhasMore
    }
    
    public override func offset() -> Int {
        return self.voffset
    }
    
    private func offset(offset: Int, callback: @escaping (Pageable<T>?, ClientError?) -> Void) throws {
        try self.load(offset: offset, limit: self.vlimit, callback: { page, error in
            if (error != nil) {
                callback(nil, error)
            } else {
                self.velements = page!.elements()
                self.vhasMore  = page!.hasMore()
                self.voffset   = page!.offset()
                self.vsize     = page!.total()
                callback(self, nil)
            }
        })
    }

   
    public override func limit() -> Int {
        return self.vlimit
    }
    
    public override func limit(limit: Int) -> Pageable<T> {
        self.vlimit = limit
        return self
    }
    
    public override func elements() -> [T] {
        return self.velements
    }
    
    
    public override func next(callback: @escaping (Pageable<T>?, ClientError?) -> Void) throws {
        try self.at(offset: self.voffset + self.velements.count, callback: { page, error in
            callback(page, error)
        })
    }
    
    public override func first(callback: @escaping (Pageable<T>?,ClientError?) -> Void) throws {
        try self.at(offset: 0, callback: { page, error in
            callback(page, error)
        })
    }
    

    public override func last(callback: @escaping (Pageable<T>?,ClientError?) -> Void) throws {
        try self.at(offset: self.vsize - self.vlimit, callback: { page, error in
            callback(page, error)
        })
    }
    
    public override func at(offset: Int, callback: @escaping (Pageable<T>?, ClientError?) -> Void) throws {
        try self.offset(offset: offset, callback: { page, error in
            callback(page, error)
        })
    }
    
    
}

/**
 * A response to a REST call supporting paging.
 *
 * - Parameter <T>: The type of elements in the page.
 */
public class PagedResponse<T> {
    
    private let velements: [T]
    private let vhasMore: Bool
    private let voffset: Int
    private let vtotal: Int
    private let vlinks: [String : String]
    
    public static func empty() -> PagedResponse<T>  {
        return PagedResponse<T>(elements: [], offset: 0, total: 0, hasMore: false, links: [:])
    }
    
    init(elements: [T], offset: Int, total: Int, hasMore: Bool, links: [String : String]) {
        self.velements = elements
        self.vhasMore  = hasMore
        self.voffset   = offset
        self.vtotal    = total
        self.vlinks    = links
    }
    
    /**
     * Get the elements in the page
     *
     * - Returns: The elements of the page
     */
    public func elements() -> [T] {
        return self.velements
    }
    
    /**
     * Get the links of the page *(first, last, next, ...)*.
     *
     * - Returns: A map of the links to navigate from this page.
     */
    public func links() -> [String : String] {
        return self.vlinks
    }
    
    /**
     * Whether the collection has more elements.
     *
     * - Returns: *true* if the collection has more elements, *false* otherwise.
     */
    public func hasMore() -> Bool {
        return self.vhasMore
    }
    
    /**
     * Return the offset of the first element.
     *
     * - Returns: The offset of the first element of this page.
     */
    public func offset() -> Int {
        return self.voffset
    }
    
    /**
     * Get the number of elements in this response.
     *
     * - Returns: The element count in this response.
     */
    public func count() -> Int {
        return self.elements().count
    }
    
    /**
     * Get the total number of items in the remote collection.
     *
     * - Returns: The total count.
     */
    public func total() -> Int {
        return self.vtotal
    }
    
    /**
     * Get from the response the item named "canonical" in the set of "links".
     *
     * - Returns: The canonical link representing this response, or *nil* if not found.
     */
    public func canonical() -> String {
        return self.vlinks["canonical"]!
    }
    
    /**
     * Get from the response the item named "self" in the set of "links".
     *
     * - Returns: The link representing this response, or *nil* if not found.
     */
    public func sself() -> String {
        return self.vlinks["self"]!
    }
    
    /**
     * Get from the response the item named "next" in the set of "links".
     *
     * - Returns: The link to access next response, or *nil* if not found.
     */
    public func next() -> String {
        return self.vlinks["next"]!
    }
    
    /**
     * Get from the response the item named "next" in the set of "links".
     *
     * - Returns: The link to access first response, or *nil* if not found.
     */
    public func first() -> String {
        return self.vlinks["first"]!
    }
    
    /**
     * Get from the response the item named "next" in the set of "links".
     *
     * - Returns: The link to access last response, or *nil* if not found.
     */
    public func last() -> String {
        return self.vlinks["last"]!
    }
    
}
