/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation

/**
 * The Resource class represents a device resource.
 */
public class Resource { 
    private let secureConnection: HttpSecureConnection
    private let appId: String
    private let deviceId: String
    private let id: String
    private let description: String
    private let url: String
    private let methods: [String]
    
    init(secureConnection: HttpSecureConnection, appId: String, deviceId: String,
         id: String, description: String, url: String, methods: [String]) {
        self.secureConnection = secureConnection
        self.appId = appId
        self.deviceId = deviceId
        self.id = id
        self.description = description
        self.url = url
        self.methods = methods
    }
    
    /**
     * Get the device identifier this resource belongs to.
     *
     * - Returns: The device identifier that owns this resource.
     */
    public func getDeviceId() -> String {
        return self.deviceId
    }
    
    /**
     * Get the resource identifier.
     *
     * - Returns: The resource id.
     */
    public func getId() -> String {
        return self.id
    }
    
    /**
     * Get the free form description of this resource.
     *
     * - Returns: The resource description.
     */
    public func getDescription() -> String {
        return self.description
    }
    
    /**
     * Get the relative url to access this resource.
     *
     * - Returns: The resource url.
     */
    public func getURL() -> String {
        return self.url
    }
    
    /**
     * Get the methods allowed on this resource.
     8
     * - Returns: The methods.
     */
    public func getMethods() -> [String] {
        return self.methods
    }
    
    /**
     * Invoke the resource `GET` method and return a response when ready.
     *
     * - Parameter queryparam: `Dictionary` of `String`, `String` query.
     *
     * - Parameter headers: `Dictionary` of `String`, `String` the request
     *                      headers.
     *
     * - Returns: Response when ready.
     *
     * - throws: `ClientError.state` if `GET` is not one of the supported 
     *          `methods`.
     */
    public func get(queryparam: [String : String],  headers: [String : String],
                    response: @escaping (Response?, ClientError?) -> ()) throws  {
        if (!self.methods.contains("GET")) {
            let errorCause = "GET not supported."
            print(errorCause)
            throw ClientError.state(NSLocalizedString(errorCause, comment: ""))
        }
        
        let path = combineApiAndQuery(queryParams: queryparam)
        
        try self.secureConnection.get(restApi: path, headers: headers,
                                      response: { httpResponse, error in
            do {
                if (error != nil) {
                    response(nil, error)
                } else {
                    if let resObject = httpResponse {
                        let res = try ResponseImpl.from(path: path,
                            reqHeaders: headers, res: resObject,
                            secureConnection: self.secureConnection)
                        response(res, nil)
                    } else {
                        response(nil, ClientError.argument(NSLocalizedString("Resource get() httpResponse argument should not be nil", comment: "")))
                    }
                }
            } catch {
                response(nil, ClientError.network(NSLocalizedString("\(error)", comment: "")))
            }
        })
    }
    

    /**
     * Update a resource of a device with the `PUT` method and returns the
     * response when ready.
     *
     * - Parameter queryparam: `Dictionary` of `String`, `String` query.
     *
     * - Parameter headers: `Dictionary` of `String`, `String` the request
     *                      headers.
     *
     * - Parameter body: The request body content.
     *
     * - Returns: The response when ready.
     *
     * - throws: `ClientError.state` if `PUT` is not one of the supported
     *          `methods`.
     */
    public func put(queryparam: [String : String], headers: [String : String],
                    body: Data, response: @escaping (Response?, ClientError?) -> ()) throws {
        if (!self.methods.contains("PUT")) {
            let errorCause = "PUT not supported."
            print(errorCause)
            throw ClientError.state(NSLocalizedString(errorCause, comment: ""))
        }
        
        let path = combineApiAndQuery(queryParams: queryparam)
        
        try self.secureConnection.put(restApi: path, payload: body,
                                      headers: headers,
                                      response: { httpResponse, error in
            do {
                if (error != nil) {
                    response(nil, error)
                } else {
                    if let resObject = httpResponse {
                        let res = try ResponseImpl.from(path: path,
                            reqHeaders: headers, res: resObject,
                            secureConnection: self.secureConnection)
                        response(res, nil)
                    } else {
                        response(nil, ClientError.argument(NSLocalizedString("Resource put() httpResponse argument should not be nil", comment: "")))
                    }
                }
            } catch {
                response(nil, ClientError.network(NSLocalizedString("\(error)", comment: "")))
            }
        })
    }
    
    
    /**
     * Update a resource of a device with the `POST` method and return a
     * response when ready.
     *
     * - Parameter queryparam: `Dictionary` of `String`, `String` query.
     *
     * - Parameter headers: `Dictionary` of `String`, `String` the request
     *             headers.
     *
     * - Parameter body: The request body content.
     *
     * - Returns: The response when ready.
     *
     * - throws: `ClientError.state` if `POST` is not one of the supported
     *          `methods`.
     */
    public func post(queryparam: [String : String], headers:[String : String],
                     body: Data, response: @escaping (Response?, ClientError?) -> ()) throws {
        if (!methods.contains("POST")) {
            let errorCause = "POST not supported."
            print(errorCause)
            throw ClientError.state(NSLocalizedString(errorCause, comment: ""))
        }
        
        let path = combineApiAndQuery(queryParams: queryparam)
        
        try self.secureConnection.post(restApi: path, payload: body,
                                  headers: headers, response: { httpResponse, error in
            do {
                if (error != nil) {
                    response(nil, error)
                } else {
                    if let resObject = httpResponse {
                        let res = try ResponseImpl.from(path: path,
                            reqHeaders: headers, res: resObject,
                            secureConnection: self.secureConnection)
                        response(res, nil)
                    } else {
                        response(nil, ClientError.argument(NSLocalizedString("Resource post() httpResponse argument should not be nil", comment: "")))
                    }
                }
            } catch {
                response(nil, ClientError.network(NSLocalizedString("\(error)", comment: "")))
            }
        })
    }
    

    /**
     * Delete a resource of a device with the `DELETE` method and return a
     * response when ready.
     *
     * - Parameter headers: `Dictionary` of `String`, `String` the request
     *                      headers.
     *
     * - Returns: The response when ready.
     *
     * - throws: `ClientError.state` if `DELETE` is not one of the supported
     *          `methods`.
     */
    public func delete(headers: [String : String],
                       response: @escaping (Response?, ClientError?) -> ()) throws {
        if (!methods.contains("DELETE")) {
            let errorCause = "DELETE not supported."
            print(errorCause)
            throw ClientError.state(NSLocalizedString(errorCause, comment: ""))
        }
        
        let path = combineApiAndQuery(queryParams: [:])
        
        try self.secureConnection.delete(restApi: path, headers: headers,
                                         response: { httpResponse, error in
            do {
                if (error != nil) {
                    response(nil, error)
                } else {
                    if let resObject = httpResponse {
                        let res = try ResponseImpl.from(path: path,
                            reqHeaders: headers, res: resObject,
                            secureConnection: self.secureConnection)
                        response(res, nil)
                    } else {
                        response(nil, ClientError.argument(NSLocalizedString("Resource delete() httpResponse argument should not be nil", comment: "")))
                    }
                }
            } catch {
                response(nil, ClientError.network(NSLocalizedString("\(error)", comment: "")))
            }
        })
    }    
    
    private func combineApiAndQuery(queryParams: [String : String]) -> String {
        var query: String = ""
        
        if queryParams.count != 0 {
            var first: Bool = true
            for p in queryParams {
                if (first) {
                    query = "?"
                    first = false
                } else {
                    query += ","
                }
                
                query += p.0 + "=" + p.1
            }
        }
        
        return RestApi.V2.getReqRoot()+"/apps/" + self.appId + "/devices/" +
            self.deviceId + "/resources/" + self.url + query
    }
}
