/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import SwiftyJSON

/**
 * Provides means to lookup for resources of a device.
 */
public class ResourceEnumerator {
    

    /// The device to list resources

    private let  deviceId: String

    /// The secure connection to use

    private let secureConnection: HttpSecureConnection

    /// The application identifier: namespace to use to look for resources

    private let appId: String
    
    /**
     * Creates a resource enumerator to retrieve resources of a device.
     *
     * - Parameter client: The client to use to retrieve messages.
     *
     * - Parameter deviceId: The device identifier.
     *
     */
    init(client: EnterpriseClient, deviceId: String) {
        self.deviceId = deviceId
        self.secureConnection = SecureHttpConnectionMap.getSecureHttpConnection(client: client)
        self.appId = client.getApplication().getId();
    }
    
    /**
     * Get the resources for the specified *deviceId*.
     *
     * - Return: A list of resources for the device or *nil* if no resources
     *
     * - Throws: IOException if an I/O error occurred while trying to retrieve 
     *           resources from server
     *
     * - Throws: GeneralSecurityException when key or signature algorithm class 
     *           cannot be loaded, or the key is not in the trusted assets 
     *           store, or the private key is invalid.  If User Authentication 
     *           is used, then *UserAuthenticationException* will be thrown if 
     *           the session cookie has expired.
     *
     */
    public func getResources(resources: @escaping (PagedResponse<Resource>?, ClientError?) -> Void) throws -> Void {
        let resPage = ResourcePageable(appId: self.appId, deviceId: self.deviceId, secureConnection: self.secureConnection)
        resPage.loadFirstPage(offset: resPage.voffset, limit: resPage.vlimit, iterator: { iteratorResp, error in
            if (error != nil) {
                resources(nil, error)
            } else {
                if let iteratorObject = iteratorResp {
                    resources(iteratorObject, nil)
                } else {
                    resources(nil, ClientError.argument(NSLocalizedString("ResourceEnumerator getResources() iteratorResp should not be nil", comment: "")))
                }
                
            }
        })
    }
    
}

class ResourceResponse : PagedResponse<Resource> {
    
    init(resources: [Resource], offset: Int, total: Int, hasMore: Bool, links: [String : String]) {
        super.init(elements: resources, offset: offset, total: total, hasMore: hasMore, links: links)
    }
    
    
    static func fromJson(jsonObject: JSON, appId: String, secureConnection: HttpSecureConnection) -> PagedResponse<Resource> {
        
        let items = jsonObject.dictionaryValue["items"]?.arrayValue
        let links = jsonObject.dictionaryValue["links"]?.arrayValue
        var hasMore = jsonObject.dictionaryValue["hasMore"]?.boolValue
        if hasMore == nil {
            hasMore = false
        }
        
        var offset = jsonObject.dictionaryValue["offset"]?.intValue
        if offset == nil {
            offset = -1
        }
        
        var total = jsonObject.dictionaryValue["total"]?.intValue
        if total == nil {
            total = -1
        }
        
        var resources : [Resource] = [Resource]()
        
        do {
            if items != nil {
                for r in items! {
                    let newResource = try ResourceResponse.createResource(resource: r, appId: appId, secureConnection: secureConnection)
                    resources.append(newResource)
                }
            }
        } catch { }
        
        var map : [String : String] = [String : String]()
        if links != nil {
            for value in links! {
                let rel = value["rel"].stringValue
                let href = value["href"].stringValue
                if rel.isEmpty == false && href.isEmpty == false {
                    map[rel] = href
                }
            }
        }
        
        return ResourceResponse(resources: resources, offset: offset!, total: total!, hasMore: hasMore!, links: map)
    }

    
    private static func createResource(resource: JSON, appId: String, secureConnection: HttpSecureConnection) throws -> Resource {
        let endpointId = resource["endpointId"].stringValue
        let id = resource["id"].stringValue
        let description = resource["description"].stringValue
        let url = resource["url"].stringValue
        
        var mList : [String] = [String]()
        let methods = resource["methods"].arrayValue
        let size = methods.count
        for i in 0..<size {
            let method = methods[i].stringValue
            mList.append(method)
        }
        
        return Resource(secureConnection: secureConnection, appId: appId, deviceId: endpointId, id: id, description: description, url: url, methods: mList)
    }
}

class ResourcePageable : PageableEnumerator<Resource> {
    
    private let appId: String
    private let deviceId: String
    private let secureConnection: HttpSecureConnection
    
    var firstPage: PagedResponse<Resource>?
    
    init(appId: String, deviceId: String, secureConnection: HttpSecureConnection) {
        
        self.appId = appId
        self.deviceId = deviceId
        self.secureConnection = secureConnection
        
        super.init(offs: 0, lim: 200)
    }
    
    func loadFirstPage(offset: Int, limit: Int, iterator: @escaping (PagedResponse<Resource>?, ClientError?) -> Void) {
        // Since we don't know the total results, pre-fetch the first page
        do {
            try self.load(offset: self.voffset, limit: self.vlimit, response: { httpIterator, error in
                if (error != nil) {
                    iterator(nil, error)
                } else {
                    if let iteratorObject = httpIterator {
                        self.firstPage = iteratorObject
                        self.vhasMore = !self.firstPage!.elements().isEmpty
                        iterator(iteratorObject, nil)
                    } else {
                        iterator(nil, ClientError.argument(NSLocalizedString("ResourceEnumerator loadFirstPage() httpIterator should not be nil", comment: "")))
                    }
                }
            })
        } catch {
            iterator(nil, ClientError.network(NSLocalizedString("Error while retrieving first page of resources", comment: "")))
        }
        
    }
    
    func load(offset: Int, limit: Int, response: @escaping (PagedResponse<Resource>?, ClientError?) -> Void) throws -> Void {
        if self.firstPage != nil {
            let page = self.firstPage!
            if offset == 0 {
                response(page, nil)
            }
        }
        
        let headers = ["Accept" : "application/json"] 
        
        var reqResource: String = RestApi.V2.getReqRoot() + "/apps/" + appId
        reqResource += "/devices/" + deviceId
        reqResource += "/resources?offset=" + String(offset) + "&limit=" + String(limit)
        
        self.vhasMore = false
        
        try self.secureConnection.get(restApi: reqResource, headers: headers, response: { getResponse, error in
            if (error != nil) {
                response(nil, ClientError.network(NSLocalizedString("Unable to retrieve page of resources for offset " + offset.description, comment: "")))
            } else {
                if let countObject = getResponse {
                    let data = countObject.getData()
                
                    var jsonObject:JSON!
                    do {
                        jsonObject = try JSON(data: data as Foundation.Data)
                    } catch {
                        response(nil, ClientError.network(String(format:NSLocalizedString(
                            "%@: JSON threw exception: %@",
                                comment: ""), "ResourceEnumerator.load()", "\(error)")))
                        return
                    }
                    let resResponse = ResourceResponse.fromJson(jsonObject: jsonObject, appId: self.appId, secureConnection: self.secureConnection)
                    response(resResponse, nil)
                } else {
                    response(nil, ClientError.argument(NSLocalizedString("ResourceEnumerator load() getResponse should not be nil", comment: "")))
                }
            }
        })
    }
}


