/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import SwiftyJSON

/**
 * `VirtualDevice` for an enterprise-client adds methods to execute a write-only
 * or executable attribute.
 */

public protocol VirtualDevice : AbstractVirtualDevice {
    
    /**
     * Execute an action without an argument to the client
     *
     * - parameter actionName: The name of the action
     *
     * - throws: `ClientError.argument` if `actionName` is not a valid action
     *                                for this virtual device or if `actionName`
     *                                requires and argument.
     *
     */
    func call(actionName: String) throws
    
    /**
     * Execute an action with data to the target client
     *
     * - parameter actionName: The name of the action
     *
     * - parameter argument: argument for the action
     *
     * - throws: `ClientError.argument` if `actionName` is not a valid action
     *                                for this virtual device or if `argument`
     *                                is not the correct type for `name`.
     */
    func call(actionName: String, argument: AnyObject) throws

    /**
     * Set a callback that is invoked when any alert in the device model 
     * is received.
     *
     * Note that it is possible to set a callback for a all alerts
     * via `setOnAlert(callback:)`.
     *
     * - parameter formatURN: the format URN for the alert
     *
     * - parameter callback: a callback to invoke when an alert is received
     *
     * - seealso: `setOnAlert(callback:)`
     */
    func setOnAlert(formatURN: String, callback: @escaping (AlertEvent)->())
    
    /**
     * Set a callback that is invoked when any alert in the device model
     * is received.
     *
     * Note that it is possible to set a callback for a specific alert
     * via `setOnAlert(formatURN:callback:)`.
     *
     * - parameter callback: a callback for all alerts.
     *
     * - seealso: `setOnAlert(formatURN:callback:)`
     */
    func setOnAlert(callback: @escaping (AlertEvent) -> ())
    
    /**
     * Set a callback that is invoked when custom data for the given
     * format URN is received.
     *
     * Note that it is possible to set a callback for a specific data format and
     * to set a callback for all formats via `setOnData(callback:)`.
     * If there is a callback for the specific data and for all formats,
     * both callbacks will be invoked.
     *
     * - parameter formatURN: The data format URN from the device model
     *
     * - parameter callback: A callback to invoke when custom data is received with
     * the format URN
     *
     * - seealso: `setOnData(callback:)`
     */
    func setOnData(formatURN: String, callback: @escaping (DataEvent)->())
    
    /**
     * Set a callback that is invoked when any custom data in the device model
     * is received.
     *
     * Note that it is possible to set a callback for a specific data format
     * via `setOnData(formatURN:callback:)`.
     *
     * - parameter callback: a callback to invoke when custom data is received
     *
     * - seealso: `setOnData(formatURN:callback:)`
     */
    func setOnData(callback: @escaping (DataEvent)->())
}

/**
 * An event passed to the `Alert` callback.
 *
 * - seealso: `VirtualDevice.setOnAlert(callback:)`
 */
public protocol AlertEvent : Event {
    
    /**
     * Get the virtual device that is the source of the event
     *
     * - returns: The `VirtualDevice` that is the source of the event.
     */
    func getVirtualDevice() -> VirtualDevice
    /**
     * Get the name-value pair from the event.
     *
     * - returns: The `NamedValue` containing attributes related to the event.
     */
    func getNamedValue() -> NamedValue
    /**
     * Get the alert's device model format URN
     *
     * - returns: The format URN.
     */
    func getURN() -> String
    /**
     * Get the time the event was raised by the device-client
     *
     * - returns: The time of the event.
     */
    func getEventTime() -> Date
}


class AlertEventImpl : AlertEvent {
    
    var virtualDevice: VirtualDevice
    var urn: String
    var value: NamedValue
    var eventTime: Date
    
    init(virtualDevice: VirtualDevice, urn: String, value: NamedValue, eventTime: Date) {
        self.virtualDevice = virtualDevice
        self.urn = urn
        self.value = value
        self.eventTime = eventTime
    }
    

    func getVirtualDevice() -> VirtualDevice {
        return self.virtualDevice
    }

    func getNamedValue() -> NamedValue {
        return self.value
    }
    

    func getURN() -> String {
        return self.urn
    }
    
    func getEventTime() -> Date {
        return self.eventTime
    }
}

/**
 * An event passed to the `Data` callback.
 *
 * - seealso: `VirtualDevice.setOnData(callback:)`
 */
public protocol DataEvent : Event {
    
    /**
     * Get the virtual device that is the source of the event
     *
     * - returns: The `VirtualDevice` that is the source of the event.
     */
    func getVirtualDevice() -> VirtualDevice
    /**
     * Get the name-value pair from the event.
     *
     * - returns: The `NamedValue` containing attributes related to the event.
     */
    func getNamedValue() -> NamedValue
    /**
     * Get the format URN of the data
     *
     * - returns: the format URN
     */
    func getURN() -> String
    /**
     * Get the time the event was raised by the device-client
     *
     * - returns: The time of the event
     */
    func getEventTime() -> Date
}


class DataEventImpl : DataEvent {
    
    var virtualDevice: VirtualDevice
    var urn: String
    var value: NamedValue
    var eventTime: Date
    
    init(virtualDevice: VirtualDevice, urn: String, value: NamedValue, eventTime: Date) {
        self.virtualDevice = virtualDevice
        self.urn = urn
        self.value = value
        self.eventTime = eventTime
    }
    
    func getVirtualDevice() -> VirtualDevice {
        return self.virtualDevice
    }
    
    func getNamedValue() -> NamedValue {
        return self.value
    }
    
    func getURN() -> String {
        return self.urn
    }
    
    func getEventTime() -> Date {
        return self.eventTime
    }
}

class VirtualDeviceImpl : VirtualDevice {
    
    private let secureConnection: HttpSecureConnection
    var base: VirtualDeviceBase?
    private let enterpriseClient: EnterpriseClient
    private var visDeviceApp: Bool?
    
    var attributeMap: [String : VirtualDeviceAttribute] = [ : ]
    var formatMap: [String : VirtualDeviceFormat] = [ : ]
    
    private var updatedAttributes = [(VirtualDeviceAttribute , AnyObject)]()
    private var updateMode : Bool = false
    
    private var onAlert: ((AlertEvent) -> ())?
    private var onData: ((DataEvent) -> ())?
    
    /*
     * first arg - application id
     * second arg - device id
     * third arg - model urn
     */
    private let DEVICE_ATTRIBUTES_CONTROL_FORMAT: String =
        RestApi.V2.getReqRoot() + "/apps/%@/devices/%@/deviceModels/%@/attributes"
    private let DEVICE_APP_ATTRIBUTES_CONTROL_FORMAT: String =
        RestApi.V2.getReqRoot() + "/apps/%@/deviceApps/%@/deviceModels/%@/attributes"
    
    /*
     * first arg - application id
     * second arg - device id
     * third arg - model urn
     * fourth arg - attribute name
     */
    private let DEVICE_ACTION_FORMAT: String = RestApi.V2.getReqRoot() +
        "/apps/%@/devices/%@/deviceModels/%@/actions/%@"
    
    private let DEVICE_APP_ACTION_FORMAT: String = RestApi.V2.getReqRoot() +
        "/apps/%@/deviceApps/%@/deviceModels/%@/actions/%@"
    
    /**
     * first arg - actionName
     * second arg - value
     */
    private static let PAYLOAD_FORMAT: String = "{ \"%@\" : \"%@\" }"
    
    /**
     * first arg - actionName
     * second arg - value
     */
    private static let NUMBER_PAYLOAD_FORMAT: String = "{ \"%@\" : %@ }"
    
    let sync = DispatchQueue(label: "vdqueue", attributes: [])
    
    /**
     * Initialize a new `VirtualDevice`
     *
     * - parameter enterpriseClient: The client using this virtual device.
     *
     * - parameter secureConnection: The secure connection to be used by this 
     *                               virtual device.
     *
     * - parameter endpointId: The identifier of the device (or deviceApp).
     *
     * - parameter model: The model implemented by the device.
     */
    init(enterpriseClient: EnterpriseClient, secureConnection: HttpSecureConnection,
            endpointId: String, model: DeviceModelImpl) {
        self.secureConnection = secureConnection
        self.enterpriseClient = enterpriseClient;
        self.base = VirtualDeviceBase(endpointId: endpointId, deviceModel: model)
        
        self.attributeMap = VirtualDeviceImpl.createAttributeMap(virtualDevice: self, deviceModel: model)
        self.formatMap = VirtualDeviceImpl.createFormatMap(virtualDevice: self, deviceModel: model)
    }

    private static func createAttributeMap(virtualDevice: VirtualDeviceImpl,
            deviceModel: DeviceModelImpl) -> [String : VirtualDeviceAttribute] {
        
        var map: [String : VirtualDeviceAttribute] = [ : ]
        
        for attribute in deviceModel.getDeviceModelAttributes().values {
            let vda: VirtualDeviceAttribute = VirtualDeviceAttribute(model: (attribute))
            map[attribute.getName()!] = vda
            let alias: String? = attribute.getAlias()
            if let alias = alias, !alias.isEmpty {
                map[alias] = vda
            }
        }
        return map
    }

    private static func createFormatMap(virtualDevice: VirtualDeviceImpl,
            deviceModel: DeviceModelImpl) -> [String : VirtualDeviceFormat] {
        
        var map: [String : VirtualDeviceFormat] = [ : ]
        
        for format in deviceModel.getDeviceModelFormats().values {
            let NAME = format.getName()
            let DESC = format.getDescription()
            let TYPE = format.getType()
            let URN = format.getURN()
            let FIELDS = format.getFields()
            
            if TYPE == DeviceModelFormatType.ALERT {
                let dmf = DeviceModelFormat(urn: URN, name: NAME,
                                            description: DESC, type: TYPE,
                                            fields: FIELDS)
                let vdf: VirtualDeviceFormat = VirtualDeviceFormat(dmFormats: dmf)
            
                map[URN] = vdf
            }
            else {
                if TYPE == DeviceModelFormatType.DATA {
                    let dmf = DeviceModelFormat(urn: URN, name: NAME,
                                                description: DESC, type: TYPE,
                                                fields: FIELDS)
                    let vdf: VirtualDeviceFormat = VirtualDeviceFormat(dmFormats: dmf)
                
                    map[URN] = vdf
                }
            }
        }
        
        return map
    }

    func getEnterpriseClient() -> EnterpriseClient {
        return self.enterpriseClient
    }
    
    func getSecureConnection() -> HttpSecureConnection {
        return self.secureConnection
    }
    
    func getAttribute(name: String) throws -> VirtualDeviceAttribute {
        let virtualDeviceAttribute: VirtualDeviceAttribute? = self.attributeMap[name]
        
        if (virtualDeviceAttribute == nil) {
            throw ClientError.argument(String(format: NSLocalizedString(
                "Unexpected attribute %@", comment: ""), name))
        }
        
        return virtualDeviceAttribute!
    }

    func get(attributeName: String) -> AnyObject? {
        var attribute: VirtualDeviceAttribute?
        for item in self.attributeMap {
            if item.1.model.getName() == attributeName {
                attribute = item.1
            }
        }
        return attribute!.get()
    }
    
    func getLastKnown(attributeName: String) throws -> AnyObject? {
        let virtualDeviceAttribute: VirtualDeviceAttribute? =
            try self.getAttribute(name: attributeName)
        
        return virtualDeviceAttribute?.getLastKnown()
    }
    
    func getFormat(formatName: String) throws -> VirtualDeviceFormat? {
        if self.formatMap[formatName] == nil {
            throw ClientError.argument(String(format: NSLocalizedString(
                "Format %@ is not a format of this device", comment: ""),
                formatName))
        }
        
        if (self.formatMap.count == 0) {
            throw ClientError.argument(String(format: NSLocalizedString(
                "Unexpected format %@", comment: ""), formatName))
        }
        
        let format: VirtualDeviceFormat? = self.formatMap[formatName]!
        
        return format
    }
    
    func getDeviceModel() -> DeviceModel {
        return self.base!.getDeviceModel()
    }

    func getEndpointId() -> String {
        return self.base!.getEndpointId()
    }

    @discardableResult
    func initValue(attributeName: String, value: AnyObject) -> VirtualDeviceImpl {
        for item in self.attributeMap {
            if item.1.model.getName() == attributeName {
                item.1.set(value: value)
                item.1.lastKnownValue = item.1.value
            }
        }
        return self
    }
    
    func set(attributeName: String, attributeValue: AnyObject) throws -> VirtualDevice {
        let virtDevAttr = try self.getAttribute(name: attributeName)

        if !virtDevAttr.isSettable() {
            throw ClientError.argument(String(format: NSLocalizedString(
                "Virtual Device: attempt to modify read-only attribute",
                comment: ""), attributeName))
        }

        if let deviceModelAttribute = virtDevAttr.getDeviceModelAttribute() {
            try virtDevAttr.validate(attribute: deviceModelAttribute,
                                     value: attributeValue)
        } else {
            throw ClientError.argument(String(format: NSLocalizedString(
                "Attribute name %@ does not exist on this device model",
                comment: ""), attributeName))
        }
        
        if self.updateMode {
            sync.sync {
                self.updatedAttributes.append((virtDevAttr, attributeValue))
            }
        } else {
            self.updateValue(virtDevAttr: virtDevAttr, name: attributeName,
                             value: attributeValue)
        }
        return self
    }

    func update() -> VirtualDevice {
        sync.sync {
            self.updateMode = true
        }
        return self
    }
    
    func finish() throws {
        sync.sync {
            if (self.updateMode) {
                self.updateFields(updatedAttributes: self.updatedAttributes)
            }
        }
    }
    
    func setOnChange(attributeName: String, callback: @escaping  (ChangeEvent) -> ()) throws {
        let attribute = try self.getAttribute(name: attributeName)
        attribute.setOnChange(callback: callback)
    }
    
    func setOnChange(callback: @escaping (ChangeEvent) -> ())  {
        self.base!.setOnChange(callback: callback)
    }
  
    func setOnError(attributeName: String, callback: @escaping (ErrorEvent) -> ()) throws {
        let attribute = try self.getAttribute(name: attributeName)
        attribute.setOnError(callback: callback)
    }

    func setOnError(callback: @escaping (ErrorEvent) -> ())  {
        self.base!.setOnError(callback: callback)
    }
    
    func setOnAlert(formatURN: String, callback: @escaping (AlertEvent) -> ()) {
        do {
            let alert = try self.getFormat(formatName: formatURN)
            alert!.setOnAlert(callback: callback)
        } catch { }
    }
    
    private var syncQueue:DispatchQueue = DispatchQueue.global(qos: .utility)
    
    func setOnAlert(callback: @escaping (AlertEvent) -> ()){
        syncQueue.sync(execute: {
            self.onAlert = callback
        })
    }
    
    func getOnAlertCallback() -> ((AlertEvent) -> ())? {
        return syncQueue.sync(execute: {
            return self.onAlert
        })
    }
    
    func setOnData(formatURN: String, callback: @escaping (DataEvent) -> ()) {
        do {
            let customData = try self.getFormat(formatName: formatURN)
            customData!.setOnData(callback: callback)
        } catch {
            // TODO: Are we sure we don't want to forward this error ?
        }
    }
    
    func setOnData(callback: @escaping (DataEvent) -> ()) {
        syncQueue.sync(execute: {
            self.onData = callback
        })
    }
    
    func getOnDataCallback() -> ((DataEvent) -> ())? {
        return syncQueue.sync(execute: {
            return self.onData
        })
    }
    
    func getURN() -> String {
        return (self.getDeviceModel().getURN())
    }
    
    func getEventTime() -> Date {
        return Date()
    }
    
    func triggerActionErrorEvent(actionName: String, error: ClientError) {
        let nameValue = NamedValueImpl(name: actionName, value: nil)
        let errEvent: ErrorEventImpl = ErrorEventImpl(virtualDevice: self,
                        value: nameValue, message: error.localizedDescription)
        
        if self.base!.getOnErrorCallback() != nil {
            self.base!.getOnErrorCallback()!(errEvent)
        }
    }
    
    func call(actionName: String) throws {
        var action: DeviceModelAction? = nil
        let actionDict = (self.getDeviceModel() as! DeviceModelImpl).getDeviceModelActions()
        if actionDict.count > 0 {
            action = actionDict[actionName]
            if (nil == action) {
                for item in actionDict {
                    if (item.1.getAlias() == actionName) {
                        action = item.1
                        break
                    }
                }
            }
        }
        
        if (nil == action) {
            throw ClientError.argument(String(format: NSLocalizedString(
                "Action name: %@ not found in device model", comment: ""),
                actionName))
        }
        
        if (action?.getArgType()) != nil {
            throw ClientError.argument(String(format: NSLocalizedString(
                "Action name: %@ requires and argument", comment: ""),
                actionName))
        }
        
        var format : String = DEVICE_ACTION_FORMAT
        self.isDeviceApp(virtDeviceResponse: {response, error in
            if (error != nil) {
                self.triggerActionErrorEvent(actionName: actionName, error: error!)
                return
            }
            
            if (true == self.visDeviceApp) {
                format = self.DEVICE_APP_ACTION_FORMAT
            }
            let resource = String(format: format,
                                   self.enterpriseClient.getApplication().getId(),
                                   self.getEndpointId(),
                                   self.base!.getDeviceModel().getURN(),
                                   actionName)
            Controller.update(virtualDevice: self, method:"POST",
                              resource:resource, payload:"{}",
                              callback: { error in
                if (error != nil) {
                    self.triggerActionErrorEvent(actionName: actionName, error: error!)
                }
            })
        })
    }

   func call(actionName: String, argument: AnyObject) throws {
        var action: DeviceModelAction? = nil
        let actionDict = (self.getDeviceModel() as! DeviceModelImpl).getDeviceModelActions()
        if actionDict.count > 0 {
            action = actionDict[actionName]
            if (nil == action) {
                for item in actionDict {
                    if (item.1.getAlias() == actionName) {
                        action = item.1
                        break
                    }
                }
            }
        }
        if (nil == action) {
            throw ClientError.argument(String(format: NSLocalizedString(
                "Action name: %@ not found in device model", comment: ""),
                actionName))
        }
        try checkActionDataType(type: action?.getArgType(), name: actionName,
                                value: argument)
        var payload = ""
        let argType = action?.getArgType()
        if argType != nil {
            var argString: String
            switch argType! {
            case .BOOLEAN:
                argString = String(argument as! Bool)
                break
            case .URI:
                // checked and it is an ExternalObject
                argString = (argument as! ExternalObject).getURI()
                break
            default:
                argString = String(describing: argument)
            }
            payload = String("{\"value\":" + argString + "}")
        }

        let performUpdate = {
            var format : String = self.DEVICE_ACTION_FORMAT
            self.isDeviceApp(virtDeviceResponse: { response, error in
                if (error != nil) {
                    self.triggerActionErrorEvent(actionName: actionName,
                                                 error: error!)
                    return
                }
                
                if (true == self.visDeviceApp) {
                    format = self.DEVICE_APP_ACTION_FORMAT
                }
                let resource = String(format: format,
                                      self.enterpriseClient.getApplication().getId(),
                                      self.getEndpointId(),
                                      self.base!.getDeviceModel().getURN(),
                                      actionName)
                
                Controller.update(virtualDevice: self, method: "POST",
                                  resource: resource, payload: payload,
                                  callback: { error in
                    if (error != nil) {
                        self.triggerActionErrorEvent(actionName: actionName,
                                                     error: error!)
                    }
                })
            })
        }

        if argType == DeviceModelAttributeType.URI {
            self.syncStorageObject(name:actionName,
                value:argument, completeUpdate: performUpdate)
        } else {
            performUpdate()
        }
    }

    // TODO: make this simple request synchronous
    /// Performs a REST call to check if this virtualDevice is a device.
    ///
    /// - Parameter virtDeviceResponse: Callback for getting the virtual device.
    func isDeviceApp(virtDeviceResponse: @escaping (VirtualDevice?, ClientError?) -> ()) {
        if self.visDeviceApp == nil {
            do {
                let headers: [String:String] = [
                    "Content-Type" : "application/json",
                    "Accept" : "application/json"
                ]
                let request = RestApi.V2.getReqRoot() + "/apps/" +
                    self.enterpriseClient.getApplication().getId() +
                    "/deviceApps?fields=type&q=" + "{\"id\":{\"$eq\":\"" +
                    self.getEndpointId() + "\"}}"

                self.visDeviceApp = false
                try self.httpGet(secureConnection: secureConnection,
                                 headers: headers, request: request,
                                 jsonResponse:  { responseJson, error in
                    if (error != nil) {
                        virtDeviceResponse(nil, error)
                        return
                    } else {
                        if let jsonObject = responseJson {
                            let items = jsonObject.dictionaryValue["items"]?.arrayValue
                            
                            // if the endpoint id is a device and not a deviceApp,
                            // the list will be empty
                            if items != nil && items?.count != 0 {
                                let type = items?.first!["type"].stringValue
                                if type == "DEVICE_APPLICATION" {
                                    self.visDeviceApp = true
                                }
                            }
                            virtDeviceResponse(self, nil)
                            return
                        } else {
                            virtDeviceResponse(nil, ClientError.argument(NSLocalizedString(
                                "VirtualDevice isDeviceApp() responseJson should not be nil",
                                comment: "")))
                            return
                        }
                    }
                })
            } catch {
                virtDeviceResponse(nil, ClientError.network(NSLocalizedString(
                    "Error while retrieving deviceApps response", comment: "")))
                return
            }
        } else {
            virtDeviceResponse(self, nil)
        }
    }
    
    
    func getIsDeviceApp() -> Bool? {
        return self.visDeviceApp
    }
    
    
    private func httpGet(secureConnection: HttpSecureConnection,
                         headers: [String : String],
                         request: String,
                         jsonResponse: @escaping ((JSON?, ClientError?) -> ())) throws {
        try secureConnection.get(restApi: request, headers: headers,
                                 response: { (getHttp, error) -> () in
            if (error != nil) {
                jsonResponse(nil, error)
            } else {
                if let httpResponse = getHttp {
                    let status = httpResponse.getStatus()
                    if status != StatusCode.OK.getCode() {
                        let s = StatusCode.valueOf(code: status)
                        jsonResponse(nil, ClientError.argument(NSLocalizedString(
                            "VirtualDevice httpGet() http get failed with \(s)",
                            comment: "")))
                        return
                    }
                    
                    let dataResponse = httpResponse.getData()
                    if dataResponse.count == 0 {
                        jsonResponse(nil, ClientError.argument(NSLocalizedString(
                            "VirtualDevice httpGet() http get failed: no data received.",
                            comment: "")))
                        return
                    }
                    var dataObject:JSON!
                    do {
                        dataObject = try JSON(data: dataResponse)
                    } catch {
                        jsonResponse(nil, ClientError.network(String(format:NSLocalizedString(
                            "%@: JSON threw exception: %@",
                            comment: ""),"VirtualDevice.httpGet()", "\(error)")))
                        return
                    }
                    jsonResponse(dataObject, nil)
                    return
                } else {
                    jsonResponse(nil, ClientError.argument(NSLocalizedString(
                        "VirtualDevice httpGet() httpResponse should not be nil",
                        comment: "")))
                }
            }
        })
    }

    func triggerUpdateValueErrorEvent(virtDevAttr: VirtualDeviceAttribute,
                                      error: ClientError) {
        virtDevAttr.value = virtDevAttr.lastKnownValue
        virtDevAttr.localUpdateRequest = false
        
        let nameValue = NamedValueImpl(name: (virtDevAttr.model.getName())!,
                value: (virtDevAttr.value == nil ? nil : virtDevAttr.value!))
        let errEvent: ErrorEventImpl = ErrorEventImpl(virtualDevice: self,
                value: nameValue, message: error.localizedDescription)
        
        if virtDevAttr.getOnError() != nil {
            virtDevAttr.getOnError()!(errEvent)
        }
        
        if self.base!.getOnErrorCallback() != nil {
            self.base!.getOnErrorCallback()!(errEvent)
        }
    }

    /**
     * Sets the the value for attribute.
     *
     * - parameter deviceAttribute: the device attribute name
     *
     * - parameter value: the device attribute value
     */
    private func updateValue(virtDevAttr: VirtualDeviceAttribute, name: String,
                             value: AnyObject) {
        
        let type = virtDevAttr.getDeviceModelAttribute()?.getType()
        let payload = "{\"value\":" + String(describing: value) + "}"
        let performUpdate = {
            self.isDeviceApp(virtDeviceResponse: { response, error in
                if error != nil {
                    self.triggerUpdateValueErrorEvent(virtDevAttr: virtDevAttr, error: error!)
                    return
                }
                var format : String = self.DEVICE_ATTRIBUTES_CONTROL_FORMAT
                if true == self.visDeviceApp {
                    format = self.DEVICE_APP_ATTRIBUTES_CONTROL_FORMAT
                }
                var resource:String = String(format: format,
                                             self.enterpriseClient.getApplication().getId(),
                                             self.getEndpointId(),
                                             self.base!.getDeviceModel().getURN())
                resource.append("/")
                resource.append(name)
                
                virtDevAttr.localUpdateRequest = true
                virtDevAttr.set(value: value)
                
                Controller.update(virtualDevice: self, method: "PUT",
                                  resource: resource, payload: payload,
                                  callback: {error in
                    if (error != nil) {
                        self.triggerUpdateValueErrorEvent(virtDevAttr: virtDevAttr,
                                                          error: error!)
                    } else {
                        virtDevAttr.localUpdateRequest = false
                    }
                })
            })
        }
        
        if type == .URI {
            syncStorageObject(name: name, value: value,
                              completeUpdate: performUpdate)
        } else {
            performUpdate()
        }
    }
    
    func triggerUpdateFieldsErrorEvent(virtDevAtrrArray: [VirtualDeviceAttribute],
                                       error: ClientError) {
        var root: NamedValue? = nil
        var last: NamedValueImpl? = nil
        
        for virtDevAttr in virtDevAtrrArray {
            virtDevAttr.value =  virtDevAttr.lastKnownValue;
            virtDevAttr.localUpdateRequest = false
            
            let nameValue = NamedValueImpl(name: (virtDevAttr.model.getName())!,
                                           value: virtDevAttr.value)
            let errEvent: ErrorEventImpl = ErrorEventImpl(virtualDevice: self,
                        value: nameValue, message: error.localizedDescription)
            
            if virtDevAttr.getOnError() != nil {
                virtDevAttr.getOnError()!(errEvent)
            }
            
            if (last != nil) {
                last!.setNext(next: nameValue)
                last = nameValue
            } else {
                root = nameValue
                last = nameValue
            }
        }
        
        if self.base!.getOnErrorCallback() != nil {
            let errDeviceEvent: ErrorEventImpl =
                ErrorEventImpl(virtualDevice: self, value: root!,
                               message: error.localizedDescription)
            self.base!.getOnErrorCallback()!(errDeviceEvent)
        }
        
        self.updatedAttributes.removeAll()
        self.updateMode = false
    }
    
    /**
     * Sets the the values for multiple attributes.
     *
     * - parameter updatedAttributes: the device attributes to be updated
     *                                including their new values
     */
    private func updateFields(updatedAttributes : [(VirtualDeviceAttribute, AnyObject)]) {
        var haveURI:Bool = false
        var urlAttrName:String?
        var urlAttrObj:VirtualDeviceAttribute?
        
        var payload:String = "{"
        
        var virtDevAtrrArray = [VirtualDeviceAttribute]()
        for (virtDevAttrObj, virtDevAttrVal) in updatedAttributes {
            let virtDevAttrName:String =
                (virtDevAttrObj.getDeviceModelAttribute()?.getName())!
            //var attrValue:String?
            let attrType:DeviceModelAttributeType =
                (virtDevAttrObj.getDeviceModelAttribute()?.getType())!
            
            if attrType == .URI {
                // Add the uri to the payload now, sync will occur before update
                haveURI = true
                urlAttrName = virtDevAttrName
                urlAttrObj = virtDevAttrObj
                break
            }

            payload = payload + getPayload(type:attrType, name: virtDevAttrName,
                                           value: virtDevAttrVal)
            
            virtDevAttrObj.localUpdateRequest = true
            virtDevAttrObj.set(value: virtDevAttrVal)
            virtDevAtrrArray.append(virtDevAttrObj)
        }
        // TODO: Why is the end of the string calculated ?
        /*
        payload = (payload?.substring(to: (payload?.index(
            before: (payload?.characters.index(before: (payload?.endIndex)!))!))!))! + "}"
        */
        payload = payload + "}"
        
        let performUpdate = {
            self.isDeviceApp(virtDeviceResponse: { response, error in
                if (error != nil) {
                    self.triggerUpdateFieldsErrorEvent(virtDevAtrrArray: virtDevAtrrArray,
                                                       error: error!)
                    return
                }
                var format : String = self.DEVICE_ATTRIBUTES_CONTROL_FORMAT
                if (true == self.visDeviceApp) {
                    format = self.DEVICE_APP_ATTRIBUTES_CONTROL_FORMAT
                }
                let resource = String(format: format,
                              self.enterpriseClient.getApplication().getId(),
                              self.getEndpointId(),
                              self.base!.getDeviceModel().getURN())
                
                Controller.update(virtualDevice: self, method: "PATCH",
                                  resource: resource, payload: payload,
                                  callback: { error in
                    if (error != nil) {
                        self.triggerUpdateFieldsErrorEvent(virtDevAtrrArray: virtDevAtrrArray,
                                                           error: error!)
                    } else {
                        for virtDevAttr in virtDevAtrrArray {
                            virtDevAttr.localUpdateRequest = false
                        }
                    }
                })
            })
        }
        
        if haveURI {
            syncStorageObject(name:urlAttrName!, value: urlAttrObj!,
                               completeUpdate: performUpdate)
        } else {
            performUpdate()
        }
    }

    private func checkActionDataType(type: DeviceModelAttributeType?, name: String, value: AnyObject?) throws {
        switch (type) {
        case .INTEGER?:
            if (!(value is Int)) {
                throw ClientError.argument(String(format: NSLocalizedString(
                    "%@ requires an Integer argument", comment: ""), name))
            }
            break
        case .NUMBER?:
            if (!(value is NSNumber)) {
                throw ClientError.argument(String(format: NSLocalizedString(
                    "%@ requires a Number argument", comment: ""), name))
            }
            break
        case .STRING?:
            if (!(value is String)) {
                throw ClientError.argument(String(format: NSLocalizedString(
                    "%@ requires a String argument", comment: ""), name))
            }
            break
        case .BOOLEAN?:
            if (!(value is Bool)) {
                throw ClientError.argument(String(format: NSLocalizedString(
                    "%@ requires a Boolean argument", comment: ""), name))
            }
            break
        case .DATETIME?:
            if (!(value is Int64)) {
                throw ClientError.argument(String(format: NSLocalizedString(
                    "%@ requires a Long argument", comment: ""), name))
            }
            break
        case .URI?:
            if (!(value is ExternalObject)) {
                throw ClientError.argument(String(format: NSLocalizedString(
                    "%@ requires an ExternalObject argument", comment: ""), name))
            }
            break;
        default:
            throw ClientError.argument(String(format: NSLocalizedString(
                "Unknown type for %@", comment: ""), name))
        }
    }
    
    private func syncStorageObject(name:String, value:AnyObject,
                                   completeUpdate: @escaping ()->()) {
        let storageObject:StorageObjectImpl = value as! StorageObjectImpl
        storageObject.setSyncEventInfo(virtualDevice: self, nameForSyncEvent:name)

        do {
            try storageObject.sync(callback: { (error) in
                if error != nil {
                    self.triggerActionErrorEvent(actionName: name,
                        error: ClientError.network(NSLocalizedString(
                            "Content sync failed.", comment: "")))
                    return
                } else {
                    completeUpdate()
                }
            })
        } catch {
            print(error)
            self.triggerActionErrorEvent(actionName: name,
                            error: ClientError.network(NSLocalizedString(
                            "Content sync failed.", comment: "")))
        }
        return
    }
    
    private func getPayload(type:DeviceModelAttributeType?, name:String,
                            value:AnyObject?) -> String {
        // JSON nil value is literal "null"
        var payload:String = "{\"" + name + "\": null}"
        if let type = type, let value = value {
            switch(type) {
            case .INTEGER, .NUMBER, .BOOLEAN:
                payload = String(format: VirtualDeviceImpl.NUMBER_PAYLOAD_FORMAT,
                                name, String(describing: value))
                break
            case .DATETIME:
                var dateTime: String
                if value is Date {
                    dateTime = String((value as! Date).timeIntervalSince1970)
                } else {
                    dateTime = String(describing: value)
                }
                payload = String(format:VirtualDeviceImpl.NUMBER_PAYLOAD_FORMAT,
                                        name, dateTime);
                break
            case .URI:
                if value is ExternalObject {
                    payload = String(format:VirtualDeviceImpl.PAYLOAD_FORMAT,
                            name, (value as! ExternalObject).getURI())
                }
                break;
            case .STRING:
                fallthrough
            default:
                payload = String(format:VirtualDeviceImpl.PAYLOAD_FORMAT,
                                        name, value as! String)
                break;
            }
        }
        return payload
    }
}
