/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import QuartzCore
import SwiftyJSON

/**
 * Access Token
 */
class AccessToken {
    var expires: Int64
    var tokenType: String
    var token: String
    var expirationTime: Int64
    
    init(expires: Int64, tokenType: String, token: String) {
        self.expires = expires
        self.tokenType = tokenType
        self.token = token
        self.expirationTime = TimeManager.currentTimeMillis() + expires
    }
    
    func hasExpired() -> Bool {
        return TimeManager.currentTimeMillis() >= self.expirationTime
    }
    
    func getExpires() -> Int64 {
        return self.expires
    }
    
    func getTokenType() -> String {
        return self.tokenType
    }
    
    func getToken() -> String {
        return self.token
    }
    
    func toString() -> String {
        var result:String = "Accessor.Token{"
        result += "expires=" + expires.description
        result += ", tokenType='" + tokenType + "'"
        result += ", token='" + token + "'" + "}"
        return  result
    }
    
    func clone() -> AccessToken {
        let newtoken = AccessToken(expires: expires, tokenType: tokenType,
                                   token: token)
        // Make sure to set the original expirationTime
        newtoken.expirationTime = expirationTime
        return newtoken
    }
    
    static func fromJSON(jsonObject : JSON) -> AccessToken {
        return AccessToken(expires: Int64(jsonObject["expires_in"].intValue),
                           tokenType: jsonObject["token_type"].string!,
                           token: jsonObject["access_token"].string!)
    }
    
    static func fromJSON(data : Foundation.Data?) throws -> AccessToken? {
        if data == nil {
            return nil
        }
        let jsonStr: String = String(data: data!, encoding: .utf8)!
        let jsonDict: [String: AnyObject]? = try parseJsonStr(str: jsonStr)
        if(jsonDict == nil) {return nil }
        let exp = Int64(jsonDict!["expires_in"]!.int32Value)
        let tokentype = jsonDict!["token_type"]! as! String
        let token = jsonDict!["access_token"]! as! String
        return AccessToken(expires: exp, tokenType: tokentype,token: token)
    }

    private static func parseJsonStr(str: String) throws -> [String : AnyObject]? {
        if let data = str.data(using: .utf8) {
            return try JSONSerialization.jsonObject(with: data, options: []) as? [String : AnyObject]
        }
        return nil
    }
}
