/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import Heimdall
import Security
import Alamofire

/**
 * The DefaultTrustedAssetManager class implementation for iOS
 */
class DefaultTrustedAssetsManager : TrustedAssetsManagerImpl {

    /// the default Name of the provisioning file
    @objc let provisionFileName:String = TAMconstants.provisionFileName
    /// the default file extension of the provisioning file
    @objc let fileExtension:String = TAMconstants.fileExtension
    // The Prefix for the tags in the keychain for the private and public keys
    @objc let tamPrefix  = "com.oracle.iot"

    private override init() {
        super.init()
        self.isActive = false
        self.isProvisioned = false
        self.tagPublic  = tamPrefix + keyPublic
        self.tagPrivate = tamPrefix + keyPrivate
    }
    
    @objc convenience init(context: Dictionary<String, Any>?) throws {
        self.init()
        if let context = context {
            let fileName = context[TAMconstants.fileKey] as? String
            let fileEXT = context[TAMconstants.fileExtKey] as? String
            let sub = context[TAMconstants.subdKey] as? String
            if let path = context[TAMconstants.pathKey] as? String {
                try self.load(urlToResource: URL(fileURLWithPath: path))
            } else {
                if let urlTR:URL = TrustedAssetsManagerImpl.URLForResource(
                        fileName: fileName!, withExtension: fileEXT!,
                        subdirectory: sub) {
                    try self.load(urlToResource: urlTR)
                }
            }
        }
    }
    
    /**
     * Deinitializes the TAM instance. Will call the reset method.
     */
    deinit {
        try! self.reset()
    }
    
//------------------------------------ Impl Provisioning methods ---------------

        
    private func load(urlToResource: URL) throws {
        self.isProvisioned = try provisionWithFileNameJSON(fileURL: urlToResource)
    }
    
    /**
     * Read the provisioning file in JSON format from the Resource folder of 
     * the Application Bundle
     *
     * - parameter fileName: the name of the file the file type must be json
     * - returns: true if the reading was successful
     */
    private func provisionWithFileNameJSON(fileURL: URL) throws -> Bool  {
        let provisionFileURL = fileURL
        let provisionData = try? Foundation.Data(contentsOf: provisionFileURL)
        let jsonObject: AnyObject?
        jsonObject = try JSONSerialization.jsonObject(with: provisionData!, options: JSONSerialization.ReadingOptions(rawValue: 0)) as! [String: AnyObject] as AnyObject?
        return readStringProvisionDatafromJSON(jsonData: jsonObject as! [String : AnyObject])
    }
    
    /**
     * Reads the provisioning data from the JSON file
     */
    private func readStringProvisionDatafromJSON(jsonData: [String: AnyObject]) -> Bool {
        self.deviceId = readElementAsString(jsonData: jsonData, name: "deviceId")!
            if(self.deviceId == "") {return false}
        self.endpointId = readElementAsString(jsonData: jsonData, name: "endpointId")!
        self.serverHost   = readElementAsString(jsonData: jsonData, name: "serverHost")!
             if(self.serverHost == "") {return false}
        self.serverPort   = readElementAsString(jsonData: jsonData, name: "serverPort")!
             if(self.serverPort == "") {return false}
        self.sharedSecret = readElementAsString(jsonData: jsonData, name: "sharedSecret")!
             if(self.sharedSecret == "") {return false}
        let active = readElementAsString(jsonData: jsonData, name: "activated")!
        self.isActive = Bool(String(active))!

        // read a root certificate from the Resources
        if(self.rootCACert == nil) {
            self.rootCACert = readCertificatefromDERFile(fileName: "iotca");
        }
        if(self.serverCert == nil) {
            self.serverCert = readCertificatefromDERFile(fileName: "iotserver"); // read the server certificate from the DER file in the resource folder
        }
        //----- Trust evaluation testing------------
        let trust = trustWithCertificates(certificates: [self.serverCert!,
                                           self.rootCACert!])
        let anchorSet = setRootCertificateAsLoneAnchorCertificateForTrust(trust: trust!,root: self.serverCert!)
        if(!anchorSet) {return false}
        // ---- set policies and
        let policies = [SecPolicyCreateBasicX509()]
        SecTrustSetPolicies(trust!, policies as CFTypeRef)
        let valid = trustIsValid(trust: trust!)
        if(!valid) {return false}
        var validPolicy = false
        if(valid){
            validPolicy = self.setServerTrustPolicy(hostName: self.getServerHost(), trust: trust!, certificateChain: [self.rootCACert!])
            if(!validPolicy) {return false}
        }
        // ------------------------------------------
        return true
    }
    
    /**
     * Reads a single element from the json Object
     *
     * - parameter jsonData: the provisioning file as dictionary
     *
     * - parameter name: the element to be read
     *
     * - returns: The element as String or an empty String if the element is not present
     */
    private func readElementAsString(jsonData: [String: AnyObject], name: String) -> String? {
        if jsonData[name] != nil  {
            return (jsonData[name] as? String)!
        } else {
            return ""
        }
    }
    
    /**
     * Reads a certificate from a JSON file, the certificate must be the BASE64 
     * encoded version of the DER file
     *
     * - parameter jsonData: the provisioning file as dictionary
     *
     * - parameter name: the certificate to be read
     *
     * - returns: the certificate object created from the JSON String or nil
     */
    private func readCertificatefromJSONinDERFormat(jsonData: [String:AnyObject], name: String) -> SecCertificate? {
        if jsonData[name] != nil {
            let myCertDataStr = (jsonData[name] as? String)!
            let myCertData = Foundation.Data(base64Encoded: myCertDataStr,
                options: Foundation.Data.Base64DecodingOptions(rawValue: 0))
            // create the certificate to see if the data are correct
            let certificate = SecCertificateCreateWithData(nil,myCertData! as CFData)!
            
            return certificate
        }
        return nil
    }
    
    /**
     * Reads a certificate in DER Format from a file in the Resources directory
     * used to provision the root certificate for the server.
     *
     * - parameter fileName: the name of the file to retrieve
     *
     * - returns: the certificate object created from the DER File
     */
    private func readCertificatefromDERFile(fileName: String) -> SecCertificate? {
         let certificateFileURL = URLForResource(fileName: fileName, withExtension: "der")
         let myCertData = try! Foundation.Data(contentsOf: certificateFileURL!)
         let certificate = SecCertificateCreateWithData(nil,myCertData as CFData)!
         return certificate
    }
//---- end of provisioning ---------------------------------------------------------------------------------------
}
