/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation

/**
 * The message type. Each type corresponds to one type of message.
 */
public enum DeviceModelFormatType: String {
    case DATA
    case ALERT
}

/**
 * Describes a field of a message.
 */
public class DeviceModelFormatField {
    
    internal let name: String
    internal let description: String
    internal let type: DeviceModelAttributeType
    internal let optional: Bool
    
    public init(name: String, description: String, type: String, optional: Bool) {
        self.name = name
        self.description = description
        self.optional = optional
        self.type = DeviceModelAttributeType(rawValue: type)!   // <--- this crashes if the type is not one of the strings in DMAT
    }
    
    /**
     * Get the format field name.
     *
     * - returns: the message format name
     */
    public func getName() -> String {
        return self.name
    }
    
    /**
     * The format field type.
     *
     * - returns: the field type
     */
    public func getType() -> DeviceModelAttributeType {
        return self.type
    }
    
    /**
     * If the field is optional
     *
     * - returns: *true* if optional, *false* otherwise
     */
    public func isOptional() -> Bool {
        return self.optional
    }
    
    public func toString() -> String {
        var str: String = "name = " + self.name
        str = str + ", description = " + self.description
        str = str + ", type = " + self.type.rawValue
        str = str + ", optional = " + self.optional.description
        return str
    }
}

/**
 * DeviceModelFormat is the model of an message in a
 * *DeviceModel*.
 */
public class DeviceModelFormat {
    
    internal let urn: String
    internal let name: String
    internal let description: String
    internal let type: DeviceModelFormatType
    internal let fields: [DeviceModelFormatField]
    
    public init(urn: String, name: String, description: String, type: DeviceModelFormatType, fields: [DeviceModelFormatField]) {
        self.urn = urn
        self.name = name
        self.description = description
        self.type = type //DeviceModelFormatType(rawValue: type)!
        self.fields = fields   // <--- this works even if the Array has no values
    }
    
    /**
     * Get the URN of the message format
     *
     * - returns: the message format URN
     */
    public func getURN() -> String {
        return self.urn
    }
    
    /**
     * Get the message format name.
     *
     * - returns: the message format name
     */
    public func getName() -> String {
        return self.name
    }
    
    /**
     * A human friendly description of the message format. If the format does
     * not contain a description, this method will return an empty String.
     *
     * - returns: the message format description, or an empty string
     */
     public func getDescription() -> String {
        return self.description
    }
    
    /**
     * The type of the format.
     *
     * - returns: the type of the format
     */
    public func getType() -> DeviceModelFormatType {
        return self.type
    }
    
    /**
     * Get the fields of the format.
     *
     * - returns: array of format field
     */
    public func getFields() -> [DeviceModelFormatField] {
        return self.fields
    }
    
    public func toString() -> String {
        var str: String = "name = " + self.name
        str = str + ", description = " + self.description
        str = str + ", type = " + self.type.rawValue
        
        str = str + ",\n fields = ["
        var firstItem: Bool = true
        for field in self.fields {
            if (!firstItem) {
                str = str + ","
            } else {
                firstItem = false
            }
            
            str = str + "\n {"
            str = str + field.getName()
            str = str + "}"
        }
        
        if (!firstItem) {
            str = str + "\n"
        }
        
        str = str + " ]"
        return str
    }
}
