/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import SwiftyJSON

/**
 */
class DeviceModelParser {
    
    static func fromJson(root: JSON) throws -> DeviceModel {
        guard let urn:String = root["urn"].string, !urn.isEmpty else {
            // error "urn cannot be null or empty"
            throw ClientError.state(NSLocalizedString("Device model urn is nil",
                                                      comment: ""))
        }
        
        guard let name:String = root["name"].string, !name.isEmpty else {
            // error "name cannot be null or empty"
            throw ClientError.state(NSLocalizedString("Device model name is nil",
                                                      comment: ""))
        }
        
        let description:String = (root["description"] != JSON.null) ?
                                  root["description"].stringValue : ""
        
        // Attributes
        var attributes: [DeviceModelAttribute] = [DeviceModelAttribute]()
        var existingAttrNames: [String] = [String]()
        // TODO: use "arrayValue" and loop over values, no need for "if let"
        if let jsonAtts = root["attributes"].array {
            for i in 0..<jsonAtts.count {
                let attribute: DeviceModelAttribute =
                    try createDeviceModelAttribute(urn: urn,
                                                   existingFieldNames: existingAttrNames,
                                                   root: jsonAtts[i])
                attributes.append(attribute)
                existingAttrNames.append(attribute.getName()!)
            }
        }
        
        // Actions
        var actions: [DeviceModelAction] = [DeviceModelAction]()
        var existingActionNames: [String] = [String]()
        // TODO: use "arrayValue" and loop over values, no need for "if let"
        if let jsonActions = root["actions"].array {
            for i in 0..<jsonActions.count {
                do {
                    let action: DeviceModelAction =
                        try createDeviceModelAction(existingActionNames: existingActionNames,
                                                    root: jsonActions[i])
                    actions.append(action)
                    existingActionNames.append(action.getName())
                } catch { }
            }
        }
        
        // Formats
        var formats: [DeviceModelFormat] = [DeviceModelFormat]()
        // TODO: use "arrayValue" and loop over values, no need for "if let"
        if let jsonFormats = root["formats"].array {
            for i in 0..<jsonFormats.count {
                let format: DeviceModelFormat =
                    try createDeviceModelFormat(root: jsonFormats[i])
                formats.append(format)
            }
        }
        
        return DeviceModelImpl(urn: urn, name: name, description: description,
                               deviceModelAttributes: attributes,
                               deviceModelActions: actions,
                               deviceModelFormats: formats)
    }
    
    private static func createDeviceModelAttribute(
            urn: String, existingFieldNames: [String],
            root: JSON) throws -> DeviceModelAttribute {
        let name: String = root["name"].string!
        if existingFieldNames.contains(name) {
            // error "duplicate attribute: " + name
        }
        
        let description: String? = root["description"].string
        let alias: String? = (root["alias"] != JSON.null) ? root["alias"].stringValue : nil
        let range: String? = root["range"].string
        let access: DeviceModelAttribute.Access =
            (root["writable"].boolValue ?
                DeviceModelAttribute.Access.READ_WRITE :
                DeviceModelAttribute.Access.READ_ONLY)
        
        var type: DeviceModelAttributeType?
        let t: String? = root["type"].string
        if t != nil {
            type = DeviceModelAttributeType(rawValue: t!)
            if type == nil {
                // error "Illegal type"
            }
        }
        
        var min = Double.nan
        var max = Double.nan
        
        if let range = range {
            let rangeArray = range.split{$0 == ","}.map(String.init)
            if let nu1 = Int(rangeArray[0]) {
                min = Double(nu1)
            } else {
                // error "invalid range"
            }
            if let nu2 = Int(rangeArray[1]) {
                max = Double(nu2)
            } else {
                // error "invalid range"
            }
        }
        
        var defaultValue: AnyObject? = nil
        if root["defaultValue"] != JSON.null {
            defaultValue = root["defaultValue"].object as AnyObject?
        } else {
            if type == DeviceModelAttributeType.URI && defaultValue is String {
                //TODO: Only HL API uses defaults, so only HL object is supported for now
                //TODO: StorageObject case?
                defaultValue = ExternalObject(uri: defaultValue as! String)
            }
        }
        
        return DeviceModelAttribute(urn: urn, name: name,
                                    description: description, type: type!,
                                    lowerBound: min as NSNumber?,
                                    upperBound: max as NSNumber?,
                                    access: access, alias: alias,
                                    defaultValue: defaultValue)
    }
    
    private static func createDeviceModelAction(existingActionNames: [String], root: JSON)throws -> DeviceModelAction {
        let name: String = root["name"].stringValue
        if existingActionNames.contains(name) {
            // error "duplicate action:" + name
        }
        
        let description: String? = root["description"].stringValue
        let alias: String? = root["alias"].stringValue
        
        var type: DeviceModelAttributeType?
        let argType: String? = root["argType"].stringValue
        if argType?.isEmpty == false {
            type = DeviceModelAttributeType(rawValue: argType!)
        }
        
        return DeviceModelAction(name: name, description: description!, type: type, lowerBound: Double.nan as NSNumber, upperBound: Double.nan as NSNumber, alias: alias!)
    }
    
    private static func createDeviceModelFormat(root: JSON) throws -> DeviceModelFormat {
        let urn: String? = root["urn"].stringValue
        let name: String? = root["name"].stringValue
        let description: String? = root["description"].stringValue
        
        var type: DeviceModelFormatType?
        let t: String? = root["type"].string
        if t != nil {
            type = DeviceModelFormatType(rawValue: t!)
            if type == nil {
                // error "Illegal type"
            }
        }
        
        var fields: [DeviceModelFormatField] = [DeviceModelFormatField]()
        let value = root["value"].dictionary
        let jsonFields = value!["fields"]!.array
        for i in 0..<jsonFields!.count {
            let field: DeviceModelFormatField =
                try createMsgFormatField(root: jsonFields![i])
            fields.append(field)
        }
        
        return DeviceModelFormat(urn: urn!, name: name!, description: description!, type: type!, fields: fields)
        
    }
    
    private static func createMsgFormatField(root: JSON) throws -> DeviceModelFormatField {
        let name: String? = root["name"].stringValue
        let description: String? = root["description"].stringValue
        let type: String? = root["type"].stringValue
        let optional: Bool = root["optional"].boolValue
        
        return DeviceModelFormatField(name: name!, description: description!, type: type!, optional: optional)
    }
}
