/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import Alamofire

/**
 */
class HttpClient {
    
    private var url: String
    
    private static var responseTimeout: Int = Config.HTTP_RESPONSE_TIMEOUT
    
    // Static variable for injection transport implementation in unit tests
    // without real server. When not used in unit tests, variable is set from
    // HttpClientImpl class.
    var transport: Transport?
    
    /**
     * Get the HTTP response timeout.
     * The timeout The default value is 15000 milliseconds and can be set
     * with the `Config.HTTP_RESPONSE_TIMEOUT` property.
     
     * - returns: the HTTP response timeout
     */
    static func getResponseTimeout() -> Int {
        return responseTimeout
    }
    
    
    convenience init(URL: String, httpManager: Alamofire.SessionManager) throws {
        self.init(url: URL, httpManager: httpManager)

    }

    init(url: String, httpManager: Alamofire.SessionManager) {
        self.url = url
        self.transport = Transport(conManager: httpManager)
    }
    
    
    func post(data: Foundation.Data?, headers: [String : String], response: @escaping ((HttpResponse?, ClientError?) -> Void)) throws -> Void {
        try self.transport!.invokeMethod(method: "POST", data: data, headers: headers, url: url, httpResponse: { (invokeResponse, error) -> Void in
            response(invokeResponse, error)
        })
    }
    
    func put(data: Foundation.Data, headers: [String : String], response: @escaping ((HttpResponse?, ClientError?) -> Void)) throws -> Void  {
        try self.transport!.invokeMethod(method: "PUT", data: data, headers: headers, url: url, httpResponse: { (invokeResponse, error) -> Void in
            response(invokeResponse, error)
        })
    }
    
    func get(headers: [String : String], response: @escaping ((HttpResponse?, ClientError?) -> Void)) throws -> Void {
        try self.transport!.invokeMethod(method: "GET", data: Foundation.Data(bytes: UnsafePointer<UInt8>([]), count: 0), headers: headers, url: url, httpResponse: { (invokeResponse, error) -> Void in
            response(invokeResponse, error)
        })
    }
    
    func delete(headers: [String : String], response: @escaping ((HttpResponse?, ClientError?) -> Void)) throws -> Void {
        try self.transport!.invokeMethod(method: "DELETE", data: Foundation.Data(bytes: UnsafePointer<UInt8>([]), count: 0), headers: headers, url: url, httpResponse: { (invokeResponse, error) -> Void in
            response(invokeResponse, error)
        })
    }
    
    static func pipe(ins: InputStream?, outs: OutputStream?) throws {
        // Setup so that if in or out is nil but not both, then the one that
        // isn't nil will end up being closed.
        // Otherwise any failure ends up as an IOException but the streams are
        // still closed
        defer {
            if (ins != nil) {
                ins!.close()
            }
            if (outs != nil) {
                outs!.close()
            }
        }
        if (ins != nil && outs != nil) {
            let BUFFER_COUNT: Int = 8096
            let buffer: UnsafeMutablePointer<UInt8> = UnsafeMutablePointer<UInt8>.allocate(capacity: BUFFER_COUNT)
            var length: Int = 0
            while (length != -1) {
                length =  ins!.read(buffer, maxLength: BUFFER_COUNT)
                if length == -1 {
                    break
                }
                outs!.write(buffer, maxLength: BUFFER_COUNT)
            }
            // deprecated
            //buffer.deallocate(capacity: BUFFER_COUNT)
            buffer.deallocate()
        }
    }
}
