/*
 *  Copyright © 2016 Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation

protocol KeychainAttributes {
    
    var accessMode: String {get}
    var attributes: [String: AnyObject] {get}
    var dataToStore: [String: AnyObject] {get set}
    var serviceName: String {get}
    var accountName: String {get}
}

extension KeychainAttributes {
    
    var accessMode: String {
        
        return String(kSecAttrAccessibleWhenUnlocked)
    }
    
    var serviceName: String {
        
        return "oracle.keychain.service"
    }
    
    
    var attributesToSave: [String: AnyObject] {
        
        var itemAttributes = attributes
        if(!dataToStore.isEmpty){
        let archivedData = NSKeyedArchiver.archivedData(withRootObject: dataToStore)
        
        itemAttributes[String(kSecValueData)] = archivedData as AnyObject?
        }
        return itemAttributes
    }
    
    func dataFromAttributes(attributes: [String: AnyObject]) -> [String: AnyObject]? {
        
        guard let valueData = attributes[String(kSecValueData)] as? Foundation.Data else { return nil }
        
        return NSKeyedUnarchiver.unarchiveObject(with: valueData) as? [String: AnyObject] ?? nil
    }
    
    func nsdataFromAttributes(attributes: [String: AnyObject]) -> Foundation.Data? {
        guard let valueData = attributes[String(kSecValueData)] as? Foundation.Data else { return nil }
        return valueData
    }
    
    var attributesForFetch: [String: AnyObject] {
        
        var itemAttributes = attributes
        
        itemAttributes[String(kSecReturnData)] = true as AnyObject?
        itemAttributes[String(kSecReturnAttributes)] = true as AnyObject?
        
        return itemAttributes
    }
    
    var attributes: [String: AnyObject] {
        
        var attributes = [String: AnyObject]()
        
        attributes[String(kSecClass)] = kSecClassGenericPassword
        attributes[String(kSecAttrAccessible)] = accessMode as AnyObject?
        attributes[String(kSecAttrService)]    = serviceName as AnyObject?
        attributes[String(kSecAttrAccount)]    = accountName as AnyObject?
        
        return attributes
    }
}

struct KeychainEntry: KeychainAttributes {
    
    let accountName:String
    var dataToStore:[String : AnyObject]
    
    init(Name accountName: String, data: String) {
        self.accountName = accountName
        self.dataToStore = ["token": data as AnyObject]
    }
    
    init(Name accountName: String) {
        self.accountName = accountName
        self.dataToStore = [String: String]() as [String : AnyObject]
    }
}

