/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import QuartzCore
import SwiftyJSON

// FIXME: comparison operators with optionals were removed from the 
// Swift Standard Libary.
// Consider refactoring the code to use the non-optional operators.
fileprivate func < <T : Comparable>(lhs: T?, rhs: T?) -> Bool {
  switch (lhs, rhs) {
  case let (l?, r?):
    return l < r
  case (nil, _?):
    return true
  default:
    return false
  }
}

// FIXME: comparison operators with optionals were removed from the 
// Swift Standard Libary.
// Consider refactoring the code to use the non-optional operators.
fileprivate func > <T : Comparable>(lhs: T?, rhs: T?) -> Bool {
  switch (lhs, rhs) {
  case let (l?, r?):
    return l > r
  default:
    return rhs < lhs
  }
}


/**
 * Implementation of protocol Equatable
 *
 * parameter lhs: one item to be compared
 *
 * parameter rhs: the other item to be compared
 *
 * - returns: *true* if equal, *false* otherwise
 */
public func ==(lhs: Message, rhs: Message) -> Bool {
    // TODO: why not use the hashValue
    if lhs === rhs {
       return true
    }
    
    if (lhs.id != rhs.id) {
        return false
    }
    
    if (lhs.clientId != rhs.clientId) {
        return false
    }
    
    if (lhs.eventTime != rhs.eventTime) {
        return false
    }

    if (lhs.destination != rhs.destination) {
        return false
    }
    
    if (lhs.priority != rhs.priority) {
        return false
    }
    
    if (lhs.properties != rhs.properties) {
        return false
    }

    if (lhs.reliability != rhs.reliability) {
        return false
    }
    
    if (lhs.source != rhs.source) {
        return false
    }
    
    if (lhs.sender != rhs.sender) {
        return false
    }
    // TODO: Using NSDictionary here to do deep comparison.
    if !(NSDictionary(dictionary: lhs.diagnostics!).isEqual(to: rhs.diagnostics!)) {
        return false
    }
    
    if (lhs.direction != rhs.direction) {
        return false
    }
    
    if (lhs.receivedTime != rhs.receivedTime) {
        return false
    }
    
    if (lhs.sentTime != rhs.sentTime) {
        return false
    }
    
    return true
}

/**
 * An abstract class for different types of messages in messaging server.
 * This class is immutable.
 *
 * - (Abstract class)
 *
 */
public class Message: NSCopying, Hashable {
    
    public func copy(with zone: NSZone? = nil) -> Any {
        // TODO: Not GOOD
        // need "copy" to deal with "Pair" equality when Message is the Pair's
        // key.
        var message:Message = self
        do {
            let messages:[Message] =
                try Message.fromJSON(jsonString:self.toJsonString())
            message = messages.first ?? self
        } catch {
            // MUST NOT FAIL
            log(type:.fault,
                message:"%@: could not create copy: %@",
                args:"Message.copy()", "\(error)")
            message = self
        }
        return message as Any
    }
    /**
     * Implementation of protocol Hashable
     *
     * - returns: the hash value
     */
    public var hashValue : Int {
        var result: Int = (id != nil) ? (id?.hashValue)!:0
        result = 31 &* result &+ (clientId?.hashValue)!
        result = 31 &* result &+ (source?.hashValue)!
        result = 31 &* result &+ (destination?.hashValue)!
        result = 31 &* result &+ (priority?.hashValue)!
        result = 31 &* result &+ (reliability.hashValue)
        result = 31 &* result &+ (eventTime?.hashValue)!
        result = 31 &* result &+ (properties?.hashValue)!
        result = 31 &* result &+ (sender?.hashValue)!
        if  diagnostics != nil {
            // TODO: Using NSDictionary because Distionary does not have
            // an init that takes a Dictionary instance, just a literal
            result = 31 &* result &+ NSDictionary(dictionary: diagnostics!).hashValue
        }
        if direction != nil {
            result = 31 &* result &+ direction!.hashValue
        }
        if receivedTime != nil {
            result = 31 &* result &+ receivedTime!.hashValue
        }
        if sentTime != nil {
            result = 31 &* result &+ sentTime!.hashValue
        }
        return result
    }
    
    static let MAX_KEY_LENGTH: Int = 2048
    static let MAX_STRING_VALUE_LENGTH: Int = 4096
    
    /**
     * The priority of the message.
     */
    public enum Priority: Int, CustomStringConvertible {
        case lowest = 0, low, medium, high, highest
  
        // access to raw values are pre-implemented in Swift:
        // setting: Priority(rawValue: 1)
        // getting: prio.toRaw()
        
        /**
         * Implementation of valueOf method missing in Swift
         */
        public static func valueOf(string: String) -> Priority? {
            var i = 0
            while let item = Priority(rawValue: i) {
                if String(describing: item).lowercased() == string.lowercased() {
                    return item
                }
                i += 1
            }
            return nil
        }
        
        public var description: String {
            switch self {
            case .lowest: return "LOWEST";
            case .low: return "LOW";
            case .medium: return "MEDIUM";
            case .high: return "HIGH";
            case .highest: return "HIGHEST";
            }
        }
    }
    
    /**
     * Reliability of the message
     */
    public enum Reliability: Int, CustomStringConvertible {
        /** on GW: Message won't be persisted, number of of retry count if 
         * sending was unsuccessful is default value
         */
        case no_guarantee = 0
        /** on GW: Message will be persisted, number of retry count if 
         * sending was unsuccessful is (default value) * 2
         */
        case best_effort = 1
        /** on GW: Message will be persisted, number of retry count will be 
         * set to Integer.MAX_VALUE.
         */
        case guaranteed_delivery = 2
        
        /**
         * Implementation of valueOf method missing in Swift
         */
        public static func valueOf(string: String) -> Reliability? {
            var i = 0
            while let item = Reliability(rawValue: i) {
                if String(describing: item).lowercased() == string.lowercased() {
                    return item
                }
                i += 1
            }
            return nil
        }
        
        public var description: String {
            switch self {
            case .no_guarantee: return "NO_GUARANTEE"
            case .best_effort: return "BEST_EFFORT"
            case .guaranteed_delivery: return "GUARANTEED_DELIVERY"
            }
        }
    }
    
    /**
     * The message type. Each type corresponds to one type of message.
     */
    public enum MsgType: String {
        case DATA = "DATA"
        case REQUEST = "REQUEST"
        case RESPONSE = "RESPONSE"
        case RESOURCE = "RESOURCES_REPORT"
        case ALERT = "ALERT"
    }
    
    /**
     * The message direction with respect to the device.
     */
    public enum Direction: Int {
        case from_device
        case to_device
        
        /**
         * Implementation of valueOf method missing in Swift
         */
        public static func valueOf(string: String) -> Direction? {
            var i = 0
            while let item = Direction(rawValue: i) {
                if String(describing: item).lowercased() == string.lowercased() {
                    return item
                }
                i += 1
            }
            return nil
        }
        
        public var description: String {
            switch self {
            case .from_device: return "FROM_DEVICE"
            case .to_device: return "TO_DEVICE"
            }
        }
    }
    
    /**
     * Created time of the message diagnostic value. The value of the created 
     * time as an Int64 that contains the epoch value in milliseconds.
     */
    public let DIAG_CREATED_TIME = "createTime"
    /**
     * Client address diagnostic value. The address client as a String
     */
    public let DIAG_CLIENT_ADDRESS = "clientAddress"
    
    /**
     * A unique ID for each message generated by the server (on receipt of a 
     * message from the client or when a message is generated by the server)
     */
    var id: String?
    /**
     * A unique ID for each message. It is generated when the message is first
     * created or can be passed in by the user.
     */
    internal var clientId: String?
    /**
     * Source of the message.
     */
    var source: String?
    /**
     * Destination of the message.
     */
    var destination: String?
    /**
     * Message priority.
     */
    internal var priority: Message.Priority?
    /**
     * Reliability for message delivering.
     */
    internal var reliability: Message.Reliability

    /**
     * Message event time.
     */
    internal var eventTime: Int64?
    /**
     * Extra information for the message.
     */
    internal var properties: MessageProperties?
    /**
     * The message sender (should correspond to endpoint id).
     */
    internal var sender: String?
    /**
     * The message diagnostics
     */
    internal var diagnostics: [String : AnyObject]? = [:]
    /**
     * The message direction (with respect to the server)
     */
    internal var direction: Message.Direction?
    /**
     * The time when the message was received by the server (epoch value in 
     * milliseconds)
     */
    internal var receivedTime: Int64?
    /**
     * The time when the message was sent by the server (epoch value in 
     * milliseconds)
     */
    internal var sentTime: Int64?
    
    private static let BASIC_NUMBER_OF_RETRIES:Int =
        Config.BASIC_NUMBER_OF_RETRIES >= 3 ? Config.BASIC_NUMBER_OF_RETRIES : 3
    
    /**
     * The number of send retries remaining for this message.
     */
    private var remainingRetries:Int = 0
    
    /**
     * Message constructor takes message builder and set values to each field.
     * If the value is `nil`, set a default value. The default value for
     * priority is `Priority.LOW`. The default value for reliability is
     * `Reliability.BEST_EFFORT`
     *
     * - parameter builder: A `MessageBuilder` instance
     * 
     * - throws: `ClientError.state` if `builder.source` is `nil` or is the
     *                                empty string.
     */
    init(builder: MessageBuilder) throws {

        // Don't require clientId to be set, auto-generate one if it isn't
        self.clientId = builder.clientId ?? UUID().uuidString
        // TODO: does not check if builder.id is nil what is possible
        self.id = builder.id
        
        guard let source = builder.source, !source.isEmpty else {
            throw ClientError.state(NSLocalizedString(
                "Source cannot be nil or empty", comment: ""))
        }
        self.source = source

        // check if this is correct it is done like that in Java
        // the fromJson methods fail if the direction is not present in the 
        // message but we have to look into this
    
        self.destination = builder.destination ?? ""
        self.sender = builder.sender ?? ""
        self.eventTime = builder.eventTime ?? Int64(Date().timeIntervalSince1970 * 1000)
        self.priority = builder.priority ?? Priority.low
        self.reliability = builder.reliability ?? Reliability.best_effort
        
        switch (self.reliability) {
        case .no_guarantee:
            remainingRetries = Message.BASIC_NUMBER_OF_RETRIES
            break
        case .best_effort:
            remainingRetries = Message.BASIC_NUMBER_OF_RETRIES * 3
            break
        case .guaranteed_delivery:
            remainingRetries = Int.max
            break
        }
        
        self.properties = builder.properties ?? MessagePropertiesBuilder().build()
        self.diagnostics = builder.diagnostics ?? nil
        self.direction = builder.direction
        self.receivedTime = builder.receivedTime
        self.sentTime = builder.sentTime
    }
    
    /**
     * Get unique ID for a message.
     *
     * - returns: Message id (string), never `nil`
     */
    public func getId() -> String {
        // TODO: fails if builder had no id
        return self.id ?? ""
    }
    
    // TODO: enterprise version returned type, but should probably only
    // be in subclass.
    /**
     * Abstract method to return message type.
     *
     * - (Abstract method)
     *
     * - returns: type, never `nil`
     */
    public func getType() -> MsgType {
        // TODO: need to have a throw statement has an effect on the
        // toString toJson method in this class and in all subclasses
        fatalError("Subclass need to implement the 'getType()' method.")
    }
    
    /**
     * Get client's unique ID for a message.
     *
     * - returns: Message id (string), never `nil`
     */
    public func getClientId() -> String {
        return self.clientId!
    }
    
    /**
     * Get message source (Endpoint Id from which the message is originated).
     *
     * - returns: source, may be `nil` if the message was sent from server.
     */
    public func getSource() -> String {
        return self.source!
    }
    
    /**
     * Get message destination (Endpoint Id to which the message is originated).
     *
     * - returns: destination, may be `nil` if the message was sent to server.
     */
    public func getDestination() -> String {
        return self.destination!
    }
    
    /**
     * Get message priority.
     *
     * - returns: priority, never `nil`
     */
    public func getPriority() -> Priority {
        return self.priority!
    }
    
    /**
     * Get message reliability
     *
     * - returns: reliability, never `nil`
     */
    public func getReliability() -> Reliability {
        return self.reliability
    }
    
    /**
     * Get event time of the message
     *
     * - returns: eventTime as Int64, never `nil`
     */
    public func getEventTime() -> Int64? {
        return self.eventTime
    }
    
    /**
     * Get message properties.
     *
     * - returns: properties, never `nil`
     */
    public func getProperties() -> MessageProperties {
        return self.properties!
    }
    
    /**
     * Get message sender (Endpoint Id from which the message is sent from).
     *
     * - returns: sender
     */
    public func getSender() -> String {
        return self.sender!
    }
    
    /**
     * Not intended for general use.
     * Used internally by the message dispatcher implementation.
     * - returns: the number of remaining retries
     */
    public func getRemainingRetries() -> Int {
        return reliability == Reliability.guaranteed_delivery ?
            Int.max : remainingRetries
    }
    
    /**
     * Not intended for general use.
     * Used internally by the message dispatcher implementation.
     * - parameter remainingRetries: the new number of remaining retries
     */
    public func setRemainingRetries(remainingRetries:Int) {
        self.remainingRetries = remainingRetries
    }
    
    /**
     * Get message diagnostics
     *
     * - returns: message diagnostics, can be `nil`
     */
    public func getDiagnostics() -> [String : AnyObject] {
        return self.diagnostics!
    }
    
    public func getDiagnosticValue(diagName: String?) -> AnyObject? {
        if (diagName != nil) && (self.diagnostics != nil) {
            return self.diagnostics![diagName!]
        }
        return nil
    }
    
    /**
     * Get message direction.
     *
     * - returns: message direction, can be `nil`
     */
    public func getDirection() -> Direction {
        return self.direction!
    }
    
    /**
     * Get message received time.
     *
     * - returns: message received time as Int64, can be `nil`
     */
    public func getReceivedTime() -> Int64? {
        return self.receivedTime
    }
    
    /**
     * Get message sent time.
     *
     * - returns: message sent time as Int64, can be `nil`
     */
    public func getSentTime() -> Int64?{
        return self.sentTime
    }
    
    /**
     * Exports common data from Message to String using JSON interpretation 
     * of the message.
     *
     * - returns: JSON interpretation of the message as String.
     */
    public func toString() throws -> String {
        return try toJsonString()
    }
    
    /**
     * Export the full properties of the message as JSON string
     *
     * - returns: message fields as Json string
     */
    public func toJsonString() throws -> String {
        return toJsonString(expand: true, justBeginning: false)
    }
    
    /**
     * Export the full or just basic properties of the message as JSON string
     *
     * - parameter expand: determine whether full properties shall be exported
     * - parameter justBeginning: if false a closing brace is appended
     
     * - returns: message fields as Json string
     */
    public func toJsonString(expand: Bool, justBeginning: Bool) -> String {
        var resultStr: String = "{"
        if id != nil {
            resultStr = resultStr + "\"id\":\"" + id! + "\","
        }
        if id == nil {
            resultStr = resultStr + "\"id\":\"" + "\","
        }
        if clientId != nil {
            resultStr = resultStr + "\"clientId\":\"" + clientId! + "\","
        }
        
        resultStr = resultStr + "\"source\":\"" + source! + "\","
        resultStr = resultStr + "\"sender\":\"" + sender! + "\","
        resultStr = resultStr + "\"destination\":\"" + destination! + "\","
        resultStr = resultStr + "\"eventTime\":" + eventTime!.description + ","
        resultStr = resultStr + "\"priority\":\"" + priority!.description + "\","
        resultStr = resultStr + "\"reliability\":\"" + reliability.description + "\","
        resultStr = resultStr + "\"type\":\"" + getType().rawValue + "\""
        if expand {
            if properties != nil && !(properties?.getAllProperties().isEmpty)! {
                resultStr = resultStr + ",\"properties\":" + (properties?.toJsonString())!
            }
        }
        if direction != nil {
            resultStr = resultStr + ",\"direction\":\"" + direction!.description + "\""
        }
        if receivedTime != nil {
            resultStr = resultStr + ",\"receivedTime\":" + receivedTime!.description
        }
        if sentTime != nil {
            resultStr = resultStr + ",\"sentTime\":" + sentTime!.description
        }
        if (diagnostics != nil) && (diagnostics!.count > 0) {
            resultStr = resultStr + ",\"diagnostics\":{"
            if diagnostics?.count > 0 {
                var z: Int = 0
                for (key, val) in diagnostics! {
                    // TODO what if val is no String??
                    resultStr = resultStr + "\"" + key + "\":\"" + val.description + "\""
                    z += 1
                    if z < diagnostics!.count {
                        resultStr = resultStr + ","
                    }
                }
            }
            resultStr = resultStr + "}"
        }
        if !justBeginning {
            resultStr = resultStr + "}"
        }
        return resultStr
    }
    
    /**
     * Export the full or just basic properties of the message, completed by
     * an additional key value pair, as JSON string
     *
     * - parameter expand: determine whether full properties shall be exported
     *
     * - parameter newKeys: an array of keys to be added
     *
     * - parameter newValues: an array of values to be added
     *
     * - returns: message fields as Json string
     */
    public func toJsonString(expand: Bool, additionalItems: String) -> String {
        var resultStr: String = toJsonString(expand: expand, justBeginning: true)
        resultStr = resultStr + "," + additionalItems
        resultStr = resultStr + "}"
        return resultStr
    }
        
    /**
     * Convert an array of Messages into a Json string
     *
     * - parameter messages: The array of Messages.
     *
     * - parameter expand: boolean flag indicating if properties to be included or not.
     *
     * - returns: A JSON string representing the list of Messages
     *
     * - Throws:
     */
    public static func toJsonString(messages: Array<Message?>, expand: Bool) throws -> String {
        var resultStr: String = "["
        var z: Int = 0
        for m in messages {
            if m != nil {
                if expand {
                    try resultStr = resultStr + m!.toJsonString()
                } else {
                    resultStr = resultStr + m!.toJsonString(expand: false, justBeginning: false)
                }
                z += 1
                if z < messages.count {
                    resultStr = resultStr + ","
                }
            }
        }
        resultStr = resultStr + "]"
        
        return resultStr
    }

    // TODO: from Enterprise version. Need to check to see if methods above
    // can suffice for these
    /**
     * Export the full properties of the message to JSON.
     *
     * - returns: Message fields in Json format, never `nil`.
     */
    public func toJSON() throws -> JSON {
        return toJSON(expand: true)
    }
    
    /**
     * Export the full or just basic properties of the message to JSON.
     *
     * - parameter expand: Determine whether full properties shall be exported.
     *
     * - returns: Message fields in Json format, never `nil`.
     */
    public func toJSON(expand: Bool) -> JSON {
        let jsonDict: [String : AnyObject] = buildJsonDict(expand: expand)
        
        return JSON(jsonDict)
    }
    
    /**
     * Export the full or just basic properties of the message, completed by
     * an additional key value pair, to JSON
     *
     * - parameter expand: determine whether full properties shall be exported
     *
     * - parameter newKeys: an array of keys to be added
     *
     * - parameter newValues: an array of values to be added
     *
     * - returns: message fields in Json format, never `nil`
     */
    public func toJSON(expand: Bool, newKeys: [String], newValues: [AnyObject]) -> JSON {
        var jsonDict: [String : AnyObject] = buildJsonDict(expand: expand)
        for i in 0..<newKeys.count {
            jsonDict[newKeys[i]] = newValues[i]
        }
        
        return JSON(jsonDict)
    }
    
    public static func fromJSON(jsonString: String) -> [Message] {
        let jsonData = jsonString.data(using: .utf8)
        let mb: MessageBuilder = MessageBuilder()
        do {
            let msgs:[Message] = try mb.fromJSON(jsonObject: jsonData!) as! [Message]
            return msgs
        } catch {
            print(String(format:NSLocalizedString("%@: threw exception. %@",
                                                  comment:""),
                         "MessageBuilder.fromJSON()", "\(error)"))
        }
        return [Message]()
    }

    /**
     * Convert `Foundation.Data` to a list of `Message`.
     *
     * - parameter messages: `Foundation.Data` containing json
     *           interpretation of `Message`
     * - returns: The `[Message]` from the `Foundation.Data`
     * - throws `ClientError.argument` if `messages` is empty.
     */
    public static func fromJSON(messages:Foundation.Data) throws -> [Message] {
        guard !messages.isEmpty else {
            throw ClientError.argument(NSLocalizedString(
                "Message.fromJson(): messages is empty", comment:""))
        }

        let jsonString:String? = String(data:messages, encoding:.utf8)
        return Message.fromJSON(jsonString:jsonString!)
    }
    
    private func buildJsonDict(expand: Bool) -> [String : AnyObject] {
        var jsonDict: [String : AnyObject] = [ : ]
        if id != nil {
            jsonDict["id"] = id as AnyObject?
        }
        if clientId != nil {
            jsonDict["clientId"] = clientId as AnyObject?
        }
        jsonDict["source"] = source as AnyObject?
        jsonDict["destination"] = destination as AnyObject?
        jsonDict["priority"] = priority?.rawValue.description as AnyObject?
        jsonDict["reliability"] = reliability.rawValue.description as AnyObject?
        jsonDict["eventTime"] = eventTime as AnyObject?
        jsonDict["sender"] = sender as AnyObject?
        jsonDict["type"] = self.getType().rawValue as AnyObject?
        if expand {
            if properties != nil && !(properties?.getAllProperties().isEmpty)! {
                jsonDict["properties"] = properties?.toJSON() as AnyObject?
            }
        }
        if direction != nil {
            jsonDict["direction"] = direction?.rawValue as AnyObject?
        }
        if receivedTime != nil {
            jsonDict["receivedTime"] = receivedTime  as AnyObject?
        }
        if sentTime != nil {
            jsonDict["sentTime"] = sentTime as AnyObject?
        }
        if diagnostics != nil {
            jsonDict["diagnostics"] = diagnostics as AnyObject?
        }
        return jsonDict
    }
    

    
    /**
     * Convert an array of Messages into a Json array
     *
     * - parameter messages: The array of Messages.
     *
     * - parameter expand: boolean flag indicating if properties to be included 
     *                     or not.
     *
     * - returns: An array of JSON representing the list of Messages
     */
    public static func toJSON(messages: Array<Message?>, expand: Bool) throws -> JSON {
        var arr: [JSON] = []
        for m in messages {
            if m != nil {
                if expand {
                    try arr.append(m!.toJSON())
                } else {
                    arr.append(m!.toJSON(expand: false))
                }
            }
        }
        return JSON(arr)
    }
}

/**
 * This is an abstract class for different message builders. Message classes
 * are immutable. A builder is required when creating messages.
 *
 * - (Abstract class)
 */
public class MessageBuilder {
    
    /**
     * A unique ID for each message that is assigned by the server (on a receipt 
     * of a message from a client or when the server generates a message)
     */
    var id: String?
    /**
     * A unique ID for each message assigned by the client for tracking 
     * purposes.
     */
    var clientId: String?
    /**
     * Source of the message (should correspond to endpoint id)
     */
    var source: String?
    /**
     * Destination of the message.
     */
    var destination: String?
    /**
     * Message priority.
     */
    var priority: Message.Priority?
    /**
     * Reliability for message delivering.
     */
    
    var reliability: Message.Reliability?
    /**
     * Message event time.
     */
    var eventTime: Int64?
    /**
     * Extra information for the message.
     */
    var properties: MessageProperties?
    /**
     * The message sender (should correspond to endpoint id).
     */
    var sender: String?
    /**
     * The message diagnostics
     */
    var diagnostics: [String : AnyObject]? = [:]
    /**
     * The message direction (with respect to the server)
     */
    var direction: Message.Direction?
    /**
     * The time when the message was received by the server (epoch value in 
     * nanoseconds)
     */
    var receivedTime: Int64?
    /**
     * The time when the message was sent by the server (epoch value in 
     * nanoseconds)
     */
    var sentTime: Int64?
    
    public init() {
    }
    
    /**
     * Sets message ID. If no id is provided during creation of the
     * Message, random one is created.
     *
     * - parameter id: UUID of the message that has to be assigned
     */
    public func setId(id: String) {
        self.id = id
    }
    /**
     * Sets message client ID. The client message id should be unique within the 
     * client. If no id is provided during creation of the Message, random one 
     * is created.
     *
     * - parameter clientId: the client ID
     */
    public func setClientId(clientId: String) {
        self.clientId = clientId
    }
    
    /**
     * Set message source. The value should correspond with endpoint ID where 
     * the message was originated.
     *
     * - parameter source: source (source endpoint ID) of the message
     */
    public func setSource(source: String) {
        self.source = source
    }
    
    /**
     * Set message destination. The value should correspond with endpoint ID 
     * where the message should be delivered.
     *
     * - parameter destination: destination (destination endpoint ID) of the 
     *                          message
     */
    public func setDestination(destination: String?) {
        if destination != nil {
            self.destination = destination
        }
    }
    
    /**
     * Set message priority.
     *
     * - parameter priority: priority of the message
     */
    public func setPriority(priority: Message.Priority) {
        self.priority = priority
    }
    
    /**
     * Sets reliability for message delivering.
     *
     * - parameter reliability: reliability for message delivering.
     */
    public func setReliability(reliability: Message.Reliability) {
        self.reliability = reliability
    }

    /**
     * Set message creation time.
     *
     * - parameter eventTime: event time of the message
     */
    public func setEventTime(eventTime: Int64) {
        self.eventTime = eventTime
    }
    
    /**
     * Set message properties.
     *
     * - parameter properties: extra information for the message
     */
    public func setProperties(properties: MessageProperties) {
        self.properties = properties
    }
    
    /**
     * Set message sender. The value should correspond with endpoint ID 
     * where the message was sent from.
     *
     * - parameter sender: sender (sender endpoint ID) of the message
     */
    public func setSender(sender: String?) {
        if sender != nil {
            self.sender = sender
        }
    }
    
    /**
     * Set message direction.
     *
     * - parameter direction: message direction
     */
    public func setDirection(direction: Message.Direction) {
        self.direction = direction
    }
    
    /**
     * Set the time when the message was received by the server.
     *
     * - parameter receivedTime: time epoch value in nano seconds
     */
    public func setReceivedTime(receivedTime: Int64) {
        self.receivedTime = receivedTime
    }
    
    /**
     * Set the time when the message was sent by the server.
     *
     * - parameter sentTime: epoch value in nano seconds
     */
    public func setSentTime(sentTime: Int64) {
        self.sentTime = sentTime
    }
    
    /** Set a diagnostic object value.
     *
     * - parameter name: diagnostic name
     *
     * - parameter value: diagnostic value
     */
    public func setDiagnostic(name: String, value: AnyObject) {
        if self.diagnostics == nil {
            self.diagnostics = [ : ]
        }
        self.diagnostics![name] = value
    }
    
    private func getString(data: Foundation.Data) throws -> String {
        return try JSONSerialization.jsonObject(with: data,
                        options: JSONSerialization.ReadingOptions()) as! String
    }
    
    /**
     * Method to deserialization of the Message from a JsonObject.
     *
     * - parameter jsonObject: the jsonObject as NSData
     *
     * - returns: extracted item as AnyObject
     *
     * - throws: `ClientError.state` if `priority`, `reliability` or `direction`
     *           is `nil` or not valid, or if `eventTime`, `receiveTime`, or 
     *           `sentTime` is 0 or if `source` or `destination` are `nil` or
     *           empty.
     */
    public func fromJSON(data: Foundation.Data?) throws -> Dictionary<String,Any> {
        if data == nil {
            return [String : Any]()
        }
        // Why isn't this using SwiftyJSON JSON(data!).dictionaryValue ?
        var jsonObject:Any? = nil
        do {
            jsonObject = try JSONSerialization.jsonObject(
                with: data!, options: JSONSerialization.ReadingOptions())
        } catch {
            print(error)
        }
        if jsonObject == nil {
            print("Message JSON structure is nil")
            return [String:Any]()
        }

        let jsonDict:Dictionary<String,Any>? = jsonObject as? Dictionary<String,Any>
        if jsonDict == nil {
            print("Message JSON structure is nil")
            return [String:Any]()
        }
        
        var source: String? = nil
        var destination: String? = nil
        var sender: String? = nil
        var priority: String? = nil
        var direction: String? = nil
        for (key, value) in jsonDict! {
            if key == "id" {
                self.id = (value as! String)
                continue
            }
            if key == "clientId" {
                self.clientId = (value as! String)
                continue
            }
            if key == "source" {
                source = (value as! String)
                continue
            }
            if key == "destination" {
                destination = (value as! String)
                continue
            }
            if key == "sender" {
                sender = (value as! String)
                continue
            }
            if key == "priority" {
                priority = (value as! String)
                if priority == nil {
                    throw ClientError.state(NSLocalizedString(
                        "Message property is nil", comment: ""))
                    
                }
                let priorityItem: Message.Priority? =
                    Message.Priority.valueOf(string: priority!)
                if priorityItem == nil {
                    throw ClientError.state(NSLocalizedString(
                        "Message property is illegal", comment: ""))
                    
                }
                setPriority(priority: priorityItem!)
                continue
            }
            if key == "reliability" {
                let reliability: String? = (value as! String)
                if reliability == nil {
                    throw ClientError.state(NSLocalizedString(
                        "Message reliability is nil", comment: ""))
                    
                }
                let reliabilityItem: Message.Reliability? =
                    Message.Reliability.valueOf(string: reliability!)
                if reliabilityItem == nil {
                    throw ClientError.state(NSLocalizedString(
                        "Message reliability is illegal", comment: ""))
                }
                setReliability(reliability: reliabilityItem!)
                continue
            }
            // NOTE: 0 is not a valid time
            if key == "eventTime" {
                let eventTime: NSNumber = value as! NSNumber
                if eventTime == 0 {
                    throw ClientError.state(NSLocalizedString(
                        "Message event time is wrong", comment: ""))
                }
                setEventTime(eventTime: eventTime.int64Value)
                continue
            }
            // TODO: handle properties
            
            // TODO: handle Diagnostics
            if key == "direction" {
                direction = (value as! String)
                if direction == nil {
                    throw ClientError.state(NSLocalizedString(
                        "Message direction is nil", comment: ""))
                }
                continue
            }
            if key == "receivedTime" {
                let receivedTime: NSNumber? = (value as! NSNumber)
                if receivedTime == 0 {
                    throw ClientError.state(NSLocalizedString(
                        "Message received time is wrong", comment: ""))
                }
                setReceivedTime(receivedTime: receivedTime!.int64Value)
                continue
            }
            if key == "sentTime" {
                let sentTime: NSNumber? = (value as! NSNumber)
                if sentTime == 0 {
                    throw ClientError.state(NSLocalizedString(
                        "Message sent time is wrong", comment: ""))
                }
                setSentTime(sentTime: sentTime!.int64Value)
                continue
            }
        }
        // TODO: remove second condition after RequestMessage from 
        // Server has source
        if (source == nil || source!.isEmpty) &&
              (destination == nil || destination!.isEmpty){
            throw ClientError.state(NSLocalizedString(
                "Source AND destination must not be both nil", comment: ""))
        }
        // TODO: Make source and destination consistent
        // setDestination accepts String? but setSource does not
        // orinally it was forcing the unwrap but the condition above
        // says that source could be nil if destination isn't.
        setSource(source: source ?? "")
        setDestination(destination: destination)
        setSender(sender: sender)
        if (direction != nil) {
           let directionItem: Message.Direction? =
            Message.Direction.valueOf(string: direction!)
           if directionItem == nil {
                throw ClientError.state(NSLocalizedString(
                    "Message direction is illegal", comment: ""))
           }
           setDirection(direction: directionItem!)
        }
        return jsonDict!
    }
    
    /**
     * Method to deserialization of the Message from a JsonObject.
     *
     * - parameter jsonObject: the jsonObject as NSData
     *
     * - returns: extracted item as AnyObject
     */
    public func fromJSON(jsonObject: Foundation.Data) throws -> AnyObject {
        let item = try fromJSONArray(jsonObject: jsonObject)
        
        return item as AnyObject
    }

    public func fromJSONArray(jsonObject: Foundation.Data) throws -> [Message] {
        //var json: Array<AnyObject>!
        var json: JSON?
        
        json = try JSON(data: jsonObject)
        
        let item = json?.array
        var messageCollection:[Message] = [Message]()
        
        if (item != nil) {
            for i in 0 ..< item!.count {
                let mb:MessageBuilder = MessageBuilder()
                
                let id: String = (item![i].dictionaryObject!["id"] as? String)!
                mb.id = id
                
                let clientId: String = (item![i].dictionaryObject!["clientId"] as? String)!
                mb.clientId = clientId
                
                let source: String = (item![i].dictionaryObject!["source"] as? String)!
                mb.source = source
                
                let destination: String = (item![i].dictionaryObject!["destination"] as? String)!
                mb.destination = destination
                
                var priority: Message.Priority
                priority = Message.Priority.valueOf(string: item![i].dictionaryObject!["priority"] as! String)!
                mb.priority = priority
                
                messageCollection.append(try Message(builder: mb))
            }
        }
        return messageCollection
    }

    /**
     * Abstract build method
     *
     * - (Abstract method)
     *
     * - returns: message
     */
    public func build() throws -> Message {
        fatalError("Subclass need to implement the 'build()' method.")
    }
}

