/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import SwiftyJSON

// FIXME: comparison operators with optionals were removed from the Swift Standard Libary.
// Consider refactoring the code to use the non-optional operators.
fileprivate func < <T : Comparable>(lhs: T?, rhs: T?) -> Bool {
  switch (lhs, rhs) {
  case let (l?, r?):
    return l < r
  case (nil, _?):
    return true
  default:
    return false
  }
}

// FIXME: comparison operators with optionals were removed from the Swift Standard Libary.
// Consider refactoring the code to use the non-optional operators.
fileprivate func > <T : Comparable>(lhs: T?, rhs: T?) -> Bool {
  switch (lhs, rhs) {
  case let (l?, r?):
    return l > r
  default:
    return rhs < lhs
  }
}


/**
 * MessagePropertiesBuilder is the builder for MessageProperties.
 */
public class MessagePropertiesBuilder {
    
    /** A dictionary for message properties. It is a key/list of value pair  */
    fileprivate var propertiesTable: [String : [String]] = [ : ]
    
    public init() {
    }
    
    /**
     * Add a new key/value pair. If the key exists, add the value to the list of
     * values. If the key does not exist, add the key and value to the 
     * properties table.
     * `Key` or `value` cannot be `nil` or empty. `Key` length cannot exceed
     * `MAX_KEY_LENGTH` bytes. `Value` length cannot exceed
     * `MAX_STRING_VALUE_LENGTH` bytes. The length is measured
     * after the string is encoded using UTF-8 encoding.
     *
     * - parameter key: property key
     *
     * - parameter value: property value
     *
     * - throws: `ClientError.argument` if a parameter is `nil`, or zero length,
     *           or `key` exceeds `Message.MAX_KEY_LENGTH`, or `value` exceeds
     *           `Message.MAX_STRING_VALUE_LENGTH`
     */
    public func addValue(key: String?, value: String?) throws {
        try validateStringKeyOrValue(string: key,
                                     maxLength: Message.MAX_KEY_LENGTH,
                                     keyOrValue: "key")
        try validateStringKeyOrValue(string: value,
                                     maxLength: Message.MAX_STRING_VALUE_LENGTH,
                                     keyOrValue: "value")
        var values: [String]
        values = propertiesTable[key!] ?? []
        values.append(value!)
        propertiesTable[key!] = values
    }
    
    /**
     * Add a new key/values pair. If the key exists, add the values to the list
     * of existing values. If the key does not exist, add the key and values to
     * the properties table. Key or values cannot be `nil`. `Key` cannot be empty
     * or exceed `MAX_KEY_LENGTH` bytes. A value in `values` cannot exceed
     * `MAX_STRING_VALUE_LENGTH` bytes. The length is measured after the string
     * is encoded using UTF-8 encoding.
     *
     * - parameter key: property key
     *
     * - parameter values: property values
     *
     * - Throws: `ClientError.argument` if a parameter is `nil`, or if the `key`
     *           parameter is zero length, or `key` exceeds 
     *           `Message.MAX_KEY_LENGTH`, or any value exceeds
     *           `Message.MAX_STRIN_VALUE_LENGTH` or is zero length
     */
    public func addValues(key: String?, values: [String]?) throws {
        // TODO: If key or value cannot be nil, why are the parameters 
        // optionals ?
        try validateStringKeyOrValue(string: key,
                                     maxLength: Message.MAX_KEY_LENGTH,
                                     keyOrValue: "key")
        if values == nil {
            throw ClientError.argument(NSLocalizedString(
                "Illegal Message property, values is nil", comment: ""))
        }
        for v in values! {
            try validateStringKeyOrValue(string: v,
                                    maxLength: Message.MAX_STRING_VALUE_LENGTH,
                                    keyOrValue: "value")
        }
        
        var existingValues: [String]
        existingValues = propertiesTable[key!] ?? []
    
        for v in values! {
            existingValues.append(v)
            propertiesTable[key!] = existingValues
        }
    }
    
    /**
     * Copy another `MessageProperties` by adding all properties to the current 
     * `MessageProperties`
     *
     * - parameter properties: `MessageProperties` to copy
     *
     * - throws: `ClientError.argument` if a any key or value is `nil`, or zero 
     *           length, or a key exceeds `Message.MAX_KEY_LENGTH`, or value 
     *           exceeds `Message.MAX_STRING_VALUE_LENGTH`
     */
    public func copy(properties: MessageProperties) throws {
        for key in properties.getKeys() {
            try addValues(key: key, values: propertiesTable[key])
        }
    }
    
    /**
     * Creates new instance of `MessageProperties` using values from 
     * `MessagePropertiesBuilder`.
     *
     * returns: Instance of `MessageProperties`
     */
    public func  build() -> MessageProperties {
        return MessageProperties(builder: self)
    }
    
    /**
     * Validate `string` for `nil`, empty string and length.
     *
     * - parameter string: property key or value
     *
     * - parameter maxLength: max key or value length
     *
     * - parameter keyOrValue: "key" or "value" to embed in message
     *
     * - throws: `ClientError.argument` if `string` is `nil`, or zero length, or
     *           exceeds `Message.MAX_KEY_LENGTH` if `keyOrValue` is "key", 
     *           exceeds `Message.MAX_STRING_VALUE_LENGTH` if `keyOrValue` is
     *           "value".
     */
    private func validateStringKeyOrValue(string: String?,
                                          maxLength: Int,
                                          keyOrValue: String) throws {
        guard let string = string else {
            throw ClientError.argument(NSLocalizedString(
                "Illegal Message property, \(keyOrValue) is nil", comment: ""))
        }
        if string.isEmpty {
            throw ClientError.argument(NSLocalizedString(
                "Illegal Message property, \(keyOrValue) is empty", comment: ""))
        }
        if string.count > maxLength {
            throw ClientError.argument(NSLocalizedString(
                "Illegal Message property, \(keyOrValue) is longer than " +
                "\(maxLength)", comment: ""))
        }
    }
}

/**
 * MessageProperties contains a table of key and list of values pairs for extra
 * information to send. This class is immutable
 */
public class MessageProperties: Hashable {
    /** A dictionary for message properties. It is a key/list of value pair  */
    private var propertiesTable: [String : [String]] = [ : ]
    
    /**
     * `MessageProperties` constructor takes `MessagePropertiesBuilder` and set 
     * propertiesTable.
     *
     * - parameter builder: Builder containing properties.
     */
    fileprivate init(builder: MessagePropertiesBuilder) {
        self.propertiesTable = builder.propertiesTable
    }
    
    /**
     * Check if the properties contain the key.
     *
     * - parameter key: property key
     *
     * - returns: `true` if the properties contain the key
     */
    public func containsKey(key: String) -> Bool {
        return (propertiesTable[key] != nil)
    }
    
    /**
     * Get an array of the keys.
     *
     * - returns: an array of the keys, never `nil`
     */
    public func getKeys() -> [String] {
        return Array(propertiesTable.keys)
    }
    
    /**
     * Get an array of values for a particular key.
     *
     * - parameter key: property key
     *
     * - returns an array of String values
     */
    public func getProperties(key: String) -> [String] {
        return propertiesTable[key]!
    }
    
    /**
     * Get the first value for a particular key.
     *
     * - parameter key: property key
     *
     * - returns: value, may return `nil` if the key does not exist or values 
     *            assigned to this key is empty.
     */
    public func getProperty(key: String) -> String? {
        let propertyValues = propertiesTable[key]
    
        if (propertyValues != nil) && (propertyValues?.count > 0) {
            return propertyValues![0]
        } else {
            return nil
        }
    }
    
    /**
     * Get the first value for a particular key.
     *
     * - parameter key: property key
     *
     * - parameter index: index in the array of the String values
     *
     * - returns: value, may return `nil` if the key does not exist or index 
     *            is out of range
     */
    public func getProperty(key: String, index: Int) -> String? {
        let propertyValues = propertiesTable[key]
        
        if (propertyValues != nil) && (propertyValues?.count > index) {
            return propertyValues![index]
        } else {
            return nil
        }
    }
    
    /**
     * Get all properties.
     *
     * - return a dictionary of key and list of values, never `nil`
     */
    public func getAllProperties() -> [String : [String]] {
        return self.propertiesTable
    }
    
    /**
     * Method to print the message properties in JSON format.
     *
     * - returns: message properties in string format
     */
    public func toString() -> String {
        return toJsonString()
    }
    
    /**
     * Method to export the message properties as Json string.
     *
     * - returns: message properties as JSON string
     */
    public func toJsonString() -> String {
        var resultStr: String = "{"
        var z: Int = 0
        for (key, val) in self.propertiesTable {
            resultStr = resultStr + "\"" + key + "\":["
            var zz: Int = 0
            for s in val {
                resultStr = resultStr + "\"" + s + "\""
                zz += 1
                if zz < val.count {
                    resultStr = resultStr + ","
                }
            }
            resultStr = resultStr + "]"
            z += 1
            if z < self.propertiesTable.count {
                resultStr = resultStr + ","
            }
        }
        resultStr = resultStr + "}"
        return resultStr
    }

    // TODO: from enterprise version
    /**
     * Method to export the message properties to Json object.
     * - returns: Message properties in JSON format.
     */
    public func toJSON() -> JSON {
        return JSON(self.propertiesTable)
    }
    
    /**
     * Implementation of protocol Hashable
     *
     * - returns: the hash value
     */
    public var hashValue : Int {
        return NSDictionary(dictionary: self.propertiesTable).hashValue
    }
    

}

/**
 * Implementation of protocol `Equatable`
 *
 * parameter lhs: one item to be compared
 *
 * parameter rhs: the other item to be compared
 *
 * - returns: `true` if equal, `false` otherwise
 */
public func ==(lhs: MessageProperties, rhs: MessageProperties) -> Bool {
    return NSDictionary(dictionary: lhs.getAllProperties()).isEqual(to: rhs.getAllProperties())
}


