/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation

/**
 * Enum for handling HTTP status codes and tests.
 */
public enum StatusCode: String {
    case OK = "OK"
    case CREATED = "Created"
    case ACCEPTED = "Accepted"
    case NON_AUTHORITATIVE_INFORMATION = "Non Authoritative Information"
    case NO_CONTENT = "No Content"
    case FINISHED = "Finished"
    case DATA_FINISHED = "Data Finished"
    case BAD_REQUEST = "Bad Request"
    case UNAUTHORIZED = "Unauthorized"
    case PAYMENT_REQUIRED = "Payment Required"
    case FORBIDDEN = "Forbidden"
    case NOT_FOUND = "Not Found"
    case METHOD_NOT_ALLOWED = "Method Not Allowed"
    case NOT_ACCEPTABLE = "Not Acceptable"
    case REQUEST_TIMEOUT = "Request Timeout"
    case CONFLICT = "Conflict"
    case INTERNAL_SERVER_ERROR = "Internal Server Error"
    case NOT_IMPLEMENTED = "Not Implemented"
    case BAD_GATEWAY = "Bad Gateway"
    case SERVICE_UNAVAILABLE = "Service Unavailable"
    case OTHER = "Other"
    
    
    /**
     * Returns HTTP status code.
     
     * - returns: HTTP status code
     */
    public func getCode() -> Int {
        switch self {
            case .OK: return 200
            case .CREATED: return 201
            case .ACCEPTED: return 202
            case .NON_AUTHORITATIVE_INFORMATION: return 203
            case .NO_CONTENT: return 204
            case .FINISHED: return 205
            case .DATA_FINISHED: return 206
            case .BAD_REQUEST: return 400
            case .UNAUTHORIZED: return 401
            case .PAYMENT_REQUIRED: return 402
            case .FORBIDDEN: return 403
            case .NOT_FOUND: return 404
            case .METHOD_NOT_ALLOWED: return 405
            case .NOT_ACCEPTABLE: return 406
            case .REQUEST_TIMEOUT: return 408
            case .CONFLICT: return 409
            case .INTERNAL_SERVER_ERROR: return 500
            case .NOT_IMPLEMENTED: return 501
            case .BAD_GATEWAY: return 502
            case .SERVICE_UNAVAILABLE: return 503
            case .OTHER: return -1
        }
    }
    
    /**
     * Returns text description of HTTP status.
     
     * - returns: text with description
     */
    public func getDescription() -> String {
        return self.rawValue
    }
    
    /**
     * Return information about StatusCode in format: Http (code) (description).
     
     * - returns: Text of StatusCode
     */
    public func toString() -> String {
        return "HTTP " + getCode().description + ": " + getDescription()
    }
    
    static let allValues = [OK, CREATED, ACCEPTED, NON_AUTHORITATIVE_INFORMATION,
                            NO_CONTENT, FINISHED, DATA_FINISHED, BAD_REQUEST,
                            UNAUTHORIZED, PAYMENT_REQUIRED, FORBIDDEN, NOT_FOUND,
                            METHOD_NOT_ALLOWED, NOT_ACCEPTABLE, REQUEST_TIMEOUT,
                            CONFLICT, INTERNAL_SERVER_ERROR, NOT_IMPLEMENTED,
                            BAD_GATEWAY, SERVICE_UNAVAILABLE, OTHER]
    
    /**
     * Returns instance of StatusCode for given code. If the code is not found, return *OTHER*
     
     * - parameter code: status code to lookup
     
     * - returns: instance of StatusCode matching the code value.
     */
    public static func valueOf(code: Int) -> StatusCode {
        for sc in StatusCode.allValues {
            if sc.getCode() == code {
                return sc
            }
        }
        return .OTHER
    }
}


