/*
 * Copyright (c) 2017, Oracle and/or its affiliates.  All rights reserved.
 *
 * This software is dual-licensed to you under the MIT License (MIT) and 
 * the Universal Permissive License (UPL).  See the LICENSE file in the root
 * directory for license terms.  You may choose either license, or both.
 */

//package com.oracle.iot.client.util
import Foundation

public enum ProgressState {
    /**
     * Initial state after upload() or download() is called
     */
    case initiated

    /**
     * Up/download is queued and not yet started
     */
    case queued

    /**
     * Up/download is currently in progress
     */
    case inProgress

    /**
     * Up/download completed successfully
     */
    case completed

    /**
     * Up/download was canceled before it completed
     */
    case canceled

    /**
     * Up/download failed without completing
     */
    case failed
    
    /**
     * Initial state
     */
    case unknown
}

/**
 * An object for receiving progress via the {@code ProgressCallback}.
 */
public protocol Progress {

    /**
     * Get the StorageObject that was queued for which this progress event 
     * pertains.
     *
     * - returns: a `StorageObject`
     */
    func getStorageObject() -> AdvancedStorageObject

    /**
     * Get the state of the transfer
     *
     * - returns: the transfer state
     */
    func getState() -> ProgressState

    /**
     * Get the exception that resulted in the transfer failure. The 
     * exception is either a `ClientError.network` or `ClientError.security`
     * error. This is useful if the progress state is `failed`.
     *
     * - returns: the error causing the failure
     */
    func getFailureCause() -> Error?

    /**
     * Get the number of bytes transferred. This can be compared to the
     * length of content obtained by calling `StorageObject.getLength()`.
     *
     * - returns: the number of bytes transferred
     */
    func getBytesTransferred() -> Int64
}


/**
 * A callback interface for monitoring progress of queued content.
 */
open class ProgressCallback {
    public init() {
    }
    /**
     * Notify of progress for content transfer.
     *
     * - parameter progress progress data
     */
    open func progress(progress:Progress) {
        fatalError("Subclass need to implement the 'progress()' method.")
    }
}



/**
 * The StorageDispatcher queues content for automatic upload to, or download
 * from, the Oracle Storage Cloud Service.
 *
 * ```
 *   // Upload example
 *   storageConnection:Storageconnection = dcd.getStorageConnection()
 *   storageDispatcher:StorageDispatcher =
 *              StorageDispatcher.getStorageDispatcher(storageConnection:storageConnection)
 *   storageDispatcher.setProgressCallback(callback: { (progress) in
 *       storageObject:StorageObject = progress.getStorageObject()
 *       if storageObject.getName() == "lenna" {
 *           if progress.getState() == ProgressState.completed {
 *               // Can now safely remove images/Lenna.jpg, if desired.
 *           } else if progress.getState() == ProgressState.inProgress &&
 *                   takingTooLong {
 *               storageDispatcher.cancel(storageObject:storageObject)
 *           }
 *       }
 *   })
 *   lenna:StorageObject = dcd.createStorageObject(name:"lenna",
 *                                              contentType: "image/jpeg")
 *   lenna.setInputPath(inputPath:"../images/Lenna.jpg")
 *   storageDispatcher.queue(storageObject:lenna)
 *
 *   // Download example
 *   storageConnection:Storageconnection = dcd.getStorageConnection()
 *   storageDispatcher:StorageDispatcher =
 *       StorageDispatcher.getStorageDispatcher(storageConnection:storageConnection)
 *   storageDispatcher.setProgressCallback(callback: { (progress) in
 *       storageObject:StorageObject = progress.getStorageObject()
 *       if storageObject.getName() == "lenna" {
 *           if progress.getState() == ProgressState.completed {
 *               // downloads/Lenna.jpg is ready
 *           } else if progress.getState() == ProgressState.inProgress &&
 *                      takingTooLong {
 *               storageDispatcher.cancel(storageObject:storageObject)
 *           }
 *       }
 *   })
 *
 *   lenna:StorageObject = dcd.createStorageObject(name:"lenna",
 *                                              contentType:"image/jpeg")
 *   lenna.setOutputPath(outputPath:"../downloads/Lenna.jpg")
 *   storageDispatcher.queue(storageObject:lenna)
 * ```
 */
public protocol StorageDispatcher {

    /**
     * Get the `StorageDispatcher` for the given `StorageConnection`.
     * There is one instance of `StorageDispatcher` for a `StorageConnection`.
     *
     * - parameter storageConnection: the `StorageConnection`
     *
     * - returns: a `StorageDispatcher`
     */
    //static func getStorageDispatcher(storageConnection:StorageConnection) -> StorageDispatcher
    static func getStorageDispatcher(dcd:AdvancedDirectlyConnectedDevice) -> StorageDispatcher
    /**
     * Add a `StorageObject` to the queue to upload/download content to/from
     * the Storage Cloud.
     *
     * - parameter storageObject The content storageObject to be queued
     * - throws:
     *      `ClientError.argument` if `storageObject` is `nil`
     *      `ClientError.state` if the object is already queued or in progress
     */
    func queue(storageObject: AdvancedStorageObject?) throws

    /**
     * Cancel the transfer of content to or from storage. This call has no
     * effect if the transfer is completed, already canceled, has failed, or
     * `storageObject` is not queued.
     *
     * - parameter storageObject: the content `StorageObject` to be canceled.
     */
    func cancel(storageObject: AdvancedStorageObject)


    /**
     * Set a callback to be notified as the transfer progresses.
     *
     * - parameter callback: callback to invoke, if `nil`,
     *                 the existing callback will be removed
     */
    func setProgressCallback(callback:ProgressCallback?)
}
