/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import Alamofire

/**
 * The `DefaultTrustedAssetsManager` class provides a default
 * implementation of the `TrustedAssetsManager` interface. The trusted assets
 * are stored in a `KeyStore` protected by a password. The trust
 * assets are stored in the `KeyStore` as follows:
 *
 * # Shared Secret #
 *
 * The shared secret is stored as a `SecretKeyEntry` under an alias in the
 * form of a `URI` constructed from the server host, the server port, the
 * client ID and, if defined, the endpoint ID, as follows:
 * ````
 * iotcs://client-id@server-host:server-port[#endpoint-id]
 * ````
 *
 * Only one such entry should exist in the `Keystore`.
 *
 * # Private Key and Certificate #
 *
 * The private key and corresponding certificate are stored as a
 * `PrivateKeyEntry` under an alias in the form of a `URI`
 * constructed from the server host, the server port, the device ID and the
 * endpoint ID, as follows:
 *
 * ````
 * iotcs://client-id@server-host:server-port#endpoint-id
 * ````
 *
 * Only one such entry should exist in the `KeyStore`.
 *
 * # Trusted Certificates (Trust Anchors) #
 *
 * Each trusted certificate is stored as a `TrustedCertificateEntry`; the
 * alias does not carry any particular semantics and its format is unspecified.
 * Many such entries may exist in the `KeyStore`.
 *
 * For example, a trusted assets *KeyStore* may contain the following
 * entries:
 *
 * ````
 * SecretKeyEntry(alias="iotcs://0-AI@acme.foo.iot.bar.oraclecloud.com:7001", 
 *                secret-key=SecretKey("changeit"))
 * PrivateKeyEntry(alias="iotcs://0-AI@acme.foo.iot.bar.oraclecloud.com:7001#XYZ", 
 *                 private-key=RSAPrivateKey(...), certificate=X509Certificate(...))
 * TrustedCertificateEntry(alias="rootCA1", certificate=X509Certificate(...))
 * TrustedCertificateEntry(alias="rootCA2", certificate=X509Certificate(...))
 * ````
 */
public protocol TrustedAssetsManager:class {
    
    /**
     * Retrieves the IoT CS server host name.
     *
     * - Returns: The IoT CS server host name.
     */
    func getServerHost() -> String
    
    /**
     * Retrieves the IoT CS server port.
     *
     * - Returns: The IoT CS server port.
     */
     func getServerPort() -> String
    
    /**
     * Retrieves the IoT CS server scheme, default is https
     *
     * - Returns: The IoT CS server scheme.
     */
     func getServerScheme() -> String
    
    /**
     * Retrieves the ID of this client. If the client is a device the client ID
     * is the device ID; if the client is a pre-activated enterprise application
     * the client ID corresponds to the assigned endpoint ID. The client ID is
     * used along with a client secret derived from the shared secret to perform
     * secret-based client authentication with the IoT CS server.
     *
     * - returns: The ID of this client.
     */
    
    func getClientId() -> String
    
    /**
     * Retrieves the public key to be used for certificate request.
     *
     * - Returns: The public key.
     */
    
    func getPublicKey() -> Foundation.Data
    
    
    /**
     * Retrieves the trust anchor or most-trusted Certification Authority (CA) 
     * certificates to be used to validate the IoT CS server certificate chain.
     *
     * - returns: A `Foundation.Data` array of DER-encoded trust anchor 
     *            certificates (byte arrays).
     */
    func getTrustAnchorCertificates() -> [Foundation.Data]?
    
    /**
     * Returns the server trust policy that was created during the provisioning 
     * of the TAM
     *
     * - returns: an evaluated `ServerTrustPolicy` object that can be used as
     *            input for the secure connection to the cloud
     */
    func getServerTrustPolicy() -> ServerTrustPolicy?
    
    /**
     * Sets the assigned endpoint ID and certificate as returned by the 
     * activation procedure. Upon a call to this method, a compliant 
     * implementation of the `TrustedAssetsManager` interface must ensure the
     * persistence of the provided endpoint credentials. This method can only 
     * be called once; unless the trust manager has been reset.
     *
     * - parameter endpointId: The assigned endpoint ID.
     *
     * - parameter certificate: The DER-encoded certificate issued by the server
     *                          or an empty array if no certificate was provided 
     *                          by the server.
     *
     * - throws: `ClientError.state` if already activated, or private keys have
     *           not been generated, or `endpointId` is already assigned, or
     *           if certificate data has not been encoded.<br/>
     *           `ClientError.argument` if `endpointId` is an empty string.
     */
    func setEndPointCredentials(endpointId: String,
                                certificate: Foundation.Data) throws -> Void
    
    
    /**
     * Retrieves the assigned endpoint ID.
     *
     * - returns: The assigned endpoint ID.
     */
    func getEndpointId() -> String
    
    
    /**
     * Retrieves the assigned endpoint certificate.
     *
     * - returns: The DER-encoded certificate or an empty array if no
     *            certificate was assigned.
     */
    func getEndpointCertificate() -> Foundation.Data?
    
    
    func getOAuthRequestPath() -> String
    
    
    /**
     * Generates the key pair to be used for authentication with the IoT CS 
     * (OAuth2 Assertion Flow).
     *
     * - parameter algorithm: The key algorithm.
     *
     * - parameter keySize: The key size.
     *
     * - throws `ClientError.argument` if `tagPrefix` is the empty string, or
     *          `keySize` is less than or equal to zero<br/>
     *           `ClientError.state` if already activated<br/>
     *           `ClientError.security` if key pair generation fails
     */
    func generateKeyPair(algorithm: String, keySize: Int) throws -> Void
    
    
    /**
     * Signs the provided data using the specified algorithm and the private 
     * key (OAuth2 Assertion Flow).
     *
     * - parameter data: The data to sign.
     *
     * - parameter algorithm: The signature algorithm to use.
     *
     * - returns: The signature.
     */
    func signWithPrivateKey(data: Foundation.Data, algorithm: String) throws -> Foundation.Data?
    
    func signWithPrivateKeyString(str: String, algorithm: String) throws -> String?
    
    
    /**
     * Retrieves the device activation credential (shared secret) to be used for
     * post-production registration. The shared secret may be generated if not 
     * already provisioned. The returned shared secret is encrypted using some 
     * pre-provisioned key - such as a public key / certificate of the IoT CS 
     * (OAuth2 Shared Secret Signature Flow).
     *
     * - Returns: The encrypted shared secret.
     * 
     * - throws: `ClientError.state` if called since it is an unsupported 
     *           operation.
     */
    func getEncryptedSharedSecret() throws -> Foundation.Data?
    
    
    /**
     * Signs the provided data using the specified algorithm and the shared
     * secret of the device indicated by the given activation id.
     * Passing `nil` for `activationId` is identical to passing
     * `getClientId()`.
     *
     * - parameter data: the data to be hashed.
     *
     * - parameter algorithm: The hash algorithm to use.
     *
     * - parameter activationId: The activation id of the device whose shared 
     *                           secret is to be used for signing.
     *
     * - returns: The hash of the data
     *
     * - throws: `ClientError.state` if `activationId` is not `nil` and the 
     *                               shared secret for that id does not exist.
     */
    func signWithSharedSecret(data: String, algorithm: String,
                              activationId: String?) throws -> Foundation.Data
    
    func signWithSharedSecretString(data: String, algorithm: String,
                                    activationId: String?) throws -> String
    
    /**
     * Returns whether the client is activated.
     *
     * - returns: Whether the client is activated.
     */
    func isActivated() -> Bool
    
    /**
     * Returns whether the TAM is successfully provisioned.
     *
     * - returns: Whether TAM is successfully provisioned.
     */
    func isProvision() -> Bool
    
    /**
     * Resets the trust material back to its provisioning state; in particular, 
     * the key pair is erased. The client will have to go, at least, through 
     * activation again; depending on the provisioning policy in place, the 
     * client may have to go through registration again.
     * - throws: `ClientError.security` if any exception occurs.
     */
    func reset() throws -> Void
    
    
    /**
     * Closes this stream and releases any system resources associated with it. 
     * If the stream is already closed then invoking this method has no effect.
     */
    func close() -> Void
}
