/*
 *  Copyright © 2016, 2018, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation

/**
 * Factory class for the creation of a TAM for a given context.
 */

open class TrustedAssetsManagerFactory {
    /**
     * - throws:
     *       - `ClientError.security` if server policy cannot be set.<br/>
     *       - `ClientError.file` if the provision file is empty or has the
     *           wrong format version, or can't be decrypted or the
     *           server url is invalid.
     */
    static func createTAMWithClassName(file: String? = nil,
                                       fileExtension: String? = nil,
                                       subdirectory: String? = nil,
                            password: String) throws -> TrustedAssetsManager? {
        // From the input we build the context (Dictionary) for this TAM 
        // instance File name and extension are always set, if nothing is 
        // provided they are set to the default values if no subdirectory is 
        // set the TAM will look in the Resource directory for the provisioning 
        // file if no password is set the provisioning file is not encrypted but
        // just a json file
        var context = Dictionary<String,Any>()
        context[TAMconstants.fileKey] = file
        context[TAMconstants.fileExtKey] = fileExtension ?? TAMconstants.fileExtension
        if (subdirectory != nil) {
            context[TAMconstants.subdKey] = subdirectory
        }
        context[TAMconstants.paswdKey] = password
        return try UnifiedTrustedAssetsManager.init(context: context)
    }
    
    /**
     * Convenience method for the creation of the standard TAM for a given context.
     * - throws:
     *       - `ClientError.file` if the provision file is empty or has the
     *           wrong format version, or can't be decrypted or the
     *           server url is invalid.
     */
    static func getTrustedAssetsManager(context: AnyObject?) throws -> TrustedAssetsManager? {
        let tamm = try DefaultTrustedAssetsManager(context: context as? Dictionary<String, Any>)
        return tamm
    }
    
    // Prevents instantiation
    private init() {}
    
    /*
     * Factory method to create a TAM instance
     *
     * - parameter path: the path to the provisioning file
     *
     * - parameter password: the password to protect the provisioning file 
     *                       during the creation of the file
     *
     * - throws: `ClientError.file` if `path` does not exist or is a directory, 
     *            or if the provisioning file is empty or has the wrong format 
     *            version, or can't be decrypted or the server url is invalid.
     */
    static func createTrustedAssetsManager(path: String,
                            password: String) throws -> TrustedAssetsManager? {
        // check if the file exist
        let filemgr = FileManager.default
        if (!filemgr.fileExists(atPath: path)) {
            throw ClientError.file(String(format: NSLocalizedString(
                "File at given path %@ does not exist", comment: ""), path))
        }
        // check if the file a directory
        var isDir: ObjCBool = false
        filemgr.fileExists(atPath: path, isDirectory: &isDir)
        if (isDir).boolValue {
            throw ClientError.file(String(format: NSLocalizedString(
                "File at given path %@ is a directory", comment: ""), path))
        }
        // build the context 
        var context = Dictionary<String, Any>()
        context[TAMconstants.pathKey] = path
        context[TAMconstants.paswdKey] = password
        // find out if this is a json file
        let url = URL(fileURLWithPath: path)
        if(url.pathExtension.lowercased() == "json") {
            return try DefaultTrustedAssetsManager(context: context)
        } else {
            return try UnifiedTrustedAssetsManager.init(context: context)
        }
    }
    
    open static func getTrustedAssetsManager(path: String, password: String) throws -> TrustedAssetsManager? {
        let tam: TrustedAssetsManager?
        tam = try TrustedAssetsManagerFactory.createTrustedAssetsManager(path: path, password: password)
        return tam!
    }
}
