/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import Heimdall
import Security
import Alamofire


class TAMconstants {
    static let provisionFileName:String = "truststore"
    static let fileExtension:String = "json"
    // The keys for the context Dictionary
    static let fileKey:String = "file"
    static let fileExtKey:String = "fileExtension"
    static let subdKey:String = "subdirectory"
    static let paswdKey:String = "password"
    static let pathKey:String  = "path"
}

// TODO: Does this class need to inherit NSObject ?
/**
 * The DefaultTrustedAssetManager class implementation
 */
public class TrustedAssetsManagerImpl :NSObject, TrustedAssetsManager {
    
    @objc var serverHost:String = ""
    @objc var serverPort:String = ""
    @objc var endpointId:String = String()
    @objc var clientId:String = ""
    @objc var deviceId:String = ""
    @objc var sharedSecret:String = ""
    /// Indicates if the TAM is activated
    @objc var isActive = false
    /// Indicates if the TAM is successfully provisioned
    @objc var isProvisioned = false
    // for the prototype to test against Server
    @objc var serverScheme = "https"
    @objc let oauthRequestPath = "/oauth2/token"
    
    // References to the certificates
    @objc var serverCert: SecCertificate?
    @objc var rootCACert: SecCertificate?
    @objc var endpointCert: SecCertificate?
    
    internal var serverPolicy:ServerTrustPolicy?
    
    // The Prefix for the tags in the keychain for the private and public keys
    @objc let keyPublic  = "public"
    @objc let keyPrivate = "private"
    @objc var tagPublic  = String()
    @objc var tagPrivate = String()
    
    @objc let RSAkeySize = 2048
    // the class to perform all the RSA computation
    var heimdall: Heimdall?
    
    // The ICD map
    @objc var icdMap: Dictionary<String, String> = Dictionary()

    public override init() {
        super.init()
    }
    
    /**
     * Returns the Server Host, must be a name that can be resolved by DNS,
     * no IP number. The Server Host name is provisioned with the
     * provisioning file.
     *
     * -returns: server Host as String
     */
    @objc public func getServerHost() -> String { return self.serverHost }

    /**
     * Returns the Server Port. Server Port is provisioned with the 
     * provisioning file.
     *
     * -returns: the server port as String
     */
    @objc public func getServerPort() -> String { return self.serverPort }
    
    /**
     * Returns the Server scheme. Server scheme is provisioned with the 
     * provisioning file.
     *
     *  -returns: the server scheme as String
     */
    @objc public func getServerScheme() -> String { return self.serverScheme }

    /**
     * Returns the activation state of the TAM
     *
     * -returns: `false` if the TAM is not activated, or `true` if the TAM is
     *           already activated
     */
    @objc public func isActivated() -> Bool { return self.isActive }
    
    /**
     * Returns the provisioning state of the TAM
     *
     * -returns: `false` if the TAM is not activated, or `true` if the TAM is
     *           already activated
     */
    @objc public func isProvision() -> Bool { return self.isProvisioned }
    
    /**
     * Sets the state of the TAM to Activated will only work of the
     * Provisioning and initialization worked correctly.
     */
    @objc public func setToActivated() -> Void {
        if(self.isProvisioned){
            self.isActive = true
        }
    }
    
    /**
     * Returns the Endpoint ID that is provisioned with the provisioning file
     *
     * -returns: the endpoint ID
     */
    @objc public func getEndpointId() -> String {return self.endpointId }

    /**
     * Returns the Client ID that is provisioned with the provisioning file
     *
     * -returns: the client ID
     */
    @objc public func getClientId() -> String {
        if(!self.endpointId.isEmpty) {return self.endpointId}
        return self.deviceId
    }
    
    /**
     * Returns the Oauth Request Path
     *
     * -returns: the oauthRequestPath
     */
    @objc public func getOAuthRequestPath() -> String { return self.oauthRequestPath }
    
    /**
     * Resets the trust material back to its provisioning state; in particular,
     * the key pair is erased. The client will have to go, at least,through
     * activation again;
     * depending on the provisioning policy in place, the client may have to go
     * through registration again.
     */
    @objc public func reset() throws {
        // deletes the key pair and sets the internal state of the TAM to false
        // the TAM has to be reprovisioned and activated
        if(heimdall != nil) {
            let _ = heimdall!.destroy()
        }
        self.isProvisioned = false
        self.isActive = false
        return
    }
    
    /**
     * Closes the TAM.
     */
    @objc public func close() {
        return
    }
    
    /**
     * Sets the assigned endpoint ID and certificate as returned by the
     * activation procedure. Upon a call to this method, a compliant
     * implementation of the `TrustedAssetsManager` interface must ensure
     * the persistence of the provided endpoint credentials. This method can
     * only be called once; unless the *TrustedAssetsManager* has been
     * reset.
     *
     * If the client is a pre-activated enterprise application, the endpoint ID
     * has already been provisioned and calling this method MUST fail.
     *
     * - parameter endpointId: the assigned endpoint ID.
     *
     * - parameter certificate: the DER-encoded certificate issued by the server    
     *                          or an empty array if no certificate was
     *                          provided by the server.
     *
     * - throws `ClientError.state` if already activated, or private keys have
     *          not been generated, or `endpointId` is already assigned, or if 
     *          certificate data has not been encoded.<br/>
     *          `ClientError.argument` if `endpointId` is an empty string.
     *
     */
    @objc public func setEndPointCredentials(endpointId: String, 
                                certificate: Foundation.Data) throws {
        // Error handling needed here
        if(self.heimdall == nil) {
            throw ClientError.state(NSLocalizedString(
                "Private key not generated", comment: ""))
        }
        if(!self.endpointId.isEmpty) {
            throw ClientError.state(NSLocalizedString(
                "EndpointId already assigned.", comment: ""))
        }
        if (endpointId.isEmpty) {
            throw ClientError.argument(NSLocalizedString(
                "Endpoint must not be an empty string.", comment: ""))
        }
        self.endpointId = endpointId
        if(certificate.count != 0) {
            if let endpointCertificate =
                SecCertificateCreateWithData(nil, certificate as CFData) {
                self.endpointCert = endpointCertificate
            } else {
                throw ClientError.security(NSLocalizedString(
                    "Certificate data are not DER encoded", comment: ""))
            }
        }
        self.setToActivated()
        return
    }
    
    /**
     * Retrieves the trust anchor or most-trusted Certification Authority (CA)
     * certificates to be used to validate the IoT CS server certificate chain.
     *
     * - Returns: An array of DER-encoded trust anchor certificates
     *            (byte arrays).
     */
    @objc public func getTrustAnchorCertificates() -> [Foundation.Data]? {
        var certArray = [CFData]()
        var certData:CFData?
        if(self.rootCACert != nil) {
            certData = SecCertificateCopyData(self.rootCACert!)
            if (certData != nil ) {
                certArray.append(certData! )
            }
        }
        if(self.serverCert != nil) {
            certData = SecCertificateCopyData(self.serverCert!)
            if (certData != nil ) {
                certArray.append(certData! )
            }
        }
        return certArray as [Foundation.Data]
    }
    
    /**
     * Retrieves the assigned endpoint certificate.
     *
     * - Returns: The DER-encoded certificate or an empty array if no 
     *            certificate was assigned.
     */
    @objc public func getEndpointCertificate() -> Foundation.Data? {
        if(self.endpointCert != nil) {
            return SecCertificateCopyData(self.endpointCert!) as Foundation.Data
        }
        return nil
    }
    
    /**
     * Retrieves the device activation credential (shared secret) to be used 
     * for post-production registration. The shared secret may be generated 
     * if not already provisioned. The returned shared secret is encrypted 
     * using some pre-provisioned key - such as a public key / certificate of 
     * the IoT CS (OAuth2 Shared Secret Signature Flow).
     *
     * - Returns: The encrypted shared secret.
     * - Throws: errorSecurityIssue if this method is called prior the
     *           *TrustedAssetsManager* is fully initialized.
     */
    // TODO: FIXME Why is this here if unsupported ?
    public func getEncryptedSharedSecret() throws -> Foundation.Data? {
        // do we need this method here it is not supported anywhere
        throw ClientError.security(NSLocalizedString(
            "Unsupported operation...", comment: ""))
    }

    /**
     * Returns the server trust policy that was created during the
     * provisioning of the TAM
     *
     * - returns: an evaluated ServerTrustPolicy object that can be used as
     *            input for the secure connection to the cloud
     */
    public func getServerTrustPolicy() -> ServerTrustPolicy? {
        return self.serverPolicy
    }


    /**
     * Signs the provided data using the specified algorithm and the private
     * key. This method is only use for assertion-based client authentication
     * with the IoT CS.
     * - parameter data: the data to sign.
     * - parameter algorithm: the signature algorithm to use. (only supported
     *                        algorithm is PKCS1SHA256) (supported in iOS are
     *                        None, PKCS1, OAEP, PKCS1SHAXXX)
     * - returns: the signature.
     */
    public func signWithPrivateKey(data: Foundation.Data, algorithm: String) throws ->
            Foundation.Data? {
        if self.heimdall == nil {
            self.heimdall = self.getHeimdallCrypto()
        }
        // No error for the algorithm because we have PKCS1SHA256 hardwired
        if(self.heimdall != nil) {
            let sigData = self.heimdall!.sign(data)
            if(sigData == nil) {
                throw ClientError.security(NSLocalizedString("Could not create signature.", comment: ""))
            }
            return sigData
        }
        return nil
    }
    

    public func signWithPrivateKeyString(str: String, algorithm: String) throws ->
            String? {
        if self.heimdall == nil {
            self.heimdall = self.getHeimdallCrypto()
        }
        let sigStr: String? = self.heimdall!.sign(str)
        if sigStr == nil {
            throw ClientError.security(NSLocalizedString("Could not create signature.", comment: ""))
        }
        return sigStr
    }

    /**
     * Signs the provided data using the specified algorithm and the shared
     * secret of the device indicated by the given activation id.
     * Passing `nil` for `activationId` is identical to passing
     * `getClientId()`.
     *
     * - parameter data: the data to be hashed.
     *
     * - parameter algorithm: The hash algorithm to use.
     *
     * - parameter activationId: The activation id of the device whose shared
     *                           secret is to be used for signing.
     *
     * - returns: The hash of the data
     *
     * - throws: `ClientError.state` if `activationId` is not `nil` and the
     *                               shared secret for that id does not exist.
     */
    @objc public func signWithSharedSecret(data: String, algorithm: String,
                    activationId: String? = nil) throws -> Foundation.Data {

        let signature:String = try signWithSharedSecretString(data: data,
            algorithm: algorithm, activationId: activationId)

        return signature.data(using: .utf8)!
    }
    
    @objc public func signWithSharedSecretString(data: String, algorithm: String,
                                activationId:String? = nil) throws -> String {
        var secretKey:String?
        if (activationId == nil || activationId! == getClientId()) {
            secretKey = self.sharedSecret
        } else {
            secretKey = icdMap[activationId!]
        }
        
        if (secretKey == nil || secretKey!.isEmpty) {
            throw ClientError.state(String(format: NSLocalizedString(
                "Shared secret for %@ is empty", comment: ""),
                    activationId != nil ? activationId! : getClientId()))
        }

        let signature = HMAC.sign(message: data,
                algorithm: HMAC.getAlgorithm(algorithm: algorithm),
                key: secretKey!)
        return signature!
    }
    
    /**
     * Creates an RSA key-pair and stores them in the keychain (RSA is the only 
     * algorithm supported by the iOS API)
     *
     * - parameter keysize: keylength in bits
     *
     * - parameter tagPrefix: for the keys to find them in the keychain
     *
     * - throws: `ClientError.argument` if `tagPrefix` is the empty string, or
     *           `keySize` is less than or equal to zero<br/>
     *           `ClientError.security` if key pair generation fails
     *
     */
    @objc public func generateKeyPair(algorithm: String, keySize: Int) throws -> Void {
        if(algorithm.isEmpty) {
            throw ClientError.argument(NSLocalizedString(
                "Prefix to identify the keys in the keychain must be set",
                    comment: ""))
        }
        if(keySize <= 0) {
            throw ClientError.argument(NSLocalizedString(
                "Key size cannot be negative or zero", comment: ""))
        }
        
        let tagPublic = self.tagPublic + algorithm
        let tagPrivate = self.tagPrivate + algorithm
        self.heimdall = Heimdall(publicTag: tagPublic,
                privateTag: tagPrivate, keySize: keySize)
        if(self.heimdall == nil) {
            throw ClientError.security(NSLocalizedString(
                "KeyPair generation failed", comment: ""))
        }
    }
    
    /**
     * Retrieves the public key to be used for certificate request.
     *
     * - returns: the public key.
     */
    @objc public func getPublicKey() -> Foundation.Data {
        return (heimdall?.publicKeyDataX509())!
    }
    
//---- Impl Trust evaluation ------------------------------------------------
 

    /**
     * Creates a SecTrust object from the given array of certificates based
     * on the X509 policies
     *
     * - parameter certificates: An array of SecCertificates that build the
     *                            chain of certificates from the target server
     *                            to the rot CA certificate
     * - returns: a SecTrust object
     */
    @objc func trustWithCertificates(certificates: [SecCertificate]) -> SecTrust? {
        let policy = SecPolicyCreateBasicX509()
        var trust: SecTrust?
        SecTrustCreateWithCertificates(certificates as CFTypeRef, policy, &trust)
        
        return trust!
    }
    
    /**
     * Set the root CA certificate as a trusted Anchor
     *
     * - parameter trust: the SecTrust object
     *
     * - parameter root: the certificate of the root CA as a
     *                    SecCertificate object
     *
     * - returns: `true`if setting the root CA as trust anchors completed
     *            with no error
     */
    @objc func setRootCertificateAsLoneAnchorCertificateForTrust(
                                trust: SecTrust, root: SecCertificate) -> Bool {
        var error =  errSecSuccess
        var rootArray = [SecCertificate]()
        rootArray.append(root)
        error = SecTrustSetAnchorCertificates(trust, rootArray as CFArray)
        error = SecTrustSetAnchorCertificatesOnly(trust, true)
        if(error == errSecSuccess) {return true}
        return false
    }
    
    /**
     * Evaluates the SecTrust object for a server given in the SecTrust object
     * by the all the certificates
     * starting at the root CA the server certificate
     *
     * - parameter trust: is the SecTrust object build from the the array of
     *                     certificates starting with the target host
     *                     certificate to the root CA certificate
     * - returns: `true` is the SecTrust object is valid
     */
    @objc func trustIsValid(trust: SecTrust) -> Bool {
        var isValid = false
        
        var result =
            SecTrustResultType(rawValue: SecTrustResultType.invalid.rawValue)
        let status = SecTrustEvaluate(trust, &result!)
        
        if status == errSecSuccess {
            let unspecified = SecTrustResultType(
                rawValue: SecTrustResultType.unspecified.rawValue)
            let proceed = SecTrustResultType(
                rawValue: SecTrustResultType.proceed.rawValue)
            
            isValid = result == unspecified || result == proceed
        }
        
        return isValid
    }
    
    /**
     * Set the Server trust policy and validate it against the given host name
     *
     *  - parameter hostName: the host to define and validate the policies for
     *  - parameter trust: the SecTrust object for this host
     *  - parameter certificateChain: the certificate chain for the host
     *  - returns: `true` if the server trust policy with the validated
     *             certificate chain and the validated host is valid
     */
    @objc func setServerTrustPolicy(hostName: String, trust: SecTrust,
                              certificateChain: [SecCertificate]) -> Bool {
        let host = hostName
        let serverTrust = trust
        let certificates = certificateChain
        let serverTrustPolicy = ServerTrustPolicy.pinCertificates(
            certificates: certificates,
            validateCertificateChain: true,
            validateHost: true
        )
        
        let serverTrustIsValid = serverTrustPolicy.evaluate(serverTrust,
                                                            forHost: host)
        if(serverTrustIsValid) {
            self.serverPolicy = serverTrustPolicy
        }
        return serverTrustIsValid
    }
    
    
// Impl ertificate keychain helper methods ---------------------------
    
    func setCertificateinKeyChain(certificate: SecCertificate,
                                  tagName: String) throws -> Bool {
        let secCertData:Foundation.Data =
            SecCertificateCopyData(certificate) as Foundation.Data
        let queryFilter = [
            String(kSecClass)              : kSecClassCertificate,
            String(kSecAttrLabel)          : tagName,
            String(kSecValueData)          : secCertData,
            ] as [String : Any]
        let result = SecItemAdd(queryFilter as CFDictionary, nil)
        if( result == errSecDuplicateItem) {
            throw ClientError.security(NSLocalizedString("Certificate already in keychain \(tagName).", comment: ""))
        }
        if result != noErr {
            throw ClientError.security(NSLocalizedString("Error in setting a certificate into the keychain.", comment: ""))
        }
        return true
    }
    
    /**
     * Retrieves a certificate as SecCertificate object from the keychain
     *
     * - returns: a certificate with the `tagName` in the keychain else `nil`
     */
    @objc func getCertificateFromKeyChain(tagName: String) ->
        SecCertificate? {
            let certAkc = getRawCertificateFromKeyChain(tagName: tagName)
            if (certAkc == nil) {
                return nil
            }
            return SecCertificateCreateWithData(nil,certAkc! as CFData)!
    }
    
    /**
     *  Retrieves the certificate data as raw data from the keychain
     * - returns: the data representing the certificate
     */
    private func getRawCertificateFromKeyChain(tagName: String) -> Foundation.Data? {
        let queryFilter = [
            String(kSecClass)              : kSecClassCertificate,
            String(kSecAttrLabel)          : tagName,
            String(kSecReturnData)         : true
            ] as [String : Any]
        
        var keyPtr: AnyObject?
        let error = withUnsafeMutablePointer(to: &keyPtr) {
            SecItemCopyMatching(queryFilter as CFDictionary,
                                UnsafeMutablePointer($0))
        }
        
        if(error != noErr) {
            print(error)
        }
        
        return keyPtr as? Foundation.Data
    }
    
    @objc func removeCertificateFromKeyChain(tagName: String) -> Bool {
        let queryFilter = [
            String(kSecClass)              : kSecClassCertificate,
            String(kSecAttrLabel)          : tagName,
            String(kSecAttrSerialNumber)   : tagName
            ] as [String : Any]
        
        let statusCode = SecItemDelete(queryFilter as CFDictionary)
        
        if statusCode != errSecSuccess {
            
            return false
        } else {
            return true
        }
    }
  
//--------------------Impl Cryptofunctions------------------------------
    
    /**
     * Generate a signature for a String message
     *
     * - parameter string: Message to generate the signature for
     * - parameter urlEncode: `true` if the resulting Base64 data should be
     *                         URL encoded
     * - returns: Signature as a Base64 string
     */
    @objc func sign(stringToSign: String, urlEncode: Bool = false) -> String? {
        return self.heimdall!.sign(stringToSign,urlEncode: urlEncode)
    }
    
    /**
     * Generate a signature for an arbitrary payload, first hashes the data
     * with SHA-256 then generates
     * signature with the private key stored in the keychain
     * - parameter data: Data to sign
     * - returns: Signature as Foundation.Data
     */
    @objc func sign(data: Foundation.Data) -> Foundation.Data? {
        return self.heimdall!.sign(data)
    }
    
    /**
     * Performs a HMAC operation over the given String and returns the
     * signature as a String.
     * The shared secret is provisioned with the Provisioning file.
     *
     * - parameter message: Message to be signed.
     * - returns: The signature of the message.
     */
    @objc func signWithSharedSecret(message: String) -> String {
        let signature = HMAC.sign(message: message,
                                  algorithm: .sha256,
                                  key: self.sharedSecret)
        return signature!
    }
    
    @objc func signWithSharedSecret(secretKey: [UInt8], message: String)-> String {
        let signature = HMAC.sign(message: message,
                                  algorithm: .sha256,
                                  key: String.init(bytes: secretKey,
                                                   encoding: .utf8)!)
        return signature!
    }

    private func getHeimdallCrypto() -> Heimdall? {
        let tagPrefix = "RSA"
        let keySize = 2048
        let tagPublic = self.tagPublic + tagPrefix
        let tagPrivate = self.tagPrivate + tagPrefix
        
        self.heimdall = Heimdall(publicTag: tagPublic,
                                 privateTag: tagPrivate, keySize: keySize)
        
        return self.heimdall
    }

//--------------------- Impl Provisioning methods ---------------------------
    
    /**
     * A more platform independent way of retrieving the URL to the file in
     * the Resources directors works in the unit test framework as well as on
     * the device
     */
    @objc func URLForResource(fileName: String,
                        withExtension: String) -> URL? {
        let url:URL? = TrustedAssetsManagerImpl.URLForResource(fileName: fileName,
                                          withExtension: withExtension,
                                          subdirectory: nil)
        return url
    }
    
    @objc class func URLForResource(fileName: String, withExtension: String,
                        subdirectory: String? ) -> URL? {
        
        // Note that tests "must" use a full path to the assets file.
        // Once a bundle is loaded, dynamically downloaded assets files
        // will not be seen. This is why tests use the same endpoint id
        // for all tests
        
        // If the filename is a full path just return a URL for it.
        // it is assumed the extension is part of the path.
        // Surprisingly the full path is used during the running of
        // DirectlyConnectedDeviceSample, so we have to assume that it
        // is "ok" to reference full paths on the the device. Maybe as
        // long as the root of the path is obtained from the bundle.
        if fileName.hasPrefix("/") {
            return URL(fileURLWithPath: fileName)
        }
        
        // ask for the application bundle
        var bundle:Bundle? = Bundle.main
        
        // This code is no longer necessary for determining a simulator
        // context for tests, since we define an app for running tests.
        // At least I think this is what this code was for. As far as I
        // can tell on the device, there is always a ".app" extension on
        // the bundle container.
        if let path = bundle?.bundlePath {
            let pathArr = path.components(separatedBy: ".")
            // if we have a path to the bundle that does not end with app
            if pathArr.isEmpty || pathArr[pathArr.count - 1] != "app" {
                // we retrieve the Bundle that is available in the test environment
                class Bundle {}
                bundle = Foundation.Bundle(for: Bundle.self)
            }
        }
        
        // if we have a bundle we can retrieve the URL
        if let bundle = bundle {
            if subdirectory != nil {
                if let url:URL = bundle.url(forResource: fileName,
                                             withExtension: withExtension,
                                             subdirectory: subdirectory) {
                    return url
                }
            } else {
                if let url2:URL = bundle.url(forResource: fileName,
                                              withExtension: withExtension) {
                    return url2
                }
            }
        }
        return nil
    }
    
    @objc func addShareSecret(hardwareId: String, sharedSecret: String) {
        icdMap[hardwareId] = sharedSecret
    }
    
//---- Impl Keychain methods
    
    @objc func insertItemWithAttributes(attributes: [String: AnyObject]) throws {
        
        let statusCode = SecItemAdd(attributes as CFDictionary, nil)
        
        if statusCode == errSecDuplicateItem {
            throw errorForStatusCode(statusCode: statusCode)
        }
        
        if statusCode != errSecSuccess {
            throw errorForStatusCode(statusCode: statusCode)
        }
    }
    
    @objc func removeItemWithAttributes(attributes: [String: AnyObject]) throws {
        
        let statusCode = SecItemDelete(attributes as CFDictionary)
        
        if statusCode != errSecSuccess {
            throw errorForStatusCode(statusCode: statusCode)
        }
    }
    
    func fetchItemWithAttributes(attributes: [String: AnyObject]) throws -> [String: AnyObject]? {
        
        var result: AnyObject?
        
        let statusCode = withUnsafeMutablePointer(to: &result) { pointer in
            SecItemCopyMatching(attributes as CFDictionary, UnsafeMutablePointer(pointer))
        }
        
        if statusCode != errSecSuccess {
            throw errorForStatusCode(statusCode: statusCode)
        }
        
        if let result = result as? [String: AnyObject] {
            return result
        }
        
        return nil
    }
    
    func errorForStatusCode(statusCode: OSStatus) -> ClientError {
        return ClientError.security(String(format: NSLocalizedString(
            "swift.keychain.error %@", comment: ""), "\(statusCode)"))
    }

}
